/**
 * Copyright (C) 2000, 2001 Red Hat, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/**
 * @brief Adm input/output functions
 *
 * Basic input/output functions for the Adm. In order to be able to store and
 * retrive XML encoded AdmContexts we use an own IO interface to the
 * filesystem.
 *
 * That way we can store the contexts either directly in files or compress,
 * encrypt or even store them in a database on the way.
 *
 * @file        adlio.c
 * @author      Crutcher Dunnavant <crutcher@redhat.com>
 * @date        05/06/2001
 * @version     1.0
 */

#define _BSD_SOURCE /* needed for fchown */
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>

#include <zlib.h>
#include <glib.h>

#include "alchemist.h"
#include "admerror.h"

/**
 * Internal buffer size for lowlevel IO operations
 */
#define BUFSIZE 1024

/**
 * Initial size of XML g_string for context read operation
 */
#define OUTPUT_STRING_INITIAL_SIZE 4096

/**
 * Reads a context from a given file.
 * @brief         Read context from file
 * @param context The pointer to the AdmContext pointer in which to return the
 *                loaded context.
 * @param path    The string of the path of the stored context.
 * @param error   Optional pointer to an AdmError pointer.
 * @return        @c 0 on success, @c -1 on failure
 * @see           AdmContext_writeToFile
 */
int
AdmContext_readFromFile (AdmContext **context,
			 const char  *path,
			 AdmError **error)
{
	gzFile gfile;
	int retval, size;
	GString *gxmlstr;
	char buf[BUFSIZE + 1];

	_adm_error_init_check(error);

	*context = NULL;
	if ((gfile = gzopen(path, "rb")) == NULL) {
		retval = -1;
		AdmError_create(error, ADM_EACCESSFAILED, _("Could not open file \"%s\""), path);
	} else {
		gxmlstr = g_string_sized_new(OUTPUT_STRING_INITIAL_SIZE);
loop:
		switch ((size = gzread(gfile, buf, BUFSIZE))) {
			default:
				buf[size] = '\0';
				g_string_append(gxmlstr, buf);
				goto loop;
				break;

			case 0:
				retval = ((*context = AdmContext_fromXML(gxmlstr->str, error)) != NULL) ? 0 : -1;
				break;

			case -1:
				AdmError_create(error, ADM_EACCESSFAILED, _("Read failed on file \"%s\""), path);
				retval = -1;
				break;
		}

		gzclose(gfile);
		g_string_free(gxmlstr, 1);
	}
	return retval;
}

/**
 * Writes a context to a given file.
 * @brief         Write context to file
 * @param context The AdmContext pointer which we want to store.
 * @param path    The string of the path where we want to store the context.
 * @param mode    The permissions of file to be saved.
 * @param uid     The User ID of the file, or @c -1.
 * @param gid     The Group ID of the file, or @c -1.
 * @param error   Optional pointer to an AdmError pointer.
 * @return        @c 0 on success, @c -1 on failure
 * @see           AdmContext_readFromFile
 */
int
AdmContext_writeToFile (AdmContext *context,
			const char *path,
			int	    mode,
			uid_t	    uid,
			gid_t       gid,
			AdmError **error)
{
	gzFile gfile;
	int retval, fd;
	char *xmlstr;

	_adm_error_init_check(error);

	if ((xmlstr = AdmContext_toXML(context, error)) == NULL) {
		/* the encode failed :( */
		retval = -1;
	} else {
		/* the encode succeded :) */

		/* Be paranoid, and zap the file */
		unlink(path);

		if ((fd = open(path, O_CREAT | O_WRONLY | O_EXCL, mode)) == -1) {
			/* the open failed ;( */
			AdmError_create(error, ADM_EACCESSFAILED, _("Could not open file \"%s\""), path);
			retval = -1;
		} else {
			/* the open succeded ;( */

			/* only try the chown if we care */
			if ((uid != -1) || (gid != -1))
				fchown(fd, uid, gid);

			if ((gfile = gzdopen(fd, "wb9")) == NULL) {
				/* The gzopen failed, odd. */
				AdmError_create(error, ADM_EACCESSFAILED, _("Could not initialize zlib for file \"%s\""), path);
				retval = -1;
			} else {
				if (gzwrite(gfile, xmlstr, strlen(xmlstr)) == 0) {
					AdmError_create(error, ADM_EACCESSFAILED, _("Write failed on file \"%s\""), path);
					retval = -1;
				} else {
					retval = 0;
				}
				gzclose(gfile);
			}
			close(fd);
		}

		/* zap the string */
		free(xmlstr);
	}
	return retval;
}
