#include <string.h>
#include <stdio.h> /* for sscanf */

#include <glib.h>

#include "alchemist.h"

#include "translation.h"
#include "alchemist-private.h"
#include "admerror.h"
#include "admcore.h"

/**
* A helper function for lifecyle checks with error setting
* @param data an AdmData node to type check
* @param error a pointer to an AdmError pointer
* @return @c 0 on failure, @c 1 on success
*/
int
_adm_lifecycle_check		(AdmData	*data,
				 AdmError	**error)
{
	if (!ADM_NODE_ISALIVE(data)) {
		AdmError_create(error, ADM_EDEADNODE, _("Attempting to access a dead node"));
		return 0;
	}
	return 1;
}

/**
* A helper function for type and lifecyle checks with error setting
* @param data an AdmData node to type check
* @param type the type to compare it against
* @param error a pointer to an AdmError pointer
* @return @c 0 on failure, @c 1 on success
*/
int
_adm_type_and_lifecycle_check		(AdmData	*data,
					 AdmDataType	type,
					 AdmError	**error)
{
	const char *str;

	if (!_adm_lifecycle_check(data, error)) {
		return 0;

	} else if (data->type != type) {
		switch (type) {
			case ADM_TYPE_BOOL:
				str = "AdmBool";
				break;
			case ADM_TYPE_INT:
				str = "AdmInt";
				break;
			case ADM_TYPE_FLOAT:
				str = "AdmFloat";
				break;
			case ADM_TYPE_STRING:
				str = "AdmString";
				break;
			case ADM_TYPE_BASE64:
				str = "AdmBase64";
				break;
			case ADM_TYPE_COPY:
				str = "AdmCopy";
				break;
			case ADM_TYPE_LIST:
				str = "AdmList";
				break;
			default:
				str = "AdmDataUnknown";
				break;
		}
		AdmError_create(error, ADM_EBADTYPE, _("The node is not an %s node"), str);
		return 0;
	}
	return 1;
}

/**
 * Takes a string, name, and checks to see if it is an
 * ADM conformant name attribute, only checks len characters.
 *
 * ADM name strings must match the regex : '[a-zA-Z_][-a-zA-Z0-9_.]*'
 # They must also be <= 255 chars in length.
 *
 * @brief       Internal Adm name attribute validation
 * @param  name The name string to be verified.
 * @param  len  The Length of the string.
 * @return      @c TRUE if name is a valid Adm name, @c FALSE otherwise.
 * @note        We don't use ctype.h functions here because we don't want name
 *              definitions to vary with locale
 */
gboolean _adm_valid_name_by_length(const gchar *name, guint len) {
	guint i;
	gchar c;
	/* we don't try to check length here, it is assumed
	 * to have happened outside, and we don't want to take
	 * the speed hit */
#ifdef __ADM_DEBUG__
	g_assert(name != NULL);
	g_assert(len <= strlen(name));
#endif
	
	/* blank names are invalid (reserved for the root data node) */
	if ((len == 0) || (len > 255))
		return FALSE;

	/* we use xml names, but do not allow use of the colon. */
	c = name[0];
	if (!(ADM_ISALPHA(c) || (c == '_')))
		return FALSE;

	/* check the rest of the characters */
	for (i = 1, c = name[1]; i < len; c = name[++ i])
		if (!ADM_ISNAMECHAR(c))
			return FALSE;

	return TRUE;
}

int AdmData_validName (const char *name)
{
	if (name == NULL)
		return 0;
	return _adm_valid_name(name);
}


/**
 * Takes a string, path, and checks to see if it is an
 * ADM conformant path attribute.
 * 
 * @brief      Internal Adm path validation
 * @param path The path string to be validated.
 * @return     @c TRUE if path is a valid Adm data path, @c FALSE otherwise.
 * @note       This function is NOT to be confused with filename paths. Path in
 *             the Adm means the access path to the data inside the Adm data
 *             structure.
 */
gboolean _adm_valid_path(const gchar *path) {
	guint i,j,k;
#ifdef __ADM_DEBUG__	
	g_assert(path != NULL);
#endif
	/* a blank path is valid */
	if (path[0] == '\0')
		return TRUE;
	
	/* paths must start with a '/', and must have at least one name */
	if ((path[0] != '/') || (path[1] == '\0'))
		return FALSE;

	/* we walk down the path, examining the names */
	k = strlen(path);
	for (i = 1, j = 1; i < k; i++) {
		if ((path[i] == '/') || (i == k - 1)){
			/* this catches the end of names on slashes,
			 * and at the end of the string */

			/* check the name */
			if (!_adm_valid_name_by_length(path + j,i-j))
				return FALSE;

			/* place the new start marker one character
			 * ahead (to avoid this '/') */
			j = i + 1;
		}
	}
	
	return TRUE;
}

/**
 * Scan an AdmData tree and return a pointer to the AdmData element matching
 * the given path. Return NULL if no AdmData element is found.
 *
 * @brief       Internal Adm data path lookup
 * @param  list The Adm list where we want to look up the path.
 * @param  path The Adm data path we want to look for in the list.
 * @return      The pointer to the AdmData if it was found or @c NULL if no
 *              such data exists or is unaccessible (e.g. anonymous lists).
 */
AdmData *
_adm_path_lookup (AdmList *list,
		  const char *path)
{
	AdmRefStr *name_tok;
	AdmData *data;
	gchar c;
	guint i, k;
#ifdef __ADM_DEBUG__
	/* supposedly, this is well armored elsewhere, and will always succed.
	 * This function is based upon this assumption */
	g_assert(_adm_valid_path(path));
#endif

	/* handle blank paths */
	if (path[0] == '\0')
		return NULL;

	/* if we are searching an anonymous list, we automatically fail */
	if (list->meta.flags && ADM_DATA_FLAG_ANONYMOUS)
		return NULL;

	while (1) {
		/* we are searching for the length of the name.
		 * we start at index 1, because we know that the first character is '/' */
//		for (i = 0, c = path[1]; (c != '\0') && (c != '/'); c = path[++i + 1]);

		i = 0;
		while (c = path[i + 1], (c != '\0') && (c != '/'))
			++i;

		/* a length based recreation of AdmList_getDataByName_b. */
		data = NULL;
		name_tok = AdmRefStr_nget(path + 1, i, 0);
		if (name_tok == NULL) 
			break;
		
		/* iterate over the list, find the first data node with the right name */
		data = list->meta.child;
		for (k = list->num_children; k > 0; data = data->next, -- k)
			if (data->name == name_tok)
				goto apl_found;
		data = NULL;
apl_found:
		/* drop the reference */
		AdmRefStr_unref(name_tok);

		/* is this the last name in the path? If it is, whatever we just got
		 * is the correct return value, even if it is NULL */
		if ((data == NULL) || (path[i + 1] == '\0'))
			break;

		/* if it is a list, and it is not anonyoms, recurse, and shorten the path. */
		if ((data->type == ADM_TYPE_LIST) && !(data->flags & ADM_DATA_FLAG_ANONYMOUS)) {
			list = (AdmList *) data;
			path = path + i + 1;
			continue;
		} else {
			data = NULL;
			break;
		}
	}

	return data;
}

/**
 * Allocates new Adm data type of the given type.
 *
 * AdmData malloc and free data are simple wrappers at the moment, but can be
 * replaced in the future by free list based functions for speed.
 *
 * @brief       Internal Adm data type malloc
 * @param  type The Adm data type to be allocated.
 * @return      The Newly allocated Adm data structure of specified type or @c
 *              NULL on error.
 */
AdmData * _adm_malloc_data(AdmDataType type) {
	AdmData *data;

	switch (type) {
	case ADM_TYPE_LIST:
		data = g_malloc(sizeof (AdmList));
		break;
			
	case ADM_TYPE_INT:
		data = g_malloc(sizeof (AdmInt));
		break;
			
	case ADM_TYPE_FLOAT:
		data = g_malloc(sizeof (AdmFloat));
		break;
			
	case ADM_TYPE_BOOL:
		data = g_malloc(sizeof (AdmBool));
		break;
			
	case ADM_TYPE_STRING:
	case ADM_TYPE_BASE64:
	case ADM_TYPE_COPY:
		data = g_malloc(sizeof (AdmString));
		break;

	default:
		data = NULL; /* to shut up the compiler */
#ifdef __ADM_DEBUG__
		/* This is a debug only event, because we /should/ be doing
		 * type checking outside this function. */
		g_error("improper AdmDataType passed to _adm_malloc_data()\n");
#endif
	}

	return data;
}

/**
 * Deallocates the given Adm data structure.
 *
 * AdmData malloc and free data are simple wrappers at the moment, but can be
 * replaced in the future by free list based functions for speed
 *
 * @brief      Internal Adm data type free
 * @param data The Adm data structure to be freed.
 */
 
void _adm_free_data(AdmData *data) {
	g_free(data);
}


