/**
 * Copyright (C) 2000, 2001 Red Hat, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/**
 * @brief Adm Error handling
 *
 * In order to have a extendable Error handling, including states and error
 * texts, and to be thread safe we use a custom error structure.
 *
 * The basic principle is similar to the *NIX style of error handling (see
 * errno(3)) with the addition of construction and destruction of error
 * structures as well as accessors to the information contained within these
 * structures.
 *
 * @file        admerror.c
 * @author      Crutcher Dunnavant <crutcher@redhat.com>
 * @date        05/06/2001
 * @version     1.0
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>

#include <glib.h>

#include "alchemist.h"

#include "admerror.h"

struct AdmError {
	unsigned int refcount;
	int errno;
	AdmRefStr *strerr;
};


void
AdmError_create	(AdmError	**error,
		int 		  num,
		const char	 *format,
				  ...)
{
	va_list args;
	gchar *str;
	AdmError *tmp;

	va_start(args, format);

	/* This is guaranteed */
	if (error == NULL) {
		va_end(args);
		return;
	}

	tmp = (AdmError *) malloc(sizeof (struct AdmError));
	tmp->errno = num;

/* Buffer alloc --> */
	str = g_strdup_vprintf(format, args);
	tmp->strerr = AdmRefStr_get(str, 1);
	g_free(str);
/* Buffer free <-- */

	*error = tmp;

	/* Initial refcount is 1 */
	(*error)->refcount = 1;

	va_end(args);
}

void
AdmError_ref (AdmError *error)
{
	++ error->refcount;
}

void
AdmError_unref (AdmError *error)
{
	if ((-- error->refcount) == 0) {
		AdmRefStr_unref(error->strerr);
		free(error);
	}
}

const char *
AdmError_strerr	(AdmError *error)
{
	return AdmRefStr_str(error->strerr);
}

int
AdmError_errno (AdmError *error)
{
	return error->errno;
}
