/**
 * Copyright (C) 2000, 2001, 2003 Red Hat, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/**
 * @brief Alchemist Data Model (Adm) base functionality
 *
 * This file contains the heart of the Alchemist, the core functionality on
 * the various data types and internal helper functions.
 *
 * First a brief overview of the whole logic behind the Alchemist:
 *
 * The Alchemist is a backend component which can simply put merge and manage
 * data structures from different sources.
 *
 * In our case we use it to retrive, store, manage, merge configurations for
 * various applications, including but not limited to apache, bind, printing
 * and network setup.
 *
 * The idea is based on a very general nested representation of data and
 * precisely defined operations upon these data types.
 *
 * There are 6 basic Alchemist data types available:
 *  - @c ADM_TYPE_LIST
 *  - @c ADM_TYPE_INT
 *  - @c ADM_TYPE_FLOAT
 *  - @c ADM_TYPE_BOOL
 *  - @c ADM_TYPE_STRING
 *  - @c ADM_TYPE_BASE64
 *  - @c ADM_TYPE_COPY
 *
 * Each of these data types has 3 attributes:
 *  - @c name
 *  - @c type
 *  - @c value
 *
 * The @c value attribute may sometimes have no meaning or might be unused,
 * at least for the @c ADM_TYPE_LIST and the @c ADM_TYPE_COPY they are not
 * being used as there are either other operators to handle the value as in the
 * case of @c ADM_TYPE_LIST or where there is no real value associated with the
 * data type as with the @c ADM_TYPE_COPY data type.
 *
 * All of the data types should be self explanatory. Only two of the data types
 * are a little special, so let's explain the two special ones.
 *
 * First the @c ADM_TYPE_LIST. It is our only container data type. No other
 * data type can contain other types (except base64 or string, but then it is
 * application specific structured or binary data which may or may not be able
 * to be reasonably represented in the Adm data structure).
 *
 * The other special data type is the @c ADM_TYPE_COPY. It is usually only used
 * in combination with the merge operation. The purpose of this data type is
 * to copy some part of a data structure to another part. Useful examples are
 * e.g. to copy the IP address of the machine (which can be obtained in the
 * future from a special data source called proc) to an IP address field of
 * your application specific data tree.
 *
 * @file        alchemist.c
 * @author      Crutcher Dunnavant <crutcher@redhat.com>
 * @date        05/06/2001
 * @version     1.0
 */

#include <string.h>
#include <stdio.h> /* for sscanf */

#include <glib.h>

#include "alchemist.h"

#include "translation.h"
#include "alchemist-private.h"
#include "admerror.h"
#include "admcore.h"

/**
 * Private helper function. Creates a AdmData element with the given
 * context, parent, name, and source. NO input validation is done,
 * because of some of the special cases which call this function.
 *
 * (And the fact that it is supposedly behind the armor of the external funcs.)
 *
 * @brief          Internal Adm data structure constructor
 * @param  context The AdmContext into which the new Adm data structure will be
 *                 put.
 * @param  parent  The AdmList parent of the new Adm data strucure.
 * @param  type    The Adm data type of the new Adm data strucure.
 * @param  name    The name of the new Adm data strucure.
 * @param  source  The source tag of the new Adm data strucure.
 * @return         Newly created and integrated Adm data structure or @c NULL
 *                 on failure.
 */
AdmData * _adm_create_data(AdmContext *context, AdmData *parent,
				  AdmDataType type, AdmRefStr *name, AdmRefStr *source) {
	AdmData * data;

	switch (type) {
	case ADM_TYPE_LIST:
		data = _adm_malloc_data(type);
		((AdmList *) data)->num_children = 0;
		break;
			
	case ADM_TYPE_INT:
		data = _adm_malloc_data(type);
		((AdmInt *) data)->value = 0;
		break;
			
	case ADM_TYPE_FLOAT:
		data = _adm_malloc_data(type);
		((AdmFloat *) data)->value = 0.0;
		break;
			
	case ADM_TYPE_BOOL:
		data = _adm_malloc_data(type);
		((AdmBool *) data)->value = FALSE;
		break;

	case ADM_TYPE_STRING:
	case ADM_TYPE_BASE64:
	case ADM_TYPE_COPY:
		data = _adm_malloc_data(type);
		/* keep ourselves from having to deal with null
		 * value strings */
		((AdmString *) data)->value = AdmRefStr_get("", 1);
		break;

	default:
		data = NULL; /* to shut up the compiler */
#ifdef __ADM_DEBUG__
		g_error("improper AdmDataType passed to _adm_create_data()\n");
#endif
		return NULL;
	}

	data->name = AdmRefStr_ref(name);
	data->source = AdmRefStr_ref(source);
	
	data->context = context;
	data->parent = parent;
	data->child = NULL;
	data->next = data->prev = data; /* keeps it tidy */
	data->type = type;
	data->flags = 0; /* Objects should have no flags set by default */
	data->refcount = 1; /* we could /NOT/ refcount here, and force ourselves
			       to ref this element when we actually link the parent
			       to /it/, but presumably, the parent is linking off
			       this function call, and so it is more convenient to do
			       this here. */

	return data; /* end of normal flow */
}

/**
 * This private helper function takes a context pointer, a parent pointer, and
 * a pointer to an AdmData node, and depth copies the data tree held by the
 * node, linked against the given context, with the root linked against the
 * given parent.
 *
 * @brief Internal Adm data tree copy function
 * @param  context The AdmContext into which the copied Adm data structure will
 *                 be put.
 * @param  parent  The AdmList parent of the copied Adm data strucure.
 * @param  data    The Adm data structure to be copied to the given context and
 *                 parent.
 * @return         The AdmData pointer to the copy of the given data or NULL on
 *                 failure.
 */
AdmData * _adm_copy_data_tree(AdmContext *context, AdmData *parent, AdmData *data, gboolean strip) {
	guint i, j;
	AdmData *foo;
	AdmData *data_copy;

#ifdef __ADM_DEBUG__
	g_assert(context != NULL);
	g_assert(data != NULL);
#endif

	/* We do NOT check that parent != NULL, because there are some cases
	 * when we need to have it == NULL (the roots of data trees, so we can
	 * link it back on itself later */

	/* don't copy AdmCopy elements when strip is on. */
	if ((strip == TRUE) && (data->type == ADM_TYPE_COPY))
		return NULL;
	
	/* create a copy of the data node */
	data_copy = _adm_create_data(context, parent, data->type,
				    data->name, data->source);

	/* copy the flags over */
	data_copy->flags = data->flags;
	
	/* copy the type specific values */
	switch (data_copy->type) {
	case ADM_TYPE_LIST:
		/* iterate over the old list, copy every element */
		foo = data->child;
		j = ((AdmList *) data)->num_children;
		for (i = 0; i < j; foo = foo->next, ++ i)
			/* dont copy AdmCopy elements when strip is true */
			if ((strip == FALSE) || (foo->type != ADM_TYPE_COPY))
				AdmList_copyData_b( (AdmList *) data_copy, foo, NULL);
		break;

	case ADM_TYPE_STRING:
	case ADM_TYPE_BASE64:
	case ADM_TYPE_COPY:
		AdmRefStr_unref(((AdmString *) data_copy)->value);
		((AdmString *) data_copy)->value = AdmRefStr_ref(((AdmString *) data)->value);
		break;
			
	case ADM_TYPE_INT:
		((AdmInt *) data_copy)->value = ((AdmInt *) data)->value;
		break;

	case ADM_TYPE_FLOAT:
		((AdmFloat *) data_copy)->value = ((AdmFloat *) data)->value;
		break;

	case ADM_TYPE_BOOL:
		((AdmBool *) data_copy)->value = ((AdmBool *) data)->value;
		break;

	default:
#ifdef __ADM_DEBUG__
		g_error("Trying to copy an AdmData element of invalid type!\n");
#endif
		break;
	}
	
	return data_copy;
}


/* Id tree functions */

/**
 * Allocates a new AdmIdentity structure.
 *
 * AdmIdentity malloc and free id are simple wrappers at the moment, but can be
 * replaced in the future by free list based functions for speed.
 *
 * @brief  Internal AdmIdentity malloc
 * @return The pointer to a newly allocated AdmIdentity structure or @c NULL on
 *         failure.
 */
AdmIdentity * _adm_malloc_id (void) {
	return g_malloc(sizeof (AdmIdentity));
}

/**
 * Deallocates a given AdmIndentiy structure.
 *
 * AdmIdentity malloc and free id are simple wrappers at the moment, but can be
 * replaced in the future by free list based functions for speed.
 *
 * @brief    Internal AdmIdentity free
 * @param id The AdmIdentity structure to be deallocated
 */
void _adm_free_id (AdmIdentity *id) {
	g_free(id);
}


/**
 * Creates a new AdmIdentity structure in the given AdmContext.
 *
 * Private helper function. creates a new AdmIdentity function, linked against
 * the given context, with the given name and serial. No input validation.
 *
 * @brief          Interal AdmIdentity structure constructor
 * @param  context The AdmContext in which the new AdmIdentity will be put.
 * @param  name    The name of the new AdmIdentity.
 * @param  serial  The serial number of the new AdmIdentity.
 * @return         The pointer to the newly created AdmIdentity or @c NULL on
 *                 failure.
 */
AdmIdentity * _adm_create_identity(AdmContext *context, AdmRefStr *name, guint serial) {
	AdmIdentity *id;
#ifdef __ADM_DEBUG__
	g_assert(name != NULL);
#endif
	
	id = _adm_malloc_id();

	id->name = AdmRefStr_ref(name);
	id->context = context;
	id->parentA = NULL;
	id->parentB = NULL;
	id->refcount = 1;
	id->serial = serial;

	return id;
}

/**
 * Removes an AdmIdentity from the enclosed AdmContext.
 *
 * This private helper function walks a id tree, unlinking
 * each node from it's context, from its children, and unrefing
 * it.
 *
 * @brief          Internal AdmIdentity recursive unlink
 * @param identity The AdmIdentity to be removed with it's children from the
 *                 enclosed AdmContext.
 */
void _adm_unlink_id_tree(AdmIdentity *identity) {

#ifdef __ADM_DEBUG__
	g_assert(identity != NULL);
	g_assert(identity->context != NULL);
#endif

	identity->context = NULL; /* this marks it dead */
	
	/* recurse over the tree */
	if (identity->parentA != NULL) {
		_adm_unlink_id_tree(identity->parentA);
		identity->parentA = NULL;
	}
	if (identity->parentB != NULL) {
		_adm_unlink_id_tree(identity->parentB);
		identity->parentB = NULL;
	}

	/*unref on the identity */
	AdmIdentity_unref(identity, NULL);
}


/**
 * Copies the given AdmIdentity to the target context.
 *
 * This private helper function takes a context pointer and an identity pointer,
 * and returns a depth copy of the id tree held by the pointer, but linked
 * against the context given.
 *
 * @brief          Internal AdmIdentity tree copy
 * @param  context The AdmContext into which the copy will be put.
 * @param  id      The AdmIdentity to be copied.
 * @return         The pointer to the copy of the new AdmIdentity.
 */
AdmIdentity * _adm_copy_id_tree(AdmContext *context, AdmIdentity *id) {
	AdmIdentity *id_copy;

#ifdef __ADM_DEBUG__
	g_assert(context != NULL);
	g_assert(id != NULL);
#endif

	/* create an id copy of id */	
	id_copy = _adm_create_identity(context, id->name, id->serial);

	/* depth copy the parents */
	id_copy->parentA =
		(id->parentA == NULL) ? NULL : _adm_copy_id_tree(context, id->parentA);
	
	id_copy->parentB =
		(id->parentB == NULL) ? NULL : _adm_copy_id_tree(context, id->parentB);

	return id_copy;
}

/**
 * Collects the the AdmCopy elements of a AdmData structure and places them in
 * a linked list of AdmCopy elements.
 *
 * A very dangerous helper function and never to be used by applications. It
 * traverses the AdmData element recursively if the AdmData element is an
 * AdmList.
 *
 * @brief      Internal AdmCopy collector
 * @param data The AdmData structure to be examined and search for AdmCopy
 *             elements.
 * @param head The pointer which will contain to the first AdmCopy structure
 *             of the found AdmCopy elements.
 * @param tail The pointer which will contain the last AdmCopy structure of the
 *             found AdmCopy elements.
 */
void _adm_collect_copies (AdmData *data, AdmCopy **head, AdmCopy **tail) {
	AdmData *foo;
	guint i;

	if (data->type == ADM_TYPE_COPY) {
		if (*tail != NULL) {
			(*tail)->meta.child = data;
			*tail = (AdmCopy *) data;
		} else {
			*head = (AdmCopy *) data;
			*tail = (AdmCopy *) data;
		}
	} else if (data->type == ADM_TYPE_LIST) {
		foo = data->child;
		i = ((AdmList *) data)->num_children;
		for (; i > 0; foo = foo->next, -- i)
			_adm_collect_copies(foo, head, tail);
	}
}

/**
 * This function implements the List Merge Algorithm.
 * It is the soul of the Alchemist, touch with care.
 *
 * The List Merge Algorithm { LMA(X <- Y) }
 *
 * ----------------------------------------
 *
 * Note: The 'insertion point' is a notional location, and moves down
 * every time an insertion is made, such that successive insertions are
 * stored in insertion order.
 *
 * @code
 * Step 0: Assertions
 *         If X or Y do not exist, fail
 *         If X or Y are not lists, fail
 *         If X or Y are atomic, fail
 *         If X and Y do not share names, fail
 * 
 * Step 1: If X or Y is prepend:
 *                 set the insetion point at the begining of X's child set
 *         Else:
 *                 set the insetion point at the end of X's child set
 * 
 * Step 2: If X is anonymous:
 *                 in order, for every child I of Y:
 *                         recursively copy I to (the holding value) Z
 *                         insert Z at the insertion point
 *         Else:
 *                 in order, for every child I of Y:
 *                         If I's name matches the name of an element J in X:
 *                                 If I and J are lists and neither is atomic:
 *                                 	If I is anonymous and J is not:
 *                                 		Anonymity collide, J wins
 *                                 		do nothing
 *                                 	Else:
 *                                         LMA(J <- I)
 *                                         If X or Y is prepend:
 *                                                 move J to the insertion point
 *                                 Else If J is not protected:
 *                                         recursively copy I to Z
 *                                         delete J from X
 *                                         If X or Y is prepend:
 *                                                 insert Z at the insertion point
 *                                         else:
 *                                                 insert Z at J's old position
 *                         Else:
 *                                 recursively copy I to Z
 *                                 insert Z at the insertion point
 * 
 * End
 * @endcode
 *
 * This function also mangles every AdmCopy which it merges in from listY,
 * using their child pointers to create a linked list of AdmCopy elements.
 * This is because it is the only efficient way of collecting a set of all
 * copy elements which are from context B in a context merge.
 *
 * THIS IS DANGEROUS, the list must be repaired, and each child pointer set
 * back to NULL, by AdmContext_merge. THe list is singly linked, and the last
 * element is linked to NULL.
 *
 * @brief       Internal AdmList merge operator
 * @param listX The first AdmList to be merged
 * @param listY The second AdmList to be merged
 * @param head  Will contain the pointer to the first AdmCopy element after the
 *              two AdmList elements are merged.
 * @param tail  Will contain the pointer to the last AdmCopy element after the
 *              two AdmList elements are merged.
 */
void     _adm_list_merge         (AdmList  *listX,
                                         AdmList  *listY,
					 AdmCopy **head,
					 AdmCopy **tail) {
	/* a cavet.
	 * We assume that the lists passed to us are well formed, in that
	 * non-anonymous lists are actually unique in their names.
	 * Bad, BAD things happen if this is not true, but it is far to
	 * expensive to validate here, and, more to the point, should be
	 * impossible. If it turns up that they are being created, it is
	 * not this fucntions fault, and their is no way to recover by the
	 * time we get here. */

	AdmData *foo, *bar, *baz, *x_tmp, *y_tmp, *r_ring, *x_ring, *y_ring;
	gboolean prepend, x_kill;
	guint i, j, sizeX, sizeY, sizeR;

#ifdef __ADM_DEBUG__
	/* Step 0 */
	g_assert(listX != NULL);
	g_assert(listY != NULL);
	g_assert(ADM_NODE_ISALIVE(listX));
	g_assert(ADM_NODE_ISALIVE(listY));
	g_assert(listX->meta.type == ADM_TYPE_LIST);
	g_assert(listY->meta.type == ADM_TYPE_LIST);
	g_assert(!(listX->meta.flags & ADM_DATA_FLAG_ATOMIC));
	g_assert(!(listY->meta.flags & ADM_DATA_FLAG_ATOMIC));
	g_assert(strcmp(listX->meta.name, listY->meta.name) == 0);
#endif


	/* figure out some values we are gonna touch often */
	prepend = (listX->meta.flags | listY->meta.flags) & ADM_DATA_FLAG_PREPEND;
	sizeX = listX->num_children;
	sizeY = listY->num_children;

	/* annoying special cases */
	if ((sizeX == 0) || (sizeY == 0)) {
		for (i = sizeY, foo = listY->meta.child; i > 0; foo = foo->next, -- i) {
			bar = AdmList_copyData_b(listX, foo, NULL);
			/* catch all the AdmCopy elements that come from listY */
			_adm_collect_copies(bar, head, tail);
		}
		return;
	}

	/* the anonymous case */
	if (listX->meta.flags & ADM_DATA_FLAG_ANONYMOUS) {
		/* listX is anonymous, so we run the anonymous merge.
		 * copy every element in listY into listX. But we need to
		 * catch the first element, so that we can adjust the list
		 * in cases of prepend, and we need to catch AdmCopy elements */
		bar = AdmList_copyData_b(listX, listY->meta.child, NULL);
		i = sizeY - 1;
		foo = listY->meta.child->next;
		for (; i > 0; foo = foo->next, -- i) {
			baz = AdmList_copyData_b(listX, foo, NULL);
			_adm_collect_copies(baz, head, tail);
		}

		/* now, since this is a ring list, and since we have the first
		 * of the new inserts laying arround, in bar, we can adjust
		 * the list fast in cases of prepend merges. */
		if (prepend) 
			listX->meta.child = bar;

		return;
	}

	/* okay, we know several things at this point.
	 * 1) listX and listY have children
	 * 2) neither is atomic
	 * 3) listX isnt anonymous.
	 *
	 * We are going to work in terms of rings, and fix up the list
	 * when we are done with this merge, to avoid needing to keep
	 * up with an insertion point */

	r_ring = NULL;
	sizeR = 0;

	x_ring = listX->meta.child;
	y_ring = listY->meta.child;


	/* A note to future maintainers.
	 * I initially implemented this algorithm without gotos. It was
	 * imposiible to read that way. This is a simpler representation,
	 * and does not duplicate code. please think twice before reworking
	 * it the other way. */
	
	/* iterate over listY's children */
	for (i = sizeY, y_tmp = y_ring; i > 0; y_tmp = y_tmp->next, -- i) {

		x_kill = FALSE;
		x_tmp = x_ring;
		/* search for name collides in listX */
		for (j = sizeX; j > 0; x_tmp = x_tmp->next, -- j) 
			if (y_tmp->name == x_tmp->name)
				goto merge_loop_collide;

		/* we have no name collide, this element of listY is
		 * unique, and we need only copy it, and insert it
		 * at the end of the r_ring */
		foo = _adm_copy_data_tree(listX->meta.context, (AdmData *) listX, y_tmp, FALSE);
		_adm_collect_copies(foo, head, tail);
		goto merge_loop_jump2;

merge_loop_collide:
		if (	(y_tmp->type == ADM_TYPE_LIST) &&
			(x_tmp->type == ADM_TYPE_LIST) &&
		    	!((y_tmp->flags | x_tmp->flags) & ADM_DATA_FLAG_ATOMIC) ) {
			/* we have a name collide, and it is between two lists.
			 * neither of which is atomic. Now we check for anonymity collide */
			if (!((~x_tmp->flags & y_tmp->flags) & ADM_DATA_FLAG_ANONYMOUS)) {
				/* we call _adm_list_merge
				 * on these lists, and then jump to the prepend handler,
				 * because x_tmp is already in the right place in the
				 * non-prepend case */
				_adm_list_merge((AdmList *) x_tmp, (AdmList *) y_tmp, head, tail);
				foo = x_tmp;
				goto merge_loop_jump1;
			} else
				/* we have an anonymity collide. x_tmp wins */
				goto merge_loop_continue;

		} else if (!(x_tmp->flags & ADM_DATA_FLAG_PROTECTED)) {
			/* we have a name collide, it is not a list merge, and the
			 * x_tmp element is not protected. we copy the y_tmp element,
			 * and set x_kill so that x_tmp will get destroyed */
			foo = _adm_copy_data_tree(listX->meta.context, (AdmData *) listX, y_tmp, FALSE);
			_adm_collect_copies(foo, head, tail);
			x_kill = TRUE;

		} else {
			/* we have a name collide, it is not a list merge, and the
			 * x_tmp element is protected, so we don't do anything */
			goto merge_loop_continue;
		}

		if (!prepend) {
			/* we splice foo into x_tmp's position */
			if (sizeX == 1) {
				x_ring = foo;
			} else {
				/* list head cases */
				if (x_ring == x_tmp)
					x_ring = foo;
				
				foo->next = x_tmp->next;
				foo->prev = x_tmp->prev;
				x_tmp->prev->next = foo;
				x_tmp->next->prev = foo;
			}

		}

merge_loop_jump1:
		if (prepend) {
			/* we remove x_tmp from x_ring */
			-- sizeX;
			if (sizeX == 0)
				x_ring = NULL;
			else if (x_ring == x_tmp)
				x_ring = x_tmp->next;
					
			x_tmp->prev->next = x_tmp->next;
			x_tmp->next->prev = x_tmp->prev;
merge_loop_jump2:
			/* splice foo into r_ring */
			++ sizeR;
			if (sizeR == 1) {
				r_ring = foo;
			} else {
				foo->next = r_ring;
				foo->prev = r_ring->prev;
				r_ring->prev = foo;
				foo->prev->next = foo;
			}
		} 

		if (x_kill) {
			x_tmp->parent = NULL;
			AdmData_unlink(x_tmp, NULL);
		}
merge_loop_continue:;
	}

	/* we now have to paste the x_ring and r_ring together. */

	/* anoying special cases */
	if (sizeX == 0) {
		listX->meta.child = r_ring;
		listX->num_children = sizeR;
		return;
	} else if (sizeR == 0) {
		listX->meta.child = x_ring;
		listX->num_children = sizeX;
		return;
	}
       
	/* splice the rings */
	x_ring->prev->next = r_ring;
	r_ring->prev->next = x_ring;
	foo = x_ring->prev;
	x_ring->prev = r_ring->prev;
	r_ring->prev = foo;
		
	/* restore a proper size to the num_children count */
	listX->num_children = sizeX + sizeR;

	/* adjust the list head for prepend, and done :) */
	listX->meta.child = (prepend) ? r_ring : x_ring;
}

void
AdmContext_flatten (AdmContext *context, AdmError **error)
{
	AdmCopy *copy_head, *copy_tmp;
	GSList *gslist_p;
	
	_adm_error_init_check(error);
	_adm_abort_if_fail(context != NULL);

	/* strip the mask set */
	for (gslist_p = context->mask_set; gslist_p != NULL; gslist_p = gslist_p->next)
		AdmRefStr_unref((AdmRefStr *) gslist_p->data);

	g_slist_free ( context->mask_set );
	context->mask_set = NULL;
	context->mask_set_size = 0;
	
	/* strip the AdmCopy elements out */
	copy_head = copy_tmp = NULL;
	_adm_collect_copies(context->data_tree_root, &copy_head, &copy_tmp);
	while (copy_head != NULL) {
		copy_tmp = (AdmCopy *) copy_head->meta.child;

		copy_head->meta.child = NULL;
		AdmData_unlink((AdmData *) copy_head, NULL);
		
		copy_head = copy_tmp;
	}
}

void
AdmContext_strip (AdmContext *context, AdmError **error)
{
	AdmIdentity *id;

	_adm_error_init_check(error);
	_adm_abort_if_fail(context != NULL);

	id = context->id_tree_root;
	if (id->parentA) {
		_adm_unlink_id_tree(id->parentA);
		id->parentA = NULL;
	}
	if (id->parentB) {
		_adm_unlink_id_tree(id->parentB);
		id->parentB = NULL;
	}
}


AdmContext *
AdmContext_merge (const char   *name,
		  unsigned int  serial,
		  AdmContext   *contextA,
		  AdmContext   *contextB,
		  AdmError **error)
{

	AdmContext *contextC;
	AdmIdentity *id;
	AdmRefStr *r_name, *path;
	guint i;
	GSList *gslist_p;
	AdmData *data, *data_list_head, *bar, *next;
	AdmCopy *copy_list_head, *copy_list_tail, *copy_tmp;

	
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(name != NULL);
	_adm_abort_if_fail(contextA != NULL);
	_adm_abort_if_fail(contextB != NULL);

	/* name validity check */
	if (!_adm_valid_name(name)) {
		AdmError_create(error, ADM_ENOPATH, _("\"%s\" not a valid name"), name);
		return NULL;
	}

	/* this function handles the Context Merge Algorithm from the ADM
	 * it goes out of order, where appropriate for efficiency */

	/* first, get a copy of A */	
	contextC = AdmContext_copy(contextA, NULL);

	/* adjust the id tree, and clone off B's branch */
	r_name = AdmRefStr_get(name, 1);
	id = _adm_create_identity(contextC, r_name, serial);
	id->parentA = contextC->id_tree_root;
	contextC->id_tree_root = id;
	
	contextC->id_tree_root->parentB = _adm_copy_id_tree(contextC, contextB->id_tree_root);

	/* adjust the data root source */
	AdmRefStr_unref(contextC->data_tree_root->source);
	contextC->data_tree_root->source = AdmRefStr_ref(r_name);

	/* drop our extra reference to r_name */
	AdmRefStr_unref(r_name);
	
	/* at this point, we are at the end of 'Step 4' from CMA */
	
	/* do 'Step 5' of the CMA */
	gslist_p = contextB->mask_set;
	for (i = contextB->mask_set_size; i > 0; i--) {
#ifdef __ADM_DEBUG__
		/* the context is malformed if this is true! */
		g_assert(gslist_p != NULL);
#endif
		
		/* get the next path */
		path = (AdmRefStr *) gslist_p->data;
		
		/* find the node, if it exists */
		data = _adm_path_lookup((AdmList *) contextC->data_tree_root, AdmRefStr_str(path));

		if (data != NULL) {
			/* it does exist, try to unlink it */
			if (!(data->flags & ADM_DATA_FLAG_PROTECTED))
				AdmData_unlink(data, NULL);
		} else {
			/* it does not exist, so this mask gets added to the
			 * new context's mask set. */
			AdmContext_setMask(contextC, AdmRefStr_str(path), NULL);
		}
		
		/* done with this path */
		gslist_p = gslist_p->next;
	}
	/* done with 'Step 5' from CMA */

	/* Step 6 and 7 */
	copy_list_head = NULL;
	copy_list_tail = NULL;
	_adm_list_merge((AdmList *) contextC->data_tree_root,
			(AdmList *) contextB->data_tree_root,
			&copy_list_head, &copy_list_tail);

	/* step 8 
	 * we create a singly linked list of AdmData trees,
	 * linked to the next tree on their parent pointer,
	 * and linked to their causal copy element on their
	 * context pointers. */
	data_list_head = NULL;
	while (copy_list_head != NULL) {
/*printf("copy:'%s','%s'\n", copy_list_head->meta.name, copy_list_head->value);*/

		data = AdmContext_getDataByPath_b(contextC, AdmRefStr_str(copy_list_head->value), NULL);
		if (data != NULL) {
			bar = _adm_copy_data_tree(contextC, NULL, data, TRUE);

			AdmRefStr_unref(bar->name);
			bar->name = AdmRefStr_ref(copy_list_head->meta.name);
			
			bar->parent = (AdmData *) copy_list_head;
			bar->context = (AdmContext *) data_list_head;
			data_list_head = bar;
		}

		/* repair the copy elements, and iterate the list head */
		copy_tmp = (AdmCopy *) copy_list_head->meta.child;
		copy_list_head->meta.child = NULL;
		copy_list_head = copy_tmp;
	}

	/* step 9 */
	while (data_list_head != NULL) {
		next = (AdmData *) data_list_head->context;
		copy_tmp = (AdmCopy *) data_list_head->parent;

		/* splice the data copy into the list occupied by the AdmCopy */
		data_list_head->context = contextC;
		data_list_head->parent = copy_tmp->meta.parent;

		data_list_head->next = copy_tmp->meta.next;
		data_list_head->prev = copy_tmp->meta.prev;
		data_list_head->next->prev = data_list_head;
		data_list_head->prev->next = data_list_head;

		if (copy_tmp->meta.parent->child == (AdmData *) copy_tmp)
			copy_tmp->meta.parent->child = data_list_head;

		copy_tmp->meta.parent = NULL;
		AdmData_unlink((AdmData *) copy_tmp, NULL);
		
		data_list_head = next;
	}
	
	return contextC;
}


AdmContext *
AdmContext_new (const char  *name,
		   unsigned int serial,
		   AdmError **error)
{
	AdmContext *context;
	AdmRefStr *r_root, *r_name;

	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(name != NULL);

	/* name validity check */
	if (!_adm_valid_name(name)) {
		AdmError_create(error, ADM_ENOPATH, _("\"%s\" not a valid name"), name);
		return NULL;
	}

	context = (AdmContext *) g_malloc (sizeof (AdmContext));

	/* the caller gets a ref */
	context->refcount = 1;

	/* Init the mask set */
	context->mask_set = NULL;
	context->mask_set_size = 0;

	/* get a refed string copy of the name and blank root node strings,
	 * we will have an extra reference to these strigns that must be
	 * disposed of. */
	r_name = AdmRefStr_get(name, 1);
	r_root = AdmRefStr_get("", 1);

	/* Init the id tree root node */
	context->id_tree_root = _adm_create_identity(context,r_name,serial);
	
	/* Init the data tree root node */
	context->data_tree_root = _adm_create_data(context, NULL, ADM_TYPE_LIST, r_root, r_name);
	context->data_tree_root->parent = context->data_tree_root;

	/* drop the reference the we have to the strings */
	AdmRefStr_unref(r_name);
	AdmRefStr_unref(r_root);

	return context;
}

AdmContext *
AdmContext_copy (AdmContext  *context,
		 AdmError **error)
{
	AdmContext *context_copy;
	GSList *gslist_p;

	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(context != NULL);
	
	context_copy = g_malloc(sizeof (AdmContext));

	/* the caller gets a ref */
	context_copy->refcount = 1;
	
	/* deep copy the id tree */
	context_copy->id_tree_root =
		_adm_copy_id_tree(context_copy, context->id_tree_root);
	
	/* deep copy the data tree */
	context_copy->data_tree_root =
		_adm_copy_data_tree(context_copy, NULL, context->data_tree_root, FALSE);
	/* set the self containment of the data tree's root node */
	context_copy->data_tree_root->parent = context_copy->data_tree_root;
	
	/* copy the mask set */
	context_copy->mask_set_size = context->mask_set_size;
	context_copy->mask_set = NULL;
	for (gslist_p = context->mask_set; gslist_p != NULL; gslist_p = gslist_p->next) {
		context_copy->mask_set =
			g_slist_append(	context_copy->mask_set,
					(gpointer) AdmRefStr_ref((AdmRefStr *) gslist_p->data) );
	}
	
	return context_copy;
}


AdmIdentity *
AdmContext_getIdentityRoot (AdmContext  *context,
			    AdmError **error)
{
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(context != NULL);

	return AdmIdentity_ref(context->id_tree_root, error);
}

AdmData *
AdmContext_getDataRoot (AdmContext  *context,
			AdmError **error)
{
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(context != NULL);

	return AdmData_ref(context->data_tree_root, error);
}

AdmData *
AdmContext_getDataByPath (AdmContext *context,
                          const char *path,
			  AdmError **error)
{
	AdmData *data;
	
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(context != NULL);
	_adm_abort_if_fail(path != NULL);

	/* path validity check */
	if (!_adm_valid_path(path)) {
		AdmError_create(error, ADM_ENOPATH, _("\"%s\" not a valid path"), path);
		return NULL;
	}

	if ((data = _adm_path_lookup((AdmList *) context->data_tree_root, path)) == NULL) {
		AdmError_create(error, ADM_ENONAME, _("The no data at path \"%s\"."), path);
		return NULL;
	}
	return AdmData_ref(data, error);
}

/****************************************************************************
 * Ref and Life cycle management functions.
 * 
 * ref, unref, isalive, unlink
 ****************************************************************************/

AdmContext *
AdmContext_ref (AdmContext  *context,
		AdmError **error)
{
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(context != NULL);

	++ context->refcount;
	return context;
}

void
AdmContext_unref (AdmContext  *context,
		  AdmError **error)
{
	GSList *gslist_p;

	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(context != NULL);

	-- context->refcount;

	if (context->refcount == 0) {
		/* Now we do the deep unlink */	
		_adm_unlink_id_tree(context->id_tree_root);

		/* we have to break the self containment of the root node
		 * for this to work, because it protects it. */
		context->data_tree_root->parent = NULL;
		AdmData_unlink(context->data_tree_root, NULL);

		/* now, we just have to consume the mask set */
		for (gslist_p = context->mask_set; gslist_p != NULL; gslist_p = gslist_p->next) 
			AdmRefStr_unref((AdmRefStr *) gslist_p->data);

		g_slist_free ( context->mask_set );

		/* all our dependent structures are dead, time to die. */
		g_free(context);
	}
}

AdmContext *
AdmIdentity_getContext (AdmIdentity	*id,
			AdmError **error)
{
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(id != NULL);
	if (!_adm_lifecycle_check((AdmData *) id, error))
		return NULL;

	return AdmContext_ref(id->context, error);
}

AdmIdentity *
AdmIdentity_ref (AdmIdentity  *identity,
		 AdmError **error)
{
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(identity != NULL);

	++ identity->refcount;
	return identity;
}

void
AdmIdentity_unref (AdmIdentity  *identity,
		   AdmError **error)
{
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(identity != NULL);

	-- identity->refcount;
	if (identity->refcount == 0) {
		/* if this identity is still 'alive' at this point,
		 * someone is unreffing us when they don't have a
		 * coresponding ref. We die, because they have /bad/
		 * code problems, and we can't try to recover if they've
		 * screwed up their memory allocations */
		g_return_if_fail(!ADM_NODE_ISALIVE(identity));

		/* free the name string */
		AdmRefStr_unref(identity->name);
	
		/* all the strings are gone, time to die */
		_adm_free_id(identity);
	}
}

int
AdmIdentity_isAlive (AdmIdentity  *identity,
		     AdmError **error)
{
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(identity != NULL);
	
	if (identity->context != NULL)
		return TRUE;

	return FALSE;
}

AdmData *
AdmData_ref (AdmData  *data,
	     AdmError **error)
{
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(data != NULL);

	++ data->refcount;
	return data;
}

void
AdmData_unref (AdmData  *data,
	       AdmError **error)
{
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(data != NULL);

	-- data->refcount;
	if (data->refcount == 0) {
		/* if this data is still 'alive' at this point,
		 * someone is unreffing us when they don't have a
		 * coresponding ref. We can't return from here,
		 * because it will leak the memory. Core dump ahead! */
		if (ADM_NODE_ISALIVE(data))
			g_error("un-refing a living node!\n"
				  "A node has been un-refed by something that did not hold a ref.\n");

		/* type specific cleanup */
		switch (data->type) {
		case ADM_TYPE_STRING:
		case ADM_TYPE_BASE64:
		case ADM_TYPE_COPY:
			/* free the string value, in appropriate types */
			AdmRefStr_unref( ((AdmString *) data)->value );
		case ADM_TYPE_LIST:
		case ADM_TYPE_INT:
		case ADM_TYPE_BOOL:
		case ADM_TYPE_FLOAT:
			/* free the name and source */
			AdmRefStr_unref(data->name);
			AdmRefStr_unref(data->source);
			break;
		default:;
#ifdef __ADM_DEBUG__
			/* This isn't good if we get here */
			g_warning("We are unrefing an AdmData element with an invalid type!\n");
#endif
		}
	
		/* all our dependent structs are gone, time to die */	
		_adm_free_data(data);
	}
}

int
AdmData_isAlive (AdmData  *data,
		 AdmError **error)
{
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(data != NULL);

	if (data->context != NULL)
		return TRUE;
	
	return FALSE;
}

int
AdmData_unlink (AdmData  *data,
		AdmError **error)
{
	AdmData *foo;
	AdmData *bar;
	guint i;
	
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(data != NULL);
	if (!_adm_lifecycle_check(data, error))
		return -1;
	
	/* root node protection */
	if (data->parent == data) {
		AdmError_create(error, ADM_ECONSTRAINT, _("Attempting to unlink the root node."));
		return -1;
	}

	data->context = NULL; /* this kills it */

	if (data->type == ADM_TYPE_LIST) {
		/* iterate over the list, unlink each child */
		foo = data->child;
		for (i = ((AdmList *) data)->num_children; i > 0;  -- i) {
			bar = foo->next;
			AdmData_unlink(foo, NULL);
			foo = bar;
		}
		
		data->child = NULL;
		((AdmList *) data)->num_children = 0;
	}	

	/* remove this data element from it's parent, if the parent exists.
	 *
	 * You may find yourself wondering how we are unlinking a living node with
	 * no parent, but we need only make sure that such a node can never be
	 * created in the normal case, then this check lets us use AdmData_unlink
	 * from AdmContext_unref to zap the root node, after it's self containment
	 * is broken. It is also used in intermidate steps by the list merge algorithm. */
	if (data->parent != NULL) {
		-- ((AdmList *) data->parent)->num_children;
		if (((AdmList *) data->parent)->num_children == 0)
			data->parent->child = NULL;
		else {
			/* splice the list arround it */
			data->prev->next = data->next;
			data->next->prev = data->prev;

			/* we don't have to worry about beginning/end of list cases
			 * here, because we know that the list has at least 2 elements,
			 * and that this is the first one. */
			if (data->parent->child == data)
				data->parent->child = data->next;
		}
	}
		
	/* unref the data */
	AdmData_unref(data, error);	

	return 0;
}



/***************************************************************************
 * AdmContext Mask fucntions
 * set, get, get num, clear by name/index
 ***************************************************************************/

/** 
 * Takes a path, and counts it's name depth.
 *
 * @brief       Interal Adm data path depth calculation
 * @param  path The data path of which we want to determine the depth.
 * @return      The depth of the data path.
 */
inline guint _adm_path_depth(const gchar *path) {
	guint i,j, count;
	/* we assume that this is a valid path */
#ifdef __ADM_DEBUG__
	g_assert(_adm_valid_path(path));
#endif
	/* we count depth (number of '/'s) */
	count = 0;
	j = strlen(path);
	for (i = 0; i < j; i++)
		if (path[i] == '/')
			++ count;

	return count;
}

/**
 * Returns an integer less than, equal to, or greater than zero if @c a is
 * found, respectively, to be less than, to match, or be greater than @c b
 * where first the depth of the data path of the given mask set elements and
 * then the lexicographical order is being used as order.
 *
 * Private helper function for AdmContext mask set operations. Used as a
 * compare function for the glib_find() function.
 *
 * @brief    Internal glib_find() AdmContext mask set comparator function
 * @param  a The first mask set element to be compared.
 * @param  b The second mask set element to be compared.
 * @return   The order of the two element as an integer, meaning less then,
 *           equal to or greater than zero if @c a is found, respectively, to
 *           be less than, to match, or be greater than @c b.
 */
gint
_adm_set_mask_compare(gconstpointer a, gconstpointer b) {
	guint deptha, depthb;
	AdmRefStr *r_stra, *r_strb;

	r_stra = (AdmRefStr *) a;
	r_strb = (AdmRefStr *) b;

	deptha = _adm_path_depth(AdmRefStr_str(r_stra));
	depthb = _adm_path_depth(AdmRefStr_str(r_strb));

	/* we sort first by depth (highest path depth goes first */
	if (deptha > depthb)
		return -1;
	else if (deptha < depthb)
		return 1;

	/* we sort second by strcmp
	 * which is algorithmically messy, but only gives us 0 on
	 * equivalent paths, and is at least consistent */
	return strcmp(AdmRefStr_str(r_stra), AdmRefStr_str(r_strb));
}


int
AdmContext_setMask (AdmContext  *context,
		      const char  *path,
		      AdmError **error)
{
	AdmRefStr *r_path;
	GSList *gslist_p;

	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(context != NULL);
	_adm_abort_if_fail(path != NULL);

	/* path validity check */
	if (!_adm_valid_path(path)) {
		AdmError_create(error, ADM_ENOPATH, _("\"%s\" not a valid path"), path);
		return -1;
	}

	 /* do not store blank paths, the are valid, but meaningless */
	if (path[0] == '\0')
		return 0;

	r_path = AdmRefStr_get(path, 1);

	/* Is it already in the list? */
	gslist_p = g_slist_find_custom(context->mask_set, (gpointer) r_path, (GCompareFunc) _adm_set_mask_compare);
	if (gslist_p != NULL) {
		AdmRefStr_unref (r_path);
		return 0;
	}

	++ context->mask_set_size;
	context->mask_set = g_slist_insert_sorted(context->mask_set, (gpointer) r_path,
						   _adm_set_mask_compare);

	return 0;
}


AdmRefStr *
AdmContext_getMask (AdmContext  *context,
		      unsigned int index,
		      AdmError **error)
{
	AdmRefStr *refstr;

	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(context != NULL);

	/* index range check */
	if (index >= context->mask_set_size) {
		AdmError_create(error, ADM_ENOINDEX, _("Index %d is out of range"), index);
		return NULL;
	}

	refstr = (AdmRefStr *) g_slist_nth_data(context->mask_set, index);

	AdmRefStr_ref(refstr);
	return refstr;
}

int
AdmContext_clearMaskByIndex (AdmContext  *context,
			       unsigned int index,
			       AdmError **error)
{
	GSList *gslist_p;
	
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(context != NULL);

	/* index range check */
	if (index >= context->mask_set_size) {
		AdmError_create(error, ADM_ENOINDEX, _("Index %d is out of range"), index);
		return -1;
	}

	/* Get the index'th element from the list */
       	gslist_p = g_slist_nth(context->mask_set, index);

	/* free it's data */
	AdmRefStr_unref(gslist_p->data);

	/* remove it from the list */
	context->mask_set = g_slist_remove(context->mask_set, gslist_p->data);
	--context->mask_set_size;

	return 0;
}

int
AdmContext_clearMaskByPath (AdmContext  *context,
				const char  *path,
				AdmError **error)
{
	GSList *gslist_p;
	
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(context != NULL);
	_adm_abort_if_fail(path != NULL);
	
	/* path validity check */
	if (!_adm_valid_path(path)) {
		AdmError_create(error, ADM_ENOPATH, _("\"%s\" not a valid path"), path);
		return -1;
	}

	/* find the matching mask entry */
	gslist_p = g_slist_find_custom(context->mask_set, (gpointer) path, (GCompareFunc) strcmp);
	
	/* was the mask found? */
	if (gslist_p == NULL) {
		AdmError_create(error, ADM_ENOPATH, _("No mask set on path \"%s\""), path);
		return -1;
	}

	/* free it's data */
	AdmRefStr_unref(gslist_p->data);

	/* remove it from the list */
	context->mask_set = g_slist_remove(context->mask_set, gslist_p->data);
	--context->mask_set_size;

	return 0;
}

unsigned int
AdmContext_getNumMasks (AdmContext *context,
			  AdmError **error)
{
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(context != NULL);

	return context->mask_set_size;
}


/*************************************************************************** 
 * Identity functions
 *
 * get parent (A,B), get/set name, get/set serial
 * 
 ***************************************************************************/

AdmIdentity *
AdmIdentity_getParentA (AdmIdentity  *identity,
			AdmError **error)
{
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(identity != NULL);
	if (!_adm_lifecycle_check((AdmData *) identity, error))
		return NULL;

	if (identity->parentA == NULL)
		return NULL;

	return AdmIdentity_ref(identity->parentA, error);
}

AdmIdentity *
AdmIdentity_getParentB (AdmIdentity  *identity,
			AdmError **error)
{
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(identity != NULL);
	if (!_adm_lifecycle_check((AdmData *) identity, error))
		return NULL;

	if (identity->parentB == NULL)
		return NULL;

	return AdmIdentity_ref(identity->parentB, error);
}

AdmRefStr *
AdmIdentity_getName (AdmIdentity  *identity,
		     AdmError **error)
{
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(identity != NULL);
	if (!_adm_lifecycle_check((AdmData *) identity, error))
		return NULL;
	
	AdmRefStr_ref(identity->name);
	return identity->name;
}

int
AdmIdentity_setName (AdmIdentity *identity,
                     const char  *name,
		     AdmError **error)
{
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(identity != NULL);
	_adm_abort_if_fail(name != NULL);
	if (!_adm_lifecycle_check((AdmData *) identity, error))
		return -1;

	/* name validity check */
	if (!_adm_valid_name(name)) {
		AdmError_create(error, ADM_EBADDATA, _("\"%s\" is an invalid name."), name);
		return -1;
	}

	AdmRefStr_unref(identity->name);
	identity->name = AdmRefStr_get(name, 1);

	/* This bit o' evil is how we keep the data tree root's source
	 * in sync with the id tree root's name */
	if (identity == identity->context->id_tree_root) {
		AdmRefStr_unref(identity->context->data_tree_root->source);
		identity->context->data_tree_root->source = AdmRefStr_ref(identity->name);
	}
	
	return 0;
}

unsigned int
AdmIdentity_getSerial (AdmIdentity  *identity,
		       AdmError **error)
{
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(identity != NULL);
	if (!_adm_lifecycle_check((AdmData *) identity, error))
		return 0;

	return identity->serial;
}

int
AdmIdentity_setSerial (AdmIdentity  *identity,
		       unsigned int  serial,
		       AdmError **error)
{
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(identity != NULL);
	if (!_adm_lifecycle_check((AdmData *) identity, error))
		return -1;

	identity->serial = serial;

	return 0;
}

/***************************************************************************
 * AdmList child functions
 *
 * copy, add, get by index, get by name, get num children, get pos, move
 ***************************************************************************/

AdmData *
AdmList_copyData (AdmList  *list,
		  AdmData  *data,
		  AdmError **error)
{
#if 0
	AdmData *data_copy;
#endif
	
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(list != NULL);
	_adm_abort_if_fail(data != NULL);
	if (!_adm_type_and_lifecycle_check((AdmData *) list, ADM_TYPE_LIST, error))
		return NULL;
	if (!_adm_lifecycle_check(data, error))
		return NULL;

	/* root node protection
	 * Necessary, because the root node has a reserved name */
	if (data->parent == data) {
		AdmError_create(error, ADM_ECONSTRAINT, _("Attempting to copy the root node."));
		return NULL;
	}

#if 0
	/* This is the old way, before AdmList_renameCopyData() */

	/* if this is a non-anonymous list, and it is non-empty, scan it for
	 * nodes matching the name of the new node */
	if (	!(list->meta.flags & ADM_DATA_FLAG_ANONYMOUS) &&
		(list->num_children > 0) &&
		(AdmList_getDataByName_b(list, AdmRefStr_str(data->name), NULL) != NULL) ){

		/* we have an illegal name colide if we get here */
		AdmError_create(error, ADM_ECONSTRAINT,
			_("Name collide on name \"%s\", cannot copy node."), data->name);
		return NULL;
	}

	/* build the copy tree. */
	data_copy = _adm_copy_data_tree(list->meta.context, (AdmData *) list, data, FALSE);
#ifdef __ADM_DEBUG__
	g_assert(data_copy->next == data_copy);
	g_assert(data_copy->prev == data_copy);
#endif
	
	/* append it to the list */
	++ list->num_children;
	if (list->num_children == 1) {
		list->meta.child = data_copy;
		
		/* we dont do this here, because it is done by _adm_create_data,
		 * called by _adm_copy_data_tree
		* data->next = data->prev = data; */
	} else {
		data_copy->next = list->meta.child;
		data_copy->prev = list->meta.child->prev;
		list->meta.child->prev = data_copy;
		data_copy->prev->next = data_copy;
	}

	return AdmData_ref(data_copy, error);
#else
	return AdmList_renameCopyData(list, data, AdmRefStr_str(data->name), error);
#endif
}

AdmData *
AdmList_renameCopyData
	(AdmList  *list,
	 AdmData  *data,
	 const char *name,
	 AdmError **error)
{
	AdmData *data_copy;
	AdmRefStr *name_refstr;
	
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(list != NULL);
	_adm_abort_if_fail(data != NULL);
	if (!_adm_type_and_lifecycle_check((AdmData *) list, ADM_TYPE_LIST, error))
		return NULL;
	if (!_adm_lifecycle_check(data, error))
		return NULL;

	/* name validity check */
	if (!_adm_valid_name(name)) {
		AdmError_create(error, ADM_EBADNAME, _("The name \"%s\" is an invalid name."), name);
		return NULL;
	}

	/* Get the AdmRefStr for the new name (note the reference) */
	name_refstr = AdmRefStr_get(name, 1);

	/* if this is a non-anonymous list, and it is non-empty, scan it for
	 * nodes matching the name */
	if (	!(list->meta.flags & ADM_DATA_FLAG_ANONYMOUS) &&
		(list->num_children > 0) &&
		(AdmList_getDataByName_b(list, AdmRefStr_str(name_refstr), NULL) != NULL) ){

		/* we have an illegal name colide if we get here */
		AdmError_create(error, ADM_ECONSTRAINT,
			_("Name collide on name \"%s\", cannot copy node."), name);

		/* drop the name_refstr */
		AdmRefStr_unref(name_refstr);
		return NULL;
	}

	/* build the copy tree. */
	data_copy = _adm_copy_data_tree(list->meta.context, (AdmData *) list, data, FALSE);
#ifdef __ADM_DEBUG__
	g_assert(data_copy->next == data_copy);
	g_assert(data_copy->prev == data_copy);
#endif

	/* Adjust the name of the copy (we can not use AdmData_setName safely here,
	* there is no parent yet. We do not ref it here, we already have a reference. */
	AdmRefStr_unref(data_copy->name);
	data_copy->name = name_refstr;
	
	/* append it to the list */
	++ list->num_children;
	if (list->num_children == 1) {
		list->meta.child = data_copy;
		
		/* we dont do this here, because it is done by _adm_create_data,
		 * called by _adm_copy_data_tree
		* data->next = data->prev = data; */
	} else {
		data_copy->next = list->meta.child;
		data_copy->prev = list->meta.child->prev;
		list->meta.child->prev = data_copy;
		data_copy->prev->next = data_copy;
	}

	return AdmData_ref(data_copy, error);
}

AdmData *
AdmList_addData (AdmList     *list,
		  AdmDataType  type,
		  const char  *name,
		  AdmError **error)
{
	/* unfortuanately, this function creates invalid copy elements,
	 * as they end up possessing the reference "", so they MUST be
	 * set after creation. */
	AdmData *data;
	AdmRefStr *r_name;
	
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(list != NULL);
	_adm_abort_if_fail(name != NULL);
	if (!_adm_type_and_lifecycle_check((AdmData *) list, ADM_TYPE_LIST, error))
		return NULL;

	/* name validity check */
	if (!_adm_valid_name(name)) {
		AdmError_create(error, ADM_EBADDATA, _("\"%s\" is an invalid name."), name);
		return NULL;
	}
	
	switch (type) {
	case ADM_TYPE_LIST:
	case ADM_TYPE_INT:
	case ADM_TYPE_FLOAT:
	case ADM_TYPE_BOOL:
	case ADM_TYPE_STRING:
	case ADM_TYPE_BASE64:
	case ADM_TYPE_COPY:
		break;
	default:
		/* This isn't a valid child, and is a programming error */
		AdmError_create(error, ADM_EBADDATA, _("Invalid type."));
		return NULL;
	}
	
	/* if this is a non-anonymous list, AND it is non-empty, scan it for
	 * nodes matching the new name */
	if (	!(list->meta.flags & ADM_DATA_FLAG_ANONYMOUS) &&
		(list->num_children > 0) &&
		(AdmList_getDataByName_b(list, name, NULL) != NULL) ){
		/* we have a name conflict if we got here */
		AdmError_create(error, ADM_ECONSTRAINT,
			_("Name collide on name \"%s\", cannot create new node."), name);
		return NULL;
	}

	/* create the new node */
	r_name = AdmRefStr_get(name, 1);
	data = _adm_create_data(list->meta.context,
				(AdmData *) list,
				type,
				r_name,
				list->meta.context->id_tree_root->name);
	AdmRefStr_unref(r_name);

#ifdef __ADM_DEBUG__
	g_assert(data->next == data);
	g_assert(data->prev == data);
#endif
	
	/* append it to the list */
	++ list->num_children;
	if (list->num_children == 1) {
		list->meta.child = data;
		
		/* we dont do this here, because it is done by _adm_create_data
		* data->next = data->prev = data; */
	} else {
		data->next = list->meta.child;
		data->prev = list->meta.child->prev;
		list->meta.child->prev = data;
		data->prev->next = data;
	}
	
	return AdmData_ref(data, error);
}

AdmData *
AdmList_getDataByIndex (AdmList      *list,
			 unsigned int  index,
			 AdmError **error)
{
	guint i;
	AdmData *foo;
	
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(list != NULL);
	if (!_adm_type_and_lifecycle_check((AdmData *) list, ADM_TYPE_LIST, error))
		return NULL;

	/* index range check */
	if (index >= list->num_children) {
		AdmError_create(error, ADM_ENOINDEX,
			_("Index %d is larger than list \"%s\"'s %d children."), index, list->meta.name, list->num_children);
		return NULL;
	}

	/* iterate over the list, stop on the approriate index */
	foo = list->meta.child;
	if (index <= (list->num_children >> 1))
		/* in this case, index is <= num_children / 2, 
		 * so we spin up to the data element */
		for (i = index; i > 0; foo = foo->next, -- i);
	else
		/* in this case, index is > num_children / 2, 
		 * so we spin down to the data element */
		for (i = list->num_children - index; i > 0; foo = foo->prev, -- i);

	return AdmData_ref(foo, error);
}


AdmData *
AdmList_getDataByName (AdmList     *list,
			const char  *name,
			AdmError **error)
{
	guint i;
	AdmData *foo;
	AdmRefStr *name_tok;
	
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(list != NULL);
	_adm_abort_if_fail(name != NULL);
	if (!_adm_type_and_lifecycle_check((AdmData *) list, ADM_TYPE_LIST, error))
		return NULL;

	/* get the refstr, if it exists */
	name_tok = AdmRefStr_get(name, 0);

	/* the refstr doesn't exists, so no one can
	 * have that name */
	if (name_tok == NULL) {
		foo = NULL;
		goto done;
	}
	
	/* iterate over the list, find the first data node with the right name */
	for (i = list->num_children, foo = list->meta.child; i > 0; foo = foo->next, -- i)
		if (foo->name == name_tok)
			goto found;

	/* we've wrapped arround, or the list is empty 
	 * it means that a child of this name is not 
	 * to be found in this list. */
	foo = NULL;

found:
	/* drop our extra reference */
	AdmRefStr_unref(name_tok);

done:
	if (foo == NULL) {
		AdmError_create(error, ADM_ENONAME, _("The AdmList \"%s\" has no child named \"%s\"."), list->meta.name, name);
		return NULL;
	}

	return AdmData_ref(foo, error);
}

AdmData *
AdmList_getDataByPath (AdmList     *list,
			const char  *path,
			AdmError **error)
{
	AdmData *foo;
	
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(list != NULL);
	_adm_abort_if_fail(path != NULL);
	if (!_adm_type_and_lifecycle_check((AdmData *) list, ADM_TYPE_LIST, error))
		return NULL;

	foo = _adm_path_lookup(list, path);

	if (foo == NULL) {
		AdmError_create(error, ADM_ENONAME,
			_("The AdmList \"%s\" has no child at path \"%s\"."), AdmRefStr_str(list->meta.name), path);
		return NULL;
	}

	return AdmData_ref(foo, error);
}

int
AdmList_moveData (AdmList      *list,
		   AdmData      *data,
		   unsigned int  index,
		   AdmError **error)
{
	AdmData *foo;

	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(list != NULL);
	_adm_abort_if_fail(data != NULL);
	if (!_adm_type_and_lifecycle_check((AdmData *) list, ADM_TYPE_LIST, error))
		return -1;
	if (!_adm_lifecycle_check(data, error))
		return -1;

	/* FIXME Does this case ever happen? */
	/* g_return_if_fail(data->parent == (AdmData *) list); */

	/* root node protection */
	if (data->parent == data) {
		AdmError_create(error, ADM_ECONSTRAINT, _("Attempting to move the root node."));
		return -1;
	}

	/* parantage check */
	if (data->parent != (AdmData *) list) {
		AdmError_create(error, ADM_ECONSTRAINT,
			_("The node \"%s\" is not a child of the list \"%s\""), data->name, list->meta.name);
		return -1;
	}

	/* index range check */
	if (index >= list->num_children) {
		AdmError_create(error, ADM_ENOINDEX,
			_("Index %d is larger than list \"%s\"'s %d children."), index, list->meta.name, list->num_children);
		return -1;
	}

	/* okay, if the list has only one child, and we've already
	 * checked that this element is in this list, and that index
	 * is a valid range for this list, then this is a short no-op.
	 *
	 * Otherwise, we get all kinds of nasty circular list problems
	 * below, that we'd have to deal with in many special cases :) */
	if (list->num_children == 1)
		return 0;

	/* remove the data from the list. */
	data->prev->next = data->next;
	data->next->prev = data->prev;

	/* fix the list if this was the head */
	if (list->meta.child == data)
		list->meta.child = data->next;

 	/* As we have just removed a child we need to correct the number of
	 * children as getDataByIndex uses this to walk backward. */
	list->num_children--; 

	/* find the node that data will be inserted in front of. */
	foo = (index == list->num_children)
		? list->meta.child
		: AdmList_getDataByIndex_b(list, index, NULL);

	/* fix the list head if this is the new head */
	if (index == 0) 
		list->meta.child = data;

	/* insert the data node in its new place */
	data->next = foo;
	data->prev = foo->prev;
	foo->prev = data;
	data->prev->next = data;

	/* Restore the original length of the list as we have now linked the 
	 * child back in. */
	list->num_children++;

	return 0;
}

#if 0
/**
 * AdmList_reparentData: reparent an AdmData element to a different AdmList.
 * 
 * This function extracts an AdmData element from its containing list, and
 * then appends it to the list given.
 * 
 * @brief        AdmList child move operator
 * @param  list  The AdmList which we want to reparent to.
 * @param  data  The AdmData pointer of the child we want to move.
 * @param  name  The name to reparent the data with.
 * @param  index The index to where we want to reparent the data.
 * @param  error Optional pointer to AdmError pointer.
 * @return -1 on error, 0 on success
 */
int         AdmList_reparentData             (AdmList      *list,
					     AdmData      *data,
                                             const char *name,
					     unsigned int  index,
					     AdmError **error);


int
AdmList_reparentData (AdmList      *list,
		   AdmData      *data,
                   const char *name,
			unsigned int index,
		   AdmError **error)
{
	AdmData *foo;

	void
	recontext_data(AdmData *data, AdmContext *context)
	{
		AdmData *foo = data->child;

		data->context = context;

		if ((foo != NULL) && (data->parent->type == ADM_TYPE_LIST)) {
			do {
				recontext_data(foo, context);
				foo = foo->next;
			} while (foo != data->child);
		}
	}

	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(list != NULL);
	_adm_abort_if_fail(data != NULL);
	if (!_adm_type_and_lifecycle_check((AdmData *) list, ADM_TYPE_LIST, error))
		return -1;
	if (!_adm_lifecycle_check(data, error))
		return -1;

	/* root node protection */
	if (data->parent == data) {
		AdmError_create(error, ADM_ECONSTRAINT, _("Attempting to reparent the root node."));
		return -1;
	}

	/* index range check */
	if (index >= list->num_children + 1) {
		AdmError_create(
			error,
			ADM_ENOINDEX,
			_("Index %d is out of range, list \"%s\" has %d children."),
			index,
			list->meta.name,
			list->num_children
		);
		return -1;
	}

	/* name validity check */
	if (!_adm_valid_name(name)) {
		AdmError_create(error, ADM_EBADDATA, _("\"%s\" is an invalid name."), name);
		return -1;
	}


	/* collision check 
	 * if it is a non-anonymous list, scan it for
	 * nodes matching the new name that are not this node */
	if (!(list->meta.flags & ADM_DATA_FLAG_ANONYMOUS)){

		foo = AdmList_getDataByName_b(list, name, NULL);
		if ((foo != NULL) && (foo != data)) {
			/* we have a name conflict if we got here */
			AdmError_create(error, ADM_ECONSTRAINT,
				_("Name collide on name \"%s\", cannot change name."), name);
			return -1;
		}
	}


	/* remove the data from its old list */
	data->next->prev = data->prev;
	data->prev->next = data->next;
	-- ((AdmList *) data->parent)->num_children;
	if (data == data->parent->child) {
		if (((AdmList *) data->parent)->num_children == 0) {
			data->parent->child = NULL;
		} else {
			data->parent->child = data->next;
		}
	}

	/* change the data's name */
	AdmRefStr_unref(data->name);
	data->name = AdmRefStr_get(name, 1);
	
	/* append it to the end of this list */
	if (list->num_children == 0) {
		/* the trival case */
		data->next = data;
		data->prev = data;

	} else {
		/* find the node that data will be inserted in front of. */
		foo = (index == list->num_children)
			? list->meta.child
			: AdmList_getDataByIndex_b(list, index, NULL);
	
		/* insert the data node in its new place */
		data->next = foo;
		data->prev = foo->prev;
		foo->prev = data;
		data->prev->next = data;
	}
	++ list->num_children;

	/* fix the list head if this is the new head */
	if (index == 0) 
		list->meta.child = data;

	/* fix the context of this sub-tree */
	if (data->context != list->meta.context)
		recontext_data(data, list->meta.context);

	return 0;
}
#endif

unsigned int
AdmList_getNumData (AdmList  *list,
			AdmError **error)
{
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(list != NULL);
	if (!_adm_type_and_lifecycle_check((AdmData *) list, ADM_TYPE_LIST, error))
		return 0;

	return list->num_children;
}

unsigned int
AdmData_getPos (AdmData  *data,
		AdmError **error)
{
	guint pos;
	AdmData *tmp;
	
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(data != NULL);
	if (!_adm_lifecycle_check(data, error))
		return 0;

	/* root node protection */
	/* FIXME This line does NOT do root node protection, and may be used in the merge
	 * algorithms. Find out, and comment this properly, or remove it */
	g_return_val_if_fail(data->parent->type == ADM_TYPE_LIST, 0);

	/* this function is always safe, because the root node is
	 * special cased, and the child list of the parent of any
	 * other node is guaranteed to be non-empty. (because the
	 * node in question is in it) */
	
	/* the list is index counted, so we start at 0 */
	pos = 0;
	/* special case for the root node (it is always 0) */
	if (data->parent != data)
		/* itereate over the list, stop when we reach data */
		for (tmp = data->parent->child; tmp != data; tmp = tmp->next, ++ pos)
			;
	
	return pos;
}

/****************************************************************************
 * AdmData meta functions.
 *
 * name, type, source, flags
 *
 ****************************************************************************/


AdmRefStr *
AdmData_getName (AdmData  *data,
		 AdmError **error)
{
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(data != NULL);
	if (!_adm_lifecycle_check(data, error))
		return NULL;

	return AdmRefStr_ref(data->name);
}

int
AdmData_setName (AdmData     *data,
		 const char  *name,
		 AdmError **error)
{
	AdmData *foo;
	
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(data != NULL);
	_adm_abort_if_fail(name != NULL);
	if (!_adm_lifecycle_check(data, error))
		return -1;

	/* root node protection */
	if (data->parent == data) {
		AdmError_create(error, ADM_ECONSTRAINT, _("Attempting to set the name of the root node"));
		return -1;
	}

	/* name validity check */
	if (!_adm_valid_name(name)) {
		AdmError_create(error, ADM_EBADNAME, _("The name \"%s\" is an invalid name."), name);
		return -1;
	}

	/* examine the parent of this node,
	 * if it is a non-anonymous list, scan it for
	 * nodes matching the new name that are not this node */
	if (	(data->parent->type == ADM_TYPE_LIST) &&
		!(data->parent->flags & ADM_DATA_FLAG_ANONYMOUS)	){

		foo = AdmList_getDataByName_b((AdmList *) data->parent, name, NULL);
		if ((foo != NULL) && (foo != data)) {
			/* we have a name conflict if we got here */
			AdmError_create(error, ADM_ECONSTRAINT,
				_("Name collide on name \"%s\", cannot change name."), name);
			return -1;
		}
	}
	
	AdmRefStr_unref(data->name);
	data->name = AdmRefStr_get(name, 1);

	return 0;
}

AdmRefStr *
AdmData_getSource (AdmData  *data,
		   AdmError **error)
{
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(data != NULL);
	if (!_adm_lifecycle_check(data, error))
		return NULL;

	return AdmRefStr_ref(data->source);
}

int
AdmData_setSource (AdmData     *data,
		   const char  *source,
		   AdmError **error)
{
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(data != NULL);
	_adm_abort_if_fail(source != NULL);
	if (!_adm_lifecycle_check(data, error))
		return -1;

	/* root node protection */
	if (data->parent == data) {
		AdmError_create(error, ADM_ECONSTRAINT, _("Attempting to set the source of the root node"));
		return -1;
	}

	/* name validity check */
	if (!_adm_valid_name(source)) {
		AdmError_create(error, ADM_EBADNAME, _("The source \"%s\" is an invalid name."), source);
		return -1;
	}
	
	AdmRefStr_unref(data->source);
	data->source = AdmRefStr_get(source, 1);

	return 0;
}

AdmData *
AdmData_getContainer (AdmData  *data,
		      AdmError **error)
{
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(data != NULL);
	if (!_adm_lifecycle_check(data, error))
		return NULL;

	/* the root node is already stored in it's own parent */
	return AdmData_ref(data->parent, error);
}

AdmData *
AdmData_getPrev	(AdmData	*data,
		int		wrap,
		AdmError	**error)
{
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(data != NULL);
	if (!_adm_lifecycle_check(data, error))
		return NULL;

	if (data->parent == data) {
		if (wrap) 
			return AdmData_ref(data, error);
	} else {
		if ((wrap) || (data != data->parent->child))
			return AdmData_ref(data->prev, error);
	}

	return NULL;
}

AdmData *
AdmData_getNext (AdmData	*data,
		int		wrap,
		AdmError	**error)
{
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(data != NULL);
	if (!_adm_lifecycle_check(data, error))
		return NULL;

	if (data->parent == data) {
		if (wrap) 
			return AdmData_ref(data, error);
	} else {
		if ((wrap) || (data->next != data->parent->child))
			return AdmData_ref(data->next, error);
	}

	return NULL;
}

AdmContext *
AdmData_getContext (AdmData  *data,
		    AdmError **error)
{
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(data != NULL);
	if (!_adm_lifecycle_check(data, error))
		return NULL;

	return AdmContext_ref(data->context, error);
}

AdmDataType
AdmData_getType (AdmData  *data,
		 AdmError **error)
{
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(data != NULL);
	if (!_adm_lifecycle_check(data, error))
		return 0;

	return data->type;
}

int
AdmData_isProtected (AdmData  *data,
		     AdmError **error)
{
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(data != NULL);
	if (!_adm_lifecycle_check(data, error))
		return 0;

	return (data->flags & ADM_DATA_FLAG_PROTECTED) ? TRUE : FALSE;
}

int
AdmData_setProtected (AdmData  *data,
		      int       state,
		      AdmError **error)
{
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(data != NULL);
	if (!_adm_lifecycle_check(data, error))
		return -1;

	/* root node protection */
	if (data->parent == data) {
		AdmError_create(error, ADM_ECONSTRAINT, _("Attempting to make a root node protected"));
		return -1;
	}

	/* set the flag */
	if (state)
		data->flags |= ADM_DATA_FLAG_PROTECTED;
	else
		data->flags &= ~ADM_DATA_FLAG_PROTECTED;

	return 0;
}

int
AdmList_isAnonymous (AdmList  *list,
		     AdmError **error)
{
	/* sanity checks */
	_adm_error_init_check(error);
	if (!_adm_type_and_lifecycle_check((AdmData *) list, ADM_TYPE_LIST, error))
		return 0;

	return (list->meta.flags & ADM_DATA_FLAG_ANONYMOUS) ? TRUE : FALSE;
}

int
AdmList_setAnonymous (AdmList  *list,
		      int       state,
		      AdmError **error)
{
	guint i,j;
	AdmData *foo, *bar;
	
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(list != NULL);
	if (!_adm_type_and_lifecycle_check((AdmData *) list, ADM_TYPE_LIST, error))
		return -1;

	/* root node protection */
	if (list->meta.parent == (AdmData *) list) {
		AdmError_create(error, ADM_ECONSTRAINT, _("Attempting to make a root node anonymous"));
		return -1;
	}

	/* Check to make sure that we don't have a name collide in a list
	 * which was anonymous */
	if ((list->meta.flags & ADM_DATA_FLAG_ANONYMOUS) &&
			!state && (list->num_children > 1)) {
		/* This is a decaying name comparison, yes, it sucks,
		 * and its slow, but it is the /fastest/ way to do this check */
		foo = list->meta.child;
		for (i = list->num_children - 1; i > 0; foo = foo->next, -- i) {
			for (j = i, bar = foo->next; j > 0; bar = bar->next, -- j) {
				if (foo->name == bar->name) {
					/* We've got a name collide, so we
					 * cant change the flag */
					AdmError_create(error, ADM_ECONSTRAINT,
						_("Name collide on name \"%s\", cannot drop anonymity"), foo->name);
					return -1;					
				}
			}
		}

		foo = list->meta.child;
		for (i = 0; i < list->num_children - 1; foo = foo->next, ++i) {

		}
	}
	
	/* set the flag */
	if (state)
		list->meta.flags |= ADM_DATA_FLAG_ANONYMOUS;
	else
		list->meta.flags &= ~ADM_DATA_FLAG_ANONYMOUS;

	return 0;
}

int
AdmList_isAtomic (AdmList  *list,
		  AdmError **error)
{
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(list != NULL);
	if (!_adm_type_and_lifecycle_check((AdmData *) list, ADM_TYPE_LIST, error))
		return 0;

	return (list->meta.flags & ADM_DATA_FLAG_ATOMIC) ? TRUE : FALSE;
}

int
AdmList_setAtomic (AdmList  *list,
		   int       state,
		   AdmError **error)
{
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(list != NULL);
	if (!_adm_type_and_lifecycle_check((AdmData *) list, ADM_TYPE_LIST, error))
		return -1;

	/* root node protection */
	if (list->meta.parent == (AdmData *) list) {
		AdmError_create(error, ADM_ECONSTRAINT, _("Attempting to make a root node prepend"));
		return -1;
	}

	if (state)
		list->meta.flags |= ADM_DATA_FLAG_ATOMIC;
	else
		list->meta.flags &= ~ADM_DATA_FLAG_ATOMIC;

	return 0;
}

int
AdmList_isPrepend (AdmList  *list,
		   AdmError **error)
{
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(list != NULL);
	if (!_adm_type_and_lifecycle_check((AdmData *) list, ADM_TYPE_LIST, error))
		return 0;

	return (list->meta.flags & ADM_DATA_FLAG_PREPEND) ? TRUE : FALSE;
}

int
AdmList_setPrepend (AdmList  *list,
		    int       state,
		    AdmError **error)
{
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(list != NULL);
	if (!_adm_type_and_lifecycle_check((AdmData *) list, ADM_TYPE_LIST, error))
		return -1;

	/* root node protection */
	if (list->meta.parent == (AdmData *) list) {
		AdmError_create(error, ADM_ECONSTRAINT, _("Attempting to make a root node prepend"));
		return -1;
	}

	if (state)
		list->meta.flags |= ADM_DATA_FLAG_PREPEND;
	else
		list->meta.flags &= ~ADM_DATA_FLAG_PREPEND;

	return 0;
}
