/**
 * Copyright (C) 2000 Red Hat, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/**
 * @brief alchemist module - Original Python bindings for the Alchemist
 *
 * This file contains all the original Python bindings of the first release of
 * the Alchemist. They C API was basically directly exported to Python and
 * there wrapped in a class hierarchy.
 *
 * In the next release we have added another set of Python bindings which
 * resemble more closely the way Python and it's data types work (for example
 * dictionary and sequence operators).
 *
 * This module is deprecated for future use. We will still continue to ship it
 * but won't extend it in any way.
 *
 * Please use the new bindings from the pyalchemist.c module instead.
 *
 * @file        alchemistmodule.c
 * @author      Tim Waugh <twaugh@redhat.com>
 * @date        05/06/2001
 * @version     1.0
 */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>

#include "Python.h"
#include "alchemist.h"

#ifdef NDEBUG
#define dprintf(x)
#else
#define dprintf(x) do { printf x; fflush (NULL); } while (0)
#endif


static PyObject *ADMCoreError;
static PyObject *ListError;
static PyObject *DeadData;
static PyObject *DeadIdentity;
static PyObject *ContextError;
static PyObject *ParseError;

#ifndef NDEBUG
static size_t string_len = 15;
static const char *str_repr (const char *s)
{
	static char *repr;
	int i;

	if (!repr) {
		const char *len = getenv ("_ALCHEMISTMODULE_STRING_LEN_");
		if (len) {
			unsigned long l;
			char *end;
			l = strtoul (len, &end, 0);
			if (end > len)
				string_len = l;
		}
	
		repr = malloc (string_len + 6);
		if (!repr)
			return "(no memory)";
	}

	repr[0] = '\"';
	for (i = 1; *s && i < string_len; i++, s++) {
		if (isprint (*s))
			repr[i] = *s;
		else
			repr[i] = '?';
	}
	if (*s) {
		const char *const cont = "...";
		strcpy (repr + i, cont);
		i += strlen (cont);
	}
	repr[i++] = '\"';
	repr[i] = '\0';

	return repr;
}
#endif

static void destroy_context (void *ctx)
{
	dprintf (("AdmContext_unref (%p) = ", ctx));
	AdmContext_unref (ctx, NULL);
	dprintf (("<void>\n"));
}

static void destroy_identity (void *id)
{
	dprintf (("AdmIdentity_unref (%p) = ", id));
	AdmIdentity_unref (id, NULL);
	dprintf (("<void>\n"));
}

static void destroy_data (void *dat)
{
	dprintf (("AdmData_unref (%p) = ", dat));
	AdmData_unref (dat, NULL);
	dprintf (("<void>\n"));
}

static PyObject *context_create (PyObject *self, PyObject *args)
{
	AdmContext *ctx;
	const char *name;
	long serial;

	if (!PyArg_ParseTuple (args, "sl", &name, &serial))
		return NULL;

	dprintf (("AdmContext_new (%s, %ld) = ", name, serial));
	ctx = AdmContext_new (name, serial, NULL);
	dprintf (("%p\n", ctx));

	if (!ctx)
		return PyErr_Format(ContextError, "Failed to create a new Context");

	return PyCObject_FromVoidPtr (ctx, destroy_context);
}

static PyObject *context_isIdentical (PyObject *self, PyObject *args)
{
	int identical;
	PyObject *ido;
	AdmIdentity *ctx_self, *ctx_other;

	ido = PyTuple_GetItem (args, 0);
	if (!ido)
		return NULL;

	if (!PyCObject_Check (ido))
		return NULL;

	ctx_self = PyCObject_AsVoidPtr (ido);

	ido = PyTuple_GetItem (args, 0);
	if (!ido)
		return NULL;

	if (!PyCObject_Check (ido))
		return NULL;

	ctx_other = PyCObject_AsVoidPtr (ido);

	dprintf (("AdmContext_isIdentical %p == %p : ", ctx_self, ctx_other));
	identical = (ctx_self == ctx_other);
	dprintf (("%s\n", identical ? "True" : "False"));

	return Py_BuildValue ("i", identical);
}

static PyObject *context_copy (PyObject *self, PyObject *args)
{
	PyObject *ctxo;
	AdmContext *ctx;
	AdmContext *ctxc;

	ctxo = PyTuple_GetItem (args, 0);
	if (!ctxo)
		return NULL;

	if (!PyCObject_Check (ctxo))
		return NULL;

	ctx = PyCObject_AsVoidPtr (ctxo);

	dprintf (("AdmContext_copy (%p) = ", ctx));
	ctxc = AdmContext_copy (ctx, NULL);
	dprintf (("%p\n", ctxc));

	if (!ctxc)
		return PyErr_Format(ADMCoreError, "Failed to copy Context");

	return PyCObject_FromVoidPtr (ctxc, destroy_context);
}

static PyObject *context_merge (PyObject *self, PyObject *args)
{
	const char *name;
	long serial;
	PyObject *ctxo;
	PyObject *arg;
	AdmContext *ctxA, *ctxB, *ctx;

	arg = PyTuple_GetItem (args, 0);
	if (!arg)
		return NULL;

	if (!PyString_Check (arg))
		return NULL;

	name = PyString_AsString (arg);
	arg = PyTuple_GetItem (args, 1);
	if (!arg)
		return NULL;

	if (!PyInt_Check (arg))
		return NULL;

	serial = PyInt_AsLong (arg);
	ctxo = PyTuple_GetItem (args, 2);
	if (!ctxo)
		return NULL;

	if (!PyCObject_Check (ctxo))
		return NULL;

	ctxA = PyCObject_AsVoidPtr (ctxo);
	ctxo = PyTuple_GetItem (args, 3);
	if (!ctxo)
		return NULL;

	ctxB = PyCObject_AsVoidPtr (ctxo);

	dprintf (("AdmContext_merge (%s, %ld, %p, %p) = ", name, serial,
		  ctxA, ctxB));
	ctx = AdmContext_merge (name, serial, ctxA, ctxB, NULL);
	dprintf (("%p\n", ctx));

	if (!ctx)
		return PyErr_Format(ContextError, "Context merge failed");

	return PyCObject_FromVoidPtr (ctx, destroy_context);
}

static PyObject *context_flatten (PyObject *self, PyObject *args)
{
	PyObject *ctxo;
	AdmContext *ctx;

	ctxo = PyTuple_GetItem (args, 0);
	if (!ctxo)
		return NULL;

	if (!PyCObject_Check (ctxo))
		return NULL;

	ctx = PyCObject_AsVoidPtr (ctxo);

	dprintf (("AdmContext_flatten (%p) = ", ctx));
	AdmContext_flatten (ctx, NULL);
	dprintf (("<void>\n"));

	Py_INCREF (Py_None);
	return Py_None;
}

static PyObject *context_toXML (PyObject *self, PyObject *args)
{
	PyObject *ctxo;
	AdmContext *ctx;
	const char *xml;

	ctxo = PyTuple_GetItem (args, 0);
	if (!ctxo)
		return NULL;

	if (!PyCObject_Check (ctxo)) {
		Py_INCREF (Py_None);
		return Py_None;
	}

	ctx = PyCObject_AsVoidPtr (ctxo);

	dprintf (("AdmContext_toXML (%p) = ", ctx));
	xml = AdmContext_toXML (ctx, NULL);
	dprintf (("%s\n", str_repr (xml)));

	if (!xml)
		return PyErr_Format(ADMCoreError, "Failed to encode Context to XML");

	return Py_BuildValue ("s", xml);
}

static PyObject *context_fromXML (PyObject *self, PyObject *args)
{
	const char *xml;
	AdmContext *ctx;

	if (!PyArg_ParseTuple (args, "s", &xml))
		return NULL;

	dprintf (("AdmContext_fromXML (%s) = ", str_repr (xml)));
	ctx = AdmContext_fromXML (xml, NULL);
	dprintf (("%p\n", ctx));

	if (!ctx)
		return PyErr_Format(ParseError, "Failed to decode XML to Context");

	return PyCObject_FromVoidPtr (ctx, destroy_context);
}

static PyObject *context_getMask (PyObject *self, PyObject *args)
{
	long index;
	PyObject *ctxo;
	PyObject *arg;
	PyObject *retobj;
	AdmContext *ctx;
	AdmRefStr *del;

	ctxo = PyTuple_GetItem (args, 0);
	if (!ctxo)
		return NULL;

	if (!PyCObject_Check (ctxo))
		return NULL;

	ctx = PyCObject_AsVoidPtr (ctxo);

	arg = PyTuple_GetItem (args, 1);
	if (!arg)
		return NULL;

	if (!PyInt_Check (arg))
		return NULL;

	index = PyInt_AsLong (arg);

	if (index >= AdmContext_getNumMasks(ctx, NULL))
		return PyErr_Format(PyExc_IndexError, "Indexing past end of Mask Set");
	
	dprintf (("AdmContext_getMask (%p, %ld) = ", ctx, index));
	del = AdmContext_getMask (ctx, index, NULL);
	dprintf (("%s\n", str_repr (AdmRefStr_str(del))));

	if (!del)
		return PyErr_Format(ADMCoreError, "NULL delete inside of Context's index range");

	
	retobj = Py_BuildValue ("s", AdmRefStr_str(del));
	AdmRefStr_unref(del);
	
	return retobj;
}

static PyObject *context_setMask (PyObject *self, PyObject *args)
{
	PyObject *ctxo;
	PyObject *arg;
	const char *path;
	AdmContext *ctx;

	ctxo = PyTuple_GetItem (args, 0);
	if (!ctxo)
		return NULL;

	if (!PyCObject_Check (ctxo))
		return NULL;

	ctx = PyCObject_AsVoidPtr (ctxo);

	arg = PyTuple_GetItem (args, 1);
	if (!arg)
		return NULL;

	if (!PyString_Check (arg))
		return NULL;

	path = PyString_AsString (arg);

	dprintf (("AdmContext_setMask (%p, %s) = ", ctx,
		  str_repr (path)));
	AdmContext_setMask (ctx, path, NULL);
	dprintf (("<void>\n"));

	Py_INCREF (Py_None);
	return Py_None;
}

static PyObject *context_clearMaskByIndex (PyObject *self, PyObject *args)
{
	long index;
	PyObject *ctxo;
	PyObject *arg;
	AdmContext *ctx;

	ctxo = PyTuple_GetItem (args, 0);
	if (!ctxo)
		return NULL;

	if (!PyCObject_Check (ctxo))
		return NULL;

	ctx = PyCObject_AsVoidPtr (ctxo);

	arg = PyTuple_GetItem (args, 1);
	if (!arg)
		return NULL;

	if (!PyInt_Check (arg))
		return NULL;

	index = PyInt_AsLong (arg);

	if (index >= AdmContext_getNumMasks(ctx, NULL))
		return PyErr_Format(PyExc_IndexError, "Indexing past end of Mask Set");
	
	dprintf (("AdmContext_clearMaskByIndex (%p, %ld) = ", ctx,
		  index));
	AdmContext_clearMaskByIndex (ctx, index, NULL);
	dprintf (("<void>\n"));

	Py_INCREF (Py_None);
	return Py_None;
}

static PyObject *context_clearMaskByPath (PyObject *self, PyObject *args)
{
	PyObject *ctxo;
	PyObject *arg;
	const char *path;
	AdmContext *ctx;

	ctxo = PyTuple_GetItem (args, 0);
	if (!ctxo)
		return NULL;

	if (!PyCObject_Check (ctxo))
		return NULL;

	ctx = PyCObject_AsVoidPtr (ctxo);

	arg = PyTuple_GetItem (args, 1);
	if (!arg)
		return NULL;

	if (!PyString_Check (arg))
		return NULL;

	path = PyString_AsString (arg);

	dprintf (("AdmContext_clearMaskByPath (%p, %s) = ", ctx,
		  str_repr (path)));
	AdmContext_clearMaskByPath (ctx, path, NULL);
	dprintf (("<void>\n"));

	Py_INCREF (Py_None);
	return Py_None;
}

static PyObject *context_getNumMasks (PyObject *self, PyObject *args)
{
	long ndel;
	PyObject *ctxo;
	AdmContext *ctx;

	ctxo = PyTuple_GetItem (args, 0);
	if (!ctxo)
		return NULL;

	if (!PyCObject_Check (ctxo))
		return NULL;

	ctx = PyCObject_AsVoidPtr (ctxo);

	dprintf (("AdmContext_getNumMasks (%p) = ", ctx));
	ndel = AdmContext_getNumMasks (ctx, NULL);
	dprintf (("%ld\n", ndel));

	return Py_BuildValue ("l", ndel);
}

static PyObject *context_getIdentityRoot (PyObject *self, PyObject *args)
{
	PyObject *ctxo;
	AdmContext *ctx;
	AdmIdentity *id;

	ctxo = PyTuple_GetItem (args, 0);
	if (!ctxo)
		return NULL;

	if (!PyCObject_Check (ctxo))
		return NULL;

	ctx = PyCObject_AsVoidPtr (ctxo);

	dprintf (("AdmContext_getIdentityRoot_b (%p) = ", ctx));
	id = AdmContext_getIdentityRoot_b (ctx, NULL);
	dprintf (("%p\n", id));

	if (!id)
		return PyErr_Format(ADMCoreError, "NULL id tree root in Context");

	dprintf (("AdmIdentity_ref (%p) = ", id));
	AdmIdentity_ref (id, NULL);
	dprintf (("<void>\n"));

	return PyCObject_FromVoidPtr (id, destroy_identity);
}

static PyObject *identity_getParentA (PyObject *self, PyObject *args)
{
	PyObject *ido;
	AdmIdentity *id;
	AdmIdentity *parentA;

	ido = PyTuple_GetItem (args, 0);
	if (!ido)
		return NULL;

	if (!PyCObject_Check (ido))
		return NULL;

	id = PyCObject_AsVoidPtr (ido);

	if (!AdmIdentity_isAlive(id, NULL))
		return PyErr_Format(DeadIdentity, "Accessing dead Identity");

	dprintf (("AdmIdentity_getParentA_b (%p) = ", id));
	parentA = AdmIdentity_getParentA_b (id, NULL);
	dprintf (("%p\n", parentA));

	if (!parentA) {
		// Is this the right thing to do?
		/* Yes, it is */
		Py_INCREF (Py_None);
		return Py_None;
	}

	dprintf (("AdmIdentity_ref (%p) = ", parentA));
	AdmIdentity_ref (parentA, NULL);
	dprintf (("<void>\n"));

	return PyCObject_FromVoidPtr (parentA, destroy_identity);
}

static PyObject *identity_getParentB (PyObject *self, PyObject *args)
{
	PyObject *ido;
	AdmIdentity *id;
	AdmIdentity *parentB;

	ido = PyTuple_GetItem (args, 0);
	if (!ido)
		return NULL;

	if (!PyCObject_Check (ido))
		return NULL;

	id = PyCObject_AsVoidPtr (ido);

	if (!AdmIdentity_isAlive(id, NULL))
		return PyErr_Format(DeadIdentity, "Accessing dead Identity");

	dprintf (("AdmIdentity_getParentB_b (%p) = ", id));
	parentB = AdmIdentity_getParentB_b (id, NULL);
	dprintf (("%p\n", parentB));

	if (!parentB) {
		// Is this the right thing to do?
		/* Yes, it is */
		Py_INCREF (Py_None);
		return Py_None;
	}

	dprintf (("AdmIdentity_ref (%p) = ", parentB));
	AdmIdentity_ref (parentB, NULL);
	dprintf (("<void>\n"));

	return PyCObject_FromVoidPtr (parentB, destroy_identity);
}

static PyObject *identity_getName (PyObject *self, PyObject *args)
{
	AdmRefStr *name;
	PyObject *ido, *retobj;
	AdmIdentity *id;

	ido = PyTuple_GetItem (args, 0);
	if (!ido)
		return NULL;

	if (!PyCObject_Check (ido))
		return NULL;

	id = PyCObject_AsVoidPtr (ido);

	if (!AdmIdentity_isAlive(id, NULL))
		return PyErr_Format(DeadIdentity, "Accessing dead Identity");

	dprintf (("AdmIdentity_getName (%p) = ", id));
	name = AdmIdentity_getName (id, NULL);

	if (!name)
		return PyErr_Format(ADMCoreError, "NULL name in living Identity");
	else
		dprintf (("%s\n", str_repr (AdmRefStr_str(name))));

	retobj = Py_BuildValue ("s", AdmRefStr_str(name));
	AdmRefStr_unref(name);

	return retobj;
}

static PyObject *identity_setName (PyObject *self, PyObject *args)
{
	PyObject *ido;
	AdmIdentity *id;
	PyObject *arg;
	const char *name;

	ido = PyTuple_GetItem (args, 0);
	if (!ido)
		return NULL;

	if (!PyCObject_Check (ido))
		return NULL;

	id = PyCObject_AsVoidPtr (ido);

	arg = PyTuple_GetItem (args, 1);
	if (!arg)
		return NULL;

	if (!PyString_Check (arg))
		return NULL;

	name = PyString_AsString (arg);

	if (!AdmIdentity_isAlive(id, NULL))
		return PyErr_Format(DeadIdentity, "Accessing dead Identity");

	dprintf (("AdmIdentity_setName (%p, %s) = ", id,
		  str_repr (name)));
	AdmIdentity_setName (id, name, NULL);
	dprintf (("<void>\n"));

	Py_INCREF (Py_None);
	return Py_None;
}

static PyObject *identity_getSerial (PyObject *self, PyObject *args)
{
	long serial;
	PyObject *ido;
	AdmIdentity *id;

	ido = PyTuple_GetItem (args, 0);
	if (!ido)
		return NULL;

	if (!PyCObject_Check (ido))
		return NULL;

	id = PyCObject_AsVoidPtr (ido);

	if (!AdmIdentity_isAlive(id, NULL))
		return PyErr_Format(DeadIdentity, "Accessing dead Identity");

	dprintf (("AdmIdentity_getSerial (%p) = ", id));
	serial = AdmIdentity_getSerial (id, NULL);
	dprintf (("%ld\n", serial));

	return Py_BuildValue ("l", serial);
}

static PyObject *identity_setSerial (PyObject *self, PyObject *args)
{
	PyObject *ido;
	AdmIdentity *id;
	PyObject *arg;
	long serial;

	ido = PyTuple_GetItem (args, 0);
	if (!ido)
		return NULL;

	if (!PyCObject_Check (ido))
		return NULL;

	id = PyCObject_AsVoidPtr (ido);

	arg = PyTuple_GetItem (args, 1);
	if (!arg)
		return NULL;

	if (!PyInt_Check (arg))
		return NULL;

	serial = PyInt_AsLong (arg);

	if (!AdmIdentity_isAlive(id, NULL))
		return PyErr_Format(DeadIdentity, "Accessing dead Identity");

	dprintf (("AdmIdentity_setSerial (%p, %ld) = ", id, serial));
	AdmIdentity_setSerial (id, serial, NULL);
	dprintf (("<void>\n"));

	Py_INCREF (Py_None);
	return Py_None;
}

static PyObject *identity_isAlive (PyObject *self, PyObject *args)
{
	int alive;
	PyObject *ido;
	AdmIdentity *id;

	ido = PyTuple_GetItem (args, 0);
	if (!ido)
		return NULL;

	if (!PyCObject_Check (ido))
		return NULL;

	id = PyCObject_AsVoidPtr (ido);

	dprintf (("AdmIdentity_isAlive (%p) is ", id));
	alive = AdmIdentity_isAlive (id, NULL);
	dprintf (("%s\n", alive ? "True" : "False"));

	return Py_BuildValue ("i", alive);
}

static PyObject *identity_isIdentical (PyObject *self, PyObject *args)
{
	int identical;
	PyObject *ido;
	AdmIdentity *id_self, *id_other;

	ido = PyTuple_GetItem (args, 0);
	if (!ido)
		return NULL;

	if (!PyCObject_Check (ido))
		return NULL;

	id_self = PyCObject_AsVoidPtr (ido);

	ido = PyTuple_GetItem (args, 0);
	if (!ido)
		return NULL;

	if (!PyCObject_Check (ido))
		return NULL;

	id_other = PyCObject_AsVoidPtr (ido);

	dprintf (("AdmIdentity_isIdentical %p == %p : ", id_self, id_other));
	identical = (id_self == id_other);
	dprintf (("%s\n", identical ? "True" : "False"));

	return Py_BuildValue ("i", identical);
}

static PyObject *context_getDataRoot (PyObject *self, PyObject *args)
{
	AdmData *dat;
	PyObject *ctxo;
	AdmContext *ctx;

	ctxo = PyTuple_GetItem (args, 0);
	if (!ctxo)
		return NULL;

	if (!PyCObject_Check (ctxo))
		return NULL;

	ctx = PyCObject_AsVoidPtr (ctxo);

	dprintf (("AdmContext_getDataRoot_b (%p) = ", ctx));
	dat = AdmContext_getDataRoot_b (ctx, NULL);
	dprintf (("%p\n", dat));

	if (!dat)
		return PyErr_Format(ADMCoreError, "NULL data tree root in Context");

	dprintf (("AdmData_ref (%p) = ", dat));
	AdmData_ref (dat, NULL);
	dprintf (("<void>\n"));

	return PyCObject_FromVoidPtr (dat, destroy_data);
}

static PyObject *data_getContainer (PyObject *self, PyObject *args)
{
	AdmData *ctr;
	PyObject *dato;
	AdmData *dat;

	dato = PyTuple_GetItem (args, 0);
	if (!dato)
		return NULL;

	if (!PyCObject_Check (dato))
		return NULL;

	dat = PyCObject_AsVoidPtr (dato);

	if (!AdmData_isAlive(dat, NULL))
		return PyErr_Format(DeadData, "Accessing dead Data element");

	dprintf (("AdmData_getContainer_b (%p) = ", dat));
	ctr = AdmData_getContainer_b (dat, NULL);
	dprintf (("%p\n", ctr));

	if (!ctr)
		return PyErr_Format(ADMCoreError, "NULL container on living Data");

	dprintf (("AdmData_ref (%p) = ", ctr));
	AdmData_ref (ctr, NULL);
	dprintf (("<void>\n"));

	return PyCObject_FromVoidPtr (ctr, destroy_data);
}

static PyObject *data_getContext (PyObject *self, PyObject *args)
{
	AdmContext *ctx;
	PyObject *dato;
	AdmData *dat;

	dato = PyTuple_GetItem (args, 0);
	if (!dato)
		return NULL;

	if (!PyCObject_Check (dato))
		return NULL;

	dat = PyCObject_AsVoidPtr (dato);

	if (!AdmData_isAlive(dat, NULL))
		return PyErr_Format(DeadData, "Accessing dead Data element");

	dprintf (("AdmData_getContext_b (%p) = ", dat));
	ctx = AdmData_getContext_b (dat, NULL);
	dprintf (("%p\n", ctx));

	if (!ctx)
		return PyErr_Format(ADMCoreError, "NULL context on living Data");

	dprintf (("AdmContext_ref (%p) = ", ctx));
	AdmContext_ref (ctx, NULL);
	dprintf (("<void>\n"));

	return PyCObject_FromVoidPtr (ctx, destroy_context);
}

static PyObject *context_getDataByPath (PyObject *self, PyObject *args)
{
	AdmData *dat;
	PyObject *ctxo;
	AdmContext *ctx;
	PyObject *arg;
	const char *path;

	ctxo = PyTuple_GetItem (args, 0);
	if (!ctxo)
		return NULL;

	if (!PyCObject_Check (ctxo))
		return NULL;

	ctx = PyCObject_AsVoidPtr (ctxo);

	arg = PyTuple_GetItem (args, 1);
	if (!arg)
		return NULL;

	if (!PyString_Check (arg))
		return NULL;

	path = PyString_AsString (arg);

	dprintf (("AdmContext_getDataByPath_b (%p, %s) = ", ctx,
		  str_repr (path)));
	dat = AdmContext_getDataByPath_b (ctx, path, NULL);
	dprintf (("%p\n", dat));

	if (!dat) 
		return PyErr_Format(PyExc_KeyError, "No data matching path");

	dprintf (("AdmData_ref (%p) = ", dat));
	AdmData_ref (dat, NULL);
	dprintf (("<void>\n"));

	return PyCObject_FromVoidPtr (dat, destroy_data);
}

static PyObject *data_getType (PyObject *self, PyObject *args)
{
	AdmDataType typ;
	PyObject *dato;
	AdmData *dat;

	dato = PyTuple_GetItem (args, 0);
	if (!dato)
		return NULL;

	if (!PyCObject_Check (dato))
		return NULL;

	dat = PyCObject_AsVoidPtr (dato);

	if (!AdmData_isAlive(dat, NULL))
		return PyErr_Format(DeadData, "Accessing dead Data");

	dprintf (("AdmData_getType (%p) = ", dat));
	typ = AdmData_getType (dat, NULL);
	dprintf (("%d\n", typ));

	return Py_BuildValue ("i", typ);
}

static PyObject *data_isAlive (PyObject *self, PyObject *args)
{
	int alive;
	PyObject *dato;
	AdmData *dat;

	dato = PyTuple_GetItem (args, 0);
	if (!dato)
		return NULL;

	if (!PyCObject_Check (dato))
		return NULL;

	dat = PyCObject_AsVoidPtr (dato);

	dprintf (("AdmData_isAlive (%p) is ", dat));
	alive = AdmData_isAlive (dat, NULL);
	dprintf (("%s\n", alive ? "True" : "False"));

	return Py_BuildValue ("i", alive);
}

static PyObject *data_isIdentical (PyObject *self, PyObject *args)
{
	int identical;
	PyObject *ido;
	AdmIdentity *data_self, *data_other;

	ido = PyTuple_GetItem (args, 0);
	if (!ido)
		return NULL;

	if (!PyCObject_Check (ido))
		return NULL;

	data_self = PyCObject_AsVoidPtr (ido);

	ido = PyTuple_GetItem (args, 0);
	if (!ido)
		return NULL;

	if (!PyCObject_Check (ido))
		return NULL;

	data_other = PyCObject_AsVoidPtr (ido);

	dprintf (("AdmData_isIdentical %p == %p : ", data_self, data_other));
	identical = (data_self == data_other);
	dprintf (("%s\n", identical ? "True" : "False"));

	return Py_BuildValue ("i", identical);
}

static PyObject *data_unlink (PyObject *self, PyObject *args)
{
	PyObject *dato;
	AdmData *dat;

	dato = PyTuple_GetItem (args, 0);
	if (!dato)
		return NULL;

	if (!PyCObject_Check (dato))
		return NULL;

	dat = PyCObject_AsVoidPtr (dato);

	if (!AdmData_isAlive(dat, NULL))
		return PyErr_Format(DeadData, "Accessing dead Data");

	dprintf (("AdmData_unlink (%p) = ", dat));
	AdmData_unlink (dat, NULL);
	dprintf (("<void>\n"));

	Py_INCREF (Py_None);
	return Py_None;
}

static PyObject *data_getName (PyObject *self, PyObject *args)
{
	PyObject *dato, *retobj;
	AdmData *dat;
	AdmRefStr *name;

	dato = PyTuple_GetItem (args, 0);
	if (!dato)
		return NULL;

	if (!PyCObject_Check (dato))
		return NULL;

	dat = PyCObject_AsVoidPtr (dato);

	if (!AdmData_isAlive(dat, NULL))
		return PyErr_Format(DeadData, "Accessing dead Data");

	dprintf (("AdmData_getName (%p) = ", dat));
	name = AdmData_getName (dat, NULL);

	if (!name)
		return PyErr_Format(ADMCoreError, "Living Data with NULL name");

	dprintf (("%s\n", str_repr (AdmRefStr_str(name))));
	retobj = Py_BuildValue ("s", AdmRefStr_str(name));
	AdmRefStr_unref(name);
	
	return retobj;
}

static PyObject *data_setName (PyObject *self, PyObject *args)
{
	PyObject *dato;
	AdmData *dat;
	PyObject *arg;
	const char *name;

	dato = PyTuple_GetItem (args, 0);
	if (!dato)
		return NULL;

	if (!PyCObject_Check (dato))
		return NULL;

	dat = PyCObject_AsVoidPtr (dato);

	arg = PyTuple_GetItem (args, 1);
	if (!arg)
		return NULL;

	if (!PyString_Check (arg))
		return NULL;

	name = PyString_AsString (arg);

	if (!AdmData_isAlive(dat, NULL))
		return PyErr_Format(DeadData, "Accessing dead Data");

	dprintf (("AdmData_setName (%p, %s) = ", dat, str_repr (name)));
	AdmData_setName (dat, name, NULL);
	dprintf (("<void>\n"));

	Py_INCREF (Py_None);
	return Py_None;
}

static PyObject *data_getSource (PyObject *self, PyObject *args)
{
	AdmRefStr *src;
	PyObject *dato, *retobj;
	AdmData *dat;

	dato = PyTuple_GetItem (args, 0);
	if (!dato)
		return NULL;

	if (!PyCObject_Check (dato))
		return NULL;

	dat = PyCObject_AsVoidPtr (dato);

	if (!AdmData_isAlive(dat, NULL))
		return PyErr_Format(DeadData, "Accessing dead Data");

	dprintf (("AdmData_getSource (%p) = ", dat));
	src = AdmData_getSource (dat, NULL);

	if (!src)
		return PyErr_Format(ADMCoreError, "Living Data with NULL Source");
	else
		dprintf (("%s\n", str_repr (AdmRefStr_str(src))));

	retobj = Py_BuildValue ("s", AdmRefStr_str(src));
	AdmRefStr_unref(src);
	return retobj;
}

static PyObject *data_setSource (PyObject *self, PyObject *args)
{
	PyObject *dato;
	AdmData *dat;
	PyObject *arg;
	const char *src;

	dato = PyTuple_GetItem (args, 0);
	if (!dato)
		return NULL;

	if (!PyCObject_Check (dato))
		return NULL;

	dat = PyCObject_AsVoidPtr (dato);

	arg = PyTuple_GetItem (args, 1);
	if (!arg)
		return NULL;

	if (!PyString_Check (arg))
		return NULL;

	src = PyString_AsString (arg);

	if (!AdmData_isAlive(dat, NULL))
		return PyErr_Format(DeadData, "Accessing dead Data");

	dprintf (("AdmData_setSource (%p, %s) = ", dat,
		  str_repr (src)));
	AdmData_setSource (dat, src, NULL);
	dprintf (("<void>\n"));

	Py_INCREF (Py_None);
	return Py_None;
}

static PyObject *data_isProtected (PyObject *self, PyObject *args)
{
	int prot;
	PyObject *dato;
	AdmData *dat;

	dato = PyTuple_GetItem (args, 0);
	if (!dato)
		return NULL;

	if (!PyCObject_Check (dato))
		return NULL;

	dat = PyCObject_AsVoidPtr (dato);

	if (!AdmData_isAlive(dat, NULL))
		return PyErr_Format(DeadData, "Accessing dead Data");

	dprintf (("AdmData_isProtected (%p) is ", dat));
	prot = AdmData_isProtected (dat, NULL);
	dprintf (("%s\n", prot ? "True" : "False"));

	return Py_BuildValue ("i", prot);
}

static PyObject *data_setProtected (PyObject *self, PyObject *args)
{
	PyObject *dato;
	AdmData *dat;
	PyObject *arg;
	int prot;

	dato = PyTuple_GetItem (args, 0);
	if (!dato)
		return NULL;

	if (!PyCObject_Check (dato))
		return NULL;

	dat = PyCObject_AsVoidPtr (dato);

	arg = PyTuple_GetItem (args, 1);
	if (!arg)
		return NULL;

	if (arg == Py_None)
		prot = 0;
	else if (PyInt_Check (arg))
		prot = !!PyInt_AsLong (arg);
	else return NULL;

	if (!AdmData_isAlive(dat, NULL))
		return PyErr_Format(DeadData, "Accessing dead Data");

	dprintf (("AdmData_setProtected (%p, %d) = ", dat, prot));
	AdmData_setProtected (dat, prot, NULL);
	dprintf (("<void>\n"));

	Py_INCREF (Py_None);
	return Py_None;
}

#define LIST_PROPERTY(prop)						\
static PyObject *list_is##prop (PyObject *self, PyObject *args)		\
{									\
	int bl;								\
	PyObject *lsto;							\
	AdmList *lst;							\
									\
	lsto = PyTuple_GetItem (args, 0);				\
	if (!lsto)							\
		return NULL;						\
									\
	if (!PyCObject_Check (lsto))					\
		return NULL;						\
									\
	lst = PyCObject_AsVoidPtr (lsto);				\
									\
	if (!AdmData_isAlive((AdmData *)lst, NULL))			\
		return PyErr_Format(DeadData, "Accessing dead Data");	\
									\
	dprintf (("AdmList_is" #prop " (%p) is ", lst));		\
	bl = AdmList_is##prop (lst, NULL);				\
	dprintf (("%s\n", bl ? "True" : "False"));			\
									\
	return Py_BuildValue ("i", bl);					\
}									\
									\
static PyObject *list_set##prop (PyObject *self, PyObject *args)	\
{									\
	PyObject *lsto;							\
	AdmList *lst;							\
	PyObject *arg;							\
	int bl;								\
									\
	lsto = PyTuple_GetItem (args, 0);				\
	if (!lsto)							\
		return NULL;						\
									\
	if (!PyCObject_Check (lsto))					\
		return NULL;						\
									\
	lst = PyCObject_AsVoidPtr (lsto);				\
									\
	if (!AdmData_isAlive((AdmData *)lst, NULL))			\
		return PyErr_Format(DeadData, "Accessing dead Data");	\
									\
	arg = PyTuple_GetItem (args, 1);				\
	if (!arg)							\
		return NULL;						\
									\
	if (arg == Py_None)						\
		bl = 0;							\
	else if (PyInt_Check (arg))					\
		bl = !!PyInt_AsLong (arg);				\
	else return NULL;						\
									\
	dprintf (("AdmList_set" #prop " (%p, %d) = ", lst, bl));	\
	AdmList_set##prop (lst, bl, NULL);				\
	dprintf (("<void>\n"));						\
									\
	Py_INCREF (Py_None);						\
	return Py_None;							\
}

LIST_PROPERTY(Anonymous)
LIST_PROPERTY(Atomic)
LIST_PROPERTY(Prepend)

static PyObject *list_getNumData (PyObject *self, PyObject *args)
{
	unsigned long n;
	PyObject *lsto;
	AdmList *lst;

	lsto = PyTuple_GetItem (args, 0);
	if (!lsto)
		return NULL;

	if (!PyCObject_Check (lsto))
		return NULL;

	lst = PyCObject_AsVoidPtr (lsto);

	if (!AdmData_isAlive((AdmData *)lst, NULL))
		return PyErr_Format(DeadData, "Accessing dead Data");

	dprintf (("AdmList_getNumData (%p) = ", lst));
	n = AdmList_getNumData (lst, NULL);
	dprintf (("%lu\n", n));

	return Py_BuildValue ("l", n);
}

static PyObject *list_getDataByIndex (PyObject *self, PyObject *args)
{
	AdmData *dat;
	PyObject *lsto;
	AdmList *lst;
	PyObject *arg;
	long index;

	lsto = PyTuple_GetItem (args, 0);
	if (!lsto)
		return NULL;

	if (!PyCObject_Check (lsto))
		return NULL;

	lst = PyCObject_AsVoidPtr (lsto);

	arg = PyTuple_GetItem (args, 1);
	if (!arg)
		return NULL;

	if (!PyInt_Check (arg))
		return NULL;

	index = PyInt_AsLong (arg);

	if (!AdmData_isAlive((AdmData *)lst, NULL))
		return PyErr_Format(DeadData, "Accessing dead Data");

	if (index >= AdmList_getNumData(lst, NULL))
		return PyErr_Format(PyExc_IndexError, "Indexing past end of Data Set");
	
	dprintf (("AdmList_getDataByIndex_b (%p, %ld) = ", lst, index));
	dat = AdmList_getDataByIndex_b (lst, index, NULL);
	dprintf (("%p\n", dat));

	if (!dat) 
		return PyErr_Format(ADMCoreError, "NULL child of List inside of index range");

	dprintf (("AdmData_ref (%p) = ", dat));
	AdmData_ref (dat, NULL);
	dprintf (("<void>\n"));

	return PyCObject_FromVoidPtr (dat, destroy_data);
}

static PyObject *list_hasDataNamed (PyObject *self, PyObject *args)
{
	AdmData *dat;
	PyObject *lsto;
	AdmList *lst;
	PyObject *arg;
	const char *name;

	lsto = PyTuple_GetItem (args, 0);
	if (!lsto)
		return NULL;

	if (!PyCObject_Check (lsto))
		return NULL;

	lst = PyCObject_AsVoidPtr (lsto);

	arg = PyTuple_GetItem (args, 1);
	if (!arg)
		return NULL;

	if (!PyString_Check (arg))
		return NULL;

	name = PyString_AsString (arg);

	if (!AdmData_isAlive((AdmData *)lst, NULL))
		return PyErr_Format(DeadData, "Accessing dead Data");

	dprintf (("AdmList_hasDataNamed (%p, %s) = ", lst,
		  str_repr (name)));
	dat = AdmList_getDataByName_b (lst, name, NULL);
	dprintf (("%p\n", dat));

	return PyInt_FromLong(dat?1:0);
}

static PyObject *list_getDataByName (PyObject *self, PyObject *args)
{
	AdmData *dat;
	PyObject *lsto;
	AdmList *lst;
	PyObject *arg;
	const char *name;

	lsto = PyTuple_GetItem (args, 0);
	if (!lsto)
		return NULL;

	if (!PyCObject_Check (lsto))
		return NULL;

	lst = PyCObject_AsVoidPtr (lsto);

	arg = PyTuple_GetItem (args, 1);
	if (!arg)
		return NULL;

	if (!PyString_Check (arg))
		return NULL;

	name = PyString_AsString (arg);

	if (!AdmData_isAlive((AdmData *)lst, NULL))
		return PyErr_Format(DeadData, "Accessing dead Data");

	dprintf (("AdmList_getDataByName_b (%p, %s) = ", lst,
		  str_repr (name)));
	dat = AdmList_getDataByName_b (lst, name, NULL);
	dprintf (("%p\n", dat));

	if (!dat) 
		return PyErr_Format(PyExc_KeyError, "No such child");

	dprintf (("AdmData_ref (%p) = ", dat));
	AdmData_ref (dat, NULL);
	dprintf (("<void>\n"));

	return PyCObject_FromVoidPtr (dat, destroy_data);
}

static PyObject *list_addData (PyObject *self, PyObject *args)
{
	AdmData *dat;
	PyObject *lsto;
	AdmList *lst;
	PyObject *arg;
	AdmDataType typ;
	const char *name;

	lsto = PyTuple_GetItem (args, 0);
	if (!lsto)
		return NULL;

	if (!PyCObject_Check (lsto))
		return NULL;

	lst = PyCObject_AsVoidPtr (lsto);

	arg = PyTuple_GetItem (args, 1);
	if (!arg)
		return NULL;

	if (!PyInt_Check (arg))
		return NULL;

	typ = PyInt_AsLong (arg);

	arg = PyTuple_GetItem (args, 2);
	if (!arg)
		return NULL;

	if (!PyString_Check (arg))
		return NULL;

	name = PyString_AsString (arg);

	if (!AdmData_isAlive((AdmData *)lst, NULL))
		return PyErr_Format(DeadData, "Accessing dead Data");

	dprintf (("AdmList_addData_b (%p, %d, %s) = ", lst,
		  typ, str_repr (name)));
	dat = AdmList_addData_b (lst, typ, name, NULL);
	dprintf (("%p\n", dat));

	if (!dat) 
		return PyErr_Format(ListError, "addData failed");

	dprintf (("AdmData_ref (%p) = ", dat));
	AdmData_ref (dat, NULL);
	dprintf (("<void>\n"));

	return PyCObject_FromVoidPtr (dat, destroy_data);
}

static PyObject *list_moveData (PyObject *self, PyObject *args)
{
	PyObject *lsto;
	AdmList *lst;
	PyObject *arg;
	AdmData *dat;
	long index;

	lsto = PyTuple_GetItem (args, 0);
	if (!lsto)
		return NULL;

	if (!PyCObject_Check (lsto))
		return NULL;

	lst = PyCObject_AsVoidPtr (lsto);

	arg = PyTuple_GetItem (args, 1);
	if (!arg)
		return NULL;

	if (!PyCObject_Check (arg))
		return NULL;

	dat = PyCObject_AsVoidPtr (arg);

	arg = PyTuple_GetItem (args, 2);
	if (!arg)
		return NULL;

	if (!PyInt_Check (arg))
		return NULL;

	index = PyInt_AsLong (arg);

	if (!AdmData_isAlive((AdmData *)lst, NULL))
		return PyErr_Format(DeadData, "Accessing dead Data");

	if (!AdmData_isAlive(dat, NULL))
		return PyErr_Format(DeadData, "Accessing dead Data");

	if (AdmData_getContainer_b(dat, NULL) != (AdmData *) lst)
		return PyErr_Format(ListError, "Attempting to move a child not in this List.");
			
	if (index >= AdmList_getNumData(lst, NULL))
		return PyErr_Format(PyExc_IndexError, "Indexing past end of List.");

	dprintf (("AdmList_moveData (%p, %p, %ld) = ", lst,
		  dat, index));
	AdmList_moveData (lst, dat, index, NULL);
	dprintf (("<void>\n"));

	Py_INCREF (Py_None);
	return Py_None;
}

static PyObject *list_copyData (PyObject *self, PyObject *args)
{
	AdmData *copy;
	PyObject *lsto;
	AdmList *lst;
	PyObject *arg;
	AdmData *dat, *foo;

	lsto = PyTuple_GetItem (args, 0);
	if (!lsto)
		return NULL;

	if (!PyCObject_Check (lsto))
		return NULL;

	lst = PyCObject_AsVoidPtr (lsto);

	arg = PyTuple_GetItem (args, 1);
	if (!arg)
		return NULL;

	if (!PyCObject_Check (arg))
		return NULL;

	dat = PyCObject_AsVoidPtr (arg);

	if (!AdmData_isAlive((AdmData *)lst, NULL))
		return PyErr_Format(DeadData, "Accessing dead Data");

	if (!AdmData_isAlive(dat, NULL))
		return PyErr_Format(DeadData, "Accessing dead Data");

	foo = AdmData_getContainer_b(dat, NULL);
	if (foo == dat)
		return PyErr_Format(ListError, "Attempting to Copy a root Data element");
	if (foo == NULL)
		return PyErr_Format(ADMCoreError, "NULL container on a living Data");
	
	dprintf (("AdmList_copyData_b (%p, %p) = ", lst, dat));
	copy = AdmList_copyData_b (lst, dat, NULL);
	dprintf (("%p\n", copy));

	if (!copy)
		return PyErr_Format(ListError, "copyData failed");

	dprintf (("AdmData_ref (%p) = ", copy));
	AdmData_ref (copy, NULL);
	dprintf (("<void>\n"));

	return PyCObject_FromVoidPtr (copy, destroy_data);
}

static PyObject *data_getPos (PyObject *self, PyObject *args)
{
	long pos;
	PyObject *dato;
	AdmData *dat;

	dato = PyTuple_GetItem (args, 0);
	if (!dato)
		return NULL;

	if (!PyCObject_Check (dato))
		return NULL;

	dat = PyCObject_AsVoidPtr (dato);

	if (!AdmData_isAlive(dat, NULL))
		return PyErr_Format(DeadData, "Accessing dead Data");

	dprintf (("AdmData_getPos (%p) = ", dat));
	pos = AdmData_getPos (dat, NULL);
	dprintf (("%ld\n", pos));

	return Py_BuildValue ("l", pos);
}

// Well, _you_ type it all out then..
#define SCALAR(name, type, ctype, ptype, astype, fmt, bld, repr)	\
static PyObject *name##_getValue (PyObject *self, PyObject *args)	\
{									\
	ctype val;							\
	PyObject *dato;							\
	Adm##type *dat;							\
									\
	dato = PyTuple_GetItem (args, 0);				\
	if (!dato)							\
		return NULL;						\
									\
	if (!PyCObject_Check (dato))					\
		return NULL;						\
									\
	dat = PyCObject_AsVoidPtr (dato);				\
									\
	dprintf (("Adm" #type "_getValue (%p, NULL) = ", dat));		\
	val = Adm##type##_getValue (dat, NULL);				\
	dprintf (( fmt "\n", repr (val)));				\
									\
	return Py_BuildValue (bld, val);				\
}									\
									\
static PyObject *name##_setValue (PyObject *self, PyObject *args)	\
{									\
	PyObject *dato;							\
	Adm##type *dat;							\
	PyObject *arg;							\
	ctype val;							\
									\
	dato = PyTuple_GetItem (args, 0);				\
	if (!dato)							\
		return NULL;						\
									\
	if (!PyCObject_Check (dato))					\
		return NULL;						\
									\
	dat = PyCObject_AsVoidPtr (dato);				\
									\
	arg = PyTuple_GetItem (args, 1);				\
	if (!arg)							\
		return NULL;						\
									\
	if (!Py##ptype##_Check (arg))					\
		return NULL;						\
									\
	val = Py##ptype##_As##astype (arg);				\
									\
	dprintf (("Adm" #type "_setValue (%p, " fmt ", NULL) = ",	\
		  dat, repr (val)));					\
	Adm##type##_setValue (dat, val, NULL);				\
	dprintf (("<void>\n"));						\
									\
	Py_INCREF (Py_None);						\
	return Py_None;							\
}

SCALAR(int, Int, int, Int, Long, "%d", "l", )
SCALAR(float, Float, double, Float, Double, "%f", "d", )
SCALAR(bool, Bool, long, Int, Long, "%ld", "l", )

#define STRSCALAR(name, type, ptype, astype, fmt, bld, repr)	\
static PyObject *name##_getValue (PyObject *self, PyObject *args)	\
{									\
	PyObject *dato, *retobj;					\
	AdmRefStr *val;							\
	Adm##type *dat;							\
									\
	dato = PyTuple_GetItem (args, 0);				\
	if (!dato)							\
		return NULL;						\
									\
	if (!PyCObject_Check (dato))					\
		return NULL;						\
									\
	dat = PyCObject_AsVoidPtr (dato);				\
									\
	dprintf (("Adm" #type "_getValue (%p, NULL) = ", dat));		\
	val = Adm##type##_getValue (dat, NULL);				\
									\
	if (val != NULL) {						\
		dprintf (( fmt "\n", repr (AdmRefStr_str(val))));	\
		retobj = Py_BuildValue (bld, AdmRefStr_str(val));	\
		AdmRefStr_unref(val);					\
	} else {							\
		retobj = NULL;						\
	}								\
									\
	return retobj;							\
}									\
									\
static PyObject *name##_setValue (PyObject *self, PyObject *args)	\
{									\
	PyObject *dato;							\
	Adm##type *dat;							\
	PyObject *arg;							\
	const char *val;						\
									\
	dato = PyTuple_GetItem (args, 0);				\
	if (!dato)							\
		return NULL;						\
									\
	if (!PyCObject_Check (dato))					\
		return NULL;						\
									\
	dat = PyCObject_AsVoidPtr (dato);				\
									\
	arg = PyTuple_GetItem (args, 1);				\
	if (!arg)							\
		return NULL;						\
									\
	if (!Py##ptype##_Check (arg))					\
		return NULL;						\
									\
	val = Py##ptype##_As##astype (arg);				\
									\
	dprintf (("Adm" #type "_setValue (%p, " fmt ", NULL) = ",	\
		  dat, repr (val)));					\
	Adm##type##_setValue (dat, val, NULL);				\
	dprintf (("<void>\n"));						\
									\
	Py_INCREF (Py_None);						\
	return Py_None;							\
}

STRSCALAR(string, String, String, String, "%s", "s", str_repr)
STRSCALAR(base64, Base64, String, String, "%s", "s", str_repr)
STRSCALAR(copy, Copy, String, String, "%s", "s", str_repr)

static PyMethodDef AlchemistMethods[] = {
    { "context_create", (PyCFunction) context_create, METH_VARARGS, NULL },
    { "context_copy", (PyCFunction) context_copy, METH_VARARGS, NULL },
    { "context_merge", (PyCFunction) context_merge, METH_VARARGS, NULL },
    { "context_flatten", (PyCFunction) context_flatten, METH_VARARGS, NULL },
    { "context_toXML", (PyCFunction) context_toXML, METH_VARARGS, NULL },
    { "context_fromXML", (PyCFunction) context_fromXML, METH_VARARGS, NULL },
    { "context_getDelete", (PyCFunction) context_getMask,METH_VARARGS,NULL},
    { "context_setDelete", (PyCFunction) context_setMask,METH_VARARGS,NULL},
    { "context_clearDeleteByIndex", (PyCFunction) context_clearMaskByIndex,
      METH_VARARGS, NULL },
    { "context_clearDeleteByString", (PyCFunction) context_clearMaskByPath,
      METH_VARARGS, NULL },
    { "context_getNumDeletes", (PyCFunction) context_getNumMasks,
      METH_VARARGS, NULL },
    { "context_isIdentical", (PyCFunction) context_isIdentical,
      METH_VARARGS, NULL },
    { "context_getIdentityRoot", (PyCFunction) context_getIdentityRoot,
      METH_VARARGS, NULL },
    { "identity_getParentA", (PyCFunction) identity_getParentA,
      METH_VARARGS, NULL },
    { "identity_getParentB", (PyCFunction) identity_getParentB,
      METH_VARARGS, NULL },
    { "identity_getName", (PyCFunction) identity_getName,
      METH_VARARGS, NULL },
    { "identity_setName", (PyCFunction) identity_setName,
      METH_VARARGS, NULL },
    { "identity_getSerial", (PyCFunction) identity_getSerial,
      METH_VARARGS, NULL },
    { "identity_setSerial", (PyCFunction) identity_setSerial,
      METH_VARARGS, NULL },
    { "identity_isAlive", (PyCFunction) identity_isAlive,
      METH_VARARGS, NULL },
    { "identity_isIdentical", (PyCFunction) identity_isIdentical,
      METH_VARARGS, NULL },
    { "context_getDataRoot", (PyCFunction) context_getDataRoot,
      METH_VARARGS, NULL },
    { "data_getContainer", (PyCFunction) data_getContainer,
      METH_VARARGS, NULL },
    { "data_getContext", (PyCFunction) data_getContext,
      METH_VARARGS, NULL },
    { "context_getDataByPath", (PyCFunction) context_getDataByPath,
      METH_VARARGS, NULL },
    { "data_getType", (PyCFunction) data_getType,
      METH_VARARGS, NULL },
    { "data_isAlive", (PyCFunction) data_isAlive,
      METH_VARARGS, NULL },
    { "data_isIdentical", (PyCFunction) data_isIdentical,
      METH_VARARGS, NULL },
    { "data_unlink", (PyCFunction) data_unlink,
      METH_VARARGS, NULL },
    { "data_getName", (PyCFunction) data_getName,
      METH_VARARGS, NULL },
    { "data_setName", (PyCFunction) data_setName,
      METH_VARARGS, NULL },
    { "data_getSource", (PyCFunction) data_getSource,
      METH_VARARGS, NULL },
    { "data_setSource", (PyCFunction) data_setSource,
      METH_VARARGS, NULL },
    { "data_isProtected", (PyCFunction) data_isProtected,
      METH_VARARGS, NULL },
    { "data_setProtected", (PyCFunction) data_setProtected,
      METH_VARARGS, NULL },
    { "list_isAnonymous", (PyCFunction) list_isAnonymous,
      METH_VARARGS, NULL },
    { "list_setAnonymous", (PyCFunction) list_setAnonymous,
      METH_VARARGS, NULL },
    { "list_isAtomic", (PyCFunction) list_isAtomic,
      METH_VARARGS, NULL },
    { "list_setAtomic", (PyCFunction) list_setAtomic,
      METH_VARARGS, NULL },
    { "list_isPrepend", (PyCFunction) list_isPrepend,
      METH_VARARGS, NULL },
    { "list_setPrepend", (PyCFunction) list_setPrepend,
      METH_VARARGS, NULL },
    { "list_getNumChildren", (PyCFunction) list_getNumData,
      METH_VARARGS, NULL },
    { "list_getChildByIndex", (PyCFunction) list_getDataByIndex,
      METH_VARARGS, NULL },
    { "list_getChildByName", (PyCFunction) list_getDataByName,
      METH_VARARGS, NULL },
    { "list_hasChildNamed", (PyCFunction) list_hasDataNamed,
      METH_VARARGS, NULL },
    { "list_addChild", (PyCFunction) list_addData,
      METH_VARARGS, NULL },
    { "list_moveChild", (PyCFunction) list_moveData,
      METH_VARARGS, NULL },
    { "list_copyData", (PyCFunction) list_copyData,
      METH_VARARGS, NULL },
    { "data_getPos", (PyCFunction) data_getPos,
      METH_VARARGS, NULL },
    { "int_getValue", (PyCFunction) int_getValue,
      METH_VARARGS, NULL },
    { "int_setValue", (PyCFunction) int_setValue,
      METH_VARARGS, NULL },
    { "float_getValue", (PyCFunction) float_getValue,
      METH_VARARGS, NULL },
    { "float_setValue", (PyCFunction) float_setValue,
      METH_VARARGS, NULL },
    { "bool_getValue", (PyCFunction) bool_getValue,
      METH_VARARGS, NULL },
    { "bool_setValue", (PyCFunction) bool_setValue,
      METH_VARARGS, NULL },
    { "string_getValue", (PyCFunction) string_getValue,
      METH_VARARGS, NULL },
    { "string_setValue", (PyCFunction) string_setValue,
      METH_VARARGS, NULL },
    { "base64_getValue", (PyCFunction) base64_getValue,
      METH_VARARGS, NULL },
    { "base64_setValue", (PyCFunction) base64_setValue,
      METH_VARARGS, NULL },
    { "copy_getValue", (PyCFunction) copy_getValue,
      METH_VARARGS, NULL },
    { "copy_setValue", (PyCFunction) copy_setValue,
      METH_VARARGS, NULL },
    { NULL, NULL, 0, NULL }
};

void init_alchemist (void)
{
	PyObject *m, *d, *p;
	m = Py_InitModule("_alchemist", AlchemistMethods);
	d = PyModule_GetDict(m);

	p = PyDict_New();
	PyDict_SetItemString(p, "__doc__",
		Py_BuildValue("s",
		"An ADMCoreError exception is raised when a conflict has been "
	        "detected in the underlying Alchemist Data Core. It should be "
		"reported to maintainers."));
	ADMCoreError = PyErr_NewException("Alchemist.ADMCoreError", NULL, p);
	PyDict_SetItemString(d, "ADMCoreError", ADMCoreError);

	p = PyDict_New();
	PyDict_SetItemString(p, "__doc__",
		Py_BuildValue("s",
		"A ListError exception is raised when a ListData method fails "
		"do to illegal operations on a List."));
	ListError = PyErr_NewException("Alchemist.ListError", NULL, p);
       	PyDict_SetItemString(d, "ListError", ListError);
	
	p = PyDict_New();
	PyDict_SetItemString(p, "__doc__",
		Py_BuildValue("s",
		"A DeadData exception is raised when an attempt is made to ",
	        "manipulate a Data node that has been unlinked, and fails ",
		"isAlive()"));
	DeadData = PyErr_NewException("Alchemist.DeadData", NULL, p);
	PyDict_SetItemString(d, "DeadData", DeadData);

	p = PyDict_New();
	PyDict_SetItemString(p, "__doc__",
		Py_BuildValue("s",
		"A DeadIdentity exception is raised when an attempt is made to ",
	        "manipulate an Identity node that has been unlinked, and fails ",
		"isAlive()"));
	DeadIdentity = PyErr_NewException("Alchemist.DeadIdentity", NULL, p);
	PyDict_SetItemString(d, "DeadIdentity", DeadIdentity);

	p = PyDict_New();
	PyDict_SetItemString(p, "__doc__",
		Py_BuildValue("s",
		"A ContextError is raised when a merge or create Context fails. "
		"This is usally the result of passing an invalid name."));
	ContextError = PyErr_NewException("Alchemist.ContextError", NULL, p);
	PyDict_SetItemString(d, "ContextError", ContextError);

	p = PyDict_New();
	PyDict_SetItemString(p, "__doc__",
		Py_BuildValue("s",
		"A ParseError exception is raised when fromXML fails to parse "
		"an XML string into a Context."));
	ParseError = PyErr_NewException("Alchemist.ParseError", NULL, p);
	PyDict_SetItemString(d, "ParseError", ParseError);

}
