/**             
 * Copyright (C) 2000, 2001 Red Hat, Inc.
 *              
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *      
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */
 
/**
 * @brief Adm BlackBox functions
 *
 * The functions here present the upper layer of the BlackBox API. A BlackBox
 * is simply put any kind of data source respectively dats sink.
 *
 * The API provides these parts:
 *  - Reference handling of BlackBoxes (lifecycle)
 *  - Read operation and verification
 *  - Write operation and verification
 *  - History handling
 *  - Tagging and dirty handling
 *
 * @file        bbwrappers.c
 * @author      Crutcher Dunnavant <crutcher@redhat.com>
 * @date        05/06/2001
 * @version     1.0
 */ 

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <glib.h>

#include "alchemist.h"
#include "blackbox.h"

#include "switchboard.h"
#include "admerror.h"


void
AdmBlackBox_ref (AdmBlackBox *box,
		 AdmError **error)
{
	_adm_error_init_check(error);
	g_assert(box != NULL);

	/* bump the refcount of the box */
	++ box->refcount;
}

void
AdmBlackBox_unref (AdmBlackBox *box,
		   AdmError **error)
{
	_adm_error_init_check(error);
	g_assert(box != NULL);

	/* decrement the refcount on the box, and check for a ref of 0 */
	if ((-- box->refcount) == 0) {
		/* ask the interface to clean itself up */
		if (box->methods->free != NULL)
			box->methods->free(box->data);

		/* zap our cache id, if we were keeping one */
		if (box->cache_id != NULL)
			AdmRefStr_unref(box->cache_id);

		/* and free the box structure */
		free(box);
	}
}

AdmContext *
AdmBlackBox_read (AdmBlackBox *box,
		  AdmError **error)
{
	AdmContext *ctx;
	_adm_error_init_check(error);
	g_assert(box != NULL);

	if (!ADM_BBCAP_SUP_READ(box->caps)) {
		AdmError_create(error, ADM_ENOREAD, _("reads unsuported by this %s"), "AdmBlackBox");
		ctx = NULL;
	} else { 
		ctx = box->methods->read(box->data, error);

		/* cache cases */
		if (ADM_BBCAP_CACHED(box->caps)) {
			if (ctx == NULL) {
				/* if we are cacheing, and the read failed, try the cache */
				ctx = _adm_cache_readSerial(AdmRefStr_str(box->cache_id), 0);
				
				/* error? what error? */
				if ((ctx != NULL) && (error != NULL)) {
					AdmError_unref(*error);
					*error = NULL;
				}

			} else {
				/* cache the read */
				_adm_cache_write(ctx, AdmRefStr_str(box->cache_id),
						box->cache_uid, box->cache_gid, box->cache_depth);
			}
		}
	}

	return ctx;
}

int
AdmBlackBox_listSerial (AdmBlackBox   *box,
			unsigned int  *array,
			unsigned int   array_len,
			unsigned int   top_serial,
			AdmError **error)
{
	AdmError *err_p;
	int count;
	_adm_error_init_check(error);

	if ((box->cache_id == NULL) && (box->methods->listSerial == NULL)) {
		AdmError_create(error, ADM_ENOLISTSERIAL, _("listSerials unsuported by this %s"), "AdmBlackBox");
		return 0;
	}

	count = 0;

	if (ADM_BBCAP_SUP_LISTSERIAL(box->caps)) {
		err_p = NULL;
		count = box->methods->listSerial(box->data, array, array_len, top_serial, &err_p);

		/* propagate errors */
		if (err_p != NULL) {
			_adm_error_prop(err_p, error);
			return 0;
		}
	}

	/* if we are cacheing, ask the cache what it has */
	if (ADM_BBCAP_CACHED(box->caps)) {
		count = _adm_cache_listSerial(AdmRefStr_str(box->cache_id), array, array_len, top_serial, count);
	}

	return count;
}

AdmContext *
AdmBlackBox_readSerial (AdmBlackBox  *box,
			unsigned int  serial,
			AdmError **error)
{
	AdmContext *ctx;
	_adm_error_init_check(error);
	g_assert(box != NULL);

	ctx = NULL;
	return ctx;
}

adm_bb_cap_t
AdmBlackBox_caps (AdmBlackBox *box,
			AdmError **error)
{
	_adm_error_init_check(error);
	_adm_abort_if_fail(box != NULL);

	return box->caps;
}

int
AdmBlackBox_write (AdmBlackBox *box,
		   AdmContext  *context,
		   AdmError **error)
{
	int retval;
	_adm_error_init_check(error);
	g_assert(box != NULL);

	if (!ADM_BBCAP_SUP_WRITE(box->caps)) {
		AdmError_create(error, ADM_ENOWRITE, _("writes unsuported by this %s"), "AdmBlackBox");
		retval = -1;

	} else {
		retval = box->methods->write(box->data, context, error);

		/* if we are cacheing, and the write succeded, cache it */
		if ((retval == 0) && ADM_BBCAP_CACHED(box->caps))
			_adm_cache_write(context, AdmRefStr_str(box->cache_id),
					box->cache_uid, box->cache_gid, box->cache_depth);
	}
	return retval;
}

unsigned int
AdmBlackBox_getMark (AdmBlackBox  *box,
			  const char   *mark,
			  AdmError **error)
{
	_adm_error_init_check(error);

	if (!ADM_BBCAP_CACHED(box->caps)) {
		AdmError_create(error, ADM_ENOCACHE, _("this %s is not cached"), "AdmBlackBox");
		return 0;
	} else {
		return _adm_cache_get_mark(AdmRefStr_str(box->cache_id), mark);
	}
}

int
AdmBlackBox_setMark (AdmBlackBox  *box,
			  const char   *mark,
			  AdmError **error)
{
	_adm_error_init_check(error);
	if (!ADM_BBCAP_CACHED(box->caps)) {
		AdmError_create(error, ADM_ENOCACHE, _("this %s is not cached"), "AdmBlackBox");
		return -1;
	} else {
		return _adm_cache_set_mark(AdmRefStr_str(box->cache_id), box->cache_uid, box->cache_gid, mark);
	}
}

int
AdmBlackBox_clearMark (AdmBlackBox  *box,
			  const char   *mark,
			  AdmError **error)
{
	_adm_error_init_check(error);
	if (!ADM_BBCAP_CACHED(box->caps)) {
		AdmError_create(error, ADM_ENOCACHE, _("this %s is not cached"), "AdmBlackBox");
		return -1;
	} else {
		return _adm_cache_clear_mark(AdmRefStr_str(box->cache_id), mark);
	}
}
