/**
 * Copyright (C) 2000, 2001 Red Hat, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/**
 * @brief Adm cache BlackBox layer
 *
 * This file implements the functions for the cache BlackBox layer.
 * It has only internal functions as the caching of BlackBoxes is completely
 * transparent and only adjustable via the switchboard and the namespace
 * configuration.
 *
 * @file        cache.c
 * @author      Crutcher Dunnavant <crutcher@redhat.com>
 * @date        05/06/2001
 * @version     1.0
 */

#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <dirent.h>
#include <stdio.h>
#include <string.h>

#include <glib.h>
#include <gmodule.h>

#include "alchemist.h"
#include "blackbox.h"
#include "switchboard.h"
#include "admerror.h"

static const char def_cache_dir_path[] = ALCHEMIST_CACHE_DIR;
static AdmRefStr *active_cache_dir_path = NULL;

AdmRefStr *
AdmBlackBox_getCacheDir (AdmError **error)
{
	_adm_error_init_check(error);

	if (active_cache_dir_path == NULL)
		active_cache_dir_path = AdmRefStr_get(def_cache_dir_path, 1);

	AdmRefStr_ref(active_cache_dir_path);
	return active_cache_dir_path;
}

int
AdmBlackBox_setCacheDir (const char *path,
			 AdmError **error)
{
	_adm_error_init_check(error);
	
	if (active_cache_dir_path != NULL)
		AdmRefStr_unref(active_cache_dir_path);

	active_cache_dir_path = (path == NULL) ? NULL : AdmRefStr_get(path, 1);
	return 0;
}

/**
 * Private helper function to verify if a given string consists only of valid
 * hexcodes.
 *
 * @brief      Internal hexadecimal string format check
 * @param  str The string to be checked.
 * @return     @c TRUE if string is a valid hexstring, @c FALSE otherwise.
 */
static int
_adm_cache_ishex (const char *str) 
{
	char c;
	if ((str == NULL) || (*str == '\0'))
		return 0;
	while ((c = *(str++)) != '\0') {
		if (! (((c >= '0') && (c <= '9')) || ((c >= 'a') && (c <= 'f'))) )
			return 0;
	}
	return 1;
}

/**
 * Read a cached context for the given cache_id and serial number.
 *
 * @brief           Internal read cached context by serial operation
 * @param  cache_id The unique cache id for the namespace cache blackbox.
 * @param  serial   The serial number of the cached context we want to read.
 * @return          The pointer to the read AdmContext or @c NULL on error.
 * @see             AdmBlackBox_readSerial
 */
AdmContext *
_adm_cache_readSerial (const char    *cache_id,
		       unsigned int   serial)
{
	GString *path;
	AdmContext *ctx;
	AdmRefStr *dir;
	int count;

	ctx = NULL;

	if (serial == 0) {
		/* special case, get the top one available */
		count = _adm_cache_listSerial(cache_id, &serial, 1, 0, 0);
		if (count == 0)
			goto ret;
	}

	dir = AdmBlackBox_getCacheDir(NULL);
	path = g_string_new(AdmRefStr_str(dir));
	AdmRefStr_unref(dir);

	g_string_sprintfa(path, "/%s/%x", cache_id, serial);

	AdmContext_readFromFile(&ctx, path->str, NULL);

	g_string_free(path, 1);

ret:
	return ctx;
}

/**
 * Cleans up the cache directory for the given cache_id. Keeps only the
 * @c depth newest cached entries. Used for expiring old entries.
 *
 * @brief          Internal cache directory expire operation
 * @param cache_id The unique cache id for the namespace cache blackbox.
 * @param depth    The maximum number of cached entries we want to keep.
 */
static void
_adm_cache_enforce_cache_depth (const char	*cache_id,
				unsigned int     depth)
{
	DIR *dir;
	struct dirent *d_ent;
	GString *path;
	AdmRefStr *cdir;
	unsigned int serial, serial_list[depth], i, tmp;

	/* contsruct the path */
	cdir = AdmBlackBox_getCacheDir(NULL);
	path = g_string_new(AdmRefStr_str(cdir));

	g_string_sprintfa(path, "/%s", cache_id);

	/* open the dir */
	dir = opendir(path->str);

	if (dir == NULL) {
		/* well, this sucks, but we cant fix it */
		g_string_free(path, 1);
		AdmRefStr_unref(cdir);
		return;
	}

	/* zap the list */
	for (i = 0; i < depth; ++ i) serial_list[i] = 0;

	/* icky, scary, nasty loop */
	while ((d_ent = readdir(dir)) != NULL) {
	
		/* ignore all files that are not valid hex characters */
		if (! _adm_cache_ishex(d_ent->d_name))
			continue;

		/* pull the number off the name */
		sscanf(d_ent->d_name, "%x", &serial);
		
		for (i = 0; i < depth; ++i) {
			if (serial_list[i] < serial) {
				for (; (i < depth) && (serial != 0); ++ i) {
					tmp = serial_list[i];
					serial_list[i] = serial;
					serial = tmp;
				}
				if (serial != 0) {
					g_string_sprintf(path, "%s/%s/%x", AdmRefStr_str(cdir), cache_id, serial);
					unlink(path->str);
				}
				goto next_loop;
			}
		}

		/* we fell off the end, zap this number */
		g_string_sprintf(path, "%s/%s/%x", AdmRefStr_str(cdir), cache_id, serial);
		unlink(path->str);
next_loop:;
	}
	closedir(dir);
	AdmRefStr_unref(cdir);
	g_string_free(path, 1);
}

/**
 * Cache the given context for the given cache_id. Also ensures that the cache
 * directory is appropriately expired as well.
 *
 * @brief          Internal cache context storage operation
 * @param context  The context we want to cache.
 * @param cache_id The unique cache id for the namespace cache blackbox.
 * @param depth    The maximum number of cached entries we want to keep.
 * @see             AdmBlackBox_write
 */
void
_adm_cache_write (AdmContext    *context,
		  const char    *cache_id,
		  uid_t		 uid,
		  gid_t		 gid,
		  unsigned int   depth)
{
	GString *path;
	AdmRefStr *cdir;
	unsigned int serial;

	cdir = AdmBlackBox_getCacheDir(NULL);
	path = g_string_new(AdmRefStr_str(cdir));
	AdmRefStr_unref(cdir);

	g_string_sprintfa(path, "/%s/", cache_id);

	if (access(path->str, W_OK) == -1) {
		/* the cache directory don't exist, try to make it. */
		mkdir(path->str, 0700);
		chown(path->str, gid, uid);

		if (access(path->str, W_OK) == -1)
			/* oh well */
			goto ret;
	}

	serial = AdmIdentity_getSerial(AdmContext_getIdentityRoot_b(context, NULL), NULL);
	g_string_sprintfa(path, "%x", serial);

	AdmContext_writeToFile(context, path->str, 0600, gid, uid, NULL);

	_adm_cache_enforce_cache_depth(cache_id, depth);

ret:
	g_string_free(path, 1);
}

/**
 * Fills in the give array with the serial numbers of the cached contexts. As
 * the given array might have already been prefilled by the real BlackBox we
 * need to be able to specify that in contrast to the listSerial() function of
 * the 'frontend' BlackBox.
 *
 * @brief                Internal serial list operation
 * @param  cache_id      The unique cache id for the namespace cache blackbox.
 * @param  array         The pointer to the preallocated array of @c int values
 *                       to be filled with the serial numbers.
 * @param  array_len     The size of the perallocated array.
 * @param  top_serial    The highest serial number that we request. Only serial
 *                       numbers below that will be returned. Use @c 0 to
 *                       request the newest (highest) serial numbers.
 * @param  prefill_count The number of already filled in array elements by the
 *                       related BlackBox.
 * @return               The amount of serial numbers filled in. Always less or
 *                       equal to the array length.
 * @see                  AdmBlackBox_listSerial
 */
int
_adm_cache_listSerial	    (const char    *cache_id,
			     unsigned int  *array,
			     unsigned int   array_len,
			     unsigned int   top_serial,
			     unsigned int   prefill_count)
{
	DIR *dir;
	struct dirent *d_ent;
	GString *path;
	AdmRefStr *cdir;
	unsigned int count = prefill_count;
	unsigned int foo, bar, i;

	/* contsruct the path */
	cdir = AdmBlackBox_getCacheDir(NULL);
	path = g_string_new(AdmRefStr_str(cdir));
	AdmRefStr_unref(cdir);

	g_string_sprintfa(path, "/%s", cache_id);

	/* open the dir */
	dir = opendir(path->str);

	/* trash the path before we check the dir */
	g_string_free(path, 1);

	if (dir == NULL) {
		/* crap. we can't read the cache dir. guess that means there isnt
		* anything in it. */
		return count;
	}

	/* paranoia! */
	for (i = prefill_count; i < array_len; ++ i) array[i] = 0;

	/* icky, scary, nasty loop */
	while ((d_ent = readdir(dir)) != NULL) {
		/* ignore all files that are not valid hex characters */
		if (! _adm_cache_ishex(d_ent->d_name))
			continue;

		/* pull the number off the name */
		sscanf(d_ent->d_name, "%x", &foo);

		/* ignore high values, unless the sentry is set to 0 */
		if ((top_serial != 0) && (foo > top_serial))
			continue;

		/* we now have a number in 'foo'. we need to scan the array for the
		* first number less than, or equall to, foo */
		for (i = 0; i < array_len; ++i) if (array[i] <= foo) break;
		if ((i == array_len) || (foo == array[i])) {
			/* we fell off the end, or we matched a number. no change */
			continue;
		}

		/* we now have an index to a number in a sorted array that is less than foo.
		* we need to insert foo here, and shift */
		do {
			bar = array[i];
			array[i] = foo;
			foo = bar;
		} while ((foo != 0) && (++i < array_len));

		/* if foo has a non-zero value, we inserted a number, but one fell off the end, so
		* we dont bump count. otherwise, we bump count */
		if (foo == 0)
			++ count;
	}
	closedir(dir);

	return count;
}

/**
 * @see                  AdmBlackBox_setMark
 */
int
_adm_cache_set_mark (const char   *cache_id,
			uid_t	uid,
			gid_t	gid,
			  const char   *mark)
{
	GString *path;
	unsigned int serial;
	AdmRefStr *cdir;
	int fd, retval;

	if (_adm_cache_listSerial(cache_id, &serial, 1, 0, 0) == 0)
		return -1;

	cdir = AdmBlackBox_getCacheDir(NULL);
	path = g_string_new(AdmRefStr_str(cdir));
	AdmRefStr_unref(cdir);

	g_string_sprintfa(path, "/%s/wm", cache_id);

	if (access(path->str, W_OK) == -1) {
		/* the cache mark directory doesn't exist, try to make it. */
		mkdir(path->str, 0700);
		chown(path->str, uid, gid);

		retval = -1;
		if (access(path->str, W_OK) == -1)
			/* oh well */
			goto ret;
	}

	g_string_sprintfa(path, "/%s", mark);

	if ((fd = open(path->str, O_WRONLY | O_TRUNC | O_CREAT, 0600)) != -1) {
		fchown(fd, uid, gid);

		if (write(fd, &serial, sizeof serial) == -1)
			retval = -1;
		else
			retval = 0;
	} else {
		retval = -1;
	}
	
ret:
	g_string_free(path, 1);
	return retval;
}

/**
 * @see                  AdmBlackBox_clearMark
 */
int
_adm_cache_clear_mark (const char   *cache_id,
			    const char   *mark)
{
	GString *path;
	AdmRefStr *cdir;
	int retval;

	cdir = AdmBlackBox_getCacheDir(NULL);
	path = g_string_new(AdmRefStr_str(cdir));
	AdmRefStr_unref(cdir);

	g_string_sprintfa(path, "/%s/wm/%s", cache_id, mark);

	retval = unlink(path->str);

	g_string_free(path, 1);

	return retval;
}

/**
 * @see                  AdmBlackBox_getMark
 */
unsigned int
_adm_cache_get_mark (const char   *cache_id,
			  const char   *mark)
{
	GString *path;
	AdmRefStr *cdir;
	unsigned int serial, retval;
	int fd;

	cdir = AdmBlackBox_getCacheDir(NULL);
	path = g_string_new(AdmRefStr_str(cdir));
	AdmRefStr_unref(cdir);

	g_string_sprintfa(path, "/%s/wm/%s", cache_id, mark);

	if ((fd = open(path->str, O_RDONLY)) != -1) {
		if (read(fd, &serial, sizeof serial) == -1)
			retval = 0;
		else
			retval = serial;
	} else {
		retval = 0;
	}
	g_string_free(path, 1);

	return retval;
}
