/**
 * Copyright (C) 2000, 2001 Red Hat, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/**
 * @brief Adm file BlackBox implementation
 *
 * This file implements the functions for the file BlackBox.
 *
 * The file BlackBox is the simple most of the BlackBoxes as it only stores and
 * reads local files.
 *
 * All required functions for a BlackBox are implemented, so it is a fully
 * functional BlackBox which is capable of reading and writing.
 *
 * @file        fileblackbox.c
 * @author      Crutcher Dunnavant <crutcher@redhat.com>
 * @date        05/06/2001
 * @version     1.0
 */

#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <pwd.h>
#include <grp.h>

#include <glib.h>

#include "alchemist.h"
#include "blackbox.h"

#define BUFSIZE 1024

struct fileblackbox_instance {
	AdmBlackBoxMethods methods;
	char *path;
	char *uname;
	char *gname;
	int flags;
	int mode;
	uid_t uid;
	gid_t gid;
};

static void		fileblackbox_free	(void			 *data);

static AdmContext *	fileblackbox_read	(void			 *data,
						 AdmError **error);

unsigned int   		fileblackbox_listSerial	(void	  		 *data,
						 unsigned int  		 *array,
						 unsigned int  		  array_len,
						 unsigned int  		  top_serial,
						 AdmError **error);

static int		fileblackbox_write	(void			 *data,
						 AdmContext		 *context,
						 AdmError **error);

int			_adm_fileblackbox_getBox(AdmBlackBoxMethods	**methods,
						 void			**data,
						 AdmList		 *box_list,
						 AdmError **error);

static AdmBlackBoxMethods fileblackbox_methods = {
	free:		fileblackbox_free,
	read:		fileblackbox_read,
	listSerial:     fileblackbox_listSerial,
	readSerial:     NULL,
	write:          fileblackbox_write,
};

#define FILEBB_SETUSER_FLAG	(1 << 0)
#define FILEBB_SETGROUP_FLAG	(1 << 1)

static void
fileblackbox_free (void     *data)
{
	struct fileblackbox_instance *in_p = data;

	if (in_p->path)
		free(in_p->path);

	if (in_p->uname)
		free(in_p->uname);

	if (in_p->gname)
		free(in_p->gname);

	free(in_p);
}

static AdmContext *
fileblackbox_read (void        *data,
		   AdmError **error)
{
	struct fileblackbox_instance *in_p = data;
	AdmContext *ctx;

	AdmContext_readFromFile(&ctx, in_p->path, error);
	return ctx;
}

unsigned int
fileblackbox_listSerial	(void	  	 *data,
			 unsigned int	 *array,
			 unsigned int  	  array_len,
			 unsigned int 	  top_serial,
			 AdmError **error)
{
	AdmContext *ctx;
	AdmIdentity *id;
	unsigned int serial, i;

	/* zero out the array */
	for (i = 0; i < array_len; ++i) array[i] = 0;

	serial = 0;
	/* unfortunately, the only way to check the version of an on disk
	* context is to read it. so thats what we do. */
	ctx = fileblackbox_read(data, NULL);
	if (ctx != NULL) {
		id = AdmContext_getIdentityRoot_b(ctx, NULL);
		serial = AdmIdentity_getSerial(id, NULL);
		AdmContext_unref(ctx, NULL);
	}

	if (serial == 0)
		return 0;

	if ((serial <= top_serial) || (top_serial == 0)) {
		array[0] = serial;
		/* signal that we have something */
		return 1;
	} else {
		return 0;
	}
}

static int
fileblackbox_write (void        *data,
		    AdmContext  *context,
		    AdmError **error)
{
	struct fileblackbox_instance *in_p = data;
	uid_t uid;
	gid_t gid;

	uid = -1;
	gid = -1;

	if (in_p->flags & FILEBB_SETUSER_FLAG) {
		if (in_p->uname) {
			struct passwd ps, *ps_p;
			char buf[1024];

			getpwnam_r(in_p->uname, &ps, buf, 1024, &ps_p);
			if (ps_p != NULL) {
				uid = ps_p->pw_uid;
			}
		} else {
			uid = in_p->uid;
		}
	}

	if (in_p->flags & FILEBB_SETGROUP_FLAG) {
		if (in_p->gname) {
			struct group gs, *gs_p;
			char buf[1024];

			getgrnam_r(in_p->gname, &gs, buf, 1024, &gs_p);
			if (gs_p != NULL) {
				gid = gs_p->gr_gid;
			}
		} else {
			gid = in_p->gid;
		}
	}

	return AdmContext_writeToFile(context, in_p->path, in_p->mode, uid, gid, error);
}

int
_adm_fileblackbox_getBox (AdmBlackBoxMethods	**methods,
		     void		**data,
		     AdmList		 *box_list,
		     AdmError		**error)
{
	AdmRefStr *path, *uname, *gname;
	int flags;
	int mode;
	int readable, writable;
	uid_t uid;
	gid_t gid;
	AdmData *childA, *childB;
	struct fileblackbox_instance *in_p;
	
	flags = 0;
	*methods = NULL;
	*data = NULL;

	path = NULL;
	uname = NULL;
	gname = NULL;

	/* path */
	childA = AdmList_getDataByName_b(box_list, "path", NULL);
	if (childA == NULL) {
		AdmError_create(error, ADM_EBADDATA, "box_cfg/box/path does not exist");
		return -1;
	} else if (AdmData_getType(childA, NULL) != ADM_TYPE_STRING) {
		AdmError_create(error, ADM_EBADDATA, "box_cfg/box/path is not a string");
		return -1;
	}
	path = AdmString_getValue((AdmString *) childA, NULL);

	/* readable */
	readable = 0;
	if ((childA = AdmList_getDataByName_b(box_list, "readable", NULL))) {
		if (AdmData_getType(childA, NULL) != ADM_TYPE_BOOL) {
			AdmError_create(error, ADM_EBADDATA, "box_cfg/box/readable is not a bool");
			return -1;
		}

		if (AdmBool_getValue((AdmBool *) childA, NULL))
			readable = 1;
	}

	/* writable */
	writable = 0;
	if ((childA = AdmList_getDataByName_b(box_list, "writable", NULL))) {
		if (AdmData_getType(childA, NULL) != ADM_TYPE_BOOL) {
			AdmError_create(error, ADM_EBADDATA, "box_cfg/box/readable is not a bool");
			return -1;
		}

		if (AdmBool_getValue((AdmBool *) childA, NULL))
			writable = 1;
	}

	/* write file mode */
	if ((childA = AdmList_getDataByName_b(box_list, "mode", NULL))) {
		if (AdmData_getType(childA, NULL) != ADM_TYPE_INT) {
			AdmError_create(error, ADM_EBADDATA, "box_cfg/box/mode is not an int");
			return -1;
		}

		mode = AdmInt_getValue((AdmInt *) childA, NULL);
	} else {
		mode = 0600;
	}

	/* uid/uname (mutually exclusive) */
	uid = 0;
	uname = NULL;
	childA = AdmList_getDataByName_b(box_list, "uid", NULL);
	childB = AdmList_getDataByName_b(box_list, "uname", NULL);

	if ((childA != NULL) && (childB != NULL)) {
		AdmError_create(error, ADM_EBADDATA, "box_cfg/box/{uid,uname} are mutually exclusive");
		goto fail;
	}
	if (((childA != NULL) || (childB != NULL)) && !writable) {
		AdmError_create(error, ADM_EBADDATA, "box_cfg/box/{uid,uname} can only be set if writable is set");
		goto fail;
	}

	if (childA) {
		if (AdmData_getType(childA, NULL) != ADM_TYPE_INT) {
			AdmError_create(error, ADM_EBADDATA, "box_cfg/box/uid is not an int");
			goto fail;
		}

		uid = AdmInt_getValue((AdmInt *) childA, NULL);
		flags |= FILEBB_SETUSER_FLAG;

	} else if (childB) {
		if (AdmData_getType(childB, NULL) != ADM_TYPE_STRING) {
			AdmError_create(error, ADM_EBADDATA, "box_cfg/box/uname is not a string");
			goto fail;
		}

		uname = AdmString_getValue((AdmString *) childB, NULL);
		flags |= FILEBB_SETUSER_FLAG;
	}

	/* gid/gname (mutually exclusive) */
	gid = 0;
	gname = NULL;
	childA = AdmList_getDataByName_b(box_list, "gid", NULL);
	childB = AdmList_getDataByName_b(box_list, "gname", NULL);
	if ((childA != NULL) && (childB != NULL)) {
		AdmError_create(error, ADM_EBADDATA, "box_cfg/box/{gid,gname} are mutually exclusive");
		goto fail;
	}
	if (((childA != NULL) || (childB != NULL)) && !writable) {
		AdmError_create(error, ADM_EBADDATA, "box_cfg/box/{gid,gname} can only be set if writable is set");
		goto fail;
	}

	if (childA) {
		if (AdmData_getType(childA, NULL) != ADM_TYPE_INT) {
			AdmError_create(error, ADM_EBADDATA, "box_cfg/box/gid is not an int");
			goto fail;
		}

		gid = AdmInt_getValue((AdmInt *) childA, NULL);
		flags |= FILEBB_SETGROUP_FLAG;
	} else if (childB) {
		if (AdmData_getType(childB, NULL) != ADM_TYPE_STRING) {
			AdmError_create(error, ADM_EBADDATA, "box_cfg/box/gname is not a string");
			goto fail;
		}

		gname = AdmString_getValue((AdmString *) childB, NULL);
		flags |= FILEBB_SETGROUP_FLAG;
	}

	/* allocate instance */
	in_p = (struct fileblackbox_instance *) malloc(sizeof (struct fileblackbox_instance));

	in_p->path = strdup(AdmRefStr_str(path));
	in_p->flags = flags;
	in_p->mode = mode;
	in_p->uid = uid;
	in_p->gid = gid;
	in_p->uname = (flags & FILEBB_SETUSER_FLAG) ? strdup(AdmRefStr_str(uname)) : NULL;
	in_p->gname = (flags & FILEBB_SETGROUP_FLAG) ? strdup(AdmRefStr_str(gname)) : NULL;
	
	/* mangle the methods */
	in_p->methods = fileblackbox_methods;
	if (!readable) in_p->methods.read = NULL;
	if (!writable) in_p->methods.write = NULL;
	*methods = &in_p->methods;

	*data = (void *) in_p;
	return 0;
fail:
	if (path)
		AdmRefStr_unref(path);
	if (uname)
		AdmRefStr_unref(uname);
	if (gname)
		AdmRefStr_unref(gname);
	return -1;
}

