/**
 * Copyright (C) 2000, 2001 Red Hat, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/**
 * @brief Adm forge BlackBox implementation
 *
 * This file implements the functions for the forge BlackBox.
 *
 * The forge BlackBox is used to transform the XML/Alchemist representation of
 * the configuration into the native file formats using XSLT.
 *
 * The forge Blackbox is a write-only BlackBox for now as it can't as of yet
 * read/parse native configuration formats and transform them into our XML
 * representation.
 *
 * @file        forgeblackbox.c
 * @author      Harald Hoyer <harald@redhat.de>
 * @date        05/06/2001
 * @version     1.0
 */

#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <pwd.h>
#include <grp.h>

#include <glib.h>
#include <libgen.h>

#include <libxslt/xslt.h>
#include <libxslt/xsltInternals.h>
#include <libxslt/transform.h>
#include <libxslt/xsltutils.h>

#include "alchemist.h"
#include "blackbox.h"

#define BUFSIZE 1024

struct forgeblackbox_instance {
  AdmBlackBoxMethods methods;
  char *path;
  char *uname;
  char *gname;
  int flags;
  int mode;
  uid_t uid;
  gid_t gid;
  char *stylesheetpath;
};

static void forgeblackbox_free (void *data);

static int forgeblackbox_write (void *data,
				AdmContext *context,
				AdmError **error);

static int  forgeblackbox_getBox (AdmBlackBoxMethods **methods,
				 void   **data,
				 AdmList *box_list,
				 AdmError **error);

static AdmBlackBoxMethods forgeblackbox_methods = {
  free: forgeblackbox_free,
  read: NULL,
  listSerial: NULL,
  readSerial: NULL,
  write: forgeblackbox_write,
};


struct AdmBlackBoxModuleVector bb_mod_vector = {
  abi_major_ver: ALCHEMIST_ABI_MAJOR_VER,
  abi_minor_ver: ALCHEMIST_ABI_MINOR_VER,
  getBox: forgeblackbox_getBox
};

#define FILEBB_SETUSER_FLAG (1 << 0)
#define FILEBB_SETGROUP_FLAG (1 << 1)

static void
forgeblackbox_free (void *data)
{
  struct forgeblackbox_instance *in_p = data;
  free(in_p->path);
  free(in_p->uname);
  free(in_p->gname);
  free(in_p->stylesheetpath);
  free(in_p);
}

static int
forgeblackbox_write (void *data,
		     AdmContext *context,
		     AdmError **error)
{
  struct forgeblackbox_instance *in_p = data;
  int retval;
  uid_t uid;
  gid_t gid;
  struct passwd *ps;
  struct group *gr;
  xsltStylesheetPtr cur = NULL;
  xmlDocPtr doc = NULL;
  xmlDocPtr res = NULL;
  char *xmlstr;
  mode_t oldmode = 0;
  char *dname = NULL;
  char oldpath[PATH_MAX];
  struct stat st;
  unsigned char is_dir = 0;

  uid_t myid = getuid();
  uid_t mygid = getgid();

  if(getcwd(oldpath, PATH_MAX)) {
    return 10;
  }

  /* set mode for the output files */
  oldmode = umask(in_p->mode);

  if(stat(in_p->path, &st)) {
    retval = 10;
    goto badwrite;
  }

  /* change directory to the <path> */
  if(!S_ISDIR(st.st_mode)) {
    free(dname);
    dname = strdup(dirname(in_p->path));
  } else {
    dname = strdup(in_p->path);  
    is_dir = 1;
  }

  if(chdir(dname)) {
    retval = 10;
    goto badwrite;
  }
  
  if (in_p->flags & FILEBB_SETUSER_FLAG) {
    if (in_p->uname) {
#ifdef CARP
#warning NOT THREAD SAFE!!! use getpwnam_r instead
#endif
      if ((ps = getpwnam(in_p->uname))) {
	uid = ps->pw_uid;
      } else {
	uid = -1;
      }
    } else {
      uid = in_p->uid;
    }
  } else {
    uid = -1;
  }

  if (in_p->flags & FILEBB_SETGROUP_FLAG) {
    if (in_p->gname) {
#ifdef CARP
#warning NOT THREAD SAFE!!! use getgrnam_r instead
#endif
      if ((gr = getgrnam(in_p->gname))) {
	gid = gr->gr_gid;
      } else {
	gid = -1;
      }
    } else {
      gid = in_p->gid;
    }
  } else {
    gid = -1;
  }  

  /* Call libxslt here */

  xmlSubstituteEntitiesDefault(1);

  /*  xmlLoadExtDtdDefaultValue = 0; */

  cur = xsltParseStylesheetFile((const xmlChar *) in_p->stylesheetpath);

  if (cur == NULL)
    goto badwrite;

  if (cur->indent == 1)
    xmlIndentTreeOutput = 1;
  else
    xmlIndentTreeOutput = 0;

  xmlstr = AdmContext_toXML (context, error);

  doc = xmlParseMemory(xmlstr, strlen(xmlstr));
  if (doc == NULL) {
    fprintf(stderr, "unable to parse context\n");
    goto badwrite;
  }

  if (in_p->flags & FILEBB_SETUSER_FLAG)
    setreuid(uid, -1);
  if (in_p->flags & FILEBB_SETGROUP_FLAG)
    setregid(gid, -1);

  /* This will write file, if splitting is set */
  res = xsltApplyStylesheet(cur, doc, 0);

  xmlFreeDoc(doc);

  if (res == NULL) {
    goto badwrite;
  }

  /* We should specify the output file in the xsl stylesheet */
  if(!is_dir) {
    xsltSaveResultToFilename(in_p->path, res, cur, 0);
  } 

 badwrite:

  chdir(oldpath);

  if (in_p->flags & FILEBB_SETUSER_FLAG)
    setreuid(myid, -1);
  if (in_p->flags & FILEBB_SETGROUP_FLAG)
    setregid(mygid, -1);

  if (res)
    xmlFreeDoc(res);

  if (cur)
    xsltFreeStylesheet(cur);

  xmlCleanupParser();
  xmlMemoryDump();

  if(oldmode)
    umask(oldmode);

  if(dname)
    free(dname);

  return retval;
}

static int
forgeblackbox_getBox (AdmBlackBoxMethods **methods,
		      void  **data,
		      AdmList *box_list,
		      AdmError **error)
{
	AdmRefStr *path, *uname, *gname, *stylesheetpath;
	int flags;
	int mode;
	uid_t uid;
	gid_t gid;
	AdmData *childA, *childB;
	struct forgeblackbox_instance *in_p;

	/* to shut up the compiler */
	forgeblackbox_methods.read = NULL;

	flags = 0;
	*methods = NULL;
	*data = NULL;

	path = NULL;
	stylesheetpath = NULL;
	uname = NULL;
	gname = NULL;

	/* path */
	childA = AdmList_getDataByName_b(box_list, "path", NULL);
	if ((childA == NULL) || (AdmData_getType(childA, NULL) != ADM_TYPE_STRING)) {

		/* if old configfilepath is set, override path */
		childA = AdmList_getDataByName_b(box_list, "configfilepath", NULL);

		if ((childA == NULL) || (AdmData_getType(childA, NULL) != ADM_TYPE_STRING)) {
			goto badconf;
		}
		path = AdmString_getValue((AdmString *) childA, NULL);
	}
	path = AdmString_getValue((AdmString *) childA, NULL);

	/* write file mode */
	if ((childA = AdmList_getDataByName_b(box_list, "mode", NULL))) {
		if (AdmData_getType(childA, NULL) != ADM_TYPE_INT) 
			goto badconf;

		mode = AdmInt_getValue((AdmInt *) childA, NULL);
	} else {
		goto badconf;
	}

	/* uid/uname (mutually exclusive) */
	uid = 0;
	uname = NULL;
	childA = AdmList_getDataByName_b(box_list, "uid", NULL);
	childB = AdmList_getDataByName_b(box_list, "uname", NULL);

	if (childA) {
		if (childB || (AdmData_getType(childA, NULL) != ADM_TYPE_INT))
			goto badconf;

		uid = AdmInt_getValue((AdmInt *) childA, NULL);
		flags |= FILEBB_SETUSER_FLAG;
	} else if (childB) {
		if (AdmData_getType(childB, NULL) != ADM_TYPE_STRING)
			goto badconf;

		uname = AdmString_getValue((AdmString *) childB, NULL);
		flags |= FILEBB_SETUSER_FLAG;
	}

	/* gid/gname (mutually exclusive) */
	gid = 0;
	gname = NULL;
	childA = AdmList_getDataByName_b(box_list, "gid", NULL);
	childB = AdmList_getDataByName_b(box_list, "gname", NULL);

	if (childA) {
		if (childB || (AdmData_getType(childA, NULL) != ADM_TYPE_INT))
			goto badconf;

		gid = AdmInt_getValue((AdmInt *) childA, NULL);
		flags |= FILEBB_SETGROUP_FLAG;
	} else if (childB) {
		if (AdmData_getType(childB, NULL) != ADM_TYPE_STRING)
			goto badconf;

		gname = AdmString_getValue((AdmString *) childB, NULL);
		flags |= FILEBB_SETGROUP_FLAG;
	}

	/* path */
	childA = AdmList_getDataByName_b(box_list, "xslstylesheetpath", NULL);
	if ((childA == NULL) || (AdmData_getType(childA, NULL) != ADM_TYPE_STRING)) {
		goto badconf;
	}
	stylesheetpath = AdmString_getValue((AdmString *) childA, NULL);

	/* allocate instance */
	in_p = (struct forgeblackbox_instance *) malloc(sizeof (struct forgeblackbox_instance));

	in_p->path = strdup(AdmRefStr_str(path));
	in_p->stylesheetpath = strdup(AdmRefStr_str(stylesheetpath));
	in_p->flags = flags;
	in_p->mode = mode;
	in_p->uid = uid;
	in_p->gid = gid;
	in_p->uname = (flags & FILEBB_SETUSER_FLAG) ? strdup(AdmRefStr_str(uname)) : NULL;
	in_p->gname = (flags & FILEBB_SETGROUP_FLAG) ? strdup(AdmRefStr_str(gname)) : NULL;
 
	*data = (void *) in_p;
	return 0;

badconf:
 	if (path)
		AdmRefStr_unref(path);
 	if (stylesheetpath)
		AdmRefStr_unref(stylesheetpath);
	if (uname)
		AdmRefStr_unref(uname);
	if (gname)
		AdmRefStr_unref(gname);
	return -1;

}
