/**
 * Copyright (C) 2000, 2001 Red Hat, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/**
 * @brief Adm dynamic blackbox loader
 *
 * In order to use and work with extentable BlackBoxes we use a dynamic loader
 * so other forms of BlackBoxes can be written and simply dynamically loaded.
 *
 * All the functions in this file except the getBox() are private and used
 * for housekeeping.
 *
 * @file        getBox.c
 * @author      Crutcher Dunnavant <crutcher@redhat.com>
 * @date        05/06/2001
 * @version     1.0
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <glib.h>
#include <gmodule.h>

#include "alchemist.h"
#include "blackbox.h"
#include "switchboard.h"
#include "admerror.h"
#include "admcore.h"
#include "internalbb.h"

#define INLINE __inline__

/*****************************************************************************/
/* Private Prototypes */

/* my list type, for keeping up with modules */
typedef struct _mod_refstr mod_refstr;
struct _mod_refstr {
	mod_refstr			*next;
	char				*mod_type;	/* case squashed modtype */
	char				*mod_path;	/* no real reason to keep this */
	GModule				*gmod_p;		/* pointer to the module */
	struct AdmBlackBoxModuleVector  *mod_vector;	/* pointer to the mod's vector */
};

static INLINE mod_refstr * _adm_swb_get_mod		(const char *mod_type);
static INLINE mod_refstr * _adm_swb_seek_mod_nolock	(const char *mod_type);
static INLINE mod_refstr * _adm_swb_load_mod_nolock	(const char *mod_type);

static INLINE int         _adm_swb_cmp_istr_lstr	(const char *i_str,
							 const char *l_str);
static INLINE void        _adm_swb_lower_str	(char       *str);

static INLINE AdmBlackBox *
_adm_swb_new_bb (AdmBlackBoxMethods	*bb_methods,
		void *bb_data,
		AdmRefStr *cache_id,
		uid_t	cache_uid,
		gid_t	cache_gid,
		unsigned int cache_depth);
/*****************************************************************************/
/* local Globals */

/* my list of modules */
static mod_refstr *mod_list_head = NULL;

/* a mutex to protect my module list from threads */
static GStaticMutex mod_mutex = G_STATIC_MUTEX_INIT;
#define MOD_LOCK()	g_static_mutex_lock(&mod_mutex)
#define MOD_UNLOCK()	g_static_mutex_unlock(&mod_mutex)


struct {
	const char *box_type;
	int   (*getBox) (AdmBlackBoxMethods    **mod,
			 void		       **data,
			 AdmList		*box_list,
			 AdmError **error);
} _adm_swb_getBox_list[] = {
	{"triggerblackbox", _adm_triggerblackbox_getBox},
	{"fileblackbox", _adm_fileblackbox_getBox},
	{NULL, NULL}
};

/*****************************************************************************/
/* private functions */ 

/*
* _adm_swb_cmp_istr_lstr
*
*  performs string comparisons. the 'i' string is treated case insesitvely, and
*  the 'l' string is assumed to be lower case. This is used for comparing
*  mod_types, because:
*  A) some OSs have case insensitive file names, and the mod type is used to
*     construct a path to the library.
*  B) the C lib caseless comparisons are locale aware, and having library
*     lookup change behavior in different locales is not acceptable.
*/
#define ISUPPERASCII(C) ((C >= 'A') && (C <= 'Z'))
static INLINE int
_adm_swb_cmp_istr_lstr (const char *i_str, const char *l_str) {
	char i, l;
	while (i = *i_str, l = *l_str, i && l) {
		if (ISUPPERASCII(i)) i |= 0x20;
		if (i != l) break;
		++ i_str; ++ l_str;
	}
	return i - l;
}

/*
* _adm_swb_lower_str
*
* lower cases a string. implemented because the C libs change their
* understanding of this based upon locale.
*/
static INLINE void
_adm_swb_lower_str (char *str) {
	char c;
	for ( ; (c = *str) ; ++ str) if (ISUPPERASCII(c)) *str = c | 0x20;
}

/*
* _adm_swb_seek_mod_nolock
*
* scans the mod list, looking for a module with a matching mod type.
* not thread safe, so needs to be inside a lock.
*/
static INLINE mod_refstr *
_adm_swb_seek_mod_nolock (const char *mod_type) {
	mod_refstr *mod_tok_p;

	mod_tok_p = mod_list_head;
	while (mod_tok_p && _adm_swb_cmp_istr_lstr(mod_type, mod_tok_p->mod_type))
	       mod_tok_p = mod_tok_p->next;
	return mod_tok_p;
}

/*
* _adm_swb_load_mod_nolock
*
* trys to load a module matching modtype. adds it to the list if available,
* and returns a pointer to the mods taken. not thread safe, so needs to be
* inside a lock.
 */
static INLINE mod_refstr *
_adm_swb_load_mod_nolock (const char *mod_type) {
	struct AdmBlackBoxModuleVector  *vec;
	GModule				*gmod_p;
	mod_refstr			*mod_tok_p;
	char				*l_mod_type,
					*path;
	
	/* copy the module type, and lowercase the copy */
	l_mod_type = strdup(mod_type);
	_adm_swb_lower_str(l_mod_type);
	
	/* construct a path to where the module should be */
	path = g_module_build_path(ALCHEMIST_BLACKBOX_MOD_PATH, l_mod_type);

	/* try to load the module
	 * We use G_MODULE_BIND_LAZY so that we have a chance of cleanly dropping a
	 * module that has the wrong ABI nuwbers. Otherwise, it would always be nasty */
	if ((gmod_p = g_module_open(path, G_MODULE_BIND_LAZY)) == NULL)
		goto failload;

	/* try to find the 'bb_mod_vector' symbol */
	if (!g_module_symbol(gmod_p, "bb_mod_vector", (gpointer *) &vec))
		goto failsymbol;

	/* check the abi version */
	if (vec->abi_major_ver != ALCHEMIST_ABI_MAJOR_VER)
		goto failsymbol;

	if (vec->abi_minor_ver > ALCHEMIST_ABI_MINOR_VER)
		goto failsymbol;

	/* Yay! we've got a valid module.
	* Make a new mod refstr */
	mod_tok_p = (mod_refstr *) malloc(sizeof (mod_refstr));

	/* setup the new refstr */
	mod_tok_p->mod_type = l_mod_type;
	mod_tok_p->gmod_p = gmod_p;
	mod_tok_p->mod_path = path;
	mod_tok_p->mod_vector = vec;

	/* stick the new refstr on the list */
	mod_tok_p->next = mod_list_head;
	mod_list_head = mod_tok_p;

	return mod_tok_p;

failsymbol:
	/* unload the module */
	g_module_close(gmod_p);
failload:
	/* free my strings */
	free(l_mod_type);
	free(path);

	return NULL;
}

/*
* _adm_swb_get_mod
*
* locks the mod list, then seeks for a module to match mod_type,
* or trys to load one if that fails.
*/
static INLINE mod_refstr *
_adm_swb_get_mod (const char *mod_type) {
	mod_refstr *mod_tok_p;

	/* call this to protect the module list */
	MOD_LOCK();

	/* check for the existance of the module in the list */
	mod_tok_p = _adm_swb_seek_mod_nolock(mod_type);

	/* if it isnt there, try to load it */
	if (!mod_tok_p)
		mod_tok_p = _adm_swb_load_mod_nolock(mod_type);

	/* drop the module lock */
	MOD_UNLOCK();

	return mod_tok_p;
}

static int
_adm_swb_valid_box_type (const char *box_type)
{
#ifdef FIXME
#warning VALIDATE THIS
#endif
	return 1;
}


static int
_adm_swb_valid_cache_id (const char *cache_id)
{
	return AdmData_validName(cache_id);
}

/*
 * _adm_swb_get_getBox
 *
 * trys to find a getBox function to handle the box_type passed.
 * searches the compiled in list, then goes off and asks the module layer
 */
static int (*_adm_swb_get_getBox(const char *box_type)) (AdmBlackBoxMethods **mod,
							 void		     **data,
							 AdmList	      *box_list,
							 AdmError **error)
{
	unsigned int i;
	mod_refstr *mod_tok_p;
	int   (*getBox) (AdmBlackBoxMethods    **mod,
			 void		       **data,
			 AdmList		*box_list,
			 AdmError **error);

	/* we are searching for a getBox function for this box_type
	* there are 2 places to look: */
	do {
		/* 1) the static list */
		for (i = 0; _adm_swb_getBox_list[i].box_type != NULL; ++i)
		if (_adm_swb_cmp_istr_lstr(box_type, _adm_swb_getBox_list[i].box_type) == 0) {
			getBox = _adm_swb_getBox_list[i].getBox;
			goto ret;
		}
	
		/* 2) the dynamic modules */
		mod_tok_p = _adm_swb_get_mod(box_type);
		if (mod_tok_p != NULL) {
			getBox = mod_tok_p->mod_vector->getBox;
			goto ret;
		}

		getBox = NULL;
	} while (0);
ret:
	return getBox;
}

AdmBlackBox *
_adm_swb_new_bb (AdmBlackBoxMethods	*bb_methods,
		void *bb_data,
		AdmRefStr *cache_id,
		uid_t	cache_uid,
		gid_t	cache_gid,
		unsigned int cache_depth)
{
	AdmBlackBox *box;
	adm_bb_cap_t caps;

	/* Allocate the box */
	box = malloc(sizeof (AdmBlackBox));

	/* clear it */
	*box = (AdmBlackBox) { NULL, };

	/* Give it an initial reference */
	box->refcount = 1;

	/* Attach methods and data */
	box->methods = bb_methods;
	box->data = bb_data;

	/* set up cacheing */
	box->cache_id = (cache_id) ? AdmRefStr_ref(cache_id) : NULL;
	box->cache_depth = cache_depth;
	box->cache_uid = cache_uid;
	box->cache_gid = cache_gid;

	/* Calculate the box's capabilities */
	caps = (adm_bb_cap_t) {{0},};
	
	if (box->methods->write != NULL) {
		ADM_BBCAP_SET_SUP_WRITE(caps);
	}

	if (box->methods->read != NULL) {
		ADM_BBCAP_SET_SUP_READ(caps);

		if (box->methods->listSerial != NULL) {
			ADM_BBCAP_SET_SUP_LISTSERIAL(caps);
		}

		if (box->methods->readSerial != NULL) {
			ADM_BBCAP_SET_SUP_READSERIAL(caps);
		}
	}

	if (box->cache_id != NULL) {
		ADM_BBCAP_SET_CACHED(caps);
	}

	box->caps = caps;

	return box;
}


/*****************************************************************************/
/* Public Functions */

AdmBlackBox *
AdmBlackBox_getBox (AdmList      *box_config,
		    AdmError **error)
{
	AdmBlackBox *box;
	AdmBlackBoxMethods *bb_methods;
	void *bb_data;
	AdmList *cache_list, *box_list;
	AdmRefStr *box_type, *cache_id;
	AdmString *string;
	AdmData *data;
	unsigned int cache_depth;
	uid_t cache_uid;
	gid_t cache_gid;
	int   (*getBox) (AdmBlackBoxMethods    **mod,
			 void		       **data,
			 AdmList		*box_list,
			 AdmError **error);

	/* Sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(box_config != NULL);
	if (!_adm_type_and_lifecycle_check((AdmData *) box_config, ADM_TYPE_LIST, error))
		return NULL;

	box = NULL;
	box_type = NULL;
	cache_id = NULL;
	cache_uid = -1;
	cache_gid = -1;

	/* validate the optional cache list */
	cache_list = (AdmList *) AdmList_getDataByName_b(box_config, "cache", NULL);
	if (cache_list == NULL) {
		cache_id = NULL;
		cache_depth = 0;
	} else {
		if (AdmData_getType((AdmData *) cache_list, NULL) != ADM_TYPE_LIST) {
			AdmError_create(error, ADM_EBADDATA, _("box_cfg/cache element is not a list"));
			goto fail;
		}

		
		string = (AdmString *) AdmList_getDataByName_b(cache_list, "id", NULL);
		if (string == NULL) {
			AdmError_create(error, ADM_EBADDATA, _("box_cfg/cache/id element does not exist, but box_cfg/cache does"));
			goto fail;

		} else if (AdmData_getType((AdmData *) string, NULL) != ADM_TYPE_STRING) {
			AdmError_create(error, ADM_EBADDATA, _("box_cfg/cache/id element is not a string"));
			goto fail;
		}
		
		cache_id = AdmString_getValue(string, NULL);
		if (! _adm_swb_valid_cache_id(AdmRefStr_str(cache_id)) ) {
			AdmError_create(error, ADM_EBADDATA, _("box_cfg/cache/id invalid cache id"));
			goto fail;
		}

		
		data = AdmList_getDataByName_b(cache_list, "depth", NULL);
		if (data == NULL) {
			AdmError_create(error, ADM_EBADDATA,
				_("box_cfg/cache/depth element does not exist, but box_cfg/cache does"));
			goto fail;

		} else if (AdmData_getType(data, NULL) != ADM_TYPE_INT) {
			AdmError_create(error, ADM_EBADDATA, _("box_cfg/cache/depth element is not a int"));
			goto fail;
		}
		cache_depth = AdmInt_getValue((AdmInt *) data, NULL);

		
		/* cache uid/gid goes here */
		

	}

	/* validate the mandatory box list */
	box_list = (AdmList *) AdmList_getDataByName_b(box_config, "box", NULL);
	if (box_list == NULL) {
		AdmError_create(error, ADM_EBADDATA, _("box_cfg/box does not exist"));
		goto fail;
	} else {
		if (AdmData_getType((AdmData *) box_list, NULL) != ADM_TYPE_LIST) {
			AdmError_create(error, ADM_EBADDATA, _("box_cfg/box element is not a list"));
			goto fail;
		}

		string = (AdmString *) AdmList_getDataByName_b(box_list, "box_type", NULL);
		if (string == NULL) {
			AdmError_create(error, ADM_EBADDATA, _("box_cfg/box_box_type element does not exist"));
			goto fail;

		} else if (AdmData_getType((AdmData *) string, NULL) != ADM_TYPE_STRING) {
			AdmError_create(error, ADM_EBADDATA, _("box_cfg/box/box_type element is not a string"));
			goto fail;
		}

		box_type = AdmString_getValue(string, NULL);
		if (! _adm_swb_valid_box_type(AdmRefStr_str(box_type)) ) {
			AdmError_create(error, ADM_EBADDATA, _("box_cfg/box/box_type invalid box_type"));
			goto fail;
		}

	}

	/* we've got a box_type and everything looks kosher. ask for a getBox function */
	getBox = _adm_swb_get_getBox(AdmRefStr_str(box_type));
	if (getBox == NULL) {
		/* crap. We dont have anything to handle this box_type */
		AdmError_create(error, ADM_ENOBOXTYPE, _("No handler found for box_type: \"%s\""), box_type);
		goto fail;
	}

	/* attempt to pull a box out of the getBox function */
	bb_methods = NULL;
	bb_data = NULL;
	if (getBox(&bb_methods, &bb_data, box_list, error) != 0)
		goto fail;


	/* we got a box! build it */
	box = _adm_swb_new_bb(bb_methods, bb_data, cache_id, cache_uid, cache_gid, cache_depth);

	/* success! */
	return box;

fail:
	if (box_type)
		AdmRefStr_unref(box_type);
	if (cache_id)
		AdmRefStr_unref(cache_id);

	return NULL;
}

