/**
 * Copyright (C) 2000, 2001, 2004 Red Hat, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/**
 * @brief Adm 2nd generation Python bindings
 *
 * As the 1st generation of Pyton bindings for the Alchemist turned out to be
 * usable but not very Python like we have decided to implement a new set of
 * bindings for the next release.
 *
 * The advantage of the new bindings is the more Python like style of access
 * and use of the Alchemist instead of a 1:1 export of the original C API.
 *
 * The new features include:
 *  - Direct object interface (no need for a Python class module anymore).
 *  - Python operators for sequences, lists and attributes.
 *  - Hiding of the original Alchemist C API.
 *
 * @file        pyalchemist.c
 * @author      Crutcher Dunnavant <crutcher@redhat.com>
 * @author      Philipp Knirsch <pknirsch@redhat.com>
 * @date        05/06/2001
 * @version     1.0
 */

#include "alchemist.h"
#include "Python.h"

static PyObject		*pyAdmContext_constructor	(AdmContext		 *ctx);

static PyTypeObject pyAdmBool_object_type;
#define pyAdmBool_Check(op) ((op)->ob_type == &pyAdmBool_object_type)
static PyTypeObject pyAdmInt_object_type;
#define pyAdmInt_Check(op) ((op)->ob_type == &pyAdmInt_object_type)
static PyTypeObject pyAdmFloat_object_type;
#define pyAdmFloat_Check(op) ((op)->ob_type == &pyAdmFloat_object_type)
static PyTypeObject pyAdmString_object_type;
#define pyAdmString_Check(op) ((op)->ob_type == &pyAdmString_object_type)
static PyTypeObject pyAdmBase64_object_type;
#define pyAdmBase64_Check(op) ((op)->ob_type == &pyAdmBase64_object_type)
static PyTypeObject pyAdmCopy_object_type;
#define pyAdmCopy_Check(op) ((op)->ob_type == &pyAdmCopy_object_type)
static PyTypeObject pyAdmList_object_type;
#define pyAdmList_Check(op) ((op)->ob_type == &pyAdmList_object_type)

#define pyAdmData_Check(__OP) \
({ \
	PyObject *op = (__OP); \
	pyAdmBool_Check(op) || pyAdmInt_Check(op) || pyAdmFloat_Check(op) || \
	pyAdmString_Check(op) || pyAdmBase64_Check(op) || pyAdmCopy_Check(op) || pyAdmList_Check(op); \
})

static PyTypeObject pyAdmIdentity_object_type;
#define pyAdmIdentity_Check(op) ((op)->ob_type == &pyAdmIdentity_object_type)
static PyTypeObject pyAdmContext_object_type;
#define pyAdmContext_Check(op) ((op)->ob_type == &pyAdmContext_object_type)

static PyTypeObject pyAdmBlackBox_object_type;
#define pyAdmBlackBox_Check(op) ((op)->ob_type == &pyAdmBlackBox_object_type)

/* ================================================================================================================== */
/** AdmError helpers **/

int		pyAdmError_check_and_raise	(AdmError *error);

int
pyAdmError_check_and_raise (AdmError *error)
{
	PyObject *exc;
	const char *str;
	int num;
	int ret;

	ret = 0;

	if (error == NULL)
		return ret;

	num = AdmError_errno(error);
	str = AdmError_strerr(error);
	switch (num) {
		case 0:
			/* wtf? a non-error error? */
			AdmError_unref(error);
			return 0;
			break;

		case ADM_ENOINDEX:
			exc = PyExc_IndexError;
			break;
		case ADM_ENONAME:
		case ADM_ENOPATH:
		case ADM_ENOSUBSPACE:
		case ADM_ENOBOXTYPE:
		case ADM_ENONAMESPACE:
			exc = PyExc_KeyError;
			break;
		case ADM_EBADTYPE:
		case ADM_ENOWRITE:
		case ADM_ENOREAD:
		case ADM_ENOREADSERIAL:
		case ADM_ENOLISTSERIAL:
		case ADM_ENOCACHE:
			exc = PyExc_TypeError;
			break;
		case ADM_EBADDATA:
		case ADM_EPARSE:
			exc = PyExc_ValueError;
			break;
		case ADM_EACCESSFAILED:
			exc = PyExc_IOError;
			break;
		default:
			exc = PyExc_RuntimeError;
			break;
	}

	PyErr_SetString(exc, str);
	AdmError_unref(error);
	return -1;
}

/* ================================================================================================================== */
/** AdmData **/

typedef struct {
	PyObject_HEAD
	AdmData *data;
	AdmContext *ctx;
} pyAdmData_object;
/* funcction prototypes */

static PyObject		*pyAdmData_constructor			(AdmData		 *data);

static void		 pyAdmData_free				(pyAdmData_object	 *self);

static long		 pyAdmData_hash				(pyAdmData_object	 *self);

static int		 pyAdmData_cmp				(pyAdmData_object	 *self,
								 PyObject		 *obj);

static int		 pyAdmData_getattr			(pyAdmData_object	 *self,
								 char			 *name,
								 PyObject		**retobj);

static int		 pyAdmData_setattr			(pyAdmData_object	 *self,
								 char			 *name,
								 PyObject		 *args);

static char		 pyAdmData_validName_doc[];
static PyObject		*pyAdmData_validName			(PyObject		*self,
								 PyObject		*args,
								 PyObject		*kwargs);

static char		 pyAdmData_unlink_doc[];
static PyObject		*pyAdmData_unlink			(pyAdmData_object	 *self,
								 PyObject		 *args);

static char		 pyAdmData_getPrev_doc[];
static PyObject		*pyAdmData_getPrev			(pyAdmData_object	 *self,
								 PyObject		 *args,
								 PyObject		 *kwargs);

static char		 pyAdmData_getNext_doc[];
static PyObject		*pyAdmData_getNext			(pyAdmData_object	 *self,
								 PyObject		 *args,
								 PyObject		 *kwargs);

#define pyAdmData_doc_macro \
"	In addition, they posses the following common AdmData attributes\n" \
"	and functions:\n" \
"\n" \
"	data.alive:\n" \
"	is this data dead; are accesses to this data legal?\n" \
"\n" \
"	data.source: assignable\n" \
"	data.name: assignable\n" \
"	the data element's name\n" \
"\n" \
"	data.source: assignable\n" \
"	the name of the source context of the data element\n" \
"\n" \
"	data.context:\n" \
"	the context of the data element\n" \
"\n" \
"	data.pos: assignable\n" \
"	the position that the data element holds in its parent list\n" \
"\n" \
"	data.protected: assignable\n" \
"	the protection status of the data element. At merge time, a\n" \
"	protected 'A' element will override a 'B' element, and an\n" \
"	unprotected 'A' element will be overriden by a 'B' element.\n" \
"\n" \
"	data.unlink()\n" \
"	unlinks the data element from its parent and context\n" \
"\n" \
"	data.prev(wrap = 0) -> prevdata\n" \
"	returns the previous sibling of an element, wraps arround the\n" \
"	begining of the list if wrap is set\n" \
"\n" \
"	data.next(wrap = 0) -> nextdata\n" \
"	returns the next sibling of an element, wraps arround the end\n" \
"	of the list if wrap is set"; 

static struct PyMethodDef pyAdmData_methods[] = {
	{"unlink",	(PyCFunction) pyAdmData_unlink,		METH_VARARGS,
								pyAdmData_unlink_doc},
	{"prev",	(PyCFunction) pyAdmData_getPrev,	METH_VARARGS | METH_KEYWORDS,
								pyAdmData_getPrev_doc},
	{"next",	(PyCFunction) pyAdmData_getNext,	METH_VARARGS | METH_KEYWORDS,
								pyAdmData_getNext_doc},
	{NULL, NULL}
};

static struct PyMethodChain pyAdmData_method_chain = {
	methods:	pyAdmData_methods,
	link:		NULL,
};


/* ------------------------------------------------------------------*/
/* Base Functions for pyAdmData_object */

static PyObject *
//pyAdmData_constructor (AdmData *data, PyTypeObject *type)
pyAdmData_constructor (AdmData *data)
{
	AdmContext *ctx;
	pyAdmData_object *pydata;
	PyTypeObject *type;
	AdmDataType dtype;
	AdmError *error;

	error = NULL;
	dtype = AdmData_getType(data, &error);
	if (pyAdmError_check_and_raise(error))
		return NULL;

	switch (dtype) {
		case ADM_TYPE_BOOL:
			type = &pyAdmBool_object_type;
			break;
		case ADM_TYPE_INT:
			type = &pyAdmInt_object_type;
			break;
		case ADM_TYPE_FLOAT:
			type = &pyAdmFloat_object_type;
			break;
		case ADM_TYPE_STRING:
			type = &pyAdmString_object_type;
			break;
		case ADM_TYPE_BASE64:
			type = &pyAdmBase64_object_type;
			break;
		case ADM_TYPE_COPY:
			type = &pyAdmCopy_object_type;
			break;
		case ADM_TYPE_LIST:
			type = &pyAdmList_object_type;
			break;
		default:
			return PyErr_Format(PyExc_RuntimeError, "Unknown AdmType");
	}

	ctx = AdmData_getContext_b(data, &error);
	if (pyAdmError_check_and_raise(error))
		return NULL;

	if (error != NULL) {
		PyErr_SetString(PyExc_RuntimeError, AdmError_strerr(error));
		AdmError_unref(error);
		return NULL;
	}

	pydata = PyObject_NEW(pyAdmData_object, type);
	if (pydata == NULL) {
		PyErr_SetString(PyExc_RuntimeError, "Failed to allocate a pyAdmData_object");
		return NULL;
	}

	pydata->data = data;
	AdmData_ref(data, NULL);

	pydata->ctx = ctx;
	AdmContext_ref(pydata->ctx, NULL);

	return (PyObject *) pydata;
}

/**
* pyAdmData_free : the destructor for pyAdmData_object
**/
static void
pyAdmData_free (pyAdmData_object *self)
{
	/* drop the ref on the data */
	AdmData_unref((AdmData *) self->data, NULL);
	AdmContext_unref(self->ctx, NULL);
	PyMem_DEL(self);
}


/**
* pyAdmData_hash : give a value we can hash against
**/
static long
pyAdmData_hash (pyAdmData_object *self)
{
	return (long) self->data;
}

/**
* pyAdmData_cmp : comparison handler
**/
static int
pyAdmData_cmp (pyAdmData_object	 *self,
	       PyObject		 *obj)
{
	if (!pyAdmData_Check(obj))
		return (int) self - (int) obj;
	else
		return (int) self->data - (int) ((pyAdmData_object *) obj)->data;
}


/**
* pyAdmData_getattr : handles x.attr calls
**/
static int
pyAdmData_getattr (pyAdmData_object *self, char *name, PyObject **retobj) 
{
	AdmError *error;
	int i;
	AdmContext *ctx;
	AdmRefStr *refstr;

	*retobj = NULL;
	error = NULL;
	if (!strcmp(name, "name")) {

		refstr = AdmData_getName(self->data, &error);
		if (pyAdmError_check_and_raise(error)) {
			*retobj = NULL;
			goto ret_handled;
		} else {
			*retobj = Py_BuildValue("s", AdmRefStr_str(refstr));
			AdmRefStr_unref(refstr);
			goto ret_handled;
		}

	} else if (!strcmp(name, "source")) {

		refstr = AdmData_getSource(self->data, &error);
		if (pyAdmError_check_and_raise(error)) {
			*retobj = NULL;
			goto ret_handled;
		} else {
			*retobj = Py_BuildValue("s", AdmRefStr_str(refstr));
			AdmRefStr_unref(refstr);
			goto ret_handled;
		}

	} else if (!strcmp(name, "alive")) {

		i = AdmData_isAlive(self->data, &error);
		if (pyAdmError_check_and_raise(error)) {
			*retobj = NULL;
			goto ret_handled;
		} else {
			*retobj = Py_BuildValue("i", i);
			goto ret_handled;
		}

	} else if (!strcmp(name, "protected")) {

		i = AdmData_isProtected(self->data, &error);
		if (pyAdmError_check_and_raise(error)) {
			*retobj = NULL;
			goto ret_handled;
		} else {
			*retobj = Py_BuildValue("i", i);
			goto ret_handled;
		}

	} else if (!strcmp(name, "context")) {

		ctx = AdmData_getContext_b(self->data, &error);
		if (pyAdmError_check_and_raise(error)) {
			*retobj = NULL;
			goto ret_handled;
		} else {
			*retobj = pyAdmContext_constructor(ctx);
			goto ret_handled;
		}
	} else if (!strcmp(name, "pos")) {

		i = AdmData_getPos(self->data, &error);
		if (pyAdmError_check_and_raise(error)) {
			*retobj = NULL;
			goto ret_handled;
		} else {
			*retobj = Py_BuildValue("i", i);
			goto ret_handled;
		}
	}

	return 0;
ret_handled:
	return 1;
}

/**
* pyAdmData_setattr : handles x.attr = v calls
**/
static int
pyAdmData_setattr (pyAdmData_object *self, char *name, PyObject *args) 
{
	AdmError *error;
	AdmList *list;
	PyObject *tmpobj;
	const char *str;
	int i;

	error = NULL;
	if (!strcmp(name, "name")) {
		if (!PyString_Check(args))
			return -1;

		str = PyString_AsString(args);

		AdmData_setName(self->data, str, &error);
		if (pyAdmError_check_and_raise(error))
			return -1;
		else
			return 1;

	} else if (!strcmp(name, "source")) {
		if (!PyString_Check(args))
			return -1;

		str = PyString_AsString(args);

		AdmData_setSource(self->data, str, &error);
		if (pyAdmError_check_and_raise(error))
			return -1;
		else
			return 1;

	} else if (!strcmp(name, "protected")) {
		/* Coerce the argument into a number we can use */
		if (!PyNumber_Check(args))
			return -1;

		if ((tmpobj = PyNumber_Int(args)) == NULL)
			return -1;

		i = PyInt_AsLong(tmpobj);
		Py_DECREF(tmpobj);

		AdmData_setProtected(self->data, i, &error);
		if (pyAdmError_check_and_raise(error))
			return -1;
		else
			return 1;

	} else if (!strcmp(name, "pos")) {
		/* Coerce the argument into a number we can use */
		if (!PyNumber_Check(args))
			return -1;

		if ((tmpobj = PyNumber_Int(args)) == NULL)
			return -1;

		i = PyInt_AsLong(tmpobj);
		Py_DECREF(tmpobj);

		/* Get the container of the element to be moved */
		list = (AdmList *)AdmData_getContainer_b(self->data, &error);
		if (pyAdmError_check_and_raise(error))
			return -1;

		/* Now move the child */
		AdmList_moveData(list, self->data, i, &error);
		if (pyAdmError_check_and_raise(error))
			return -1;
		else
			return 1;
	}
	return 0;
}

/* -------------------------------------------------------------------------- */
/* Methods for AdmData types */

static char pyAdmData_validName_doc[] =
"validName(name) -- checks to see if a name is valid for use as an AdmData name.\n"
"Names must match: Name = (Letter | '_') (Letter | Digit | '_' | '-' | '.')*";

static PyObject *
pyAdmData_validName	(PyObject		*self,
			 PyObject		*args,
			 PyObject		*kwargs)
{
	const char *name;
	const char *kw[] = {"name", NULL};

	name = NULL;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "s", (char **) kw, &name))
		return NULL;

	return Py_BuildValue("i", AdmData_validName(name));
}

static char pyAdmData_unlink_doc[] = 
"AdmData.unlink() -- unlinks the data element from its parent and context";

static PyObject *
pyAdmData_unlink	(pyAdmData_object	*self,
			 PyObject		*args)
{
	AdmError *error;

	error = NULL;

	if (!PyArg_ParseTuple(args, ""))
		return NULL;

	AdmData_unlink(self->data, &error);
	if (pyAdmError_check_and_raise(error))
		return NULL;

	return Py_BuildValue("");

}

static char pyAdmData_getPrev_doc[] =
"AdmData.prev(wrap = 0) -> prevdata -- returns the previous sibling of an element,"
" wraps arround the begining of the list if wrap is set";

static PyObject *
pyAdmData_getPrev	(pyAdmData_object	*self,
			 PyObject		*args,
			 PyObject		*kwargs)
{
	AdmError *error;
	AdmData *data;
	PyObject *retobj;
	int wrap;
	const char *kw[] = {"wrap", NULL};

	error = NULL;
	wrap = 0;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|i", (char **) kw, &wrap))
		return NULL;

	data = AdmData_getPrev_b(self->data, wrap, &error);
	if (pyAdmError_check_and_raise(error))
		return NULL;

	if (data != NULL) 
		retobj = pyAdmData_constructor(data);
	else
		retobj = Py_BuildValue("");
	return retobj;
}

static char pyAdmData_getNext_doc[] =
"AdmData.next(wrap = 0) -> nextdata -- returns the next sibling of an element,"
" wraps arround the end of the list if wrap is set";

static PyObject *
pyAdmData_getNext	(pyAdmData_object	*self,
			 PyObject		*args,
			 PyObject		*kwargs)
{
	AdmError *error;
	AdmData *data;
	PyObject *retobj;
	int wrap;
	const char *kw[] = {"wrap", NULL};

	error = NULL;
	wrap = 0;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|i", (char **) kw, &wrap))
		return NULL;

	data = AdmData_getNext_b(self->data, wrap, &error);
	if (pyAdmError_check_and_raise(error))
		return NULL;

	if (data != NULL) 
		retobj = pyAdmData_constructor(data);
	else
		retobj = Py_BuildValue("");
	return retobj;
}

/* ================================================================================================================== */
/** AdmBool **/

/* funcction prototypes */
static PyObject		*pyAdmBool_getattr			(pyAdmData_object	 *self,
								 char			 *name);

static int		 pyAdmBool_setattr			(pyAdmData_object	 *self,
								 char			 *name,
								 PyObject		 *args);

static PyObject		*pyAdmBool_repr				(pyAdmData_object	 *self);

static PyObject		*pyAdmBool_str				(pyAdmData_object	 *self);

/** the pyAdmBool_type structure is a PyTypeObject struct, and holds our type callbacks **/
static PyTypeObject pyAdmBool_object_type = {
	PyObject_HEAD_INIT(&PyType_Type)		/** Required initialization				**/
	0,						/** ob_size		: Usually 0			**/
	"AdmBool",					/** tp_name		: Type Name			**/
	sizeof(pyAdmData_object),			/** tp_basicsize	: Object Size			**/
	0,						/** tp_itemsize		: Usually 0			**/

	/* Standard Methods */
	(destructor)	pyAdmData_free,			/** tp_dealloc		: refcount = 0			**/
	(printfunc)	NULL,				/** tp_print		: print x			**/
	(getattrfunc)	pyAdmBool_getattr,		/** tp_getattr		: x.attr			**/
	(setattrfunc)	pyAdmBool_setattr,		/** tp_setattr		: x.attr = v			**/
	(cmpfunc)	pyAdmData_cmp,			/** tp_compare		: cmp(x, y)			**/
	(reprfunc)	pyAdmBool_repr,			/** tp_repr		: repr(x)			**/

	/* Type Categories */
	NULL,						/** tp_as_number	: Number Methods		**/
	NULL,						/** tp_as_sequence	: Sequence methods		**/
	NULL,						/** tp_as_mapping	: Mapping functions		**/
	(hashfunc)	pyAdmData_hash,			/** tp_hash		: dict[x]			**/
	(ternaryfunc)	NULL,				/** tp_call		: x()				**/
	(reprfunc)	pyAdmBool_str,			/** tp_str		: str(x)			**/
};

static char pyAdmBool_doc[] =
"AdmBool elements have the following attributes and functions:\n"
"\n"
"	bool.value: assignable\n"
"	holds the boolean value of the AdmBool element\n"
"\n"
pyAdmData_doc_macro;


static struct PyMethodDef pyAdmBool_methods[] = {
	{NULL, NULL}
};

static struct PyMethodChain pyAdmBool_method_chain = {
	methods:	pyAdmBool_methods,
	link:		&pyAdmData_method_chain,
};

/* ------------------------------------------------------------------*/
/* Base Functions for pyAdmBool_object */


/**
* pyAdmBool_getattr : handles x.attr calls
**/
static PyObject *
pyAdmBool_getattr (pyAdmData_object *self, char *name) 
{
	AdmError *error;
	PyObject *retobj;
	int i;
	
	error = NULL;
	
	if (pyAdmData_getattr(self, name, &retobj))
		return retobj;

	if (!strcmp(name, "value")) {
		i = AdmBool_getValue((AdmBool *) self->data, &error);
		if (pyAdmError_check_and_raise(error))
			return NULL;

		return Py_BuildValue("i", i);
	} else if (!strcmp(name, "__doc__")) {
		return PyString_FromString(pyAdmBool_doc);
	}

	return Py_FindMethodInChain(&pyAdmBool_method_chain, (PyObject *) self, name);
}

/**
* pyAdmBool_setattr : handles x.attr = v calls
**/
static int
pyAdmBool_setattr (pyAdmData_object *self, char *name, PyObject *args) 
{
	AdmError *error;
	PyObject *tmpobj;
	int i;

	error = NULL;
	
	if ((i = pyAdmData_setattr(self, name, args)) != 0)
       		return i<0?-1:0;

	if (!strcmp(name, "value")) {
		/* Coerce the argument into a number we can use */
		if (!PyNumber_Check(args))
			return -1;

		if ((tmpobj = PyNumber_Int(args)) == NULL)
			return -1;

		i = PyInt_AsLong(tmpobj);
		Py_DECREF(tmpobj);

		AdmBool_setValue((AdmBool *) self->data, i, &error);
		if (pyAdmError_check_and_raise(error))
			return -1;

		return 0;
	}

	return -1;
}

/**
* pyAdmBool_str : handles str(x) calls
**/
static PyObject *
pyAdmBool_str (pyAdmData_object *self)
{
	AdmError *error;
	int i;

	error = NULL;
	
	i = AdmBool_getValue((AdmBool *) self->data, &error);
	if (pyAdmError_check_and_raise(error))
		return NULL;

	return PyString_FromString(i ? "true" : "false");
}


/**
* pyAdmBool_repr : handles repr(x) calls
**/
static PyObject *
pyAdmBool_repr (pyAdmData_object *self)
{
	AdmError *error;
	AdmRefStr *name, *source;
	char buf[256];
	int i;

	error = NULL;
	
	name = AdmData_getName(self->data, &error);
	if (pyAdmError_check_and_raise(error))
		return NULL;

	source = AdmData_getSource(self->data, &error);
	if (pyAdmError_check_and_raise(error)) {
		AdmRefStr_unref(name);
		return NULL;
	}

	i = AdmBool_getValue((AdmBool *) self->data, &error);
	if (pyAdmError_check_and_raise(error)) {
		AdmRefStr_unref(name);
		AdmRefStr_unref(source);
		return NULL;
	}

	snprintf(buf, 255, "<AdmBool, name=\"%s\", source=\"%s\", value=%s at %p>",
			AdmRefStr_str(name), AdmRefStr_str(source), i ? "true" : "false" , self);

	AdmRefStr_unref(name);
	AdmRefStr_unref(source);

	return PyString_FromString(buf);
}



/* ================================================================================================================== */
/** AdmInt **/

/* funcction prototypes */
static PyObject		*pyAdmInt_getattr			(pyAdmData_object	 *self,
								 char			 *name);

static int		 pyAdmInt_setattr			(pyAdmData_object	 *self,
								 char			 *name,
								 PyObject		 *args);

static PyObject		*pyAdmInt_repr				(pyAdmData_object	 *self);

static PyObject		*pyAdmInt_str				(pyAdmData_object	 *self);

/** the pyAdmInt_type structure is a PyTypeObject struct, and holds our type callbacks **/
static PyTypeObject pyAdmInt_object_type = {
	PyObject_HEAD_INIT(&PyType_Type)		/** Required initialization				**/
	0,						/** ob_size		: Usually 0			**/
	"AdmInt",					/** tp_name		: Type Name			**/
	sizeof(pyAdmData_object),			/** tp_basicsize	: Object Size			**/
	0,						/** tp_itemsize		: Usually 0			**/

	/* Standard Methods */
	(destructor)	pyAdmData_free,			/** tp_dealloc		: refcount = 0			**/
	(printfunc)	NULL,				/** tp_print		: print x			**/
	(getattrfunc)	pyAdmInt_getattr,		/** tp_getattr		: x.attr			**/
	(setattrfunc)	pyAdmInt_setattr,		/** tp_setattr		: x.attr = v			**/
	(cmpfunc)	pyAdmData_cmp,			/** tp_compare		: cmp(x, y)			**/
	(reprfunc)	pyAdmInt_repr,			/** tp_repr		: repr(x)			**/

	/* Type Categories */
	NULL,						/** tp_as_number	: Number Methods		**/
	NULL,						/** tp_as_sequence	: Sequence methods		**/
	NULL,						/** tp_as_mapping	: Mapping functions		**/
	(hashfunc)	pyAdmData_hash,			/** tp_hash		: dict[x]			**/
	(ternaryfunc)	NULL,				/** tp_call		: x()				**/
	(reprfunc)	pyAdmInt_str,			/** tp_str		: str(x)			**/
};

static char pyAdmInt_doc[] =
"AdmInt elements have the following attributes and functions:\n"
"\n"
"	int.value: assignable\n"
"	holds the integer value of the AdmInt element\n"
"\n"
pyAdmData_doc_macro;

static struct PyMethodDef pyAdmInt_methods[] = {
	{NULL, NULL}
};

static struct PyMethodChain pyAdmInt_method_chain = {
	methods:	pyAdmInt_methods,
	link:		&pyAdmData_method_chain,
};

/* ------------------------------------------------------------------*/
/* Base Functions for pyAdmInteger_object */

/**
* pyAdmInt_getattr : handles x.attr calls
**/
static PyObject *
pyAdmInt_getattr (pyAdmData_object *self, char *name) 
{
	AdmError *error;
	PyObject *retobj;
	int i;
	
	error = NULL;
	
	if (pyAdmData_getattr(self, name, &retobj)) {
		return retobj;

	} else if (!strcmp(name, "value")) {
		i = AdmInt_getValue((AdmInt *) self->data, &error);
		if (pyAdmError_check_and_raise(error))
			return NULL;

		return Py_BuildValue("i", i);

	} else if (!strcmp(name, "__doc__")) {
		return PyString_FromString(pyAdmInt_doc);
	}

	return Py_FindMethodInChain(&pyAdmInt_method_chain, (PyObject *) self, name);
}

/**
* pyAdmInt_setattr : handles x.attr = v calls
**/
static int
pyAdmInt_setattr (pyAdmData_object *self, char *name, PyObject *args) 
{
	AdmError *error;
	PyObject *tmpobj;
	int i;

	error = NULL;
	
	if ((i = pyAdmData_setattr(self, name, args)))
       		return i<0?-1:0;

	if (!strcmp(name, "value")) {
		/* Coerce the argument into a number we can use */
		if (!PyNumber_Check(args))
			return -1;

		if ((tmpobj = PyNumber_Int(args)) == NULL)
			return -1;

		i = PyInt_AsLong(tmpobj);
		Py_DECREF(tmpobj);

		AdmInt_setValue((AdmInt *) self->data, i, &error);
		if (pyAdmError_check_and_raise(error))
			return -1;

		return 0;
	}

	return -1;
}

/**
* pyAdmInt_repr : handles repr(x) calls
**/
static PyObject *
pyAdmInt_repr (pyAdmData_object *self)
{
	AdmError *error;
	AdmRefStr *name, *source;
	char buf[256];
	int i;

	error = NULL;
	
	name = AdmData_getName(self->data, &error);
	if (pyAdmError_check_and_raise(error))
		return NULL;

	source = AdmData_getSource(self->data, &error);
	if (pyAdmError_check_and_raise(error)) {
		AdmRefStr_unref(name);
		return NULL;
	}

	i = AdmInt_getValue((AdmInt *) self->data, &error);
	if (pyAdmError_check_and_raise(error)) {
		AdmRefStr_unref(name);
		AdmRefStr_unref(source);
		return NULL;
	}

	snprintf(buf, 255, "<AdmInt, name=\"%s\", source=\"%s\", value=%d at %p>",
			AdmRefStr_str(name), AdmRefStr_str(source), i, self);

	AdmRefStr_unref(name);
	AdmRefStr_unref(source);

	return PyString_FromString(buf);
}


/**
* pyAdmInt_str : handles str(x) calls
**/
static PyObject *
pyAdmInt_str (pyAdmData_object *self)
{
	AdmError *error;
	char buf[256];
	int i;

	error = NULL;
	
	i = AdmInt_getValue((AdmInt *) self->data, &error);
	if (pyAdmError_check_and_raise(error))
		return NULL;

	snprintf(buf, 255, "%d", i);

	return PyString_FromString(buf);
}


/* ================================================================================================================== */
/** AdmFloat **/

/* funcction prototypes */
static PyObject		*pyAdmFloat_getattr			(pyAdmData_object	 *self,
								 char			 *name);

static int		 pyAdmFloat_setattr			(pyAdmData_object	 *self,
								 char			 *name,
								 PyObject		 *args);

static PyObject		*pyAdmFloat_repr			(pyAdmData_object	 *self);

static PyObject		*pyAdmFloat_str				(pyAdmData_object	 *self);

/** the pyAdmFloat_type structure is a PyTypeObject struct, and holds our type callbacks **/
static PyTypeObject pyAdmFloat_object_type = {
	PyObject_HEAD_INIT(&PyType_Type)		/** Required initialization				**/
	0,						/** ob_size		: Usually 0			**/
	"AdmFloat",					/** tp_name		: Type Name			**/
	sizeof(pyAdmData_object),			/** tp_basicsize	: Object Size			**/
	0,						/** tp_itemsize		: Usually 0			**/

	/* Standard Methods */
	(destructor)	pyAdmData_free,			/** tp_dealloc		: refcount = 0			**/
	(printfunc)	NULL,				/** tp_print		: print x			**/
	(getattrfunc)	pyAdmFloat_getattr,		/** tp_getattr		: x.attr			**/
	(setattrfunc)	pyAdmFloat_setattr,		/** tp_setattr		: x.attr = v			**/
	(cmpfunc)	pyAdmData_cmp,			/** tp_compare		: cmp(x, y)			**/
	(reprfunc)	pyAdmFloat_repr,		/** tp_repr		: repr(x)			**/

	/* Type Categories */
	NULL,						/** tp_as_number	: Number Methods		**/
	NULL,						/** tp_as_sequence	: Sequence methods		**/
	NULL,						/** tp_as_mapping	: Mapping functions		**/
	(hashfunc)	pyAdmData_hash,			/** tp_hash		: dict[x]			**/
	(ternaryfunc)	NULL,				/** tp_call		: x()				**/
	(reprfunc)	pyAdmFloat_str,			/** tp_str		: str(x)			**/
};

static char pyAdmFloat_doc[] =
"AdmFloat elements have the following attributes and functions:\n"
"\n"
"	float.value: assignable\n"
"	holds the floating point value of the AdmFloat element\n"
"\n"
pyAdmData_doc_macro;

static struct PyMethodDef pyAdmFloat_methods[] = {
	{NULL, NULL}
};

static struct PyMethodChain pyAdmFloat_method_chain = {
	methods:	pyAdmFloat_methods,
	link:		&pyAdmData_method_chain,
};

/* ------------------------------------------------------------------*/
/* Base Functions for pyAdmFloat_object */

/**
* pyAdmFloat_getattr : handles x.attr calls
**/
static PyObject *
pyAdmFloat_getattr (pyAdmData_object *self, char *name) 
{
	AdmError *error;
	PyObject *retobj;
	double f;

	error = NULL;
	
	if (pyAdmData_getattr(self, name, &retobj)) {
		return retobj;

	} else if (!strcmp(name, "value")) {
		f = AdmFloat_getValue((AdmFloat *) self->data, &error);
		if (pyAdmError_check_and_raise(error))
			return NULL;

		return Py_BuildValue("d", f);

	} else if (!strcmp(name, "__doc__")) {
		return PyString_FromString(pyAdmFloat_doc);
	}

	return Py_FindMethodInChain(&pyAdmFloat_method_chain, (PyObject *) self, name);
}

/**
* pyAdmFloat_setattr : handles x.attr = v calls
**/
static int
pyAdmFloat_setattr (pyAdmData_object *self, char *name, PyObject *args) 
{
	AdmError *error;
	PyObject *tmpobj;
	double f;
	int i;

	error = NULL;
	
	if ((i = pyAdmData_setattr(self, name, args)))
       		return i<0?-1:0;

	if (!strcmp(name, "value")) {
		/* Coerce the argument into a number we can use */
		if (!PyNumber_Check(args))
			return -1;

		if ((tmpobj = PyNumber_Float(args)) == NULL)
			return -1;

		f = PyFloat_AsDouble(tmpobj);
		Py_DECREF(tmpobj);

		AdmFloat_setValue((AdmFloat *) self->data, f, &error);
		if (pyAdmError_check_and_raise(error))
			return -1;

		return 0;
	}

	return -1;
}

/**
* pyAdmFloat_repr : handles repr(x) calls
**/
static PyObject *
pyAdmFloat_repr (pyAdmData_object *self)
{
	AdmError *error;
	AdmRefStr *name, *source;
	char buf[256];
	double f;

	error = NULL;
	
	name = AdmData_getName(self->data, &error);
	if (pyAdmError_check_and_raise(error))
		return NULL;

	source = AdmData_getSource(self->data, &error);
	if (pyAdmError_check_and_raise(error)) {
		AdmRefStr_unref(name);
		return NULL;
	}

	f = AdmFloat_getValue((AdmFloat *) self->data, &error);
	if (pyAdmError_check_and_raise(error)) {
		AdmRefStr_unref(name);
		AdmRefStr_unref(source);
		return NULL;
	}

	snprintf(buf, 255, "<AdmFloat, name=\"%s\", source=\"%s\", value=%la at %p>",
			AdmRefStr_str(name), AdmRefStr_str(source), f, self);

	AdmRefStr_unref(name);
	AdmRefStr_unref(source);

	return PyString_FromString(buf);
}


/**
* pyAdmFloat_str : handles str(x) calls
**/
static PyObject *
pyAdmFloat_str (pyAdmData_object *self)
{
	AdmError *error;
	char buf[256];
	double f;

	error = NULL;
	
	f = AdmFloat_getValue((AdmFloat *) self->data, &error);
	if (pyAdmError_check_and_raise(error))
		return NULL;

	snprintf(buf, 255, "%la", f);

	return PyString_FromString(buf);
}



/* ================================================================================================================== */
/** AdmString **/

/* funcction prototypes */
static PyObject		*pyAdmString_getattr			(pyAdmData_object	 *self,
								 char			 *name);

static int		 pyAdmString_setattr			(pyAdmData_object	 *self,
								 char			 *name,
								 PyObject		 *args);

static PyObject		*pyAdmString_repr			(pyAdmData_object	 *self);

static PyObject		*pyAdmString_str			(pyAdmData_object	 *self);

/** the pyAdmString_type structure is a PyTypeObject struct, and holds our type callbacks **/
static PyTypeObject pyAdmString_object_type = {
	PyObject_HEAD_INIT(&PyType_Type)		/** Required initialization				**/
	0,						/** ob_size		: Usually 0			**/
	"AdmString",					/** tp_name		: Type Name			**/
	sizeof(pyAdmData_object),			/** tp_basicsize	: Object Size			**/
	0,						/** tp_itemsize		: Usually 0			**/

	/* Standard Methods */
	(destructor)	pyAdmData_free,			/** tp_dealloc		: refcount = 0			**/
	(printfunc)	NULL,				/** tp_print		: print x			**/
	(getattrfunc)	pyAdmString_getattr,		/** tp_getattr		: x.attr			**/
	(setattrfunc)	pyAdmString_setattr,		/** tp_setattr		: x.attr = v			**/
	(cmpfunc)	pyAdmData_cmp,			/** tp_compare		: cmp(x, y)			**/
	(reprfunc)	pyAdmString_repr,		/** tp_repr		: repr(x)			**/

	/* Type Categories */
	NULL,						/** tp_as_number	: Number Methods		**/
	NULL,						/** tp_as_sequence	: Sequence methods		**/
	NULL,						/** tp_as_mapping	: Mapping functions		**/
	(hashfunc)	pyAdmData_hash,			/** tp_hash		: dict[x]			**/
	(ternaryfunc)	NULL,				/** tp_call		: x()				**/
	(reprfunc)	pyAdmString_str,		/** tp_str		: str(x)			**/
};

static char pyAdmString_doc[] =
"AdmString elements have the following attributes and functions:\n"
"\n"
"	string.value: assignable\n"
"	holds the string value of the AdmString element\n"
"\n"
pyAdmData_doc_macro;


static struct PyMethodDef pyAdmString_methods[] = {
	{NULL, NULL}
};

static struct PyMethodChain pyAdmString_method_chain = {
	methods:	pyAdmString_methods,
	link:		&pyAdmData_method_chain,
};

/* ------------------------------------------------------------------*/
/* Base Functions for pyAdmString_object */

/**
* pyAdmString_getattr : handles x.attr calls
**/
static PyObject *
pyAdmString_getattr (pyAdmData_object *self, char *name) 
{
	AdmError *error;
	PyObject *retobj;
	AdmRefStr *str;
	
	error = NULL;
	
	if (pyAdmData_getattr(self, name, &retobj)) {
		return retobj;

	} else if (!strcmp(name, "value")) {
		str = AdmString_getValue((AdmString *) self->data, &error);
		if (pyAdmError_check_and_raise(error))
			return NULL;

		retobj = Py_BuildValue("s", str);
		AdmRefStr_unref(str);
		return retobj;

	} else if (!strcmp(name, "__doc__")) {
		return PyString_FromString(pyAdmString_doc);
	}

	return Py_FindMethodInChain(&pyAdmString_method_chain, (PyObject *) self, name);
}

/**
* pyAdmString_setattr : handles x.attr = v calls
**/
static int
pyAdmString_setattr (pyAdmData_object *self, char *name, PyObject *args) 
{
	AdmError *error;
	const char *str;
	int i;

	error = NULL;
	
	if ((i = pyAdmData_setattr(self, name, args)))
       		return i<0?-1:0;

	if (!strcmp(name, "value")) {
		if (!PyString_Check(args))
			return -1;

		str = PyString_AsString(args);

		AdmString_setValue((AdmString *) self->data, str, &error);
		if (pyAdmError_check_and_raise(error))
			return -1;

		return 0;
	}

	return -1;
}

/**
* pyAdmString_repr : handles repr(x) calls
**/
static PyObject *
pyAdmString_repr (pyAdmData_object *self)
{
	AdmError *error;
	AdmRefStr *name, *source, *value;
	char buf[256];

	error = NULL;
	
	name = AdmData_getName(self->data, &error);
	if (pyAdmError_check_and_raise(error)) {
		return NULL;
	}

	source = AdmData_getSource(self->data, &error);
	if (pyAdmError_check_and_raise(error)) {
		AdmRefStr_unref(name);
		return NULL;
	}

	value = AdmString_getValue((AdmString *) self->data, &error);
	if (pyAdmError_check_and_raise(error)) {
		AdmRefStr_unref(source);
		AdmRefStr_unref(name);
		return NULL;
	}

	snprintf(buf, 255, "<AdmString, name=\"%s\", source=\"%s\", value=\"%s\" at %p>",
			AdmRefStr_str(name), AdmRefStr_str(source), AdmRefStr_str(value), self);

	AdmRefStr_unref(name);
	AdmRefStr_unref(source);
	AdmRefStr_unref(value);

	return PyString_FromString(buf);
}


/**
* pyAdmString_str : handles str(x) calls
**/
static PyObject *
pyAdmString_str (pyAdmData_object *self)
{
	AdmError *error;
	AdmRefStr *value;
	PyObject *retobj;

	error = NULL;
	
	value = AdmString_getValue((AdmString *) self->data, &error);
	if (pyAdmError_check_and_raise(error))
		return NULL;

	retobj = PyString_FromString(AdmRefStr_str(value));
	AdmRefStr_unref(value);
	return retobj;
}


/* ================================================================================================================== */
/** AdmBase64 **/

/* funcction prototypes */
static PyObject		*pyAdmBase64_getattr			(pyAdmData_object	 *self,
								 char			 *name);

static int		 pyAdmBase64_setattr			(pyAdmData_object	 *self,
								 char			 *name,
								 PyObject		 *args);

static PyObject		*pyAdmBase64_repr			(pyAdmData_object	 *self);

static PyObject		*pyAdmBase64_str			(pyAdmData_object	 *self);

/** the pyAdmBase64_type structure is a PyTypeObject struct, and holds our type callbacks **/
static PyTypeObject pyAdmBase64_object_type = {
	PyObject_HEAD_INIT(&PyType_Type)		/** Required initialization				**/
	0,						/** ob_size		: Usually 0			**/
	"AdmBase64",					/** tp_name		: Type Name			**/
	sizeof(pyAdmData_object),			/** tp_basicsize	: Object Size			**/
	0,						/** tp_itemsize		: Usually 0			**/

	/* Standard Methods */
	(destructor)	pyAdmData_free,			/** tp_dealloc		: refcount = 0			**/
	(printfunc)	NULL,				/** tp_print		: print x			**/
	(getattrfunc)	pyAdmBase64_getattr,		/** tp_getattr		: x.attr			**/
	(setattrfunc)	pyAdmBase64_setattr,		/** tp_setattr		: x.attr = v			**/
	(cmpfunc)	pyAdmData_cmp,			/** tp_compare		: cmp(x, y)			**/
	(reprfunc)	pyAdmBase64_repr,		/** tp_repr		: repr(x)			**/

	/* Type Categories */
	NULL,						/** tp_as_number	: Number Methods		**/
	NULL,						/** tp_as_sequence	: Sequence methods		**/
	NULL,						/** tp_as_mapping	: Mapping functions		**/
	(hashfunc)	pyAdmData_hash,			/** tp_hash		: dict[x]			**/
	(ternaryfunc)	NULL,				/** tp_call		: x()				**/
	(reprfunc)	pyAdmBase64_str,		/** tp_str		: str(x)			**/
};

static char pyAdmBase64_doc[] =
"AdmBase64 elements have the following attributes and functions:\n"
"\n"
"	base64.value: assignable\n"
"	holds the base64 encoded value of the AdmBase64 element,\n"
"	valid base64 is:\n"
"		an arbitrary number of 4 character groups of\n"
"		alphanumerics, where the last group's last,\n"
"		or last	2 characters, can be '='\n"
"\n"
"	base64.binvalue: assignable\n"
"	holds the binary value of the AdmBase64 element\n"
"\n"
pyAdmData_doc_macro;


static struct PyMethodDef pyAdmBase64_methods[] = {
	{NULL, NULL}
};

static struct PyMethodChain pyAdmBase64_method_chain = {
	methods:	pyAdmBase64_methods,
	link:		&pyAdmData_method_chain,
};

/* ------------------------------------------------------------------*/
/* Base Functions for pyAdmContext_object */

/**
* pyAdmBase64_getattr : handles x.attr calls
**/
static PyObject *
pyAdmBase64_getattr (pyAdmData_object *self, char *name) 
{
	AdmError *error;
	PyObject *retobj;
	AdmRefStr *refstr;
	const char *str;

	error = NULL;
	
	if (pyAdmData_getattr(self, name, &retobj)) {
		return retobj;

	} else if (!strcmp(name, "value")) {
		refstr = AdmBase64_getValue((AdmBase64 *) self->data, &error);
		if (pyAdmError_check_and_raise(error))
			return NULL;

		retobj = PyString_FromString(AdmRefStr_str(refstr));
		AdmRefStr_unref(refstr);
		return retobj;

	} else if (!strcmp(name, "binValue")) {
		size_t n;

		str = NULL;

		n = AdmBase64_getBinValue((AdmBase64 *) self->data, (void *) &str, &error);
		if (pyAdmError_check_and_raise(error))
			return NULL;

		retobj = PyString_FromStringAndSize(str, n);
		free((char *) str);

		return retobj;

	} else if (!strcmp(name, "__doc__")) {
		return PyString_FromString(pyAdmBase64_doc);
	}


	return Py_FindMethodInChain(&pyAdmBase64_method_chain, (PyObject *) self, name);
}

/**
* pyAdmBase64_setattr : handles x.attr = v calls
**/
static int
pyAdmBase64_setattr (pyAdmData_object *self, char *name, PyObject *args) 
{
	AdmError *error;
	const char *str;
	int i;

	error = NULL;

	if ((i = pyAdmData_setattr(self, name, args)))
       		return i<0?-1:0;
	
	if (!strcmp(name, "value")) {
		if (!PyString_Check(args))
			return -1;

		str = PyString_AsString(args);

		AdmBase64_setValue((AdmBase64 *) self->data, str, &error);
		if (pyAdmError_check_and_raise(error))
			return -1;

		return 0;
	} else if (!strcmp(name, "binValue")) {
		size_t n;

		if (!PyString_Check(args))
			return -1;

		str = PyString_AsString(args);
		n = PyString_Size(args);

		AdmBase64_setBinValue((AdmBase64 *) self->data, str, n, &error);
		if (pyAdmError_check_and_raise(error))
			return -1;

		return 0;
	}

	return -1;
}

/**
* pyAdmBase64_repr : handles repr(x) calls
**/
static PyObject *
pyAdmBase64_repr (pyAdmData_object *self)
{
	AdmError *error;
	AdmRefStr *name, *source, *value;
	char buf[256];

	error = NULL;

	name = AdmData_getName(self->data, &error);
	if (pyAdmError_check_and_raise(error))
		return NULL;

	source = AdmData_getSource(self->data, &error);
	if (pyAdmError_check_and_raise(error)) {
		AdmRefStr_unref(name);
		return NULL;
	}

	value = AdmBase64_getValue((AdmBase64 *) self->data, &error);
	if (pyAdmError_check_and_raise(error)) {
		AdmRefStr_unref(source);
		AdmRefStr_unref(name);
		return NULL;
	}

	snprintf(buf, 255, "<AdmBase64, name=\"%s\", source=\"%s\", value=\"%s\" at %p>",
			AdmRefStr_str(name), AdmRefStr_str(source), AdmRefStr_str(value), self);

	AdmRefStr_unref(name);
	AdmRefStr_unref(source);
	AdmRefStr_unref(value);

	return PyString_FromString(buf);
}

/**
* pyAdmBase64_str : handles str(x) calls
**/
static PyObject *
pyAdmBase64_str (pyAdmData_object *self)
{
	AdmError *error;
	PyObject *retobj;
	AdmRefStr *value;

	error = NULL;
	value = AdmBase64_getValue((AdmBase64 *) self->data, &error);
	if (pyAdmError_check_and_raise(error))
		return NULL;

	retobj = PyString_FromString(AdmRefStr_str(value));
	AdmRefStr_unref(value);
	return retobj;
}


/* ================================================================================================================== */
/** AdmCopy **/

/* funcction prototypes */
static PyObject		*pyAdmCopy_getattr		(pyAdmData_object	 *self,
							 char			 *name);

static int		 pyAdmCopy_setattr		(pyAdmData_object	 *self,
							 char			 *name,
							 PyObject		 *args);

static PyObject		*pyAdmCopy_repr			(pyAdmData_object	 *self);

static PyObject		*pyAdmCopy_str			(pyAdmData_object	 *self);

/** the pyAdmCopy_type structure is a PyTypeObject struct, and holds our type callbacks **/
static PyTypeObject pyAdmCopy_object_type = {
	PyObject_HEAD_INIT(&PyType_Type)		/** Required initialization				**/
	0,						/** ob_size		: Usually 0			**/
	"AdmCopy",					/** tp_name		: Type Name			**/
	sizeof(pyAdmData_object),			/** tp_basicsize	: Object Size			**/
	0,						/** tp_itemsize		: Usually 0			**/

	/* Standard Methods */
	(destructor)	pyAdmData_free,			/** tp_dealloc		: refcount = 0			**/
	(printfunc)	NULL,				/** tp_print		: print x			**/
	(getattrfunc)	pyAdmCopy_getattr,		/** tp_getattr		: x.attr			**/
	(setattrfunc)	pyAdmCopy_setattr,		/** tp_setattr		: x.attr = v			**/
	(cmpfunc)	pyAdmData_cmp,			/** tp_compare		: cmp(x, y)			**/
	(reprfunc)	pyAdmCopy_repr,			/** tp_repr		: repr(x)			**/

	/* Type Categories */
	NULL,						/** tp_as_number	: Number Methods		**/
	NULL,						/** tp_as_sequence	: Sequence methods		**/
	NULL,						/** tp_as_mapping	: Mapping functions		**/
	(hashfunc)	pyAdmData_hash,			/** tp_hash		: dict[x]			**/
	(ternaryfunc)	NULL,				/** tp_call		: x()				**/
	(reprfunc)	pyAdmCopy_str,			/** tp_str		: str(x)			**/
};

static char pyAdmCopy_doc[] =
"AdmCopy elements have the following attributes and functions:\n"
"\n"
"	copy.value: assignable\n"
"	holds the path of the AdmCopy element\n"
"\n"
pyAdmData_doc_macro;


static struct PyMethodDef pyAdmCopy_methods[] = {
	{NULL, NULL}
};

static struct PyMethodChain pyAdmCopy_method_chain = {
	methods:	pyAdmCopy_methods,
	link:		&pyAdmData_method_chain,
};

/* ------------------------------------------------------------------*/
/* Base Functions for pyAdmCopy_object */

/**
* pyAdmCopy_getattr : handles x.attr calls
**/
static PyObject *
pyAdmCopy_getattr (pyAdmData_object *self, char *name) 
{
	AdmError *error;
	PyObject *retobj;
	AdmRefStr *str;

	error = NULL;
	
	if (pyAdmData_getattr(self, name, &retobj)) {
		return retobj;

	} else if (!strcmp(name, "value")) {
		str = AdmCopy_getValue((AdmCopy *) self->data, &error);
		if (pyAdmError_check_and_raise(error))
			return NULL;

		retobj = PyString_FromString(AdmRefStr_str(str));
		AdmRefStr_unref(str);
		return retobj;

	} else if (!strcmp(name, "__doc__")) {
		return PyString_FromString(pyAdmCopy_doc);
	}

	return Py_FindMethodInChain(&pyAdmCopy_method_chain, (PyObject *) self, name);
}

/**
* pyAdmCopy_setattr : handles x.attr = v calls
**/
static int
pyAdmCopy_setattr (pyAdmData_object *self, char *name, PyObject *args) 
{
	AdmError *error;
	const char *str;
	int i;

	error = NULL;

	if ((i = pyAdmData_setattr(self, name, args)))
       		return i<0?-1:0;

	if (!strcmp(name, "value")) {
		if (!PyString_Check(args))
			return -1;

		str = PyString_AsString(args);

		AdmCopy_setValue((AdmCopy *) self->data, str, &error);
		if (pyAdmError_check_and_raise(error))
			return -1;

		return 0;
	}

	return -1;
}

/**
* pyAdmCopy_repr : handles repr(x) calls
**/
static PyObject *
pyAdmCopy_repr (pyAdmData_object *self)
{
	AdmError *error;
	AdmRefStr *name, *source, *value;
	char buf[256];

	error = NULL;

	name = AdmData_getName(self->data, &error);
	if (pyAdmError_check_and_raise(error))
		return NULL;

	source = AdmData_getSource(self->data, &error);
	if (pyAdmError_check_and_raise(error)) {
		AdmRefStr_unref(name);
		return NULL;
	}

	value = AdmCopy_getValue((AdmCopy *) self->data, &error);
	if (pyAdmError_check_and_raise(error)) {
		AdmRefStr_unref(name);
		AdmRefStr_unref(source);
		return NULL;
	}

	snprintf(buf, 255, "<AdmCopy, name=\"%s\", source=\"%s\", value=\"%s\" at %p>",
			AdmRefStr_str(name), AdmRefStr_str(source), AdmRefStr_str(value), self);

	AdmRefStr_unref(name);
	AdmRefStr_unref(source);
	AdmRefStr_unref(value);

	return PyString_FromString(buf);
}

/**
* pyAdmCopy_str : handles str(x) calls
**/
static PyObject *
pyAdmCopy_str (pyAdmData_object *self)
{
	AdmError *error;
	AdmRefStr *value;
	PyObject *retobj;

	error = NULL;
	value = AdmCopy_getValue((AdmCopy *) self->data, &error);
	if (pyAdmError_check_and_raise(error))
		return NULL;

	retobj = PyString_FromString(AdmRefStr_str(value));
	AdmRefStr_unref(value);
	return retobj;
}


/* ================================================================================================================== */
/** AdmList **/

/* funcction prototypes */
static PyObject		*pyAdmList_getattr		(pyAdmData_object	 *self,
							 char			 *name);

static int		 pyAdmList_setattr		(pyAdmData_object	 *self,
							 char			 *name,
							 PyObject		 *args);

static PyObject		*pyAdmList_repr			(pyAdmData_object	 *self);

static int		 pyAdmList_len			(pyAdmData_object	 *self);

static PyObject		*pyAdmList_sq_getitem		(pyAdmData_object	 *self,
							 int			  i);

static PyObject		*pyAdmList_sq_getslice		(pyAdmData_object	 *self,
							 int			  i,
							 int			  j);

static PyObject		*pyAdmList_mp_getitem		(pyAdmData_object	 *self,
							 PyObject		 *args);

static int		pyAdmList_mp_setitem		(pyAdmData_object	 *self,
							 PyObject		 *key,
							 PyObject		 *args);

static char		 pyAdmList_addData_doc[];
static PyObject		*pyAdmList_addData		(pyAdmData_object	 *self,
							 PyObject		 *args,
							 PyObject		 *kwargs);

static char		 pyAdmList_has_key_doc[];
static PyObject		*pyAdmList_has_key		(pyAdmData_object	 *self,
							 PyObject		 *args,
							 PyObject		 *kwargs);

static char		 pyAdmList_get_doc[];
static PyObject		*pyAdmList_get			(pyAdmData_object	 *self,
							 PyObject		 *args,
							 PyObject		 *kwargs);

static char		 pyAdmList_keys_doc[];
static PyObject		*pyAdmList_keys			(pyAdmData_object	 *self,
							 PyObject		 *args);

static char		 pyAdmList_copyData_doc[];
static PyObject		*pyAdmList_copyData		(pyAdmData_object	 *self,
							 PyObject		 *args,
							 PyObject		 *kwargs);

static PySequenceMethods pyAdmList_sequence_methods = {
	(inquiry)		pyAdmList_len,		/** sq_lenth		: __len__(x)			**/
	(binaryfunc)		NULL,			/** sq_concat		: __add__(x, y)			**/
	(intargfunc)		NULL,			/** sq_repeat		: __mul__(x, n)__		**/
	(intargfunc)		pyAdmList_sq_getitem,	/** sq_item		: __getitem__(x, n)__		**/
	(intintargfunc)		pyAdmList_sq_getslice,	/** sq_slice		: __getslice__(x, i, j)__	**/
	(intobjargproc)		NULL,			/** sq_ass_item		: __setitem__(x, n, v)__	**/
	(intintobjargproc)	NULL,			/** sq_ass_slice	: __setslice__(x, i, j, v)__	**/
};

static PyMappingMethods pyAdmList_mapping_methods = {
	(inquiry)		pyAdmList_len,		/** mp_lenth		: __len__(x)			**/
	(binaryfunc)		pyAdmList_mp_getitem,	/** mp_subscript	: __getitem__(x, key)		**/
	(objobjargproc)		pyAdmList_mp_setitem,	/** mp_ass_subscript	: __setitem__(x, key, value)	**/
};

/** the pyAdmList_type structure is a PyTypeObject struct, and holds our type callbacks **/
static PyTypeObject pyAdmList_object_type = {
	PyObject_HEAD_INIT(&PyType_Type)		/** Required initialization				**/
	0,						/** ob_size		: Usually 0			**/
	"AdmList",					/** tp_name		: Type Name			**/
	sizeof(pyAdmData_object),			/** tp_basicsize	: Object Size			**/
	0,						/** tp_itemsize		: Usually 0			**/

	/* Standard Methods */
	(destructor)	pyAdmData_free,			/** tp_dealloc		: refcount = 0			**/
	(printfunc)	NULL,				/** tp_print		: print x			**/
	(getattrfunc)	pyAdmList_getattr,		/** tp_getattr		: x.attr			**/
	(setattrfunc)	pyAdmList_setattr,		/** tp_setattr		: x.attr = v			**/
	(cmpfunc)	pyAdmData_cmp,			/** tp_compare		: cmp(x, y)			**/
	(reprfunc)	pyAdmList_repr,			/** tp_repr		: repr(x)			**/

	/* Type Categories */
	NULL,						/** tp_as_number	: Number Methods		**/
	&pyAdmList_sequence_methods,			/** tp_as_sequence	: Sequence methods		**/
	&pyAdmList_mapping_methods,			/** tp_as_mapping	: Mapping functions		**/
	(hashfunc)	pyAdmData_hash,			/** tp_hash		: dict[x]			**/
	(ternaryfunc)	NULL,				/** tp_call		: x()				**/
	(reprfunc)	NULL,				/** tp_str		: str(x)			**/
};

static char pyAdmList_doc[] =
"AdmList elements have the following attributes and functions:\n"
"\n"
pyAdmData_doc_macro;


static struct PyMethodDef pyAdmList_methods[] = {
	{"addData",	(PyCFunction) pyAdmList_addData,	METH_VARARGS | METH_KEYWORDS,
								pyAdmList_addData_doc},
	{"keys",	(PyCFunction) pyAdmList_keys,		METH_VARARGS,
								pyAdmList_keys_doc},
	{"get",		(PyCFunction) pyAdmList_get,		METH_VARARGS | METH_KEYWORDS,
								pyAdmList_get_doc},
	{"has_key",	(PyCFunction) pyAdmList_has_key,	METH_VARARGS | METH_KEYWORDS,
								pyAdmList_has_key_doc},
	{"copyData",	(PyCFunction) pyAdmList_copyData,	METH_VARARGS | METH_KEYWORDS,
								pyAdmList_copyData_doc},
	{NULL, NULL}
};

static struct PyMethodChain pyAdmList_method_chain = {
	methods:	pyAdmList_methods,
	link:		&pyAdmData_method_chain,
};

/* ------------------------------------------------------------------*/
/* Base Functions for pyAdmContext_object */

/**
* pyAdmList_getattr : handles x.attr calls
**/
static PyObject *
pyAdmList_getattr (pyAdmData_object *self, char *name) 
{
	AdmError *error;
	PyObject *retobj;
	int i;
	
	error = NULL;

	if (pyAdmData_getattr(self, name, &retobj)) {
		return retobj;

	} else if (!strcmp(name, "anonymous")) {
		i = AdmList_isAnonymous((AdmList *) self->data, &error);
		if (pyAdmError_check_and_raise(error))
			return NULL;

		return Py_BuildValue("i", i);

	} else if (!strcmp(name, "atomic")) {
		i = AdmList_isAtomic((AdmList *) self->data, &error);
		if (pyAdmError_check_and_raise(error))
			return NULL;

		return Py_BuildValue("i", i);

	} else if (!strcmp(name, "prepend")) {
		i = AdmList_isPrepend((AdmList *) self->data, &error);
		if (pyAdmError_check_and_raise(error))
			return NULL;

		return Py_BuildValue("i", i);


	} else if (!strcmp(name, "__doc__")) {
		return PyString_FromString(pyAdmList_doc);
	}

	return Py_FindMethodInChain(&pyAdmList_method_chain, (PyObject *) self, name);
}

/**
* pyAdmList_setattr : handles x.attr = v calls
**/
static int
pyAdmList_setattr (pyAdmData_object *self, char *name, PyObject *args) 
{
	AdmError *error;
	PyObject *tmpobj;
	int i;

	error = NULL;

	if ((i = pyAdmData_setattr(self, name, args)))
       		return i<0?-1:0;

	if (!strcmp(name, "anonymous")) {
		/* Coerce the argument into a number we can use */
		if (!PyNumber_Check(args))
			return -1;

		if ((tmpobj = PyNumber_Int(args)) == NULL)
			return -1;

		i = PyInt_AsLong(tmpobj);
		Py_DECREF(tmpobj);

		AdmList_setAnonymous((AdmList *) self->data, i, &error);
		if (pyAdmError_check_and_raise(error))
			return -1;

		return 0;

	} else if (!strcmp(name, "atomic")) {
		/* Coerce the argument into a number we can use */
		if (!PyNumber_Check(args))
			return -1;

		if ((tmpobj = PyNumber_Int(args)) == NULL)
			return -1;

		i = PyInt_AsLong(tmpobj);
		Py_DECREF(tmpobj);

		AdmList_setAtomic((AdmList *) self->data, i, &error);
		if (pyAdmError_check_and_raise(error))
			return -1;

		return 0;

	} else if (!strcmp(name, "prepend")) {
		/* Coerce the argument into a number we can use */
		if (!PyNumber_Check(args))
			return -1;

		if ((tmpobj = PyNumber_Int(args)) == NULL)
			return -1;

		i = PyInt_AsLong(tmpobj);
		Py_DECREF(tmpobj);

		AdmList_setPrepend((AdmList *) self->data, i, &error);
		if (pyAdmError_check_and_raise(error))
			return -1;

		return 0;
	}

	return -1;
}

/**
* pyAdmList_repr : handles repr(x) calls
**/
static PyObject *
pyAdmList_repr (pyAdmData_object *self)
{
	AdmError *error;
	AdmRefStr *name, *source;
	char buf[256];

	error = NULL;
	name = AdmData_getName(self->data, &error);
	if (pyAdmError_check_and_raise(error))
		return NULL;

	source = AdmData_getSource(self->data, &error);
	if (pyAdmError_check_and_raise(error)) {
		AdmRefStr_unref(name);
		return NULL;
	}

	snprintf(buf, 255, "<AdmList, name=\"%s\", source=\"%s\" at %p>",
			AdmRefStr_str(name), AdmRefStr_str(source), self);

	AdmRefStr_unref(name);
	AdmRefStr_unref(source);

	return PyString_FromString(buf);
}

/**
* pyAdmList_len : handles __len__(x)
**/
static int
pyAdmList_len (pyAdmData_object *self)
{
	AdmError *error;
	int n;

	error = NULL;
	n = AdmList_getNumData((AdmList *) self->data, &error);
	if (pyAdmError_check_and_raise(error))
		return -1;

	return n;
}

/**
* pyAdmList_sq_getitem : sequence getitem handler
**/
static PyObject	*
pyAdmList_sq_getitem (pyAdmData_object	 *self,
		      int		  i)
{
	AdmData *data;
	AdmError *error;
	unsigned int n;

	error = NULL;
	n = AdmList_getNumData((AdmList *) self->data, NULL);
	if (pyAdmError_check_and_raise(error))
		return NULL;

	/* adjust the index */
	if (i < 0) {
		i += n;
	}

	/* check for range violations */
	if ((i < 0) || (i >= n)) {
		return PyErr_Format(PyExc_IndexError, "AdmList index is out of range");
	}

	data = AdmList_getDataByIndex_b((AdmList *) self->data, i, NULL);
	if (pyAdmError_check_and_raise(error))
		return NULL;
	
	return pyAdmData_constructor(data);
}

/**
* pyAdmList_sq_getitem : sequence getitem handler
**/
static PyObject	*
pyAdmList_sq_getslice (pyAdmData_object	 *self,
		      int		  i,
		      int		  j)
{
	PyObject *retlist;
	AdmError *error;
	AdmDataType type;
	unsigned int n;

	error = NULL;

	n = AdmList_getNumData((AdmList *) self->data, &error);
	if (pyAdmError_check_and_raise(error))
		return NULL;

	if ((n == 0) && (i == 0)) {
		return PyList_New(0);
	}

	/* Adjust the range */
	if (i < 0)
		i += n;
	if (j < 0)
		j += n;
	if (j > n)
		j = n;

	/* Check for illegal ranges */
	if ((i < 0) || (i >= n) || (j < 0)) {
		return PyErr_Format(PyExc_IndexError, "AdmList index is out of range");
	}

	/* Check for empty ranges */
	if ((j - i) <= 0) {
		return PyList_New(0);
	}
	
	/* Step up to give me some scratch space off the stack */
	{
		PyObject *obj;
		int k;
		AdmData *data, *data_array[j - i];

		for (k = 0; k < j - i; ++k) {

			data = AdmList_getDataByIndex_b((AdmList *) self->data, k, &error);
			if (pyAdmError_check_and_raise(error))
				return NULL;

			data_array[k] = data;
		}

		retlist = PyList_New(0);

		for (k = 0; k < j - i; ++k) {
			data = data_array[k];

			type = AdmData_getType(data, &error);
			if (pyAdmError_check_and_raise(error)) {
				Py_DECREF(retlist);
				return NULL;
			}

			if ((obj = pyAdmData_constructor(data)) == NULL) {
				Py_DECREF(retlist);
				return NULL;
			}
			PyList_Append(retlist, obj);
			Py_DECREF(obj);
		}
		return retlist;
	}
}

/**
 * pyAdmList_mp_getitem : getitem mapping handler
 **/
static PyObject	*
pyAdmList_mp_getitem (pyAdmData_object	 *self,
		PyObject		 *args)
{
	PyObject *tmpobj;
	const char *path;
	AdmError *error;
	AdmData *data;
	int i;

	error = NULL;

	if (PyString_Check(args)) {

		path = PyString_AsString(args);

		if (path[0] == '/') {
			data = AdmList_getDataByPath_b((AdmList *) self->data, path, &error);
		} else {
			data = AdmList_getDataByName_b((AdmList *) self->data, path, &error);
		}
		if (pyAdmError_check_and_raise(error))
			return NULL;

		return pyAdmData_constructor(data);

	} else if (PyNumber_Check(args)) {
		/* Coerce the argument into a number we can use */
		if ((tmpobj = PyNumber_Int(args)) == NULL)
			return NULL;

		i = PyInt_AsLong(tmpobj);
		Py_DECREF(tmpobj);

		return pyAdmList_sq_getitem(self, i);

	} else {
		return PyErr_Format(PyExc_TypeError, "Invalid selector");
	}
}

/**
 * pyAdmList_mp_setitem : setitem mapping handler
 **/
static int
pyAdmList_mp_setitem (pyAdmData_object	 *self,
		PyObject		 *key,
		PyObject		 *args)
{
	const char *name;
	int i;
	unsigned int n;
	double f;
	AdmData *data;
	AdmError *error;
	AdmDataType type;
	PyObject *tmpobj;

	error = NULL;

	if (PyString_Check(key)) {

		name = PyString_AsString(key);

		data = AdmList_getDataByName_b((AdmList *) self->data, name, &error);
		if (data == NULL) {
			if (AdmError_errno(error) != ADM_ENONAME) {
				pyAdmError_check_and_raise(error);
				return -1;
			}

			/* Drop the ADM_ENONAME error */
			AdmError_unref(error);
			error = NULL;

			type = -1;

			if (PyInt_Check(args) || PyLong_Check(args)) {
				type = ADM_TYPE_INT;
			} else if (PyFloat_Check(args)) {
				type = ADM_TYPE_FLOAT;
			} else if (PyString_Check(args)) {
				type = ADM_TYPE_STRING;
			} else {
				return -1;
			}

			data = AdmList_addData_b((AdmList *) self->data, type, name, &error);
			if (pyAdmError_check_and_raise(error))
				return -1;

		}

	} else if (PyNumber_Check(key)) {

		/* Coerce the argument into a number we can use */
		if ((tmpobj = PyNumber_Int(key)) == NULL)
			return -1;

		i = PyInt_AsLong(tmpobj);
		Py_DECREF(tmpobj);

		n = AdmList_getNumData((AdmList *) self->data, &error);
		if (pyAdmError_check_and_raise(error))
			return -1;

		if (i < 0)
			i += n;

		if ((i < 0) || (i >= n)) {
			return -1;
		}

		data = AdmList_getDataByIndex_b((AdmList *) self->data, i, &error);
		if (pyAdmError_check_and_raise(error))
			return -1;

	} else {
		return -1;
	}

	switch (AdmData_getType(data, NULL)) {
		case ADM_TYPE_STRING:
			if (!PyString_Check(args))
				return -1;

			AdmString_setValue((AdmString *) data, PyString_AsString(args), &error);
			return pyAdmError_check_and_raise(error);

		case ADM_TYPE_BASE64:
			if (!PyString_Check(args))
				return -1;

			AdmBase64_setValue((AdmBase64 *) data, PyString_AsString(args), &error);
			return pyAdmError_check_and_raise(error);

		case ADM_TYPE_COPY:
			if (!PyString_Check(args))
				return -1;

			AdmCopy_setValue((AdmCopy *) data, PyString_AsString(args), &error);
			return pyAdmError_check_and_raise(error);

		case ADM_TYPE_BOOL:
			if (!PyNumber_Check(args))
				return -1;

			if ((tmpobj = PyNumber_Int(args)) == NULL)
				return -1;

			i = PyInt_AsLong(tmpobj);
			Py_DECREF(tmpobj);

			AdmBool_setValue((AdmBool *) data, i, &error);
			return pyAdmError_check_and_raise(error);

		case ADM_TYPE_INT:
			if (!PyNumber_Check(args))
				return -1;

			if ((tmpobj = PyNumber_Int(args)) == NULL)
				return -1;

			i = PyInt_AsLong(tmpobj);
			Py_DECREF(tmpobj);

			AdmInt_setValue((AdmInt *) data, i, &error);
			return pyAdmError_check_and_raise(error);

		case ADM_TYPE_FLOAT:
			if (!PyNumber_Check(args))
				return -1;

			if ((tmpobj = PyNumber_Float(args)) == NULL)
				return -1;

			f = PyFloat_AsDouble(tmpobj);
			Py_DECREF(tmpobj);

			AdmFloat_setValue((AdmFloat *) data, f, &error);
			return pyAdmError_check_and_raise(error);

		default:
			return -1;
	}
}

/* -------------------------------------------------------------------------------- */
/* Methods for pyAdmList */

static char pyAdmList_addData_doc[] = 
"AdmList.addData(type, name) -> data -- adds a new element to the List";

static PyObject *
pyAdmList_addData (	pyAdmData_object	*self,
			PyObject		*args,
			PyObject		*kwargs)
{
	AdmError *error;
	AdmData *data;
	PyObject *type;
	AdmDataType dtype;
	char *name;
	const char *kw[] = {"type", "name", NULL};

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O!s", (char **) kw, &PyType_Type, &type, &name))
		return NULL;

	/* C doesnt seem to want to switch on pointers, so ... */
	if (type == (PyObject *) &pyAdmList_object_type) {
		dtype = ADM_TYPE_LIST;
	} else if (type == (PyObject *) &pyAdmString_object_type) {
		dtype = ADM_TYPE_STRING;
	} else if (type == (PyObject *) &pyAdmBase64_object_type) {
		dtype = ADM_TYPE_BASE64;
	} else if (type == (PyObject *) &pyAdmCopy_object_type) {
		dtype = ADM_TYPE_COPY;
	} else if (type == (PyObject *) &pyAdmBool_object_type) {
		dtype = ADM_TYPE_BOOL;
	} else if (type == (PyObject *) &pyAdmInt_object_type) {
		dtype = ADM_TYPE_INT;
	} else if (type == (PyObject *) &pyAdmFloat_object_type) {
		dtype = ADM_TYPE_FLOAT;
	} else {
		return PyErr_Format(PyExc_TypeError, "Invalid Type");
	}

	error = NULL;
	data = AdmList_addData_b((AdmList *) self->data, dtype, name, &error);
	if (pyAdmError_check_and_raise(error))
		return NULL;

	return pyAdmData_constructor(data);
}

static char pyAdmList_has_key_doc[] = 
"AdmList.has_key(key) -> 1 or 0 -- querries a list to see if it has a key";

/**
 * pyAdmList_has_key 
 **/
static PyObject	*
pyAdmList_has_key (	pyAdmData_object	 *self,
			PyObject		 *args,
			PyObject		 *kwargs)
{
	AdmError *error;
	const char *name;
	const char *kw[] = {"key", NULL};

	error = NULL;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "s", (char **) kw, &name))
		return NULL;

	AdmList_getDataByName_b((AdmList *) self->data, name, &error);
	if (error == NULL) {
		return Py_BuildValue("i", 1);
	} else if (AdmError_errno(error) == ADM_ENONAME) {
		return Py_BuildValue("i", 0);
	} else {
		pyAdmError_check_and_raise(error);
		return NULL;
	}
}

static char pyAdmList_get_doc[] = 
"AdmList.get(key, value = None) -> AdmList[key] or value -- returns the AdmList object, or value.";

/**
 * pyAdmList_get 
 **/
static PyObject	*
pyAdmList_get (		pyAdmData_object	 *self,
			PyObject		 *args,
			PyObject		 *kwargs)
{
	AdmError *error;
	PyObject *value, *key, *retval, *exc;
	const char *kw[] = {"key", "value", NULL};

	error = NULL;
	value = NULL;
	key = NULL;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O|O", (char **) kw, &key, &value))
		return NULL;
	
	retval = pyAdmList_mp_getitem(self, key);
	if ((exc = PyErr_Occurred()) == NULL) {
		return retval;
	} else if ((exc == PyExc_IndexError) || (exc == PyExc_KeyError)) {
		PyErr_Clear();
		if (value == NULL) {
			return Py_BuildValue("");
		} else {
			Py_INCREF(value);
			return value;
		}
	} else {
		return NULL;
	}

}


static char pyAdmList_keys_doc[] = 
"AdmList.keys() -> keylist -- returns a list of the List's keys";

/**
 * pyAdmList_keys : make a list of names available
 **/
static PyObject	*
pyAdmList_keys (pyAdmData_object	 *self,
		PyObject		 *args)
{
	PyObject *retlist;
	AdmError *error;
	AdmRefStr *name;
	unsigned int n;

	PyObject *obj;
	int k;
	AdmData *data;


	error = NULL;

	if (!PyArg_ParseTuple(args, ""))
		return NULL;

	n = AdmList_getNumData((AdmList *) self->data, &error);
	if (pyAdmError_check_and_raise(error))
		return NULL;

	retlist = PyList_New(0);

	for (k = 0; k < n; ++k) {

		data = AdmList_getDataByIndex_b((AdmList *) self->data, k, &error);
		if (pyAdmError_check_and_raise(error))
			return NULL;

		name = AdmData_getName(data, &error);
		if (pyAdmError_check_and_raise(error)) {
			Py_DECREF(retlist);
			return NULL;
		}

		obj = PyString_FromString(AdmRefStr_str(name));
		AdmRefStr_unref(name);

		if (obj == NULL) {
			Py_DECREF(retlist);
			return NULL;
		}
		PyList_Append(retlist, obj);
		Py_DECREF(obj);
	}
	return retlist;
}

static char pyAdmList_copyData_doc[] = 
"AdmList.copyData(data, name = None or name) -> copy\n"
"	copies the element given, changing the name if one is supplied.\n"
"\n"
"\n"
"AdmList.copyData([data1, data2, ...], name = None or [name1, name2, ...])\n"
"	-> [copy1, copy2, ...]\n"
"\n"
"	copies the elements in the list, renaming them if a name list is\n"
"	supplied. If a name list is supplied, it _must_ be equal in length.\n";

static PyObject	*
pyAdmList_copyData		(pyAdmData_object	 *self,
				 PyObject		 *args,
				 PyObject		 *kwargs)
{
	AdmError *error;
	AdmData *data;
	PyObject *obj, *nameobj, *nameobj_i, *pydata, *retlist;
	unsigned int i, n;
	const char *kw[] = {"data", "name", NULL};

	error = NULL;
	nameobj = NULL;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O|O", (char **) kw, &obj, &nameobj))
		return NULL;

	if (pyAdmData_Check(obj)) {

		if (nameobj == NULL) {
			data = AdmList_copyData_b(
				(AdmList *) self->data,
				((pyAdmData_object *) obj)->data,
				&error
			);

		} else if (PyString_Check(nameobj)) {
			data = AdmList_renameCopyData_b(
				(AdmList *) self->data,
				((pyAdmData_object *) obj)->data,
				PyString_AsString(nameobj),
				&error
			);

		} else {
			return PyErr_Format(
				PyExc_TypeError,
				"expected type 'string' for name, got %s",
				nameobj->ob_type->tp_name
			);
		}

		if (pyAdmError_check_and_raise(error))
			return NULL;

		return pyAdmData_constructor(data);

	} else if (PyList_Check(obj)) {

		n = PyList_Size(obj);

		if (nameobj != NULL) {
			if (PyList_Check(nameobj)) {
				if (n != PyList_Size(nameobj)) {
					return PyErr_Format(
						PyExc_ValueError,
						"name list length does not equal data list length"
					);
				}
			} else {
				return PyErr_Format(
					PyExc_TypeError,
					"expected type 'list' for name, got %s",
					nameobj->ob_type->tp_name
				);
			}
		}

		retlist = PyList_New(0);

		for (i = 0; i < n; ++ i) {
			pydata = PyList_GetItem(obj, i);

			if (!pyAdmData_Check(pydata)) {
				PyErr_Format(PyExc_TypeError, "item %d of list not an AdmData element", i);
				goto undo;
			}

			if (nameobj == NULL) {
				data = AdmList_copyData_b(
					(AdmList *) self->data,
					((pyAdmData_object *) pydata)->data,
					&error
				);
			} else {
				nameobj_i = PyList_GetItem(nameobj, i);

				if (!PyString_Check(nameobj_i)) {
					PyErr_Format(PyExc_TypeError, "item %d of name list not a string", i);
					goto undo;
				}

				data = AdmList_renameCopyData_b(
					(AdmList *) self->data,
					((pyAdmData_object *) pydata)->data,
					PyString_AsString(nameobj_i),
					&error
				);
			}

			if (pyAdmError_check_and_raise(error)) {
				goto undo;
			}

			pydata = pyAdmData_constructor(data);
			PyList_Append(retlist, pydata);
			Py_DECREF(pydata);
		}
		return retlist;

undo:
		/* This is our poor man's transaction support. We delete everything we added */
		n = PyList_Size(retlist);
		for (i = 0; i < n; ++ i) {
			pydata = PyList_GetItem(retlist, i);
			AdmData_unlink(((pyAdmData_object *) pydata)->data, NULL);
		}

		Py_DECREF(retlist);
		return NULL;

	} else {
		return PyErr_Format(PyExc_TypeError, "arg 1 neither an AdmData element nor a list");
	}


}


/* ================================================================================================================== */
/** AdmIdentity **/

/** pyAdmContext_object typedef. **/
typedef struct {
	PyObject_HEAD
		AdmIdentity *id;
	AdmContext *ctx;
} pyAdmIdentity_object;

/* funcction prototypes */
static PyObject		*pyAdmIdentity_constructor		(AdmIdentity		 *id);

static void		 pyAdmIdentity_free			(pyAdmIdentity_object	 *self);

static long		 pyAdmIdentity_hash			(pyAdmIdentity_object	 *self);

static PyObject		*pyAdmIdentity_getattr			(pyAdmIdentity_object	 *self,
								char			 *name);

static int		 pyAdmIdentity_setattr			(pyAdmIdentity_object	 *self,
								char			 *name,
								PyObject		 *args);

static int		 pyAdmIdentity_cmp			(pyAdmIdentity_object	 *self,
								PyObject		 *obj);

static PyObject		*pyAdmIdentity_repr			(pyAdmIdentity_object	 *self);

/** the pyAdmIdentity_type structure is a PyTypeObject struct, and holds our type callbacks **/
static PyTypeObject pyAdmIdentity_object_type = {
	PyObject_HEAD_INIT(&PyType_Type)		/** Required initialization				**/
	0,						/** ob_size		: Usually 0			**/
	"AdmIdentity",					/** tp_name		: Type Name			**/
	sizeof(pyAdmIdentity_object),			/** tp_basicsize	: Object Size			**/
	0,						/** tp_itemsize		: Usually 0			**/

	/* Standard Methods */
	(destructor)	pyAdmIdentity_free,		/** tp_dealloc		: refcount = 0			**/
	(printfunc)	NULL,				/** tp_print		: print x			**/
	(getattrfunc)	pyAdmIdentity_getattr,		/** tp_getattr		: x.attr			**/
	(setattrfunc)	pyAdmIdentity_setattr,		/** tp_setattr		: x.attr = v			**/
	(cmpfunc)	pyAdmIdentity_cmp,		/** tp_compare		: cmp(x, y)			**/
	(reprfunc)	pyAdmIdentity_repr,		/** tp_repr		: repr(x)			**/

	/* Type Categories */
	NULL,						/** tp_as_number	: Number Methods		**/
	NULL,						/** tp_as_sequence	: Sequence methods		**/
	NULL,						/** tp_as_mapping	: Mapping functions		**/
	(hashfunc)	pyAdmIdentity_hash,		/** tp_hash		: dict[x]			**/
	(ternaryfunc)	NULL,				/** tp_call		: x()				**/
	(reprfunc)	NULL,				/** tp_str		: str(x)			**/
};

/** Lookup structure for pyAdmIdentity_getattr **/
static struct PyMethodDef pyAdmIdentity_methods[] = {
	{NULL, NULL}
};

/* ------------------------------------------------------------------*/
/* Base Functions for pyAdmContext_object */

static PyObject *
pyAdmIdentity_constructor (AdmIdentity *id)
{
	AdmError *error;
	AdmContext *ctx;
	pyAdmIdentity_object *pyid;

	error = NULL;

	ctx = AdmIdentity_getContext_b(id, &error);
	if (pyAdmError_check_and_raise(error))
		return NULL;

	AdmContext_ref(ctx, &error);
	if (pyAdmError_check_and_raise(error))
		return NULL;

	AdmIdentity_ref(id, &error);
	if (pyAdmError_check_and_raise(error))
		return NULL;

	/* create a new pyAdmContext_object, and attach the pyAdmContext_object_type struct */
	pyid = PyObject_NEW(pyAdmIdentity_object, &pyAdmIdentity_object_type);
	if (pyid == NULL) {
		AdmIdentity_unref(id, NULL);
		AdmContext_unref(ctx, NULL);
		return NULL;
	}

	pyid->id = id;
	pyid->ctx = ctx;

	return (PyObject *) pyid;
}

/**
 * pyAdmIdentity_free : the destructor for pyAdmIdentity_object
 **/
static void
pyAdmIdentity_free (pyAdmIdentity_object *self)
{
	/* drop the ref on the list */
	AdmIdentity_unref(self->id, NULL);
	AdmContext_unref(self->ctx, NULL);
	PyMem_DEL(self);
}


/**
 * pyAdmIdentity_hash : give a value we can hash against
 **/
static long
pyAdmIdentity_hash (pyAdmIdentity_object *self)
{
	return (long) self->id;
}

/**
 * pyAdmIdentity_cmp : comparison handler
 **/
static int
pyAdmIdentity_cmp (pyAdmIdentity_object	 *self,
		PyObject		 *obj)
{
	if (!pyAdmIdentity_Check(obj))
		return (int) self - (int) obj;
	else
		return (int) self->id - (int) ((pyAdmIdentity_object *) obj)->id;
}


/**
 * pyAdmIdentity_getattr : handles x.attr calls
 **/
static PyObject *
pyAdmIdentity_getattr (pyAdmIdentity_object *self, char *name) 
{
	PyObject *retobj;
	AdmError *error;
	AdmIdentity *id;
	AdmRefStr *tname;
	long serial;
	int i;

	error = NULL;
	if (!strcmp(name, "name")) {

		tname = AdmIdentity_getName(self->id, NULL);
		if (pyAdmError_check_and_raise(error))
			return NULL;

		retobj = Py_BuildValue("s", AdmRefStr_str(tname));
		AdmRefStr_unref(tname);
		return retobj;

	} else if (!strcmp(name, "serial")) {

		serial = AdmIdentity_getSerial(self->id, &error);
		if (pyAdmError_check_and_raise(error))
			return NULL;

		return Py_BuildValue("l", serial);

	} else if (!strcmp(name, "alive")) {

		i = AdmIdentity_isAlive(self->id, &error);
		if (pyAdmError_check_and_raise(error))
			return NULL;

		return Py_BuildValue("i", i);

	} else if (!strcmp(name, "context")) {
		return pyAdmContext_constructor(self->ctx);

	} else if (!strcmp(name, "A")) {

		id = AdmIdentity_getParentA_b(self->id, &error);
		if (pyAdmError_check_and_raise(error))
			return NULL;

		if (id == NULL) {
			return Py_BuildValue("");
		} else {
			return pyAdmIdentity_constructor(id);
		}
	} else if (!strcmp(name, "B")) {

		id = AdmIdentity_getParentB_b(self->id, &error);
		if (pyAdmError_check_and_raise(error))
			return NULL;

		if (id == NULL) {
			return Py_BuildValue("");
		} else {
			return pyAdmIdentity_constructor(id);
		}
	}

	return Py_FindMethod(pyAdmIdentity_methods, (PyObject *) self, name);
}

/**
 * pyAdmIdentity_setattr : handles x.attr = v calls
 **/
static int
pyAdmIdentity_setattr (pyAdmIdentity_object *self, char *name, PyObject *args) 
{
	AdmError *error;
	PyObject *tmpobj;
	const char *str;
	long serial;

	error = NULL;

	if (!strcmp(name, "name")) {
		if (!PyString_Check(args))
			return -1;

		str = PyString_AsString(args);

		AdmIdentity_setName(self->id, str, &error);
		if (pyAdmError_check_and_raise(error))
			return -1;

		return 0;

	} else if (!strcmp(name, "serial")) {
		/* Coerce the argument into a number we can use */
		if (!PyNumber_Check(args))
			return -1;

		if ((tmpobj = PyNumber_Long(args)) == NULL)
			return -1;

		serial = PyLong_AsLong(tmpobj);
		Py_DECREF(tmpobj);

		if (serial <= 0)
			return -1;

		AdmIdentity_setSerial(self->id, serial, &error);
		if (pyAdmError_check_and_raise(error))
			return -1;

		return 0;
	}

	return -1;
}

/**
 * pyAdmIdentity_repr : handles repr(x) calls
 **/
static PyObject *
pyAdmIdentity_repr (pyAdmIdentity_object *self)
{
	AdmError *error;
	AdmRefStr *name;
	unsigned long serial;
	char buf[256];

	error = NULL;

	name = AdmIdentity_getName(self->id, &error);
	if (pyAdmError_check_and_raise(error))
		return NULL;

	serial = AdmIdentity_getSerial(self->id, &error);
	if (pyAdmError_check_and_raise(error)) {
		AdmRefStr_unref(name);
		return NULL;
	}

	snprintf(buf, 255, "<AdmIdentity, name=\"%s\", serial=%lu at %p>", AdmRefStr_str(name), serial, self);

	AdmRefStr_unref(name);

	return PyString_FromString(buf);
}


/* ================================================================================================================== */
/** AdmContext **/

/** pyAdmContext_object typedef. **/
typedef struct {
	PyObject_HEAD
		AdmContext *ctx;
} pyAdmContext_object;

/* funcction prototypes */
static char		 pyAdmContext_new_doc[];
static PyObject		*pyAdmContext_new		(PyObject		 *self,
							PyObject		 *args,
							PyObject		 *kwargs);

static PyObject		*pyAdmContext_constructor	(AdmContext		 *ctx);

static void		 pyAdmContext_free		(pyAdmContext_object	 *self);

static long		 pyAdmContext_hash		(pyAdmContext_object     *self);

static PyObject		*pyAdmContext_getattr		(pyAdmContext_object	 *self,
							char			 *name);

static int		 pyAdmContext_setattr		(pyAdmContext_object	 *self,
							char			 *name,
							PyObject		 *args);

static int		 pyAdmContext_cmp		(pyAdmContext_object	 *self,
							PyObject		 *obj);

static PyObject		*pyAdmContext_str		(pyAdmContext_object	 *self);

static PyObject		*pyAdmContext_repr		(pyAdmContext_object	 *self);

static char		 pyAdmContext_copy_doc[];
static PyObject		*pyAdmContext_copy		(pyAdmContext_object	 *self,
							PyObject		 *args);

static char		 pyAdmContext_flatten_doc[];
static PyObject		*pyAdmContext_flatten		(pyAdmContext_object	 *self,
							PyObject		 *args);

static char		 pyAdmContext_strip_doc[];
static PyObject		*pyAdmContext_strip		(pyAdmContext_object	 *self,
							PyObject		 *args);

static char		 pyAdmContext_set_mask_doc[];
static PyObject		*pyAdmContext_set_mask		(pyAdmContext_object	 *self,
							PyObject		 *args,
							PyObject		 *kwargs);

static char		 pyAdmContext_clear_mask_doc[];
static PyObject		*pyAdmContext_clear_mask	(pyAdmContext_object	 *self,
							PyObject		 *args,
							PyObject		 *kwargs);

static char		 pyAdmContext_write_to_file_doc[];
static PyObject		*pyAdmContext_write_to_file	(pyAdmContext_object	 *self,
							PyObject		 *args,
							PyObject		*kwargs);

/** the pyAdmContext_type structure is a PyTypeObject struct, and holds our type callbacks **/
static PyTypeObject pyAdmContext_object_type = {
	PyObject_HEAD_INIT(&PyType_Type)		/** Required initialization				**/
	0,						/** ob_size		: Usually 0			**/
	"AdmContext",					/** tp_name		: Type Name			**/
	sizeof(pyAdmContext_object),			/** tp_basicsize	: Object Size			**/
	0,						/** tp_itemsize		: Usually 0			**/

	/* Standard Methods */
	(destructor)	pyAdmContext_free,		/** tp_dealloc		: refcount = 0			**/
	(printfunc)	NULL,				/** tp_print		: print x			**/
	(getattrfunc)	pyAdmContext_getattr,		/** tp_getattr		: x.attr			**/
	(setattrfunc)	pyAdmContext_setattr,		/** tp_setattr		: x.attr = v			**/
	(cmpfunc)	pyAdmContext_cmp,		/** tp_compare		: cmp(x, y)			**/
	(reprfunc)	pyAdmContext_repr,		/** tp_repr		: repr(x)			**/

	/* Type Categories */
	NULL,						/** tp_as_number	: Number Methods		**/
	NULL,						/** tp_as_sequence	: Sequence methods		**/
	NULL,						/** tp_as_mapping	: Mapping functions		**/
	(hashfunc)	pyAdmContext_hash,		/** tp_hash		: dict[x]			**/
	(ternaryfunc)	NULL,				/** tp_call		: x()				**/
	(reprfunc)	pyAdmContext_str,		/** tp_str		: str(x)			**/
};

/** Lookup structure for pyAdmContext_getattr **/
static struct PyMethodDef pyAdmContext_methods[] = {
	{"copy",	(PyCFunction) pyAdmContext_copy,		METH_VARARGS,
									pyAdmContext_copy_doc},
	{"flatten",	(PyCFunction) pyAdmContext_flatten,		METH_VARARGS,
									pyAdmContext_flatten_doc},
	{"strip",	(PyCFunction) pyAdmContext_strip,		METH_VARARGS,
									pyAdmContext_strip_doc},
	{"setMask",	(PyCFunction) pyAdmContext_set_mask,		METH_VARARGS | METH_KEYWORDS,
									pyAdmContext_set_mask_doc},
	{"clearMask",	(PyCFunction) pyAdmContext_clear_mask,		METH_VARARGS | METH_KEYWORDS,
									pyAdmContext_clear_mask_doc},
	{"writeToFile",	(PyCFunction) pyAdmContext_write_to_file,	METH_VARARGS | METH_KEYWORDS,
									pyAdmContext_write_to_file_doc},
	{NULL, NULL}
};

/* ------------------------------------------------------------------*/
/* Constructors for pyAdmContext_object */

static char pyAdmContext_new_doc[] =
"AdmContext(name, serial, A, B, path, xml) -> context --\n"
"	if name and serial are set:\n"
"		creates a new context of the given name and serial;\n"
"	if name and serial and A and B are set:\n"
"		merges contexts A and B, and uses the name and serial;\n"
"	if path is set:\n"
"		loads the context from the file at path;\n"
"	if xml is set:\n"
"		decodes the xml string to produce the context";

static PyObject *
pyAdmContext_new (PyObject *self, PyObject *args, PyObject *kwargs)
{
	PyObject *obj;
	pyAdmContext_object *pyctxA, *pyctxB;
	AdmContext *ctx;
	AdmError *error;
	const char *name, *path, *xml;
	long serial;

	const char *kw[] = {"name", "serial", "A", "B", "path", "xml", NULL};

	name = NULL;
	serial = 0;
	pyctxA = NULL;
	pyctxB = NULL;
	path = NULL;
	xml = NULL;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|slO!O!ss:AdmContext", (char **) kw,
				&name, &serial, &pyAdmContext_object_type, &pyctxA,
				&pyAdmContext_object_type, &pyctxB, &path, &xml))
		return NULL;

	/* Nasty check for all the construction cases */
	ctx = NULL;
	error = NULL;
	if (path != NULL) {
		if (name || serial || pyctxA || pyctxB || xml)
			return PyErr_Format(PyExc_TypeError, "'path' must be specified alone");

		AdmContext_readFromFile(&ctx, path, &error);
		if (pyAdmError_check_and_raise(error))
			return NULL;

	} else if (xml != NULL) {
		if (name || serial || pyctxA || pyctxB || path)
			return PyErr_Format(PyExc_TypeError, "'xml' must be specified alone");

		ctx = AdmContext_fromXML(xml, NULL);
		if (pyAdmError_check_and_raise(error))
			return NULL;

	} else if (name != NULL) {
		if (serial <= 0) 
			return PyErr_Format(PyExc_ValueError, "serial numbers must be positive and non-zero");

		if ((pyctxA != NULL) && (pyctxB != NULL)) {

			ctx = AdmContext_merge(name, serial, pyctxA->ctx, pyctxB->ctx, &error);
			if (pyAdmError_check_and_raise(error))
				return NULL;

		} else if ((pyctxA == NULL) && (pyctxB == NULL)) {

			ctx = AdmContext_new(name, serial, &error);
			if (pyAdmError_check_and_raise(error))
				return NULL;

		} else {
			return PyErr_Format(PyExc_ValueError, "two contexts are required for a merge");
		}
	} else {
		return PyErr_Format(PyExc_TypeError, "Insufficient Information in Constructor");
	}

	if (ctx == NULL)
		return PyErr_Format(PyExc_RuntimeError, "failed to create a context");

	/* We will have an extra ref to this */
	obj = pyAdmContext_constructor(ctx);
	AdmContext_unref(ctx, NULL);

	return obj;
}

/**
 * pyAdmContext_constructor
 **/
static PyObject	*
pyAdmContext_constructor (AdmContext *ctx)
{
	pyAdmContext_object *pyctx;
	AdmError *error;

	error = NULL;
	AdmContext_ref(ctx, &error);
	if (pyAdmError_check_and_raise(error))
		return NULL;

	/* create a new pyAdmContext_object, and attach the pyAdmContext_object_type struct */
	pyctx = PyObject_NEW(pyAdmContext_object, &pyAdmContext_object_type);
	pyctx->ctx = ctx;

	return (PyObject *) pyctx;

}

/* ------------------------------------------------------------------*/
/* Base Functions for pyAdmContext_object */

/**
 * pyAdmContext_free : the destructor for pyAdmContext_object
 **/
static void
pyAdmContext_free (pyAdmContext_object *self)
{
	/* drop the ref on the context */
	AdmContext_unref(self->ctx, NULL);
	PyMem_DEL(self);
}

/**
 * pyAdmContext_hash : give a value we can hash against
 **/
static long
pyAdmContext_hash (pyAdmContext_object *self)
{
	return (long) self->ctx;
}

/**
 * pyAdmContext_cmp : comparison handler
 **/
static int
pyAdmContext_cmp (pyAdmContext_object	 *self,
		PyObject		 *obj)
{
	if (!pyAdmContext_Check(obj))
		return (int) self - (int) obj;
	else
		return (int) self->ctx - (int) ((pyAdmContext_object *) obj)->ctx;
}

/**
 * pyAdmContext_getattr : handles x.attr calls
 **/
static PyObject *
pyAdmContext_getattr (pyAdmContext_object *self, char *name) 
{
	PyObject *retobj;
	AdmError *error;
	AdmRefStr *refstr;
	AdmIdentity *id;
	AdmData *data;
	long serial;

	error = NULL;
	if (!strcmp(name, "name")) {
		id = AdmContext_getIdentityRoot_b(self->ctx, &error);
		if (pyAdmError_check_and_raise(error))
			return NULL;

		refstr = AdmIdentity_getName(id, &error);
		if (pyAdmError_check_and_raise(error))
			return NULL;

		retobj = Py_BuildValue("s", AdmRefStr_str(refstr));
		AdmRefStr_unref(refstr);
		return retobj;

	} else if (!strcmp(name, "serial")) {
		id = AdmContext_getIdentityRoot_b(self->ctx, &error);
		if (pyAdmError_check_and_raise(error))
			return NULL;

		serial = AdmIdentity_getSerial(id, &error);
		if (pyAdmError_check_and_raise(error))
			return NULL;

		return Py_BuildValue("l", serial);

	} else if (!strcmp(name, "data")) {
		data = AdmContext_getDataRoot_b(self->ctx, &error);
		if (pyAdmError_check_and_raise(error))
			return NULL;

		return pyAdmData_constructor(data);

	} else if (!strcmp(name, "id")) {
		id = AdmContext_getIdentityRoot_b(self->ctx, &error);
		if (pyAdmError_check_and_raise(error))
			return NULL;

		return pyAdmIdentity_constructor((AdmIdentity *) id);

	} else if (!strcmp(name, "masks")) {
		PyObject *retlist, *foo;
		unsigned int i, n;

		n = AdmContext_getNumMasks(self->ctx, &error);
		if (pyAdmError_check_and_raise(error))
			return NULL;

		retlist = PyList_New(0);
		for (i = 0; i < n; ++ i) {
			refstr = AdmContext_getMask(self->ctx, i, &error);
			if (pyAdmError_check_and_raise(error)) {
				Py_DECREF(retlist);
				return NULL;
			}

			foo = PyString_FromString(AdmRefStr_str(refstr));
			PyList_Append(retlist, foo);
			Py_DECREF(foo);
			AdmRefStr_unref(refstr);
		}
		return retlist;
	}

	return Py_FindMethod(pyAdmContext_methods, (PyObject *) self, name);
}

/**
* pyAdmContext_setattr : handles x.attr = v calls
**/
static int
pyAdmContext_setattr (pyAdmContext_object *self, char *name, PyObject *args) 
{
	PyObject *tmpobj;
	AdmIdentity *id;
	AdmError *error;
	const char *str;
	long serial;

	error = NULL;
	if (!strcmp(name, "name")) {
		if (!PyString_Check(args))
			return -1;

		str = PyString_AsString(args);

		id = AdmContext_getIdentityRoot_b(self->ctx, &error);
		if (pyAdmError_check_and_raise(error))
			return -1;

		AdmIdentity_setName(id, str, &error);
		if (pyAdmError_check_and_raise(error))
			return -1;
		
		return 0;

	} else if (!strcmp(name, "serial")) {
		/* Coerce the argument into a number we can use */
		if (!PyNumber_Check(args))
			return -1;

		if ((tmpobj = PyNumber_Long(args)) == NULL)
			return -1;

		serial = PyLong_AsLong(tmpobj);
		Py_DECREF(tmpobj);

		if (serial <= 0)
			return -1;

		id = AdmContext_getIdentityRoot_b(self->ctx, &error);
		if (pyAdmError_check_and_raise(error))
			return -1;

		AdmIdentity_setSerial(id, serial, &error);
		if (pyAdmError_check_and_raise(error))
			return -1;

		return 0;
	}

	return -1;
}


/**
* pyAdmContext_repr : handles repr(x) calls
**/
static PyObject *
pyAdmContext_repr (pyAdmContext_object *self)
{
	AdmIdentity *id;
	AdmError *error;
	AdmRefStr *name;
	char buf[256];
	unsigned int serial;

	error = NULL;

	id = AdmContext_getIdentityRoot_b(self->ctx, &error);
	if (pyAdmError_check_and_raise(error))
		return NULL;

	name = AdmIdentity_getName(id, &error);
	if (pyAdmError_check_and_raise(error))
		return NULL;

	serial = AdmIdentity_getSerial(id, &error);
	if (pyAdmError_check_and_raise(error)) {
		AdmRefStr_unref(name);
		return NULL;
	}

	snprintf(buf, 255, "<AdmContext, name=\"%s\", serial=%u at %p>", AdmRefStr_str(name), serial, self);

	AdmRefStr_unref(name);

	return PyString_FromString(buf);
}


/**
* pyAdmContext_str : handles str(x) calls
* returns the XML encoding of a context
**/
static PyObject *
pyAdmContext_str (pyAdmContext_object *self)
{
	PyObject *obj;
	AdmError *error;
	char *str;

	error = NULL;
	str = AdmContext_toXML(self->ctx, &error);
	if (pyAdmError_check_and_raise(error))
		return NULL;

	obj = Py_BuildValue("s", str);
	free(str);

	return obj;
}

/* ----------------------------------------------------------------- */
/* Methods for pyAdmContext_object */

static char pyAdmContext_copy_doc[] =
"AdmContext.copy() -> copycontext -- makes a copy of the context";

static PyObject *
pyAdmContext_copy (pyAdmContext_object *self, PyObject *args)
{
	AdmError *error;
	AdmContext *ctx;
	PyObject *pyctx;

	if (!PyArg_ParseTuple(args, ""))
		return NULL;

	error = NULL;
	ctx = AdmContext_copy(self->ctx, &error);
	if (pyAdmError_check_and_raise(error))
		return NULL;

	pyctx = pyAdmContext_constructor(ctx);

	/* drop the extra ref to ctx */
	AdmContext_unref(ctx, NULL);

	return pyctx;
}

static char pyAdmContext_flatten_doc[] =
"AdmContext.flatten() -> context -- flattens a context, stripping the masks and resolving the copys";

static PyObject *
pyAdmContext_flatten (		pyAdmContext_object	*self,
				PyObject		*args)
{
	AdmError *error;
	PyObject *pyctx;

	if (!PyArg_ParseTuple(args, ""))
		return NULL;

	error = NULL;
	AdmContext_flatten(self->ctx, &error);
	if (pyAdmError_check_and_raise(error))
		return NULL;

	pyctx = pyAdmContext_constructor(self->ctx);

	return pyctx;
}

static char pyAdmContext_strip_doc[] =
"AdmContext.strip() -> context -- strips a context, stripping all non-root id nodes from its id tree";

static PyObject *
pyAdmContext_strip (		pyAdmContext_object	*self,
				PyObject		*args)
{
	AdmError *error;
	PyObject *pyctx;

	if (!PyArg_ParseTuple(args, ""))
		return NULL;

	error = NULL;
	AdmContext_strip(self->ctx, &error);
	if (pyAdmError_check_and_raise(error))
		return NULL;

	pyctx = pyAdmContext_constructor(self->ctx);

	return pyctx;
}

static char pyAdmContext_set_mask_doc[] =
"AdmContext.setMask(mask) -- sets a new mask on a context";

static PyObject	*
pyAdmContext_set_mask	(pyAdmContext_object	*self,
			 PyObject		 *args,
			 PyObject		 *kwargs) 
{
	AdmError *error;
	const char *mask;
	const char *kw[] = {"mask", NULL};

	error = NULL;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "s", (char **) kw, &mask))
		return NULL;

	AdmContext_setMask(self->ctx, mask, &error);

	if (pyAdmError_check_and_raise(error))
		return NULL;

	return Py_BuildValue(""); 
}

static char pyAdmContext_clear_mask_doc[] =
"AdmContext.clearMask(mask) -- clears a mask from a context";

static PyObject	*
pyAdmContext_clear_mask (pyAdmContext_object	 *self,
			 PyObject		 *args,
			 PyObject		 *kwargs)
{
	AdmError *error;
	char *mask;
	const char *kw[] = {"mask", NULL};

	error = NULL;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "s", (char **) kw, &mask))
		return NULL;

	AdmContext_clearMaskByPath(self->ctx, mask, &error);

	if (pyAdmError_check_and_raise(error))
		return NULL;

	return Py_BuildValue(""); 
}

static char pyAdmContext_write_to_file_doc[] =
"AdmContext.writeToFile(path, mode = 0600, uid = -1, gid = -1) -- writes the context to a file";

static PyObject *
pyAdmContext_write_to_file (pyAdmContext_object *self, PyObject *args, PyObject *kwargs)
{
	const char *path;
	int uid, gid, mode;
	AdmError *error;

	const char *kw[] = {"path", "mode", "uid", "gid", NULL};

	path = NULL;
	mode = 0600;
	uid = -1;
	gid = -1;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "s|iii", (char **) kw, &path, &mode, &uid, &gid))
		return NULL;

	error = NULL;
	AdmContext_writeToFile(self->ctx, path, mode, uid, gid, &error);
	if (pyAdmError_check_and_raise(error))
		return NULL;

	return Py_BuildValue("");
}

/* ================================================================================================================== */
/* pyAdmBlackBox prototypes */

typedef struct {
	PyObject_HEAD
	AdmBlackBox	*box;
} pyAdmBlackBox_object;

static char		 pyAdmBlackBox_new_doc[];
static PyObject		*pyAdmBlackBox_new		(PyObject		 *self,
							 PyObject		 *args,
							 PyObject		 *kwargs);

static PyObject		*pyAdmBlackBox_constructor	(AdmBlackBox		 *ctx);

static void		 pyAdmBlackBox_free		(pyAdmBlackBox_object	 *self);

static int		 pyAdmBlackBox_cmp		(pyAdmBlackBox_object	 *self,
							 PyObject		 *obj);

static PyObject		*pyAdmBlackBox_getattr		(pyAdmBlackBox_object	 *self,
							 char			 *name);

static long		 pyAdmBlackBox_hash		(pyAdmBlackBox_object	 *self);

static char		 pyAdmBlackBox_read_doc[];
static PyObject		 *pyAdmBlackBox_read		(pyAdmBlackBox_object	 *self,
							 PyObject		 *args,
							 PyObject		 *kwargs);

static char		 pyAdmBlackBox_write_doc[];
static PyObject		 *pyAdmBlackBox_write		(pyAdmBlackBox_object	 *self,
							 PyObject		 *args,
							 PyObject		 *kwargs);

static char		 pyAdmBlackBox_listSerial_doc[];
static PyObject		 *pyAdmBlackBox_listSerial	(pyAdmBlackBox_object	 *self,
							 PyObject		 *args,
							 PyObject		 *kwargs);

static PyTypeObject pyAdmBlackBox_object_type = {
	PyObject_HEAD_INIT(&PyType_Type)		/** Required initialization				**/
	0,						/** ob_size		: Usually 0			**/
	"AdmBlackBox",					/** tp_name		: Type Name			**/
	sizeof(pyAdmBlackBox_object),			/** tp_basicsize	: Object Size			**/
	0,						/** tp_itemsize		: Usually 0			**/

	/* Standard Methods */
	(destructor)	pyAdmBlackBox_free,		/** tp_dealloc		: refcount = 0			**/
	(printfunc)	NULL,				/** tp_print		: print x			**/
	(getattrfunc)	pyAdmBlackBox_getattr,		/** tp_getattr		: x.attr			**/
	(setattrfunc)	NULL,				/** tp_setattr		: x.attr = v			**/
	(cmpfunc)	pyAdmBlackBox_cmp,		/** tp_compare		: cmp(x, y)			**/
	(reprfunc)	NULL,				/** tp_repr		: repr(x)			**/

	/* Type Categories */
	NULL,						/** tp_as_number	: Number Methods		**/
	NULL,						/** tp_as_sequence	: Sequence methods		**/
	NULL,						/** tp_as_mapping	: Mapping functions		**/
	(hashfunc)	pyAdmBlackBox_hash,		/** tp_hash		: dict[x]			**/
	(ternaryfunc)	NULL,				/** tp_call		: x()				**/
	(reprfunc)	NULL,				/** tp_str		: str(x)			**/
};

static struct PyMethodDef pyAdmBlackBox_methods[] = {
	{"read",	(PyCFunction) pyAdmBlackBox_read,		METH_VARARGS | METH_KEYWORDS,
									pyAdmBlackBox_read_doc},
	{"write",	(PyCFunction) pyAdmBlackBox_write,		METH_VARARGS | METH_KEYWORDS,
									pyAdmBlackBox_write_doc},
	{"listSerial",	(PyCFunction) pyAdmBlackBox_listSerial,		METH_VARARGS | METH_KEYWORDS,
									pyAdmBlackBox_listSerial_doc},
	{NULL, NULL}
};

/* ------------------------------------------------------------------*/
/* constructor for pyAdmBlackBox_objects */

static char pyAdmBlackBox_new_doc[] =
"AdmBlackBox(box_cfg) -> blackbox -- creates a new blackbox";

static PyObject *
pyAdmBlackBox_new (PyObject *self, PyObject *args, PyObject *kwargs)
{
	AdmError *error;
	AdmBlackBox *box;
	pyAdmData_object *box_cfg;
	PyObject *obj;

	const char *kw[] = {"box_cfg", NULL};

	box_cfg = NULL;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O!", (char **) kw, &pyAdmList_object_type, &box_cfg))
		return NULL;

	error = NULL;
	box = AdmBlackBox_getBox((AdmList *) box_cfg->data, &error);
	if (pyAdmError_check_and_raise(error))
		return NULL;

	obj = pyAdmBlackBox_constructor(box);
	AdmBlackBox_unref(box, NULL);

	return obj;
}

static PyObject	*
pyAdmBlackBox_constructor (AdmBlackBox *box)
{
	pyAdmBlackBox_object *pybox;
	AdmError *error;
	
	error = NULL;
	AdmBlackBox_ref(box, &error);
	if (pyAdmError_check_and_raise(error))
		return NULL;

	/* create a new pyAdmContext_object, and attach the pyAdmContext_object_type struct */
	pybox = PyObject_NEW(pyAdmBlackBox_object, &pyAdmBlackBox_object_type);
	pybox->box = box;
	
	return (PyObject *) pybox;
}

/* ------------------------------------------------------------------*/
/* Base Functions for pyAdmBlackBox_object */

static void
pyAdmBlackBox_free (pyAdmBlackBox_object *self)
{
	/* drop the ref on the box */
	AdmBlackBox_unref(self->box, NULL);
	PyMem_DEL(self);
}

/**
* pyAdmBlackBox_hash : give a value we can hash against
**/
static long
pyAdmBlackBox_hash (pyAdmBlackBox_object *self)
{
	return (long) self->box;
}

/**
* pyAdmBlackBox_cmp : comparison handler
**/
static int
pyAdmBlackBox_cmp (pyAdmBlackBox_object	 *self,
		  PyObject		 *obj)
{
	if (!pyAdmBlackBox_Check(obj))
		return (int) self - (int) obj;
	else
		return (int) self->box - (int) ((pyAdmBlackBox_object *) obj)->box;
}

static void
quux(PyObject *dict, char *name)
{
	PyObject *string, *integer;
	integer = PyInt_FromLong(1);
	string = PyString_FromString(name);
	PyDict_SetItem(dict, string, integer);
	Py_DECREF(string);
	Py_DECREF(integer);
}

/**
* pyAdmBlackBox_getattr : handles x.attr calls
**/
static PyObject *
pyAdmBlackBox_getattr (pyAdmBlackBox_object *self, char *name) 
{
	AdmError *error;

	error = NULL;
	if (!strcmp(name, "caps")) {
		PyObject *retdict;
		adm_bb_cap_t caps;

		caps = AdmBlackBox_caps(self->box, &error);
		if (pyAdmError_check_and_raise(error))
			return NULL;

		retdict = PyDict_New();

		if (ADM_BBCAP_SUP_WRITE(caps)) 
			quux(retdict, "write");

		if (ADM_BBCAP_SUP_READ(caps))
			quux(retdict, "read");

		if (ADM_BBCAP_SUP_LISTSERIAL(caps))
			quux(retdict, "listSerial");
			
		if (ADM_BBCAP_SUP_READSERIAL(caps))
			quux(retdict, "readSerial");
			
		if (ADM_BBCAP_CACHED(caps))
			quux(retdict, "cached");

		return retdict;
	}

	return Py_FindMethod(pyAdmBlackBox_methods, (PyObject *) self, name);
}

/* ------------------------------------------------------------------*/
/* Methods for pyAdmBlackBox_object */

static char pyAdmBlackBox_read_doc[] =
"AdmBlackBox.read(serial = None) -> context -- reads a context from the AdmBlackBox, reads by serial number if"
" serial is set, reads the top if serial is 0";

static PyObject *
pyAdmBlackBox_read (	pyAdmBlackBox_object	*self,
			PyObject		*args,
			PyObject		*kwargs)
{
	AdmError *error;
	AdmContext *ctx;
	PyObject *pyctx, *serial_obj, *serial_int_obj;
	long serial;
	const char *kw[] = {"serial", NULL};

	error = NULL;
	serial_obj = NULL;
	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|O", (char **) kw, &serial_obj))
		return NULL;

	if (serial_obj != NULL) {
		if (!PyNumber_Check(serial_obj))
			return PyErr_Format(PyExc_TypeError, "serial must be a numeric type");

		serial_int_obj = PyNumber_Int(serial_obj);
		serial = PyInt_AsLong(serial_int_obj);
		Py_DECREF(serial_int_obj);	

		if (serial < 0)
			return PyErr_Format(PyExc_ValueError, "serial numbers must be positive.");

		ctx = AdmBlackBox_readSerial(self->box, serial, &error);

	} else { /* serial_obj == NULL */

		ctx = AdmBlackBox_read(self->box, &error);
	}

	if (pyAdmError_check_and_raise(error))
		return NULL;

	pyctx = pyAdmContext_constructor(ctx);
	AdmContext_unref(ctx, NULL);

	return pyctx;
}

static char pyAdmBlackBox_write_doc[] =
"AdmBlackBox.write(context) -- writes a context to the AdmBlackBox";

static PyObject *
pyAdmBlackBox_write (pyAdmBlackBox_object *self, PyObject *args, PyObject *kwargs)
{
	AdmError *error;
	PyObject *pyctx;
	const char *kw[] = {"context", NULL};

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O!", (char **) kw, &pyAdmContext_object_type, &pyctx))
		return NULL;

	error = NULL;
	AdmBlackBox_write(self->box, ((pyAdmContext_object *) pyctx)->ctx, &error);
	if (pyAdmError_check_and_raise(error))
		return NULL;

	return Py_BuildValue("");
}

static char pyAdmBlackBox_listSerial_doc[] =
"AdmBlackBox.listSerial(count, top_serial = 0) -> [topnum, secondtopnum, ...] -- lists the top count serial numbers"
" below top_serial, or the top count numbers if top_serial is 0";

static PyObject *
pyAdmBlackBox_listSerial (	pyAdmBlackBox_object	*self,
				PyObject		*args,
				PyObject		*kwargs)
{
	AdmError *error;
	PyObject *retlist, *temp;
	unsigned foo, bar;
	unsigned int n, top_serial, count;
	const char *kw[] = {"count", "top_serial", NULL};

	foo = 0;
	bar = 1;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|ll", (char **) kw, &foo, &bar))
		return NULL;

	if ((foo < 0) || (bar < 0))
		return PyErr_Format(PyExc_ValueError, "serial numbers must be positive.");

	n = foo;
	top_serial = bar;

	error = NULL;
	{
		unsigned int j, serial_array[n];


		count = AdmBlackBox_listSerial(self->box, serial_array, n, top_serial, &error);
		if (pyAdmError_check_and_raise(error))
			return NULL;

		retlist = PyList_New(n);
		for (j = 0; j < count; ++j ) {
			temp = PyInt_FromLong(serial_array[j]);
			PyList_Insert(retlist, j, temp);
			Py_DECREF(temp);
		};
	}

	return retlist;
}

/* ================================================================================================================== */
/* Switchboard stuff */

static char		 pyAdmSwitchboard_getSwbDir_doc[];
static PyObject		*pyAdmSwitchboard_getSwbDir		(PyObject		 *self,
								 PyObject		 *args);

static char		 pyAdmSwitchboard_setSwbDir_doc[];
static PyObject		*pyAdmSwitchboard_setSwbDir		(PyObject		 *self,
								 PyObject		 *args,
								 PyObject		 *kwargs);

static char		 pyAdmBlackBox_getCacheDir_doc[];
static PyObject		*pyAdmBlackBox_getCacheDir		(PyObject		 *self,
								 PyObject		 *args);

static char		 pyAdmBlackBox_setCacheDir_doc[];
static PyObject		*pyAdmBlackBox_setCacheDir		(PyObject		 *self,
								 PyObject		 *args,
								 PyObject		 *kwargs);

static char		 pyAdmSwitchboard_readNsp_doc[];
static PyObject		*pyAdmSwitchboard_readNsp		(PyObject		 *self,
								 PyObject		 *args,
								 PyObject		 *kwargs);

static char		 pyAdmSwitchboard_writeNsp_doc[];
static PyObject		*pyAdmSwitchboard_writeNsp		(PyObject		 *self,
								 PyObject		 *args,
								 PyObject		 *kwargs);

static char		 pyAdmSwitchboard_markNsp_doc[];
static PyObject		*pyAdmSwitchboard_markNsp		(PyObject		 *self,
								 PyObject		 *args,
								 PyObject		 *kwargs);

static char		 pyAdmSwitchboard_checkNspMark_doc[];
static PyObject		*pyAdmSwitchboard_checkNspMark		(PyObject		 *self,
								 PyObject		 *args,
								 PyObject		 *kwargs);

static char		 pyAdmSwitchboard_readNspCtx_doc[];
static PyObject		*pyAdmSwitchboard_readNspCtx		(PyObject		 *self,
								 PyObject		 *args,
								 PyObject		 *kwargs);

static char		 pyAdmSwitchboard_writeNspCtx_doc[];
static PyObject		*pyAdmSwitchboard_writeNspCtx		(PyObject		 *self,
								 PyObject		 *args,
								 PyObject		 *kwargs);

static char		 pyAdmSwitchboard_getNspCfg_doc[];
static PyObject		*pyAdmSwitchboard_getNspCfg		(PyObject		 *self,
								 PyObject		 *args,
								 PyObject		 *kwargs);

static char		 pyAdmSwitchboard_getSubspCfg_doc[];
static PyObject		*pyAdmSwitchboard_getSubspCfg		(PyObject		 *self,
								 PyObject		 *args,
								 PyObject		 *kwargs);

static char		 pyAdmSwitchboard_getBox_doc[];
static PyObject		*pyAdmSwitchboard_getBox		(PyObject		 *self,
								 PyObject		 *args,
								 PyObject		 *kwargs);

/* functions */

static char pyAdmSwitchboard_getSwbDir_doc[] =
"getSwbDir() -> path -- gets the current switchboard directory";

static PyObject	*
pyAdmSwitchboard_getSwbDir		(PyObject		 *self,
					 PyObject		 *args)
{
	AdmError *error;
	AdmRefStr *path;
	PyObject *retobj;

	error = NULL;

	if (!PyArg_ParseTuple(args, ""))
		return NULL;

	path = AdmSwitchboard_getSwbDir(&error);
	if (pyAdmError_check_and_raise(error))
		return NULL;

	retobj = PyString_FromString(AdmRefStr_str(path));
	AdmRefStr_unref(path);
	return retobj;
}

static char pyAdmSwitchboard_setSwbDir_doc[] =
"setSwbDir(path) -- sets the current switchboard directory";

static PyObject	*
pyAdmSwitchboard_setSwbDir		(PyObject		 *self,
					 PyObject		 *args,
					 PyObject		 *kwargs)
{
	AdmError *error;
	const char *path;
	const char *kw[] = {"path", NULL};

	error = NULL;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "s", (char **) kw, &path))
		return NULL;

	AdmSwitchboard_setSwbDir(path, &error);
	if (pyAdmError_check_and_raise(error))
		return NULL;

	return Py_BuildValue("");
}

static char pyAdmBlackBox_getCacheDir_doc[] =
"getCacheDir() -> path -- gets the current cache directory";

static PyObject	*
pyAdmBlackBox_getCacheDir		(PyObject		 *self,
					 PyObject		 *args)
{
	AdmError *error;
	PyObject *retobj;
	AdmRefStr *path;

	error = NULL;

	if (!PyArg_ParseTuple(args, ""))
		return NULL;

	path = AdmBlackBox_getCacheDir(&error);
	if (pyAdmError_check_and_raise(error))
		return NULL;

	retobj = PyString_FromString(AdmRefStr_str(path));
	AdmRefStr_unref(path);
	return retobj;
}

static char pyAdmBlackBox_setCacheDir_doc[] =
"setCacheDir(path) -- sets the current cache directory";

static PyObject	*
pyAdmBlackBox_setCacheDir		(PyObject		 *self,
					 PyObject		 *args,
					 PyObject		 *kwargs)
{
	AdmError *error;
	const char *path;
	const char *kw[] = {"path", NULL};

	error = NULL;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "s", (char **) kw, &path))
		return NULL;

	AdmBlackBox_setCacheDir(path, &error);
	if (pyAdmError_check_and_raise(error))
		return NULL;

	return Py_BuildValue("");
}

static char pyAdmSwitchboard_readNspCtx_doc[] =
"readNspCtx(namespace) -> context -- reads the given namespace context from the switchboard dir";

static PyObject	*
pyAdmSwitchboard_readNspCtx	(PyObject		 *self,
				 PyObject		 *args,
				 PyObject		 *kwargs)
{
	AdmContext *ctx;
	AdmError *error;
	PyObject *retobj;
	const char *namespace;
	const char *kw[] = {"namespace", NULL};

	error = NULL;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "s", (char **) kw, &namespace))
		return NULL;

	ctx = AdmSwitchboard_readNspCtx(namespace, &error);
	if (pyAdmError_check_and_raise(error))
		return NULL;
	
	retobj = pyAdmContext_constructor(ctx);
	AdmContext_unref(ctx, NULL);

	return retobj;
}

static char pyAdmSwitchboard_writeNspCtx_doc[] =
"writeNspCtx(namespace, context, mode = 0644, uid = -1, gid = -1) -> -- writes the given namespace context to the switchboard dir";

static PyObject	*
pyAdmSwitchboard_writeNspCtx	(PyObject		 *self,
				 PyObject		 *args,
				 PyObject		 *kwargs)
{
	AdmError *error;
	PyObject *pyctx;
	int mode;
	long uid, gid;
	const char *namespace;
	const char *kw[] = {"namespace", "context", "mode", "uid", "gid", NULL};

	error = NULL;

	mode = 0644;
	uid = -1;
	gid = -1;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "sO!|ill", (char **) kw,
			&namespace, &pyAdmContext_object_type, &pyctx, &mode, &uid, &gid))
		return NULL;

	AdmSwitchboard_writeNspCtx(namespace, ((pyAdmContext_object *) pyctx)->ctx, mode, uid, gid, &error);
	if (pyAdmError_check_and_raise(error))
		return NULL;
	
	return Py_BuildValue("");
}

static char pyAdmSwitchboard_getNspCfg_doc[] =
"getNspCfg(context, namespace) -> namespace_cfg -- gets a namespace config from a switchboard context";

static PyObject	*
pyAdmSwitchboard_getNspCfg	(PyObject		 *self,
				 PyObject		 *args,
				 PyObject		 *kwargs)
{
	AdmError *error;
	AdmList *list;
	PyObject *pyctx;
	const char *namespace;
	const char *kw[] = {"context", "namespace", NULL};
	
	error = NULL;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O!s", (char **) kw, &pyAdmContext_object_type, &pyctx, &namespace))
		return NULL;

	list = AdmSwitchboard_getNspCfg_b(((pyAdmContext_object *) pyctx)->ctx, namespace, &error);
	if (pyAdmError_check_and_raise(error))
		return NULL;

	return pyAdmData_constructor((AdmData *) list);
}

static char pyAdmSwitchboard_getSubspCfg_doc[] =
"getSubspCfg(namespace_cfg, subspace = \"config\") -> subsp_cfg -- gets a subspace_cfg from a namespace_cfg";

static PyObject	*
pyAdmSwitchboard_getSubspCfg	(PyObject		 *self,
				 PyObject		 *args,
				 PyObject		 *kwargs)
{
	AdmError *error;
	AdmList *list;
	PyObject *namespace_cfg;
	const char *subspace;
	const char *kw[] = {"namespace_cfg", "subspace", NULL};
	
	error = NULL;
	subspace = NULL;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O!|z", (char **) kw, &pyAdmList_object_type, &namespace_cfg, &subspace))
		return NULL;
	
	list = AdmSwitchboard_getSubspCfg_b((AdmList *) ((pyAdmData_object *) namespace_cfg)->data, subspace, &error);
	if (pyAdmError_check_and_raise(error))
		return NULL;

	return pyAdmData_constructor((AdmData *) list);
}

static char pyAdmSwitchboard_getBox_doc[] =
"getBox(subspace_cfg, box_name) -> blackbox -- initalizes a box in a subspace";

static PyObject	*
pyAdmSwitchboard_getBox		(PyObject		 *self,
				 PyObject		 *args,
				 PyObject		 *kwargs)
{
	AdmError *error;
	AdmBlackBox *box;
	PyObject *subspace_cfg, *retobj;
	const char *box_name;
	const char *kw[] = {"subspace_cfg", "box_name", NULL};
	
	error = NULL;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O!s", (char **) kw, &pyAdmList_object_type, &subspace_cfg, &box_name))
		return NULL;

	
	box = AdmSwitchboard_getBox((AdmList *) ((pyAdmData_object *) subspace_cfg)->data, box_name, &error);
	if (pyAdmError_check_and_raise(error))
		return NULL;

	retobj = pyAdmBlackBox_constructor(box);
	AdmBlackBox_unref(box, NULL);

	return retobj;
}

static char pyAdmSwitchboard_readNsp_doc[] =
"readNsp(namespace, subspace = \"config\", depth = 0, namespace_cfg) -> context -- reads a subspace, uses namespace or the"
" namespace configured by namespace_cfg, cascades to depth, 0 means all";

static PyObject	*
pyAdmSwitchboard_readNsp	(PyObject		 *self,
				 PyObject		 *args,
				 PyObject		 *kwargs)
{
	AdmError *error;
	AdmContext *ctx;
	PyObject *pyctx;
	pyAdmData_object *namespace_cfg;
	int depth;
	const char *namespace, *subspace;
	const char *kw[] = {"namespace", "subspace", "depth", "namespace_cfg", NULL};
	
	error = NULL;
	depth = 0;
	namespace_cfg = NULL;
	subspace = NULL;
	namespace = NULL;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|zziO!", (char **) kw, &namespace, &subspace,
				&depth, &pyAdmList_object_type, &namespace_cfg))
		return NULL;

	if (depth < 0)
		return PyErr_Format(PyExc_ValueError, "depth can not be negative");
		
	if ((namespace_cfg == NULL) && (namespace == NULL))
		return PyErr_Format(PyExc_TypeError, "Expected namespace or namespace_cfg, found neither.");

	if ((namespace_cfg != NULL) && (namespace != NULL))
		return PyErr_Format(PyExc_TypeError, "namespace and namespace_cfg are mutually exclusive");

	ctx = (namespace_cfg == NULL)
		? AdmSwitchboard_readNsp(namespace, subspace, depth, &error) 
		: AdmSwitchboard_readNspWithConfig((AdmList *) namespace_cfg->data, subspace, depth, &error);

	if (pyAdmError_check_and_raise(error))
		return NULL;

	pyctx = pyAdmContext_constructor(ctx);
	AdmContext_unref(ctx, NULL);

	return pyctx;
}

static char pyAdmSwitchboard_writeNsp_doc[] =
"writeNsp(namespace, context, subspace = \"config\", namespace_cfg) -- writes the context to namespace, or to the"
" namespace configured by namespace_cfg";

static PyObject	*
pyAdmSwitchboard_writeNsp	(PyObject		 *self,
				 PyObject		 *args,
				 PyObject		 *kwargs)
{
	AdmError *error;
	pyAdmData_object *namespace_cfg;
	pyAdmContext_object *pyctx;
	const char *namespace, *subspace;
	const char *kw[] = {"namespace", "context", "subspace", "namespace_cfg", NULL};
	
	error = NULL;
	namespace_cfg = NULL;
	subspace = NULL;
	namespace = NULL;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|zO!zO!", (char **) kw, &namespace,
			&pyAdmContext_object_type, &pyctx, &subspace, &pyAdmList_object_type, &namespace_cfg))
		return NULL;

	if (pyctx == NULL)
		return PyErr_Format(PyExc_TypeError, "context required.");

	if ((namespace_cfg == NULL) && (namespace == NULL))
		return PyErr_Format(PyExc_TypeError, "Expected namespace or namespace_cfg, found neither.");

	if ((namespace_cfg != NULL) && (namespace != NULL))
		return PyErr_Format(PyExc_TypeError, "namespace and namespace_cfg are mutually exclusive");

	if (namespace_cfg == NULL)
		AdmSwitchboard_writeNsp(namespace, subspace, pyctx->ctx, &error); 
	else
		AdmSwitchboard_writeNspWithConfig((AdmList *) namespace_cfg->data, subspace, pyctx->ctx, &error);

	if (pyAdmError_check_and_raise(error))
		return NULL;

	return Py_BuildValue("");
}

static char pyAdmSwitchboard_markNsp_doc[] =
"markNsp(namespace, mark, subspace = \"config\", namespace_cfg) -- marks the namespace, or the namespace"
" configured by namespace_cfg";

static PyObject	*
pyAdmSwitchboard_markNsp	(PyObject		 *self,
				 PyObject		 *args,
				 PyObject		 *kwargs)
{
	AdmError *error;
	pyAdmData_object *namespace_cfg;
	const char *namespace, *mark, *subspace;
	const char *kw[] = {"namespace", "mark", "subspace", "namespace_cfg", NULL};
	
	error = NULL;
	namespace_cfg = NULL;
	subspace = NULL;
	namespace = NULL;
	mark = NULL;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|zszO!", (char **) kw, &namespace, &mark, &subspace,
				&pyAdmList_object_type, &namespace_cfg))
		return NULL;

	if (mark == NULL)
		return PyErr_Format(PyExc_TypeError, "mark required.");

	if ((namespace_cfg == NULL) && (namespace == NULL))
		return PyErr_Format(PyExc_TypeError, "Expected namespace or namespace_cfg, found neither.");

	if ((namespace_cfg != NULL) && (namespace != NULL))
		return PyErr_Format(PyExc_TypeError, "namespace and namespace_cfg are mutually exclusive");

	if (namespace_cfg == NULL)
		AdmSwitchboard_markNsp(namespace, subspace, mark, &error); 
	else
		AdmSwitchboard_markNspWithConfig((AdmList *) namespace_cfg->data, subspace, mark, &error);

	if (pyAdmError_check_and_raise(error))
		return NULL;

	return Py_BuildValue("");
}

static char pyAdmSwitchboard_checkNspMark_doc[] =
"checkNspMark(namespace, mark, subspace = \"config\", namespace_cfg) -- checks a mark on the namespace, or the namespace"
" configured by namespace_cfg";

static PyObject	*
pyAdmSwitchboard_checkNspMark	(PyObject		 *self,
				 PyObject		 *args,
				 PyObject		 *kwargs)
{
	AdmError *error;
	pyAdmData_object *namespace_cfg;
	const char *namespace, *mark, *subspace;
	const char *kw[] = {"namespace", "mark", "subspace", "namespace_cfg", NULL};
	int retval;
	
	error = NULL;
	namespace_cfg = NULL;
	subspace = NULL;
	namespace = NULL;
	mark = NULL;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|zszO!", (char **) kw, &namespace, &mark, &subspace,
				&pyAdmList_object_type, &namespace_cfg))
		return NULL;

	if (mark == NULL)
		return PyErr_Format(PyExc_TypeError, "mark required.");

	if ((namespace_cfg == NULL) && (namespace == NULL))
		return PyErr_Format(PyExc_TypeError, "Expected namespace or namespace_cfg, found neither.");

	if ((namespace_cfg != NULL) && (namespace != NULL))
		return PyErr_Format(PyExc_TypeError, "namespace and namespace_cfg are mutually exclusive");

	retval = (namespace_cfg == NULL)
		? AdmSwitchboard_checkNspMark(namespace, subspace, mark, &error) 
		: AdmSwitchboard_checkNspMarkWithConfig((AdmList *) namespace_cfg->data, subspace, mark, &error);

	if (pyAdmError_check_and_raise(error))
		return NULL;

	return Py_BuildValue("i", retval);
}


/* ================================================================================================================== */
/* pyalchemist stuff */

/* methods exposed by pyalchemist */
static struct PyMethodDef pyalchemist_methods[] = {
	{"validName",		(PyCFunction) pyAdmData_validName,			METH_VARARGS | METH_KEYWORDS,
											pyAdmData_validName_doc},

	{"AdmContext",		(PyCFunction) pyAdmContext_new,				METH_VARARGS | METH_KEYWORDS,
											pyAdmContext_new_doc},
	{"AdmBlackBox",		(PyCFunction) pyAdmBlackBox_new,			METH_VARARGS | METH_KEYWORDS,
											pyAdmBlackBox_new_doc},

	{"readNspCtx",		(PyCFunction) pyAdmSwitchboard_readNspCtx,		METH_VARARGS | METH_KEYWORDS,
											pyAdmSwitchboard_readNspCtx_doc},
	{"writeNspCtx",		(PyCFunction) pyAdmSwitchboard_writeNspCtx,		METH_VARARGS | METH_KEYWORDS,
											pyAdmSwitchboard_writeNspCtx_doc},

	{"readNsp",		(PyCFunction) pyAdmSwitchboard_readNsp,			METH_VARARGS | METH_KEYWORDS,
											pyAdmSwitchboard_readNsp_doc},
	{"writeNsp",		(PyCFunction) pyAdmSwitchboard_writeNsp,		METH_VARARGS | METH_KEYWORDS,
											pyAdmSwitchboard_writeNsp_doc},
	{"markNsp",		(PyCFunction) pyAdmSwitchboard_markNsp,			METH_VARARGS | METH_KEYWORDS,
											pyAdmSwitchboard_markNsp_doc},
	{"checkNspMark",	(PyCFunction) pyAdmSwitchboard_checkNspMark,		METH_VARARGS | METH_KEYWORDS,
											pyAdmSwitchboard_checkNspMark_doc},

	{"getNspCfg",		(PyCFunction) pyAdmSwitchboard_getNspCfg,		METH_VARARGS | METH_KEYWORDS,
											pyAdmSwitchboard_getNspCfg_doc},
	{"getSubspCfg",		(PyCFunction) pyAdmSwitchboard_getSubspCfg,		METH_VARARGS | METH_KEYWORDS,
											pyAdmSwitchboard_getSubspCfg_doc},
	{"getBox",		(PyCFunction) pyAdmSwitchboard_getBox,			METH_VARARGS | METH_KEYWORDS,
											pyAdmSwitchboard_getBox_doc},

	{"getSwbDir",		(PyCFunction) pyAdmSwitchboard_getSwbDir,		METH_VARARGS,
											pyAdmSwitchboard_getSwbDir_doc},
	{"setSwbDir",		(PyCFunction) pyAdmSwitchboard_setSwbDir,		METH_VARARGS | METH_KEYWORDS,
											pyAdmSwitchboard_setSwbDir_doc},
	{"getCacheDir",		(PyCFunction) pyAdmBlackBox_getCacheDir,		METH_VARARGS,
											pyAdmBlackBox_getCacheDir_doc},
	{"setCacheDir",		(PyCFunction) pyAdmBlackBox_setCacheDir,		METH_VARARGS | METH_KEYWORDS,
											pyAdmBlackBox_setCacheDir_doc},

	{NULL, NULL}
};

/* This function is magically called at module load time */
/* AND MUST BE NAMED 'initpyalchemist' */
void initpyalchemist() {
	PyObject *mod, *dict, *impmod;

	/* Tell python about pyalchemist's methods */
	mod = Py_InitModule("pyalchemist", pyalchemist_methods);

	/* Get __dict__ to fill it with some useful globals */
	dict = PyModule_GetDict(mod);

	PyDict_SetItemString(dict, "AdmListType",	(PyObject *) &pyAdmList_object_type);
	PyDict_SetItemString(dict, "AdmStringType",	(PyObject *) &pyAdmString_object_type);
	PyDict_SetItemString(dict, "AdmBase64Type",	(PyObject *) &pyAdmBase64_object_type);
	PyDict_SetItemString(dict, "AdmCopyType",	(PyObject *) &pyAdmCopy_object_type);
	PyDict_SetItemString(dict, "AdmBoolType",	(PyObject *) &pyAdmBool_object_type);
	PyDict_SetItemString(dict, "AdmIntType",	(PyObject *) &pyAdmInt_object_type);
	PyDict_SetItemString(dict, "AdmFloatType",	(PyObject *) &pyAdmFloat_object_type);

	PyDict_SetItemString(dict, "AdmIdentityType",	(PyObject *) &pyAdmIdentity_object_type);
	PyDict_SetItemString(dict, "AdmContextType",	(PyObject *) &pyAdmContext_object_type);

	PyDict_SetItemString(dict, "AdmBlackBoxType",	(PyObject *) &pyAdmBlackBox_object_type);

	/* load python convenience module */
	impmod = PyImport_ImportModule("pyalchemist_python");
	if (impmod != NULL) {
		PyObject *key_list, *key, *value, *impdict;
		unsigned int i, k;

		impdict = PyModule_GetDict(impmod);
		key_list = PyDict_Keys(impdict);
		k = PyList_Size(key_list);
		for (i = 0; i < k; ++i) {
			key = PyList_GetItem(key_list, i);
			value = PyDict_GetItem(impdict, key);
			PyDict_SetItem(dict, key, value);
		}
		Py_DECREF(key_list);
	}
	
}


/* generic structs */
//
//typedef struct {
//	PyObject_HEAD
//	generic_data_type *data1;
//	generic_data_type *data2;
//	...
//} generic_object;
//
//static PySequenceMethods generic_sequence_methods = {
//	(inquiry)		NULL,			/** sq_lenth		: __len__(x)			**/
//	(binaryfunc)		NULL,			/** sq_concat		: __add__(x, y)			**/
//	(intargfunc)		NULL,			/** sq_repeat		: __mul__(x, n)__		**/
//	(intargfunc)		NULL,			/** sq_item		: __getitem__(x, n)__		**/
//	(intintargfunc)		NULL,			/** sq_slice		: __getslice__(x, i, j)__	**/
//	(intobjargproc)		NULL,			/** sq_ass_item		: __setitem__(x, n, v)__	**/
//	(intintobjargproc)	NULL,			/** sq_ass_slice	: __setslice__(x, i, j, v)__	**/
//};
//
//static PyMappingMethods generic_mapping_methods = {
//	(inquiry)		NULL,			/** mp_lenth		: __len__(x)			**/
//	(binaryfunc)		NULL,			/** mp_subscript	: __getitem__(x, key)		**/
//	(objobjargproc)		NULL,			/** mp_ass_subscript	: __setitem__(x, key, value)	**/
//};
//
///** the pyAdmContext_type structure is a PyTypeObject struct, and holds our type callbacks **/
//static PyTypeObject generic_object_type = {
//	PyObject_HEAD_INIT(&PyType_Type)		/** Required initialization				**/
//	0,						/** ob_size		: Usually 0			**/
//	"GenericName",					/** tp_name		: Type Name			**/
//	sizeof(generic_object),				/** tp_basicsize	: Object Size			**/
//	0,						/** tp_itemsize		: Usually 0			**/
//
//	/* Standard Methods */
//	(destructor)	NULL,				/** tp_dealloc		: refcount = 0			**/
//	(printfunc)	NULL,				/** tp_print		: print x			**/
//	(getattrfunc)	NULL,				/** tp_getattr		: x.attr			**/
//	(setattrfunc)	NULL,				/** tp_setattr		: x.attr = v			**/
//	(cmpfunc)	NULL,				/** tp_compare		: cmp(x, y)			**/
//	(reprfunc)	NULL,				/** tp_repr		: repr(x)			**/
//
//	/* Type Categories */
//	NULL,						/** tp_as_number	: Number Methods		**/
//	NULL,						/** tp_as_sequence	: Sequence methods		**/
//	NULL,						/** tp_as_mapping	: Mapping functions		**/
//	(hashfunc)	NULL,				/** tp_hash		: dict[x]			**/
//	(binaryfunc)	NULL,				/** tp_call		: x()				**/
//	(reprfunc)	NULL,				/** tp_str		: str(x)			**/
//};
//

