/**
 * Copyright (C) 2000, 2001 Red Hat, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/**
 * @brief refstrizer
 *
 * The refstrizer provides refcounted, guaranteed unique strings. This allows refstrs
 * to be 'copied' cheaply, provides a cheap strlen, and allows comparison of refstrs
 * by comparison of refstr pointers (since they are unique).
 *
 * The internal data representation is based on Red/Black trees as described
 * in Introduction to Algorithms by Cormen/Leiserson/Rivest.
 *
 * Red/Black trees are fairly efficient in time and space and therefore a good
 * form of internal representation.
 *
 * @file        refstr.c
 * @author      Crutcher Dunnavant <crutcher@redhat.com>
 * @date        05/06/2001
 * @version     1.0
 */

#include <stddef.h>
#include <stdlib.h>
#include <string.h>

/** Mutex block **/
#include <glib.h>
#define REFSTR_MUTEX_TYPE GStaticMutex
#define REFSTR_MUTEX_INIT G_STATIC_MUTEX_INIT
#define REFSTR_MUTEX_LOCK(M)	g_static_mutex_lock(M)
#define REFSTR_MUTEX_UNLOCK(M)	g_static_mutex_unlock(M)
/* End Mutex Block */
	
#include "refstr.h"

/*****************************************************************************
 *****************************************************************************
 * RefStrizer. Trust me, it's better this way.
 *
 * If you are curious, it is a pretty efficient
 * hash table of red-black trees of refcounted strings
 *
 * I would strongly suggest that you NOT touch the tree code without a copy
 * of CLR next to you. If you dont know what CLR is, /really/ dont touch it.
 * 
 * I dare you to make it faster.
 ****************************************************************************/

/* nothing to see here, move along */
typedef struct refstr_head refstr_head;
struct refstr_head {
	refstr_head **slot, *parent, *left, *right;
	size_t length;
	size_t refcount;
	enum { red, black } color;
	refstr string;
};
#define REFSTR_GETHEAD(REFSTR) \
({ \
 	refstr *REFSTR_TMP = (REFSTR); \
	(refstr_head *) ( REFSTR_TMP - &((refstr_head *) NULL)->string ); \
})

/** We special case the empty string, to avoid screwing with our hash algorithms. We set it
 * with an initial ref of 1, so it should never get de-allocated. We also use it for the
 * sentry node in the trees. */
#define NIL &empty_str
static refstr_head empty_str = {
	slot:		NULL,
	parent:		NIL,
	left:		NIL,
	right:		NIL,
	length:		0,
	refcount:	1,
	color:		black,
	string:		'\0'
};

/** With population size N, mean slot density is N/HASHSIZE,
 * since the hash slots are red-black trees, mean depth is
 * ln (N/HASHSIZE), For HASHSIZE = (1 << A) and N = (1 << B),
 * mean depth is B - A, so pick an appropriate HASHSIZE :) */
#define HASHSIZE (1 << 10)
refstr_head *refstr_hash_table[HASHSIZE] = {NULL, };

static REFSTR_MUTEX_TYPE refstr_hash_mutex = REFSTR_MUTEX_INIT;


/* Private Prototypes */
static inline refstr_head *	refstr_malloc_refstr	(const char * str, size_t s_len);
static inline void		refstr_free_refstr	(refstr_head * rstr_h_p);

static inline refstr_head **	refstr_get_hash_slot	(const char * str, size_t s_len);
static inline size_t		refstr_hash_str		(const char * str, size_t s_len);
static inline int		refstr_str_tok_cmp	(const char *str, size_t s_len, refstr_head *rstr_h_p);

static inline void		refstr_left_rotate	(refstr_head **slot, refstr_head *z_rstr_h_p);
static inline void		refstr_right_rotate	(refstr_head **slot, refstr_head *z_rstr_h_p);


/* Private Functions */
static inline refstr_head *
refstr_malloc_refstr(const char * str, size_t s_len)
{
	refstr_head *rstr_h_p; 
	char *string;

	/* malloc space for the meta data and the string */
	rstr_h_p = (refstr_head *) malloc (sizeof(refstr_head) + s_len); 

	/* copy the string, the length, and init the refcount */
	string = (char *) &rstr_h_p->string;
	strncpy(string, str, s_len); 
	string[s_len] = '\0'; /* null terminate */

	rstr_h_p->length = s_len; 
	rstr_h_p->refcount = 1; 

	return rstr_h_p; 
}

static inline void
refstr_free_refstr(refstr_head * rstr_h_p)
{
	free(rstr_h_p); 
}

static inline refstr_head **
refstr_get_hash_slot (const char * str, size_t s_len) 
{

	return &refstr_hash_table[refstr_hash_str(str, s_len)];
}

static inline size_t
refstr_hash_str(const char * str, size_t s_len)
{
	/* the important thing about this hash function is that it hashes on the
	 * end of strings, making refstr_str_tok_cmp much more likly to exit early */
	int i; 
	i = str[s_len - 1]; 
	if (s_len > 1)
		i |= ((int) str[s_len - 2]) << 8; 
	if (s_len > 2) 
		i |= ((int) str[s_len - 3]) << 16;
	i += s_len;
	return ((i >> 8) ^ i) & (HASHSIZE - 1);
}

static inline int
refstr_str_tok_cmp (const char *str, size_t s_len, refstr_head *rstr_h_p)
{
	int delta; 

	/* shorter strings are always 'less', longer ones are always 'more'
	 * this lets us skip many many strcmps, since we already have length
	 * laying arround :) */
	delta = s_len - rstr_h_p->length; 
	return (delta != 0) ? delta : memcmp(str, &rstr_h_p->string, s_len); 
}

static inline void
refstr_left_rotate (refstr_head **slot, refstr_head *z_rstr_h_p)
{
	refstr_head *y_rstr_h_p, *x_rstr_h_p; 
	x_rstr_h_p = z_rstr_h_p; 
	y_rstr_h_p = x_rstr_h_p->right; 
	
	x_rstr_h_p->right = y_rstr_h_p->left; 
	if (y_rstr_h_p->left != NIL)
		y_rstr_h_p->left->parent = x_rstr_h_p; 

	y_rstr_h_p->parent = x_rstr_h_p->parent; 

	if (x_rstr_h_p->parent == NIL) { 
		*slot = y_rstr_h_p; 
	} else if (x_rstr_h_p == x_rstr_h_p->parent->left) { 
		x_rstr_h_p->parent->left = y_rstr_h_p; 
	} else { 
		x_rstr_h_p->parent->right = y_rstr_h_p; 
	} 

	y_rstr_h_p->left = x_rstr_h_p; 
	x_rstr_h_p->parent = y_rstr_h_p; 
}

static inline void
refstr_right_rotate (refstr_head **slot, refstr_head *z_rstr_h_p)
{
 	refstr_head *x_rstr_h_p, *y_rstr_h_p; 
	y_rstr_h_p = z_rstr_h_p; 
	x_rstr_h_p = y_rstr_h_p->left; 

	y_rstr_h_p->left = x_rstr_h_p->right; 
	if (x_rstr_h_p->right != NIL)
		x_rstr_h_p->right->parent = y_rstr_h_p; 

	x_rstr_h_p->parent = y_rstr_h_p->parent; 

	if (y_rstr_h_p->parent == NIL) { 
		*slot = x_rstr_h_p; 
	} else if (y_rstr_h_p == y_rstr_h_p->parent->right) { 
		y_rstr_h_p->parent->right = x_rstr_h_p; 
	} else { 
		y_rstr_h_p->parent->left = x_rstr_h_p; 
	} 

	x_rstr_h_p->right = y_rstr_h_p;
	y_rstr_h_p->parent = x_rstr_h_p;
}

/**
* Takes a const char pointer, and gets a refstr for that string.
* Creates a refstr if force is specified and no matching refstr already exists.
* @brief Get a refstr
* @param str the string to refstrize
* @param force whether or not to create a new refstr if one does not exist
* @note a reference is returned with the refstr
* @note a strlen is performed to determine the length of the string.
* @return a refstr pointer, or NULL, if none is found
*/
refstr *
refstr_get (const char *str, int force)
{
	return refstr_nget(str, strlen(str), force);
}

/**
* Takes a const char pointer and a string length, and gets a refstr for that string.
* Creates a refstr if force is specified and no matching refstr already exists.
* @brief Get a refstr
* @param str the string to refstrize
* @param s_len the string length to use
* @param force whether or not to create a new refstr if one does not exist
* @note a reference is returned with the refstr
* @return a refstr pointer, or NULL, if none is found
*/
refstr *
refstr_nget (const char *str, size_t s_len, int force)
{
	/* Take a string, return a pointer to a refstr
	 * that holds a copy of that string,
	 * creating the refstr if necessary.
	 * (You now own a ref to the refstr) */
       	int cmp;
	refstr_head **slot, *rstr_h_p, *x_rstr_h_p, *y_rstr_h_p;

	/* break off this case to avoid sending a s_len == 0 to the hash */
	if (s_len == 0) {
		/* lock here to protect the refcount */
		REFSTR_MUTEX_LOCK(&refstr_hash_mutex);
		++ empty_str.refcount;
		rstr_h_p = &empty_str;
		goto tok_done;
	}

	/* wait till after the hash func to lock. */
	slot = refstr_get_hash_slot(str, s_len);
	REFSTR_MUTEX_LOCK(&refstr_hash_mutex);

	/* why look, its an uninitialized slot */
	*slot = *slot ?: NIL;

	/* seek on the tree */
	y_rstr_h_p = NIL;
	rstr_h_p = *slot;
	while (rstr_h_p != NIL) {
		if ((cmp = refstr_str_tok_cmp(str, s_len, rstr_h_p)) == 0) {
			++ rstr_h_p->refcount;
			goto tok_done;
		}
		y_rstr_h_p = rstr_h_p;
		rstr_h_p = (cmp < 0) ? rstr_h_p->left : rstr_h_p->right;
	}

	if (force == 0) {
		/* adjust to return nothing */
		rstr_h_p = NULL;
		goto tok_done;
	} else {
		rstr_h_p = refstr_malloc_refstr(str, s_len);
		rstr_h_p->slot = slot;
		rstr_h_p->left = NIL;
		rstr_h_p->right = NIL;
		rstr_h_p->parent = y_rstr_h_p;

		if (y_rstr_h_p == NIL) {
			*slot = rstr_h_p;
		} else {
			if (cmp < 0) 
				y_rstr_h_p->left = rstr_h_p;
			else 
				y_rstr_h_p->right = rstr_h_p;
		}

		rstr_h_p->color = red;

		x_rstr_h_p = rstr_h_p;	
		while ((x_rstr_h_p != *slot) && (x_rstr_h_p->parent->color == red)) {
			if (x_rstr_h_p->parent == x_rstr_h_p->parent->parent->left) {
				y_rstr_h_p = x_rstr_h_p->parent->parent->right;
				if (y_rstr_h_p->color == red) {
					x_rstr_h_p->parent->color = black;
					y_rstr_h_p->color = black;
					x_rstr_h_p->parent->parent->color = red;
					x_rstr_h_p = x_rstr_h_p->parent->parent;
				} else {
					if (x_rstr_h_p == x_rstr_h_p->parent->right) {
						x_rstr_h_p = x_rstr_h_p->parent;
						refstr_left_rotate(slot, x_rstr_h_p);
					}
					x_rstr_h_p->parent->color = black;
					x_rstr_h_p->parent->parent->color = red;
					refstr_right_rotate(slot, x_rstr_h_p->parent->parent);
				}
			} else {/* x_rstr_h_p->parent == x_rstr_h_p->parent->parent->right */
				y_rstr_h_p = x_rstr_h_p->parent->parent->left;
				if (y_rstr_h_p->color == red) {
					x_rstr_h_p->parent->color = black;
					y_rstr_h_p->color = black;
					x_rstr_h_p->parent->parent->color = red;
					x_rstr_h_p = x_rstr_h_p->parent->parent;
				} else {
					if (x_rstr_h_p == x_rstr_h_p->parent->left) {
						x_rstr_h_p = x_rstr_h_p->parent;
						refstr_right_rotate(slot, x_rstr_h_p);
					}
					x_rstr_h_p->parent->color = black;
					x_rstr_h_p->parent->parent->color = red;
					refstr_left_rotate(slot, x_rstr_h_p->parent->parent);
				}
			}
		}
		(*slot)->color = black;
	}
tok_done:
	REFSTR_MUTEX_UNLOCK(&refstr_hash_mutex);
	return rstr_h_p ? &rstr_h_p->string : NULL;
}

/**
* Returns the length of a refstr. Fast.
* @param rstr_p a refstr pointer
* @return the length of the refstr
*/
size_t
refstr_len (refstr *rstr_p)
{
	return REFSTR_GETHEAD(rstr_p)->length;
}

/**
* Increments the reference count of a refstr. Returns the refstr address, to allow for the sytax:
* @code
* tk_copy_p = refstr_ref(rstr_p);
* @endcode
* @param rstr_p a refstr pointer
* @return the value held in rstr_p
*/
refstr *
refstr_ref (refstr *rstr_p)
{
	/* Bump up the refcount of an existing refstr */
	REFSTR_MUTEX_LOCK(&refstr_hash_mutex);
	++ REFSTR_GETHEAD(rstr_p)->refcount;
	REFSTR_MUTEX_UNLOCK(&refstr_hash_mutex);
	return rstr_p;
}

/**
* Decrements the reference count of a refstr, and de-allocates it when the reference reaches 0.
* @param rstr_p a refstr pointer
*/
void
refstr_unref (refstr *rstr_p)
{
	/* drop the refcount of an existing refstr,
	 * destroy it if appropriate. */
	refstr_head **slot, *rstr_h_p, *x_rstr_h_p, *y_rstr_h_p, *w_rstr_h_p;
	char c;
	rstr_h_p = REFSTR_GETHEAD(rstr_p);
	slot = rstr_h_p->slot;
	REFSTR_MUTEX_LOCK(&refstr_hash_mutex);
	if (-- rstr_h_p->refcount == 0) {
		/* red-black delete, its a bitch */
		if ((rstr_h_p->left != NIL) && (rstr_h_p->right != NIL)) {
			y_rstr_h_p = rstr_h_p->right;
			while (y_rstr_h_p->left != NIL)
				y_rstr_h_p = y_rstr_h_p->left;
		} else {
			y_rstr_h_p = rstr_h_p;
		}
	
		x_rstr_h_p = (y_rstr_h_p->left != NIL) ? y_rstr_h_p->left : y_rstr_h_p->right;
		x_rstr_h_p->parent = y_rstr_h_p->parent;
		
		if (y_rstr_h_p->parent == NIL)
			*slot = x_rstr_h_p;
		else if (y_rstr_h_p == y_rstr_h_p->parent->left)
			y_rstr_h_p->parent->left = x_rstr_h_p;
		else
			y_rstr_h_p->parent->right = x_rstr_h_p;
	
		c = y_rstr_h_p->color;
		if (y_rstr_h_p != rstr_h_p) {
			/* this is non-standard, since we cannot copy keys. */
			y_rstr_h_p->parent = rstr_h_p->parent;
			if (rstr_h_p->parent != NIL) {
				if (rstr_h_p == rstr_h_p->parent->left)
					rstr_h_p->parent->left = y_rstr_h_p;
				else
					rstr_h_p->parent->right = y_rstr_h_p;
			} else {
				*slot = y_rstr_h_p;
			}
			
			y_rstr_h_p->left = rstr_h_p->left;
			rstr_h_p->left->parent = y_rstr_h_p;

			y_rstr_h_p->right = rstr_h_p->right;
			rstr_h_p->right->parent = y_rstr_h_p;

			y_rstr_h_p->color = rstr_h_p->color;
		}

		if (c == black) {
			while ((x_rstr_h_p != *slot) && (x_rstr_h_p->color == black)) {
				if (x_rstr_h_p == x_rstr_h_p->parent->left) {
					w_rstr_h_p = x_rstr_h_p->parent->right;
					if (w_rstr_h_p->color == red) {
						w_rstr_h_p->color = black;
						x_rstr_h_p->parent->color = red;
						refstr_left_rotate(slot, x_rstr_h_p->parent);
						w_rstr_h_p = x_rstr_h_p->parent->right;
					}
					if ((w_rstr_h_p->left->color == black) && (w_rstr_h_p->right->color == black)) {
						w_rstr_h_p->color = red;
						x_rstr_h_p = x_rstr_h_p->parent;
					} else {
						if (w_rstr_h_p->right->color == black) {
							w_rstr_h_p->left->color = black;
							w_rstr_h_p->color = red;
							refstr_right_rotate(slot, w_rstr_h_p);
							w_rstr_h_p = x_rstr_h_p->parent->right;
						}
						w_rstr_h_p->color = x_rstr_h_p->parent->color;
						x_rstr_h_p->parent->color = black;
						w_rstr_h_p->right->color = black;
						refstr_left_rotate(slot, x_rstr_h_p->parent);

						/* throw an exit condition */
						x_rstr_h_p = *slot;
					}
				} else {
					w_rstr_h_p = x_rstr_h_p->parent->left;
					if (w_rstr_h_p->color == red) {
						w_rstr_h_p->color = black;
						x_rstr_h_p->parent->color = red;
						refstr_right_rotate(slot, x_rstr_h_p->parent);
						w_rstr_h_p = x_rstr_h_p->parent->left;
					}
					if ((w_rstr_h_p->right->color == black) && (w_rstr_h_p->left->color == black)) {
						w_rstr_h_p->color = red;
						x_rstr_h_p = x_rstr_h_p->parent;
					} else {
						if (w_rstr_h_p->left->color == black) {
							w_rstr_h_p->right->color = black;
							w_rstr_h_p->color = red;
							refstr_left_rotate(slot, w_rstr_h_p);
							w_rstr_h_p = x_rstr_h_p->parent->left;
						}
						w_rstr_h_p->color = x_rstr_h_p->parent->color;
						x_rstr_h_p->parent->color = black;
						w_rstr_h_p->left->color = black;
						refstr_right_rotate(slot, x_rstr_h_p->parent);

						/* throw an exit condition */
						x_rstr_h_p = *slot;
					}
				}
			}
			x_rstr_h_p->color = black;
		}
		
		/* deallocate the refstr */
		refstr_free_refstr(rstr_h_p);
	}
	REFSTR_MUTEX_UNLOCK(&refstr_hash_mutex);
}

