/* Copyright (C) 2000, 2001 Red Hat, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/**
 * @brief Adm Switchboard implementation
 *
 * This file implements the functions of the Switchboard.
 *
 * The Switchboard is our highest external API for the applications to use.
 *
 * It's features are:
 *  - Easy management of the namespace configuration
 *  - Loading and saving through BlackBoxes
 *  - Generation of final merged native configuration
 *
 * @file        switchboard.c
 * @author      Crutcher Dunnavant <crutcher@redhat.com>
 * @date        05/06/2001
 * @version     1.0
 */


#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <dirent.h>
#include <stdio.h>
#include <string.h>

#include <glib.h>
#include <gmodule.h>

#include "alchemist.h"
#include "blackbox.h"

#include "switchboard.h"
#include "admerror.h"
#include "admcore.h"

static const char def_swb_dir_path[] = ALCHEMIST_SWITCHBOARD_DIR;
static AdmRefStr *active_swb_dir_path = NULL;
AdmRefStr *
AdmSwitchboard_getSwbDir (AdmError **error)
{
	_adm_error_init_check(error);

	if (active_swb_dir_path == NULL)
		active_swb_dir_path = AdmRefStr_get(def_swb_dir_path, 1);

	AdmRefStr_ref(active_swb_dir_path);
	return active_swb_dir_path;
}

int
AdmSwitchboard_setSwbDir (const char *path,
			 AdmError **error)
{
	_adm_error_init_check(error);
	
	if (active_swb_dir_path != NULL)
		AdmRefStr_unref(active_swb_dir_path);

	active_swb_dir_path = (path == NULL) ? NULL : AdmRefStr_get(path, 1);
	return 0;
}

AdmContext *
AdmSwitchboard_readNspCtx  (const char   *namespace,
			          AdmError    **error)
{
	GString *path;
	AdmContext *ctx;
	AdmRefStr *sdir;
	_adm_error_init_check(error);
	_adm_abort_if_fail(namespace != NULL);

	if (!AdmData_validName(namespace)) {
		AdmError_create(error, ADM_EBADDATA, _("Invalid namespace name \"%s\""), namespace);
		return NULL;
	}

	sdir = AdmSwitchboard_getSwbDir(NULL);
	path = g_string_new(AdmRefStr_str(sdir));
	AdmRefStr_unref(sdir);

	g_string_sprintfa(path, "/%s.swb.adl", namespace);

	ctx = NULL;
	AdmContext_readFromFile(&ctx, path->str, NULL);
	g_string_free(path, 1);

	if (ctx == NULL) {
		AdmError_create(error, ADM_ENONAMESPACE, _("Nsp \"%s\" unavailable"), namespace);
	}

	return ctx;
}

int
AdmSwitchboard_writeNspCtx (const char   *namespace,
				  AdmContext   *context,
				  int		mode,
				  uid_t		uid,
				  gid_t		gid,
			          AdmError    **error)
{
	GString *path;
	AdmRefStr *sdir;
	int retval;

	_adm_error_init_check(error);
	_adm_abort_if_fail(namespace != NULL);
	_adm_abort_if_fail(context != NULL);

	if (!AdmData_validName(namespace)) {
		AdmError_create(error, ADM_EBADDATA, _("Invalid namespace name \"%s\""), namespace);
		return -1;
	}

	sdir = AdmSwitchboard_getSwbDir(NULL);
	path = g_string_new(AdmRefStr_str(sdir));
	AdmRefStr_unref(sdir);

	g_string_sprintfa(path, "/%s.swb.adl", namespace);

	retval = AdmContext_writeToFile(context, path->str, mode, uid, gid, error);
	g_string_free(path, 1);

	return retval;
}

AdmList     *AdmSwitchboard_getNspCfg   (
				AdmContext   *config,
				const char   *namespace,
				AdmError **error)
{
	AdmData *dataA, *dataB;

	_adm_error_init_check(error);
	_adm_abort_if_fail(namespace != NULL);
	_adm_abort_if_fail(config != NULL);

	if ((dataA = AdmContext_getDataByPath_b(config, "/switchboard", NULL)) == NULL) {
		AdmError_create(error, ADM_EBADDATA, _("Invalid switchboard config, no '/switchboard' element"));
		return NULL;
	}

	if (AdmData_getType(dataA, NULL) != ADM_TYPE_LIST) {
		AdmError_create(error, ADM_EBADDATA,
			_("Invalid switchboard config, '/switchboard' element not an AdmList"));
		return NULL;
	}

	if ((dataB = AdmList_getDataByName_b((AdmList *) dataA, "namespaces", NULL)) == NULL) {
		AdmError_create(error, ADM_EBADDATA, _("Invalid switchboard config, no '/switchboard/namespaces' element"));
		return NULL;
	}

	if (AdmData_getType(dataB, NULL) != ADM_TYPE_LIST) {
		AdmError_create(error, ADM_EBADDATA,
			_("Invalid switchboard config, '/switchboard/namespaces' element not an AdmList"));
		return NULL;
	}

	if ((dataA = AdmList_getDataByName_b((AdmList *) dataB, namespace, NULL)) == NULL) {
		AdmError_create(error, ADM_ENONAMESPACE, _("Nsp \"%s\" unavailable"), namespace);
		return NULL;
	}

	if (AdmData_getType(dataA, NULL) != ADM_TYPE_LIST) {
		AdmError_create(error, ADM_EBADDATA,
			_("Invalid switchboard config, '/switchboard/namespaces/%s' element not an AdmList"), namespace);
		return NULL;
	}

	return (AdmList *) AdmData_ref(dataA, error);
}


AdmList     *AdmSwitchboard_getSubspCfg      (
					       AdmList   *namespace_cfg,
						const char   *subspace,
					       AdmError **error)
{
	AdmData *dataA, *dataB;

	_adm_error_init_check(error);
	_adm_abort_if_fail(namespace_cfg != NULL);

	subspace = subspace ?: "config";

	if ((dataB = AdmList_getDataByName_b(namespace_cfg, "subspaces", NULL)) == NULL) {
		AdmError_create(error, ADM_EBADDATA,
			_("Invalid switchboard config, no 'subspaces' child"));
		return NULL;
	}

	if (AdmData_getType(dataB, NULL) != ADM_TYPE_LIST) {
		AdmError_create(error, ADM_EBADDATA,
			_("Invalid switchboard config, '/subspaces' element not an AdmList"));
		return NULL;
	}

	if ((dataA = AdmList_getDataByName_b((AdmList *) dataB, subspace, NULL)) == NULL) {
		AdmError_create(error, ADM_ENOSUBSPACE, _("subspace '%s' unavailable"), subspace);
		return NULL;
	}

	if (AdmData_getType(dataA, NULL) != ADM_TYPE_LIST) {
		AdmError_create(error, ADM_EBADDATA,
			_("Invalid switchboard config, '/subspaces/%s' element not an AdmList"), subspace);
		return NULL;
	}

	return (AdmList *) AdmData_ref(dataA, error);
}

AdmBlackBox     *AdmSwitchboard_getBox	(
					AdmList	    *subspace_cfg,
					const char   *box_name,
					AdmError **error)
{
	AdmBlackBox *box;
	AdmData *dataA, *dataB;

	_adm_error_init_check(error);
	_adm_abort_if_fail(box_name != NULL);
	_adm_abort_if_fail(subspace_cfg != NULL);

	if ((dataB = AdmList_getDataByName_b(subspace_cfg, "box_configs", NULL)) == NULL) {
		AdmError_create(error, ADM_EBADDATA,
		_("Invalid subspace config, no 'box_configs' child"));
		return NULL;
	}

	if (AdmData_getType(dataB, NULL) != ADM_TYPE_LIST) {
		AdmError_create(error, ADM_EBADDATA,
		_("Invalid subspace config, '/box_configs' element not an AdmList"));
		return NULL;
	}

	if ((dataA = AdmList_getDataByName_b((AdmList *) dataB, box_name, NULL)) == NULL) {
		AdmError_create(error, ADM_ENOSUBSPACE,
			_("box '%s' unavailable"), box_name);
		return NULL;
	}

	if (AdmData_getType(dataA, NULL) != ADM_TYPE_LIST) {
		AdmError_create(error, ADM_EBADDATA,
		_("Invalid subspace config, '/box_configs/%s' element not an AdmList"), box_name);
		return NULL;
	}

	box = AdmBlackBox_getBox((AdmList *) dataA, error);;
	return box;
}


AdmContext  *
AdmSwitchboard_readNsp	(const char   *namespace,
				 const char   *subspace,
				 unsigned int	depth,
				 AdmError **error)
{
	AdmContext *nctx, *context;
	AdmList *config;

	_adm_error_init_check(error);
	_adm_abort_if_fail(namespace != NULL);

	subspace = subspace ?: "config";

	if ((nctx = AdmSwitchboard_readNspCtx(namespace, error)) == NULL)
		return NULL;

	if ((config = AdmSwitchboard_getNspCfg_b(nctx, namespace, error)) == NULL) {
		AdmContext_unref(nctx, NULL);
		return NULL;
	}

	context = AdmSwitchboard_readNspWithConfig(config, subspace, depth, error);
	AdmContext_unref(nctx, NULL);

	return context;
}

AdmContext  *
AdmSwitchboard_readNspWithConfig	(
					 AdmList      *namespace_cfg,
					const char   *subspace,
					 unsigned int  depth,
					 AdmError **error)
{
	AdmList *read_policy, *read_set, *subspace_cfg, *tmp_list, *read_element;
	AdmData *data;
	int cont_on_fail, exit_on_success;
	AdmContext *ctx_mix, *ctx_working, *ctx_merge;
	AdmRefStr *box_name;
	AdmBlackBox *box;

	_adm_error_init_check(error);
	_adm_abort_if_fail(namespace_cfg != NULL);

	subspace = subspace ?: "config";

	if ((subspace_cfg = AdmSwitchboard_getSubspCfg_b(namespace_cfg, subspace, error)) == NULL)
		return NULL;

	if ((read_policy = (AdmList *) AdmList_getDataByName_b(subspace_cfg, "read_policy", NULL)) == NULL) {
		AdmError_create(error, ADM_EBADDATA, _("Invalid subspace config, no 'read_policy' child"));
		return NULL;
	}

	if (AdmData_getType((AdmData *) read_policy, NULL) != ADM_TYPE_LIST) {
		AdmError_create(error, ADM_EBADDATA, _("Invalid subspace config, '/read_policy' element not an AdmList"));
		return NULL;
	}

	if ((read_set = (AdmList *) AdmList_getDataByName_b(read_policy, "read_set", NULL)) == NULL) {
		AdmError_create(error, ADM_EBADDATA, _("Invalid subspace config, no '/read_policy/read_set' child"));
		return NULL;
	}

	if (AdmData_getType((AdmData *) read_set, NULL) != ADM_TYPE_LIST) {
		AdmError_create(error, ADM_EBADDATA, _("Invalid subspace config, '/read_policy/read_set' element not an AdmList"));
		return NULL;
	}

	if ((read_element = (AdmList *) AdmList_getDataByIndex_b(read_set, 0, NULL)) == NULL) {
		AdmError_create(error, ADM_EBADDATA, _("Invalid subspace config, '/read_policy/read_set' empty"));
		return NULL;
	}

	if (depth == 0) /* 0 == read all */
		depth = AdmList_getNumData(read_set, NULL);

	ctx_mix = NULL;
	ctx_working = NULL;
	while ((read_element != NULL) && (depth-- != 0)) {
		if ((box_name = AdmData_getName((AdmData *) read_element, error)) == NULL) {
			goto drop_ctx_mix;
		}

		if (AdmData_getType((AdmData *) read_element, NULL) != ADM_TYPE_LIST) {
			AdmError_create(error, ADM_EBADDATA,
				_("Invalid subspace config, '/read_policy/read_set/%s' not a list"),
				AdmRefStr_str(box_name));
			goto drop_name;
		}

		if ((data = AdmList_getDataByName_b(read_element, "cont_on_fail", NULL)) == NULL) {
			AdmError_create(error, ADM_EBADDATA,
				_("Invalid subspace config, no '/read_policy/read_set/%s/cont_on_fail' element"),
				AdmRefStr_str(box_name));
			goto drop_name;
		}
		if (AdmData_getType(data, NULL) != ADM_TYPE_BOOL) {
			AdmError_create(error, ADM_EBADDATA,
				_("Invalid subspace config, no '/read_policy/read_set/%s/cont_on_fail' element"),
				AdmRefStr_str(box_name));
			goto drop_name;
		}
		cont_on_fail = AdmBool_getValue((AdmBool *) data, error);


		if ((data = AdmList_getDataByName_b(read_element, "exit_on_success", NULL)) == NULL) {
			AdmError_create(error, ADM_EBADDATA,
				_("Invalid subspace config, no '/read_policy/read_set/%s/exit_on_success' element"),
				AdmRefStr_str(box_name));
			goto drop_name;
		}
		if (AdmData_getType(data, NULL) != ADM_TYPE_BOOL) {
			AdmError_create(error, ADM_EBADDATA,
				_("Invalid subspace config, no '/read_policy/read_set/%s/exit_on_success' element"),
				AdmRefStr_str(box_name));
			goto drop_name;
		}
		exit_on_success = AdmBool_getValue((AdmBool *) data, error);


		if ((box = AdmSwitchboard_getBox(subspace_cfg, AdmRefStr_str(box_name), error)) == NULL) {
			goto drop_name;
		}
		if (!ADM_BBCAP_SUP_READ(AdmBlackBox_caps(box, NULL))) {
			AdmError_create(error, ADM_EBADDATA,
				_("Invalid subspace config, '/read_policy/read_set/%s' element does not support reads"),
				AdmRefStr_str(box_name));
			AdmBlackBox_unref(box, NULL);
			goto drop_name;
		}
		ctx_working = AdmBlackBox_read(box, NULL);
		AdmBlackBox_unref(box, NULL);

		
		if (ctx_working == NULL) {
			if (!cont_on_fail) {
				AdmError_create(error, ADM_EACCESSFAILED,
					_("read failed on '/read_policy/read_set/%s' element"),
					AdmRefStr_str(box_name));
				goto drop_name;
			}
		} else {
			if (ctx_mix == NULL) {
				ctx_mix = ctx_working;
			} else {
				ctx_merge = AdmContext_merge("cascade_merge", 1, ctx_mix, ctx_working, error);
				AdmContext_unref(ctx_mix, NULL);
				AdmContext_unref(ctx_working, NULL);
				if ((ctx_mix = ctx_merge) == NULL) {
					goto drop_name;
				}
			}
			if (exit_on_success) {
				AdmRefStr_unref(box_name);
				goto done;
			}
		}

		AdmRefStr_unref(box_name);
		tmp_list = (AdmList *) AdmData_getNext_b((AdmData *) read_element, 0, NULL);
		read_element = tmp_list;
		continue;


		drop_name:
		AdmRefStr_unref(box_name);

		drop_ctx_mix:
		if (ctx_mix)
			AdmContext_unref(ctx_mix, NULL);

		return NULL;
	}
	
done:
	return ctx_mix;

}

int
AdmSwitchboard_writeNsp	(const char   *namespace,
				 const char   *subspace,
				 AdmContext   *context,
				 AdmError **error)
{
	AdmContext *nctx;
	AdmList *config;
	int retval;

	_adm_error_init_check(error);
	_adm_abort_if_fail(namespace != NULL);
	_adm_abort_if_fail(context != NULL);

	subspace = subspace ?: "config";

	if ((nctx = AdmSwitchboard_readNspCtx(namespace, error)) == NULL)
		return -1;

	if ((config = AdmSwitchboard_getNspCfg_b(nctx, namespace, error)) == NULL) {
		AdmContext_unref(nctx, NULL);
		return -1;
	}

	retval = AdmSwitchboard_writeNspWithConfig(config, subspace, context, error);
	AdmContext_unref(nctx, NULL);

	return retval;
}

int
AdmSwitchboard_writeNspWithConfig	(
					 AdmList      *namespace_cfg,
					const char   *subspace,
					 AdmContext   *context,
					 AdmError **error)
{
	AdmList *write_policy, *write_set, *subspace_cfg, *tmp_list, *write_element;
	AdmData *data;
	int cont_on_fail, ret;
	AdmRefStr *box_name;
	AdmBlackBox *box;

	_adm_error_init_check(error);
	_adm_abort_if_fail(namespace_cfg != NULL);

	subspace = subspace ?: "config";

	if ((subspace_cfg = AdmSwitchboard_getSubspCfg_b(namespace_cfg, subspace, error)) == NULL)
		return -1;

	if ((write_policy = (AdmList *) AdmList_getDataByName_b(subspace_cfg, "write_policy", NULL)) == NULL) {
		AdmError_create(error, ADM_EBADDATA, _("Invalid subspace config, no 'write_policy' child"));
		return -1;
	}

	if (AdmData_getType((AdmData *) write_policy, NULL) != ADM_TYPE_LIST) {
		AdmError_create(error, ADM_EBADDATA, _("Invalid subspace config, '/write_policy' element not an AdmList"));
		return -1;
	}

	if ((write_set = (AdmList *) AdmList_getDataByName_b(write_policy, "write_set", NULL)) == NULL) {
		AdmError_create(error, ADM_EBADDATA, _("Invalid subspace config, no '/write_policy/write_set' child"));
		return -1;
	}

	if (AdmData_getType((AdmData *) write_set, NULL) != ADM_TYPE_LIST) {
		AdmError_create(error, ADM_EBADDATA, _("Invalid subspace config, '/write_policy/write_set' element not an AdmList"));
		return -1;
	}

	if ((write_element = (AdmList *) AdmList_getDataByIndex_b(write_set, 0, NULL)) == NULL) {
		AdmError_create(error, ADM_EBADDATA, _("Invalid subspace config, '/write_policy/write_set' empty"));
		return -1;
	}

	while (write_element != NULL) {
		if ((box_name = AdmData_getName((AdmData *) write_element, error)) == NULL) {
			goto fail;
		}

		if (AdmData_getType((AdmData *) write_element, NULL) != ADM_TYPE_LIST) {
			AdmError_create(error, ADM_EBADDATA,
				_("Invalid subspace config, '/write_policy/write_set/%s' not a list"),
				AdmRefStr_str(box_name));
			goto drop_name;
		}

		if ((data = AdmList_getDataByName_b(write_element, "cont_on_fail", NULL)) == NULL) {
			AdmError_create(error, ADM_EBADDATA,
				_("Invalid subspace config, no '/write_policy/write_set/%s/cont_on_fail' element"),
				AdmRefStr_str(box_name));
			goto drop_name;
		}
		if (AdmData_getType(data, NULL) != ADM_TYPE_BOOL) {
			AdmError_create(error, ADM_EBADDATA,
				_("Invalid subspace config, no '/write_policy/write_set/%s/cont_on_fail' element"),
				AdmRefStr_str(box_name));
			goto drop_name;
		}
		cont_on_fail = AdmBool_getValue((AdmBool *) data, error);


		if ((box = AdmSwitchboard_getBox(subspace_cfg, AdmRefStr_str(box_name), error)) == NULL) {
			goto drop_name;
		}
		if (!ADM_BBCAP_SUP_WRITE(AdmBlackBox_caps(box, NULL))) {
			AdmError_create(error, ADM_EBADDATA,
				_("Invalid subspace config, '/write_policy/write_set/%s' element does not support writes"),
				AdmRefStr_str(box_name));
			AdmBlackBox_unref(box, NULL);
			goto drop_name;
		}
		ret = AdmBlackBox_write(box, context, NULL);
		AdmBlackBox_unref(box, NULL);

		if (ret != 0) {
			AdmError_create(error, ADM_EACCESSFAILED,
				_("write failed on '/write_policy/write_set/%s' element"),
				AdmRefStr_str(box_name));
			goto drop_name;
		}

		AdmRefStr_unref(box_name);
		tmp_list = (AdmList *) AdmData_getNext_b((AdmData *) write_element, 0, NULL);
		write_element = tmp_list;
		continue;


		drop_name:
		AdmRefStr_unref(box_name);

		fail:
		return -1;
	}
	
	return 0;

}

int
AdmSwitchboard_markNsp (const char   *namespace,
				   const char   *subspace,
				   const char   *mark,
				   AdmError **error)
{
	AdmContext *nctx;
	AdmList *config;
	int retval;

	_adm_error_init_check(error);
	_adm_abort_if_fail(namespace != NULL);
	_adm_abort_if_fail(mark != NULL);

	subspace = subspace ?: "config";

	if ((nctx = AdmSwitchboard_readNspCtx(namespace, error)) == NULL)
		return -1;

	if ((config = AdmSwitchboard_getNspCfg_b(nctx, namespace, error)) == NULL) {
		AdmContext_unref(nctx, NULL);
		return -1;
	}

	retval = AdmSwitchboard_markNspWithConfig(config, subspace, mark, error);
	AdmContext_unref(nctx, NULL);

	return retval;
}

int
AdmSwitchboard_markNspWithConfig (
				AdmList		*namespace_cfg,
				const char   *subspace,
				const char   *mark,
				AdmError	**error)
{
	AdmList *read_policy, *read_set, *subspace_cfg, *tmp_list, *read_element;
	AdmData *data;
	AdmRefStr *box_name;
	AdmBlackBox *box;
	int ret, wm;

	_adm_error_init_check(error);
	_adm_abort_if_fail(namespace_cfg != NULL);

	subspace = subspace ?: "config";

	if ((subspace_cfg = AdmSwitchboard_getSubspCfg_b(namespace_cfg, subspace, error)) == NULL)
		return -1;

	if ((read_policy = (AdmList *) AdmList_getDataByName_b(subspace_cfg, "read_policy", NULL)) == NULL) {
		AdmError_create(error, ADM_EBADDATA, _("Invalid subspace config, no 'read_policy' child"));
		return -1;
	}

	if (AdmData_getType((AdmData *) read_policy, NULL) != ADM_TYPE_LIST) {
		AdmError_create(error, ADM_EBADDATA, _("Invalid subspace config, '/read_policy' element not an AdmList"));
		return -1;
	}

	if ((read_set = (AdmList *) AdmList_getDataByName_b(read_policy, "read_set", NULL)) == NULL) {
		AdmError_create(error, ADM_EBADDATA, _("Invalid subspace config, no '/read_policy/read_set' child"));
		return -1;
	}

	if (AdmData_getType((AdmData *) read_set, NULL) != ADM_TYPE_LIST) {
		AdmError_create(error, ADM_EBADDATA, _("Invalid subspace config, '/read_policy/read_set' element not an AdmList"));
		return -1;
	}

	if ((read_element = (AdmList *) AdmList_getDataByIndex_b(read_set, 0, NULL)) == NULL) {
		AdmError_create(error, ADM_EBADDATA, _("Invalid subspace config, '/read_policy/read_set' empty"));
		return -1;
	}

	while (read_element != NULL) {
		if ((box_name = AdmData_getName((AdmData *) read_element, error)) == NULL) {
			goto fail;
		}

		if (AdmData_getType((AdmData *) read_element, NULL) != ADM_TYPE_LIST) {
			AdmError_create(error, ADM_EBADDATA,
				_("Invalid subspace config, '/read_policy/read_set/%s' not a list"),
				AdmRefStr_str(box_name));
			goto drop_name;
		}

		if ((data = AdmList_getDataByName_b(read_element, "mark", NULL)) == NULL) {
			AdmError_create(error, ADM_EBADDATA,
				_("Invalid subspace config, no '/read_policy/read_set/%s/mark' element"),
				AdmRefStr_str(box_name));
			goto drop_name;
		}
		if (AdmData_getType(data, NULL) != ADM_TYPE_BOOL) {
			AdmError_create(error, ADM_EBADDATA,
				_("Invalid subspace config, no '/read_policy/read_set/%s/mark' element"),
				AdmRefStr_str(box_name));
			goto drop_name;
		}
		wm = AdmBool_getValue((AdmBool *) data, error);

		if (!wm) {
			goto next;
		}

		if ((box = AdmSwitchboard_getBox(subspace_cfg, AdmRefStr_str(box_name), error)) == NULL) {
			goto drop_name;
		}
		if (!ADM_BBCAP_SUP_READ(AdmBlackBox_caps(box, NULL))) {
			AdmError_create(error, ADM_EBADDATA,
				_("Invalid subspace config, '/read_policy/read_set/%s' element does not support reads"),
				AdmRefStr_str(box_name));
			AdmBlackBox_unref(box, NULL);
			goto drop_name;
		}

		ret = AdmBlackBox_setMark(box, mark, error);
		AdmBlackBox_unref(box, NULL);
		if (ret == -1) {
			goto drop_name;
		}

		next:
		AdmRefStr_unref(box_name);
		tmp_list = (AdmList *) AdmData_getNext_b((AdmData *) read_element, 0, NULL);
		read_element = tmp_list;
		continue;


		drop_name:
		AdmRefStr_unref(box_name);

		fail:
		return -1;
	}
	

	return 0;
}

int
AdmSwitchboard_checkNspMark  (const char   *namespace,
				const char   *subspace,
				const char   *mark,
				AdmError **error)
{
	AdmContext *nctx;
	AdmList *config;
	int retval;

	_adm_error_init_check(error);
	_adm_abort_if_fail(namespace != NULL);
	_adm_abort_if_fail(mark != NULL);

	subspace = subspace ?: "config";

	if ((nctx = AdmSwitchboard_readNspCtx(namespace, error)) == NULL)
		return -1;

	if ((config = AdmSwitchboard_getNspCfg_b(nctx, namespace, error)) == NULL) {
		AdmContext_unref(nctx, NULL);
		return -1;
	}

	retval = AdmSwitchboard_checkNspMarkWithConfig(config, subspace, mark, error);
	AdmContext_unref(nctx, NULL);

	return retval;
}

int
AdmSwitchboard_checkNspMarkWithConfig ( AdmList		*namespace_cfg,
					const char   *subspace,
					const char   *mark,
					AdmError	**error)
{
	AdmList *read_policy, *read_set, *subspace_cfg, *tmp_list, *read_element;
	AdmData *data;
	AdmRefStr *box_name;
	AdmBlackBox *box;
	unsigned int wm_serial, serial, wm;

	_adm_error_init_check(error);
	_adm_abort_if_fail(namespace_cfg != NULL);

	subspace = subspace ?: "config";

	if ((subspace_cfg = AdmSwitchboard_getSubspCfg_b(namespace_cfg, subspace, error)) == NULL)
		return 0;

	if ((read_policy = (AdmList *) AdmList_getDataByName_b(subspace_cfg, "read_policy", NULL)) == NULL) {
		AdmError_create(error, ADM_EBADDATA, _("Invalid subspace config, no 'read_policy' child"));
		return 0;
	}

	if (AdmData_getType((AdmData *) read_policy, NULL) != ADM_TYPE_LIST) {
		AdmError_create(error, ADM_EBADDATA, _("Invalid subspace config, '/read_policy' element not an AdmList"));
		return 0;
	}

	if ((read_set = (AdmList *) AdmList_getDataByName_b(read_policy, "read_set", NULL)) == NULL) {
		AdmError_create(error, ADM_EBADDATA, _("Invalid subspace config, no '/read_policy/read_set' child"));
		return 0;
	}

	if (AdmData_getType((AdmData *) read_set, NULL) != ADM_TYPE_LIST) {
		AdmError_create(error, ADM_EBADDATA, _("Invalid subspace config, '/read_policy/read_set' element not an AdmList"));
		return 0;
	}

	if ((read_element = (AdmList *) AdmList_getDataByIndex_b(read_set, 0, NULL)) == NULL) {
		AdmError_create(error, ADM_EBADDATA, _("Invalid subspace config, '/read_policy/read_set' empty"));
		return 0;
	}

	while (read_element != NULL) {
		if ((box_name = AdmData_getName((AdmData *) read_element, error)) == NULL) {
			goto fail;
		}

		if (AdmData_getType((AdmData *) read_element, NULL) != ADM_TYPE_LIST) {
			AdmError_create(error, ADM_EBADDATA,
				_("Invalid subspace config, '/read_policy/read_set/%s' not a list"),
				AdmRefStr_str(box_name));
			goto drop_name;
		}

		if ((data = AdmList_getDataByName_b(read_element, "mark", NULL)) == NULL) {
			AdmError_create(error, ADM_EBADDATA,
				_("Invalid subspace config, no '/read_policy/read_set/%s/mark' element"),
				AdmRefStr_str(box_name));
			goto drop_name;
		}
		if (AdmData_getType(data, NULL) != ADM_TYPE_BOOL) {
			AdmError_create(error, ADM_EBADDATA,
				_("Invalid subspace config, no '/read_policy/read_set/%s/mark' element"),
				AdmRefStr_str(box_name));
			goto drop_name;
		}
		wm = AdmBool_getValue((AdmBool *) data, error);

		if (!wm) {
			goto next;
		}

		if ((box = AdmSwitchboard_getBox(subspace_cfg, AdmRefStr_str(box_name), error)) == NULL) {
			goto drop_name;
		}
		if (!ADM_BBCAP_SUP_READ(AdmBlackBox_caps(box, NULL))) {
			AdmError_create(error, ADM_EBADDATA,
				_("Invalid subspace config, '/read_policy/read_set/%s' element does not support reads"),
				AdmRefStr_str(box_name));
			AdmBlackBox_unref(box, NULL);
			goto drop_name;
		}

		{
			AdmError *l_error;
			l_error = NULL;

			wm_serial = AdmBlackBox_getMark(box, mark, &l_error);
			if (l_error != NULL) {
				_adm_error_prop(l_error, error);
				AdmBlackBox_unref(box, NULL);
				goto drop_name;

			}
		}

		if (AdmBlackBox_listSerial(box, &serial, 1, 0, error) == -1) {
			AdmBlackBox_unref(box, NULL);
			goto drop_name;
		}

		AdmBlackBox_unref(box, NULL);

		if (serial > wm_serial) {
			goto drop_name;
		}

		next:
		AdmRefStr_unref(box_name);
		tmp_list = (AdmList *) AdmData_getNext_b((AdmData *) read_element, 0, NULL);
		read_element = tmp_list;
		continue;


		drop_name:
		AdmRefStr_unref(box_name);

		fail:
		return 0;
	}
	
	return 1;
}


