/**
 * Copyright (C) 2000, 2001 Red Hat, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/**
 * @file        triggerblackbox.c
 * @author      Crutcher Dunnavant <crutcher@redhat.com>
 * @date        05/06/2001
 * @version     1.0
 */

#include <stdlib.h>
#include <unistd.h>

#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <glib.h>

#include "alchemist.h"
#include "blackbox.h"
#include "internalbb.h"

#define BUFSIZE 1024


typedef struct triggerblackbox_instance triggerblackbox_instance;
struct triggerblackbox_instance {
	AdmRefStr *path;
	AdmRefStr **arg_array;
	AdmRefStr **env_array;
	int wait;
};

static void		triggerblackbox_free	(void			 *data);

static int		triggerblackbox_write	(void			 *data,
						 AdmContext		 *context,
						 AdmError **error);

static AdmBlackBoxMethods triggerblackbox_methods = {
	free:		triggerblackbox_free,
	read:		NULL,
	listSerial:     NULL,
	readSerial:     NULL,
	write:          triggerblackbox_write,
};

static void
clear (AdmRefStr **array)
{
	unsigned int i=0;
	while (array[i] != NULL) {
		AdmRefStr_unref(array[i]);
		++i;
	}
	free(array);
}

static void
triggerblackbox_free (void     *data)
{
	struct triggerblackbox_instance *in_p = data;

	AdmRefStr_unref(in_p->path);
	clear(in_p->arg_array);
	clear(in_p->env_array);

	free(in_p);
}

static int
triggerblackbox_write (void        *data,
		    AdmContext  *context,
		    AdmError **error)
{
	pid_t pid;
	int dev_null;
	struct triggerblackbox_instance *in_p = data;

	pid = fork();
	if (pid == -1) {
		AdmError_create(error, ADM_EACCESSFAILED, "the fork failed!");
		return -1;

	} else if (pid == 0) {
		/* Data Code */

		/* close everything */
		long max_fd;
		max_fd = sysconf(_SC_OPEN_MAX);
		while (max_fd > 0)
			close(max_fd--);

		/* drop to '/' and drop terminal */
		daemon(0, 0);

		/* stdout & stderr > /dev/null */
		dev_null = open("/dev/null", O_WRONLY);
		dup2(dev_null, 1);
		dup2(dev_null, 2);
		close(dev_null);

		/* exec! */
		execve(AdmRefStr_str(in_p->path), (char **) in_p->arg_array, (char **) in_p->env_array);

		/* paranoia! */
		exit(-1);

		/* extreme paranoia! */
		return -1;

	} else {
		if (in_p->wait) { /* do we wait on the child? */
			int status;
			waitpid(pid, &status, 0);
			if (WIFEXITED(status) && WEXITSTATUS(status) == 0) {
				return 0;
			} else {
				AdmError_create(error, ADM_EACCESSFAILED, "%s failed", in_p->path);
				return -1;
			}
		} else {
			return 0;
		}
	}

}

int
_adm_triggerblackbox_getBox (AdmBlackBoxMethods	**methods,
		     void		**data,
		     AdmList		 *box_list,
		     AdmError		**error)
{
	triggerblackbox_instance *in_p;
	AdmData *path, *foo, *bar;
	AdmRefStr **arg_array, **env_array, *tok;
	unsigned int wait, len, i;
	
	/* path */
	path = AdmList_getDataByName_b(box_list, "path", NULL);
	if (path == NULL) {
		AdmError_create(error, ADM_EBADDATA, "box/path does not exist");
		return -1;
	} else if (AdmData_getType(path, NULL) != ADM_TYPE_STRING) {
		AdmError_create(error, ADM_EBADDATA, "box/path is not a string");
		return -1;
	}

	/* wait? */
	foo = AdmList_getDataByName_b(box_list, "wait", NULL);
	if (foo == NULL) {
		AdmError_create(error, ADM_EBADDATA, "box/wait does not exist");
		return -1;
	} else if (AdmData_getType(foo, NULL) != ADM_TYPE_BOOL) {
		AdmError_create(error, ADM_EBADDATA, "box/wait is not a bool");
		return -1;
	}
	wait = AdmBool_getValue((AdmBool *) foo, NULL);


	/* arg_list */
	foo = AdmList_getDataByName_b(box_list, "arg_list", NULL);
	if (foo == NULL) {
		AdmError_create(error, ADM_EBADDATA, "box/arg_list does not exist");
		return -1;
	} else if (AdmData_getType(foo, NULL) != ADM_TYPE_LIST) {
		AdmError_create(error, ADM_EBADDATA, "box/arg_list is not a list");
		return -1;
	}

	len = AdmList_getNumData((AdmList *) foo, NULL);
	arg_array = calloc(len + 1, sizeof (AdmRefStr *));
	for (i=0;i<len;++i) {
		bar = AdmList_getDataByIndex_b((AdmList *) foo, i, NULL);
		if (AdmData_getType(bar, NULL) != ADM_TYPE_STRING) {
			int j;
			for (j = 0; j <= i; ++j)
				AdmRefStr_unref(arg_array[j]);
			free(arg_array);

			tok = AdmData_getName(bar, NULL);
			AdmError_create(error, ADM_EBADDATA, "box/arg_list/%s is not a string",
				AdmRefStr_str(tok));
			AdmRefStr_unref(tok);
			return -1;
		}

		/* we are keeping a ref */
		arg_array[i] = AdmString_getValue((AdmString *) bar, NULL);
	}

	/* env_list */
	foo = AdmList_getDataByName_b(box_list, "env_list", NULL);
	if (foo == NULL) {
		AdmError_create(error, ADM_EBADDATA, "box/env_list does not exist");
		clear(arg_array);
		return -1;
	} else if (AdmData_getType(foo, NULL) != ADM_TYPE_LIST) {
		AdmError_create(error, ADM_EBADDATA, "box/env_list is not a list");
		clear(arg_array);
		return -1;
	}

	len = AdmList_getNumData((AdmList *) foo, NULL);
	env_array = calloc(len + 1, sizeof (AdmRefStr *));
	for (i=0;i<len;++i) {
		bar = AdmList_getDataByIndex_b((AdmList *) foo, i, NULL);
		if (AdmData_getType(bar, NULL) != ADM_TYPE_STRING) {
			int j;
			clear(arg_array);
			for (j = 0; j <= i; ++j)
				AdmRefStr_unref(env_array[j]);
			free(env_array);

			tok = AdmData_getName(bar, NULL);
			AdmError_create(error, ADM_EBADDATA, "box/env_list/%s is not a string",
				AdmRefStr_str(tok));
			AdmRefStr_unref(tok);
			return -1;
		}

		/* we are keeping a ref */
		env_array[i] = AdmString_getValue((AdmString *) bar, NULL);
	}

	/* build the instance */
	in_p = malloc(sizeof (triggerblackbox_instance));
	in_p->path = AdmString_getValue((AdmString *) path, NULL); /* we are keeping a ref */
	in_p->arg_array = arg_array;
	in_p->env_array = env_array;
	in_p->wait = wait;

	/* return */
	*data = in_p;
	*methods = &triggerblackbox_methods;
	return 0;

}

