/**
 * Copyright (C) 2000, 2001 Red Hat, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/**
 * @brief Adm getValue and setValue functions 
 *
 * These functions operate upon AdmData subtyps, and assert type validitity
 * and input pointer non-NULL-ity.
 *
 * They return a NULL, 0, or FALSE value, and an error,
 * 	 upon access to an unlinked (dead) node.
 *
 * String values returned are volatile, so if the caller will need them
 * at a later point, the strings should be copied 
 * getValue and setValue functions
 *
 * @file	value.c
 * @author	Crutcher Dunnavant <crutcher@redhat.com>
 * @date	05/06/2001
 * @version 	1.0
 */

#include <string.h>
#include <stdio.h> /* for sscanf */
#include <stdlib.h> /* for malloc/free */

#include <glib.h>

#include "alchemist.h"

#include "refstr.h"
#include "translation.h"
#include "alchemist-private.h"
#include "admerror.h"
#include "admcore.h"

char * B64encode (const void * str, size_t ns);
int B64decode (const char * s, void ** datap, size_t *lenp);

/**
 * Validate a base64 string 
 * @brief Internal base64 type validation
 * @param string The character pointer to the string to be validated.
* @param len the length of the string to check
 * @return @c TRUE if the string is a valid base64 string, @c FALSE otherwise.
 * @see AdmBase64 AdmBase64_setValue AdmBase64_getValue
 */
int _adm_valid_base64(const char *string, unsigned long len) {
	unsigned long i;
	char c;
	
#ifdef __ADM_DEBUG__
	g_assert(string != NULL);
#endif
	
	/* an empty string is valid */
	if (len == 0)
		return 1;

	/* base64 encoded strings must be multiples of 4 */
	if (len % 4 != 0)
		return 0;

	for (i = 0; i < len; ++ i) {
		c = string[i];
		if (ADM_ISALPHA(c) || ADM_ISDIGIT(c) || (c == '+') || (c == '/'))
			continue;
		else if ((c == '=') && ((len - i) <= 2))
			continue;
		else
			return 0;
	}
	return 1;
}

int
AdmInt_getValue (AdmInt   *int_scalar,
		 AdmError **error)
{
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(int_scalar != NULL);
	if (!_adm_type_and_lifecycle_check((AdmData *) int_scalar, ADM_TYPE_INT, error))
		return 0;
	
	return int_scalar->value;
}

int
AdmInt_setValue (AdmInt   *int_scalar,
		 int       value,
		 AdmError **error)
{
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(int_scalar != NULL);
	if (!_adm_type_and_lifecycle_check((AdmData *) int_scalar, ADM_TYPE_INT, error))
		return -1;

	int_scalar->value = value;

	return 0;
}

double
AdmFloat_getValue (AdmFloat *float_scalar,
		   AdmError **error)
{
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(float_scalar != NULL);
	if (!_adm_type_and_lifecycle_check((AdmData *) float_scalar, ADM_TYPE_FLOAT, error))
		return 0;

	return float_scalar->value;
}

int
AdmFloat_setValue (AdmFloat *float_scalar,
		   double    value,
		   AdmError **error)
{
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(float_scalar != NULL);
	if (!_adm_type_and_lifecycle_check((AdmData *) float_scalar, ADM_TYPE_FLOAT, error))
		return -1;

	float_scalar->value = value;

	return 0;
}

int
AdmBool_getValue (AdmBool  *bool_scalar,
		  AdmError **error)
{
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(bool_scalar != NULL);
	if (!_adm_type_and_lifecycle_check((AdmData *) bool_scalar, ADM_TYPE_BOOL, error))
		return 0;

	return bool_scalar->value;
}

int
AdmBool_setValue (AdmBool  *bool_scalar,
		  int       value,
		  AdmError **error)
{
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(bool_scalar != NULL);
	if (!_adm_type_and_lifecycle_check((AdmData *) bool_scalar, ADM_TYPE_BOOL, error))
		return -1;

	bool_scalar->value = value;

	return 0;
}

AdmRefStr *
AdmString_getValue (AdmString *string_scalar,
		    AdmError **error)
{
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(string_scalar != NULL);
	if (!_adm_type_and_lifecycle_check((AdmData *) string_scalar, ADM_TYPE_STRING, error))
		return NULL;

	return AdmRefStr_ref(string_scalar->value);
}

int
AdmString_setValue (AdmString  *string_scalar,
		    const char *value,
		    AdmError **error)
{
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(string_scalar != NULL);
	_adm_abort_if_fail(value != NULL);
	if (!_adm_type_and_lifecycle_check((AdmData *) string_scalar, ADM_TYPE_STRING, error))
		return -1;

	/* UTF8 encoding check */
	if (!g_utf8_validate(value, -1, NULL)) {
		AdmError_create(error, ADM_EBADDATA, _("The value string was invalid UTF-8"));
		return -1;
	}

	AdmRefStr_unref(string_scalar->value);
	string_scalar->value = AdmRefStr_get(value, 1);

	return 0;
}

AdmRefStr *
AdmBase64_getValue (AdmBase64 *base64_scalar,
		    AdmError **error)
{
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(base64_scalar != NULL);
	if (!_adm_type_and_lifecycle_check((AdmData *) base64_scalar, ADM_TYPE_BASE64, error))
		return NULL;

	return AdmRefStr_ref(base64_scalar->value);
}

int
AdmBase64_setValue (AdmBase64  *base64_scalar,
		    const char *value,
		    AdmError **error)
{
	unsigned long len;

	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(base64_scalar != NULL);
	_adm_abort_if_fail(value != NULL);
	if (!_adm_type_and_lifecycle_check((AdmData *) base64_scalar, ADM_TYPE_BASE64, error))
		return -1;

	/* base 64 encoding check */
	len = strlen(value);
	if (!_adm_valid_base64(value, len)) {
		AdmError_create(error, ADM_EBADDATA, _("The value string was invalid base 64"));
		return -1;
	}
	
	AdmRefStr_unref(base64_scalar->value);
	base64_scalar->value = AdmRefStr_nget(value, len, 1);

	return 0;
}

size_t
AdmBase64_getBinValue (AdmBase64 *base64_scalar,
		       void     **value,
		       AdmError **error)
{
	size_t len;
	int ret;

	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(base64_scalar != NULL);
	if (!_adm_type_and_lifecycle_check((AdmData *) base64_scalar, ADM_TYPE_BASE64, error))
		return 0;

	ret = B64decode(AdmRefStr_str(base64_scalar->value), value, &len);

	if (ret == 0)
		return len;
	else
		return 0;
}

int
AdmBase64_setBinValue (AdmBase64 *base64_scalar,
		       const void *value,
		       size_t     size,
		       AdmError **error)
{
	char *tmpstr;
	int ret;

	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(base64_scalar != NULL);
	_adm_abort_if_fail(value != NULL);
	if (!_adm_type_and_lifecycle_check((AdmData *) base64_scalar, ADM_TYPE_BASE64, error))
		return -1;

	tmpstr = B64encode(value, size);

	if (tmpstr == NULL)
		return -1;

	ret = AdmBase64_setValue(base64_scalar, tmpstr, error);

	free(tmpstr);

	return ret;
}

/**
 * Base64 encode helper function.
 *
 * This code has actually been taken directly from the rpm-4.0.2 source. It
 * implements the encoding of binary data into base64 format as specified in
 * RFC1421.
 */
char * B64encode (const void * str, size_t ns)
{   
    static char b64enc[] =
        "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
    const unsigned char *s = str;
    unsigned char *t, *te;
    size_t nt;
    unsigned c;
    
    if (s == NULL)      return NULL;
    if (*s == '\0')     return strdup("");
    
    if (ns == 0) ns = strlen(s);
    nt = ((ns + 2) / 3) * 4;
    t = te = malloc(nt + 1);
    
    while (ns) {
        c = *s++;
        *te++ = b64enc[ (c >> 2) ];
        *te++ = b64enc[ ((c & 0x3) << 4) | (*s >> 4) ];
        if (--ns <= 0) {
            *te++ = '=';
            *te++ = '=';
            continue;
        } 
        c = *s++;
        *te++ = b64enc[ ((c & 0xf) << 2) | (*s >> 6) ];
        if (--ns <= 0) {
            *te++ = '=';
            continue;
        }
        *te++ = b64enc[ (int)(*s & 0x3f) ];
        s++;
        --ns;
    }
    *te = '\0';
    return t;
}

/**
 * Base64 decode helper function.
 *
 * This code has actually been taken directly from the rpm-4.0.2 source. It
 * implements the decoding of a base64 string back to it's binary form conform
 * to RFC1421.
 */
int B64decode (const char * s, void ** datap, size_t *lenp)
{
    static /*@only@*/ char * b64dec = NULL;
    unsigned char *t, *te;
    size_t ns, nt;
    unsigned a, b, c, d;

    if (s == NULL)      return 1;
    ns = strlen(s);
    if (ns & 0x3)       return 2;

    if (b64dec == NULL) {
        b64dec = malloc(255);
        memset(b64dec, 0x80, 255);
        for (c = 'A'; c <= 'Z'; c++)
            b64dec[ c ] = 0 + (c - 'A');
        for (c = 'a'; c <= 'z'; c++)
            b64dec[ c ] = 26 + (c - 'a');
        for (c = '0'; c <= '9'; c++)
            b64dec[ c ] = 52 + (c - '0');
        b64dec[(unsigned)'+'] = 62;
        b64dec[(unsigned)'/'] = 63;
        b64dec[(unsigned)'='] = 0;
    }

    nt = (ns / 4) * 3;
    t = te = malloc(nt + 1);

    while (ns > 0) {
        if ((a = b64dec[ (unsigned)*s++ ]) == 0x80)
            break;
        if ((b = b64dec[ (unsigned)*s++ ]) == 0x80)
            break;
        if ((c = b64dec[ (unsigned)*s++ ]) == 0x80)
            break;
        if ((d = b64dec[ (unsigned)*s++ ]) == 0x80)
            break;
        ns -= 4;
        *te++ = (a << 2) | (b >> 4);
        if (s[-2] == '=') break;
        *te++ = (b << 4) | (c >> 2);
        if (s[-1] == '=') break;
        *te++ = (c << 6) | d;
    }

    if (ns > 0) {
        free(t);
        return 3;
    }
    if (lenp)
        *lenp = (te - t);
    if (datap)
        *datap = t;
    else
        free(t);

    return 0;
}

AdmRefStr *
AdmCopy_getValue (AdmCopy  *copy_scalar,
		  AdmError **error)
{
	/* sanity checks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(copy_scalar != NULL);
	if (!_adm_type_and_lifecycle_check((AdmData *) copy_scalar, ADM_TYPE_COPY, error))
		return NULL;

	return AdmRefStr_ref(copy_scalar->value);
}

int
AdmCopy_setValue (AdmCopy    *copy_scalar,
		  const char *value,
		  AdmError **error)
{
	/* sanity cheacks */
	_adm_error_init_check(error);
	_adm_abort_if_fail(copy_scalar != NULL);
	_adm_abort_if_fail(value != NULL);
	if (!_adm_type_and_lifecycle_check((AdmData *) copy_scalar, ADM_TYPE_COPY, error))
		return -1;

	/* copy path check */
	if ((!_adm_valid_path(value)) || (value[0] == '\0')) {
		AdmError_create(error, ADM_EBADDATA, _("The value string was an invalid path"));
		return -1;
	}
	
	AdmRefStr_unref(copy_scalar->value);
	copy_scalar->value = AdmRefStr_get(value, 1);

	return 0;
}

