/**
 * Copyright (C) 2000, 2001 Red Hat, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/**
 * @brief Adm XML encoder
 *
 * This file implements our XML encoder.
 *
 * It uses the the gmarkup extension of the upcoming glib 2.0 to do the
 * simple form of encoding that we need. We might switch to another encoder
 * in the future, e.g. the one from libxml.
 *
 * @file        xml-encode.c
 * @author      Crutcher Dunnavant <crutcher@redhat.com>
 * @date        05/06/2001
 * @version     1.0
 */

#include <glib.h>
#include <string.h>
#include <stdio.h> /* for sscanf */

#include "alchemist.h"
#include "refstr.h"
#include "alchemist-private.h"
#include "admerror.h"


/* encoder functions */
void _adm_xml_encode_id_tree	(AdmIdentity *id,
					 GString *g_str,
					 guint depth);

void _adm_xml_encode_data_tree	(AdmData *data,
					 GString *g_str,
					 guint depth);


/* enoding functions */
void
_adm_xml_encode_id_tree	(AdmIdentity *id,
			 GString *g_str,
			 guint depth) {
	guint i;
	
	/* header and value */
	for (i = depth; i > 0; -- i)
		g_string_append(g_str, "  ");
	g_string_sprintfa(g_str, "<id NAME=\"%s\" SERIAL=\"%u\">\n", AdmRefStr_str(id->name), id->serial);

	/* parents */
	if (id->parentA != NULL)
		_adm_xml_encode_id_tree(id->parentA, g_str, depth + 1);
	else {
		for (i = depth + 1; i > 0; -- i)
			g_string_append(g_str, "  ");
		g_string_append(g_str, "<null/>\n");
	}

	if (id->parentB != NULL)
		_adm_xml_encode_id_tree(id->parentB, g_str, depth + 1);
	else {
		for (i = depth + 1; i > 0; -- i)
			g_string_append(g_str, "  ");
		g_string_append(g_str, "<null/>\n");
	}

	/* footer */	
	for (i = depth; i > 0; -- i)
		g_string_append(g_str, "  ");
	g_string_append(g_str, "  </id>\n");

	return;
}

void
_adm_xml_encode_data_tree	(AdmData *data,
				 GString *g_str,
				 guint depth) {
	guint i;
	AdmData *foo;
	gchar *tmp;


	/* offset */
	for (i = depth; i > 0; -- i)
		g_string_append(g_str, "  ");

	/* the element name is the same as the AdmData elements name */
	g_string_sprintfa(g_str, "<%s ", AdmRefStr_str(data->name));
	
	/* we only encode the source if it differs from the name of the root id node */
	if (data->context->id_tree_root->name != data->source)
		g_string_sprintfa(g_str, "SOURCE=\"%s\" ", AdmRefStr_str(data->source));

	if (data->flags & ADM_DATA_FLAG_PROTECTED)
		g_string_append(g_str, "PROTECTED=\"TRUE\" ");
	if (data->flags & ADM_DATA_FLAG_ANONYMOUS)
		g_string_append(g_str, "ANONYMOUS=\"TRUE\" ");
	if (data->flags & ADM_DATA_FLAG_ATOMIC)
		g_string_append(g_str, "ATOMIC=\"TRUE\" ");
	if (data->flags & ADM_DATA_FLAG_PREPEND)
		g_string_append(g_str, "PREPEND=\"TRUE\" ");
	
	switch (data->type) {
	case ADM_TYPE_INT:
		g_string_sprintfa(g_str, "TYPE=\"INT\" VALUE=\"%d\"/>\n", ((AdmInt *) data)->value);
		break;

	case ADM_TYPE_FLOAT:
/*buffer*/	tmp = (gchar *) g_malloc(24); /* FIXME This nonsense is a workaround for glib 1.2, fix it for glib 2.0 */
		sprintf(tmp, "%la", ((AdmFloat *) data)->value);
		g_string_sprintfa(g_str, "TYPE=\"FLOAT\" VALUE=\"%s\"/>\n", tmp);
/*end buffer*/	g_free(tmp);
		break;

	case ADM_TYPE_BOOL:
		g_string_sprintfa(g_str, "TYPE=\"BOOL\" VALUE=\"%s\"/>\n", ((AdmBool *) data)->value ? "TRUE" : "FALSE");
		break;

	case ADM_TYPE_STRING:
/*buffer*/	tmp = g_markup_escape_text(AdmRefStr_str(((AdmString *) data)->value),
						AdmRefStr_len( ((AdmString *) data)->value ) );
		g_string_sprintfa(g_str, "TYPE=\"STRING\" VALUE=\"%s\"/>\n", tmp);
/*end buffer*/	g_free(tmp);
		break;

	case ADM_TYPE_BASE64:
		g_string_sprintfa(g_str, "TYPE=\"BASE64\" VALUE=\"%s\"/>\n", AdmRefStr_str(((AdmBase64 *) data)->value));
		break;

	case ADM_TYPE_COPY:
		g_string_sprintfa(g_str, "TYPE=\"COPY\" VALUE=\"%s\"/>\n", AdmRefStr_str(((AdmCopy *) data)->value));
		break;

	case ADM_TYPE_LIST:
		g_string_append(g_str, "TYPE=\"LIST\">\n");

		i = ((AdmList *) data)->num_children;
		foo = data->child;
		for (; i > 0; foo = foo->next, -- i)
			_adm_xml_encode_data_tree(foo, g_str, depth + 1);
		
		for (i = depth + 1; i > 0; -- i)
			g_string_append(g_str, "  ");
		g_string_sprintfa(g_str, "</%s>\n", AdmRefStr_str(data->name));
		break;
	
	default:
		g_string_append(g_str, "ERROR!/>\n");
	}
	
	return;
}


/**
 * AdmContext_toXML: encodes an @AdmContext element as XML, and returns the string
 * @context: a pointer to an @AdmContext element
 * @error: an optional @GError
 *
 * This function walks an @AdmContext element, and produces an XML encoded
 * representation of it, which is returned as a string.
 *
 * Return value: a char *
 **/
char *
AdmContext_toXML (AdmContext *context, AdmError **error)
{
	GString *g_str;
	AdmData *data;
	guint i;
	GSList *gslist_p;
	gchar *str;

	_adm_error_init_check(error);
	_adm_abort_if_fail(context != NULL);

#define INITIAL_XML_BUFFER_LEN 4096
	/* XML header. */
	g_str = g_string_sized_new(INITIAL_XML_BUFFER_LEN);
	g_string_append(g_str, "<?xml version=\"1.0\"?>\n<adm_context VERSION=\"" ADM_VERSION_STRING "\">\n");

	/* id tree */
	_adm_xml_encode_id_tree(context->id_tree_root, g_str, 1);

	/* data tree */
	g_string_append(g_str, "  <datatree>\n");
	data = context->data_tree_root->child;
	i = ((AdmList *) context->data_tree_root)->num_children;
	for (; i > 0; data = data->next, -- i)
		_adm_xml_encode_data_tree(data, g_str, 2);
	g_string_append(g_str, "    </datatree>\n");

	/* mask set */
	gslist_p = context->mask_set;
	for (i = context->mask_set_size; i > 0; gslist_p = gslist_p->next, -- i)
		g_string_sprintfa(g_str, "  <mask PATH=\"%s\"/>\n", (refstr *) gslist_p->data);

	/* footer */
	g_string_append(g_str, "  </adm_context>\n");
	
	str = g_str->str;
	i = g_str->len;
	g_string_free(g_str, FALSE);

	return g_realloc(str, i + 1);
}

