/**
 * Copyright (C) 2000, 2001 Red Hat, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/**
 * @brief Adm XML parser
 *
 * This file implements our XML parser.
 *
 * Like our XML encoder we use the gmarkup of the glib 2.0. Again, this might
 * change in the future by using libxml and the included SAX parser.
 *
 * @file        xml-parse.c
 * @author      Crutcher Dunnavant <crutcher@redhat.com>
 * @date        05/06/2001
 * @version     1.0
 */

#include <glib.h>
#include <string.h>
#include <stdio.h> /* for sscanf */

#include "alchemist.h"
#include "refstr.h"
#include "alchemist-private.h"
#include "admerror.h"
#include "admcore.h"

/* Parser types */
typedef enum {
	ADM_PSTATE_INITIAL = 0,		/* waiting for an ADM_CONTEXT to start */
	ADM_PSTATE_CONTEXT_IDWAIT,	/* waiting for the root ID to start */
	ADM_PSTATE_ID_AWAIT,		/* waiting for a parent A ID to start */
	ADM_PSTATE_ID_BWAIT,		/* waiting for a parent B ID to start */
	ADM_PSTATE_NULL_ENDWAIT,	/* waiting for the completion of a NULL */
	ADM_PSTATE_ID_ENDWAIT,		/* waiting for an ID to end */
	ADM_PSTATE_CONTEXT_DTWAIT,	/* waiting for the DATATREE to start */
	ADM_PSTATE_DT_ENDWAIT,		/* waiting for the DATATREE to end, or for a data element */
	ADM_PSTATE_LIST_ENDWAIT,	/* waiting for a list to end, or for a data element */
	ADM_PSTATE_DATA_ENDWAIT,	/* waiting for a data element to end */
	ADM_PSTATE_CONTEXT_ENDWAIT,	/* waiting for a DELETE, or for the ADM_CONTEXT to end */
	ADM_PSTATE_DELETE_ENDWAIT,	/* waiting for a DELETE to end */
	ADM_PSTATE_FINAL,		/* Done parsing, additional tags are an error */
	ADM_PSTATE_ERROR		/* Throw an error */
} AdmParseState;

typedef struct {
	AdmContext *context;
	GSList *s_stack;
	GSList *d_stack;
} AdmParseRefStr;

/* parser functions */
void _adm_parse_start_element  (GMarkupParseContext *context,
                                      const gchar         *element_name,
                                      const gchar        **attribute_names,
                                      const gchar        **attribute_values,
                                      gpointer             user_data,
                                      GError             **error);

void _adm_parse_end_element    (GMarkupParseContext *context,
                                      const gchar         *element_name,
                                      gpointer             user_data,
                                      GError             **error);

void _adm_parse_error          (GMarkupParseContext *context,
                                      GError              *error,
                                      gpointer             user_data);

/* xml parser callbacks */

/* stack helper functions */
#define _adm_stack_push(STACK_P_P, DATA) \
	((void) (*STACK_P_P = g_slist_prepend(*STACK_P_P, DATA)))
inline gpointer _adm_stack_pop(GSList **stack) {
	gpointer data;
	GSList *link;
	
	data = NULL;
	link = *stack;
	if (link != NULL) {
		data = link->data;

		*stack = link->next;
		g_slist_free_1(link);
	}

	return data;
}
#define _adm_parse_push_state(PTOKEN_P, STATE) \
       	_adm_stack_push(&(PTOKEN_P)->s_stack, (gpointer) STATE)
#define _adm_parse_pop_state(PTOKEN_P) \
	((AdmParseState) _adm_stack_pop(&(PTOKEN_P)->s_stack))

#define _adm_parse_push_data(PTOKEN_P, DATA) \
	_adm_stack_push(&(PTOKEN_P)->d_stack, (gpointer) DATA)
#define _adm_parse_pop_data(PTOKEN_P) \
	((AdmData *) _adm_stack_pop(&(PTOKEN_P)->d_stack))

/* search for a particular string in a NULL terminated array of strings */
const gchar * _adm_parse_array_lookup (const gchar **key_array,
					 const gchar **value_array,
					 const gchar *match_string )
{
	guint i;
	for (i = 0; key_array[i] != NULL; ++ i)
		if (strcmp(key_array[i], match_string) == 0)
			return value_array[i];
	return NULL;
}

void _adm_parse_start_element  (GMarkupParseContext *context,
                          const gchar         *element_name,
                          const gchar        **attribute_names,
                          const gchar        **attribute_values,
                          gpointer             user_data,
                          GError             **error)
{

	AdmParseRefStr *p_token;
	AdmRefStr *refstrA, *refstrB;
	AdmParseState state;
	AdmData *data, *parent, *tmp;
	AdmDataType type;
	AdmIdentity *id;
	guint i;
	gdouble f;
	const gchar *name_str, *source_str, *type_str, *value_str, *flag_str;
	
#ifdef __ADM_DEBUG__
	g_assert(context != NULL);
	g_assert(element_name != NULL);
	g_assert(attribute_names != NULL);
	g_assert(attribute_values != NULL);
	g_assert(user_data != NULL);
#endif

	/* extract the p_token pointer from user_data */
	p_token = (AdmParseRefStr *) user_data;

	/* pop the current state off the stack */
	state = _adm_parse_pop_state(p_token);	
	
	switch (state) {
	case ADM_PSTATE_INITIAL:
		/* check for a valid root node */
		if (strcmp(element_name, "adm_context") != 0) {
			g_set_error(error, G_MARKUP_ERROR, G_MARKUP_ERROR_UNKNOWN_ELEMENT,
				"Document does not contain <adm_context> root.\n");
			break;
		}
		
		/* search for the encoding version attribute */
		value_str = _adm_parse_array_lookup(attribute_names, attribute_values, "VERSION");
		if (value_str == NULL) {
			g_set_error(error, G_MARKUP_ERROR, G_MARKUP_ERROR_INVALID_CONTENT,
				"adm_context encoding version not set.\n");
			break;
		}

		/* check the encoding version */
		if (strcmp(ADM_VERSION_STRING, value_str) != 0) {
			g_set_error(error, G_MARKUP_ERROR, G_MARKUP_ERROR_INVALID_CONTENT,
				"adm_context encoding version invalid.\n");
			break;
		}

		/* initialize the state stack */
		_adm_parse_push_state(p_token, ADM_PSTATE_FINAL);
		_adm_parse_push_state(p_token, ADM_PSTATE_CONTEXT_ENDWAIT);
		_adm_parse_push_state(p_token, ADM_PSTATE_DT_ENDWAIT);
		_adm_parse_push_state(p_token, ADM_PSTATE_CONTEXT_DTWAIT);
		_adm_parse_push_state(p_token, ADM_PSTATE_CONTEXT_IDWAIT);
		break;
		
	case ADM_PSTATE_ID_AWAIT:
	case ADM_PSTATE_ID_BWAIT:
		if (strcmp(element_name, "null") == 0) {
			/* we've found a NULL element. Wait for its end */
			_adm_parse_push_state(p_token, ADM_PSTATE_NULL_ENDWAIT);
			break;
		}

	case ADM_PSTATE_CONTEXT_IDWAIT:
		/* check for a valid ID element */
		if (strcmp(element_name, "id") != 0) {
			g_set_error(error, G_MARKUP_ERROR, G_MARKUP_ERROR_UNKNOWN_ELEMENT,
				"Waiting for an <id> element, found: %s", element_name);
			break;
		}

		name_str = _adm_parse_array_lookup(attribute_names, attribute_values, "NAME");
		if (name_str == NULL) {
			g_set_error(error, G_MARKUP_ERROR, G_MARKUP_ERROR_INVALID_CONTENT,
				"<id> element NAME attribute not set.");
			break;
		}
		
		if (!_adm_valid_name(name_str)) {
			g_set_error(error, G_MARKUP_ERROR, G_MARKUP_ERROR_INVALID_CONTENT,
				"<id> element NAME attribute invalid.");
			break;
		}
		
		value_str = _adm_parse_array_lookup(attribute_names, attribute_values, "SERIAL");
		if (value_str == NULL) {
			g_set_error(error, G_MARKUP_ERROR, G_MARKUP_ERROR_INVALID_CONTENT,
				"<id> element SERIAL attribute not set.");
			break;
		}

		if (!sscanf(value_str,"%d",&i)) {
			g_set_error(error, G_MARKUP_ERROR, G_MARKUP_ERROR_INVALID_CONTENT,
				"<id> element SERIAL attribute invalid.");
			break;
		}

		/* we have a name, and a serial number at this point */
		if (state == ADM_PSTATE_CONTEXT_IDWAIT) {
			/* If this is the root ID element, we have enough information
			 * to create the context */
			p_token->context = AdmContext_new(name_str, i, NULL);
			id = p_token->context->id_tree_root;

		} else {
			refstrA = AdmRefStr_get(name_str, 1);
			id = _adm_create_identity(p_token->context, refstrA, i);
			AdmRefStr_unref(refstrA);
			
			if (state == ADM_PSTATE_ID_AWAIT) 
				((AdmIdentity *) p_token->d_stack->data)->parentA = id;
			else /* BWAIT */
				((AdmIdentity *) p_token->d_stack->data)->parentB = id;
		}

		/* schedule our end, our B, and our A parent */
		_adm_parse_push_state(p_token, ADM_PSTATE_ID_ENDWAIT);
		_adm_parse_push_state(p_token, ADM_PSTATE_ID_BWAIT);
		_adm_parse_push_state(p_token, ADM_PSTATE_ID_AWAIT);
	       
		/* push this data element onto the stack, so that our parents can find it. */
		_adm_parse_push_data(p_token, id);
		break;
		
	case ADM_PSTATE_NULL_ENDWAIT:
		g_set_error(error, G_MARKUP_ERROR, G_MARKUP_ERROR_INVALID_CONTENT,
				"Waiting for a <null> element to end, found: %s", element_name);
		break;
		
	case ADM_PSTATE_ID_ENDWAIT:
		g_set_error(error, G_MARKUP_ERROR, G_MARKUP_ERROR_INVALID_CONTENT,
				"Waiting for an <id> element to end, found: %s", element_name);
		break;
		
	case ADM_PSTATE_CONTEXT_DTWAIT:
		if (strcmp(element_name, "datatree") != 0)
			g_set_error(error, G_MARKUP_ERROR, G_MARKUP_ERROR_UNKNOWN_ELEMENT,
				"Waiting for a <datatree>, found: %s", element_name);

		/* push the root list, so we can link against it later */
		_adm_parse_push_data(p_token, p_token->context->data_tree_root);
		break;
		
	case ADM_PSTATE_DT_ENDWAIT:
	case ADM_PSTATE_LIST_ENDWAIT:
		/* reset the current state */
		_adm_parse_push_state(p_token, state);
		
		if (!_adm_valid_name(element_name)) {
			g_set_error(error, G_MARKUP_ERROR, G_MARKUP_ERROR_INVALID_CONTENT,
				"Data element with invalid name: %s", element_name);
			break;
		}
		
		type_str = _adm_parse_array_lookup(attribute_names, attribute_values, "TYPE");
		if (type_str == NULL) {
			g_set_error(error, G_MARKUP_ERROR, G_MARKUP_ERROR_INVALID_CONTENT,
				"Data element with no TYPE attribute: %s", element_name);
			break;
		}

		if (!strcmp(type_str, "LIST")) type = ADM_TYPE_LIST;
		else if (!strcmp(type_str, "INT")) type = ADM_TYPE_INT;
		else if (!strcmp(type_str, "FLOAT")) type = ADM_TYPE_FLOAT;
		else if (!strcmp(type_str, "BOOL")) type = ADM_TYPE_BOOL;
		else if (!strcmp(type_str, "STRING")) type = ADM_TYPE_STRING;
		else if (!strcmp(type_str, "BASE64")) type = ADM_TYPE_BASE64;
		else if (!strcmp(type_str, "COPY")) type = ADM_TYPE_COPY;
		else {
			g_set_error(error, G_MARKUP_ERROR, G_MARKUP_ERROR_INVALID_CONTENT,
				"Data element with invalid TYPE attribute: %s", type_str);
			break;
		}
			
		source_str = _adm_parse_array_lookup(attribute_names, attribute_values, "SOURCE");
		if (source_str == NULL) {
			/* blank sources are implied to be equal to the base source */
			refstrB = AdmRefStr_ref(p_token->context->id_tree_root->name);
		} else if (!_adm_valid_name(source_str)) {
			g_set_error(error, G_MARKUP_ERROR, G_MARKUP_ERROR_INVALID_CONTENT,
				"Data element with invalid SOURCE attribute: %s", source_str);
			break;
		} else {
			refstrB = AdmRefStr_get(source_str, 1);
		}

		/* create the data element, and append it to its parent */
		parent = (AdmData *) p_token->d_stack->data;
		refstrA = AdmRefStr_get(element_name, 1);

		/* check for name collides */
		if (!(parent->flags & ADM_DATA_FLAG_ANONYMOUS)) {
			i = ((AdmList *) parent)->num_children;
			tmp = parent->child;
			while (i > 0) {
				if (tmp->name == refstrA) {
					g_set_error(error, G_MARKUP_ERROR, G_MARKUP_ERROR_INVALID_CONTENT,
						"Name collide: %s in non-anonymous list: %s", 
						AdmRefStr_str(refstrA), AdmRefStr_str(parent->name));
					return;			
				}
				tmp = tmp->next;
				--i;
			}
		}
		
		data = _adm_create_data(p_token->context, parent, type, refstrA, refstrB);
		AdmRefStr_unref(refstrA);
		AdmRefStr_unref(refstrB);
		
		if (++ ((AdmList *) parent)->num_children == 1) {
			parent->child = data;
		} else {
			data->next = parent->child;
			data->prev = parent->child->prev;
			data->prev->next = data;
			data->next->prev = data;
		}
		
		flag_str = _adm_parse_array_lookup(attribute_names, attribute_values, "PROTECTED");
		if (flag_str && !strcmp(flag_str, "TRUE")) data->flags |= ADM_DATA_FLAG_PROTECTED;
		
		flag_str = _adm_parse_array_lookup(attribute_names, attribute_values, "PREPEND");
		if (flag_str && !strcmp(flag_str, "TRUE")) data->flags |= ADM_DATA_FLAG_PREPEND;
		
		flag_str = _adm_parse_array_lookup(attribute_names, attribute_values, "ANONYMOUS");
		if (flag_str && !strcmp(flag_str, "TRUE")) data->flags |= ADM_DATA_FLAG_ANONYMOUS;
		
		flag_str = _adm_parse_array_lookup(attribute_names, attribute_values, "ATOMIC");
		if (flag_str && !strcmp(flag_str, "TRUE")) data->flags |= ADM_DATA_FLAG_ATOMIC;
		
		
		if (data == NULL) {
			g_set_error(error, G_MARKUP_ERROR, G_MARKUP_ERROR_INVALID_CONTENT,
				"Illegal list insertion of %s Data element %s", type_str, element_name);
			break;
		}

		/* push this element onto the data stack */
		_adm_parse_push_data(p_token, data);

		if (type == ADM_TYPE_LIST) {
			/* wait for our ending */
			_adm_parse_push_state(p_token, ADM_PSTATE_LIST_ENDWAIT);
			break;
		}
	
		value_str = _adm_parse_array_lookup(attribute_names, attribute_values, "VALUE");
		if (value_str == NULL) {
			g_set_error(error, G_MARKUP_ERROR, G_MARKUP_ERROR_INVALID_CONTENT,
				"%s Data element with no VALUE attribute: %s", type_str, element_name);
			break;
		}

		switch (type) {
		case ADM_TYPE_INT:
			if (!sscanf(value_str, "%d", &i)) {
				g_set_error(error, G_MARKUP_ERROR, G_MARKUP_ERROR_INVALID_CONTENT,
					"%s Data element with invalid VALUE attribute: %s", type_str, element_name);
				break;
			}
			((AdmInt *) data)->value = i;
			break;

		case ADM_TYPE_FLOAT:
			if (!sscanf(value_str, "%la", (double *) &f)) {
				g_set_error(error, G_MARKUP_ERROR, G_MARKUP_ERROR_INVALID_CONTENT,
					"%s Data element with invalid VALUE attribute: %s", type_str, element_name);
				break;
			}
			((AdmFloat *) data)->value = f;
			break;

		case ADM_TYPE_BOOL:
			if (strcmp(value_str, "TRUE") == 0)
				AdmBool_setValue((AdmBool *) data, TRUE, NULL);
			else if (strcmp(value_str, "FALSE") == 0)
				AdmBool_setValue((AdmBool *) data, FALSE, NULL);
			else
				g_set_error(error, G_MARKUP_ERROR, G_MARKUP_ERROR_INVALID_CONTENT,
					"%s Data element with invalid VALUE attribute: %s", type_str, element_name);

			break;

		case ADM_TYPE_STRING:
			/* the markup parser already de-escaped and validated our string */
			AdmRefStr_unref(((AdmString *) data)->value);
			((AdmString *) data)->value = AdmRefStr_get(value_str, 1);
			break;
			
		case ADM_TYPE_BASE64:
			{
				unsigned long len = strlen(value_str);
				if (!_adm_valid_base64(value_str, len)) {
					g_set_error(error, G_MARKUP_ERROR, G_MARKUP_ERROR_INVALID_CONTENT,
						"Base64 Data element with invalid payload: %s", element_name);
				} else {
					AdmRefStr_unref(((AdmBase64 *) data)->value);
					((AdmBase64 *) data)->value = AdmRefStr_nget(value_str, len, 1);
				}
			}
			break;
			

		case ADM_TYPE_COPY:
			if (!_adm_valid_path(value_str)) {
				g_set_error(error, G_MARKUP_ERROR, G_MARKUP_ERROR_INVALID_CONTENT,
					"Copy Data element with invalid VALUE: %s, %s", element_name, value_str);
			} else {
				AdmRefStr_unref(((AdmCopy *) data)->value);
				((AdmCopy *) data)->value = AdmRefStr_get(value_str, 1);
			}
			break;
	
		default:
			/* this cant happen, we checked above */
			break;
		}

		/* wait for the completion of this data element */	
		_adm_parse_push_state(p_token, ADM_PSTATE_DATA_ENDWAIT);

		break;
		
	case ADM_PSTATE_DATA_ENDWAIT:
		g_set_error(error, G_MARKUP_ERROR, G_MARKUP_ERROR_INVALID_CONTENT,
				"Waiting for a Data element to end, found: %s", element_name);
		break;
		
	case ADM_PSTATE_CONTEXT_ENDWAIT:
		if (strcmp(element_name, "mask") != 0) {
			g_set_error(error, G_MARKUP_ERROR, G_MARKUP_ERROR_UNKNOWN_ELEMENT,
				"Waiting for <adm_context> to end, or a <mask>, found: %s", element_name);
			break;
		}

		value_str = _adm_parse_array_lookup(attribute_names, attribute_values, "PATH");
		if (value_str == NULL) {
			g_set_error(error, G_MARKUP_ERROR, G_MARKUP_ERROR_INVALID_CONTENT,
				"<mask> element PATH attribute not set.");
			break;
		}
		
		if (!_adm_valid_path(value_str)) {
			g_set_error(error, G_MARKUP_ERROR, G_MARKUP_ERROR_INVALID_CONTENT,
				"<mask> element PATH attribute INVALID: %s", value_str);
			break;
		}
		
		/* set the mask */
		AdmContext_setMask(p_token->context, value_str, NULL);

		/* reset the state, and wait for the end of the mask. */
		_adm_parse_push_state(p_token, ADM_PSTATE_CONTEXT_ENDWAIT);
		_adm_parse_push_state(p_token, ADM_PSTATE_DELETE_ENDWAIT);
		break;

	case ADM_PSTATE_DELETE_ENDWAIT:
		g_set_error(error, G_MARKUP_ERROR, G_MARKUP_ERROR_INVALID_CONTENT,
				"Waiting for a <mask> element to end, found: %s", element_name);
		break;
		
	case ADM_PSTATE_FINAL:
		/* we've had a fatal error, raise the error */
		g_set_error(error, G_MARKUP_ERROR, G_MARKUP_ERROR_INVALID_CONTENT,
				"Data past document end.");
		break;
		
	default:
#ifdef __ADM_DEBUG__
		g_error("UNDEFINED STATE REACHED!\n");
#endif
		g_set_error(error, G_MARKUP_ERROR, G_MARKUP_ERROR_INVALID_CONTENT,
				"Data past document end.");
		break;
	}

	return;
}

void _adm_parse_end_element    (GMarkupParseContext *context,
                          const gchar         *element_name,
                          gpointer             user_data,
                          GError             **error)
{
	AdmParseRefStr *p_token;
	AdmParseState state;

#ifdef __ADM_DEBUG__
	g_assert(context != NULL);
	g_assert(element_name != NULL);
	g_assert(user_data != NULL);
#endif

	/* extract the p_token pointer from user_data */
	p_token = (AdmParseRefStr *) user_data;

	/* pop the current state off the stack */
	state = _adm_parse_pop_state(p_token);
	
	switch (state) {
	case ADM_PSTATE_INITIAL:
		g_set_error(error, G_MARKUP_ERROR, G_MARKUP_ERROR_INVALID_CONTENT,
				"Waiting for an <adm_context>, found a closing: %s", element_name);
		break;

	case ADM_PSTATE_CONTEXT_IDWAIT:
		g_set_error(error, G_MARKUP_ERROR, G_MARKUP_ERROR_INVALID_CONTENT,
				"Waiting for an <id>, found a closing: %s", element_name);
		break;

	case ADM_PSTATE_ID_AWAIT:
		g_set_error(error, G_MARKUP_ERROR, G_MARKUP_ERROR_INVALID_CONTENT,
				"Waiting for an A parent <id>, found a closing: %s", element_name);
		break;

	case ADM_PSTATE_ID_BWAIT:
		g_set_error(error, G_MARKUP_ERROR, G_MARKUP_ERROR_INVALID_CONTENT,
				"Waiting for an B parent <id>, found a closing: %s", element_name);
		break;

	case ADM_PSTATE_NULL_ENDWAIT:
		if (strcmp("null", element_name) != 0)
			g_set_error(error, G_MARKUP_ERROR, G_MARKUP_ERROR_INVALID_CONTENT,
				"Waiting for the end of a <null> element, found a closing: %s", element_name);
		break;

	case ADM_PSTATE_ID_ENDWAIT:
		/* pop the id element */
		_adm_parse_pop_data(p_token);
		break;
		
	case ADM_PSTATE_CONTEXT_DTWAIT:
		g_set_error(error, G_MARKUP_ERROR, G_MARKUP_ERROR_INVALID_CONTENT,
				"Waiting for a <datatree>, found a closing: %s", element_name);
		break;

	case ADM_PSTATE_DT_ENDWAIT:
		if (strcmp("datatree", element_name) != 0) {
			g_set_error(error, G_MARKUP_ERROR, G_MARKUP_ERROR_INVALID_CONTENT,
				"Waiting for the end of the <datatree>, found a closing: %s", element_name);
			break;
		}
		/* pop the data list */
		_adm_parse_pop_data(p_token);
		break;

	case ADM_PSTATE_LIST_ENDWAIT:
	case ADM_PSTATE_DATA_ENDWAIT:
		if (strcmp( AdmRefStr_str(((AdmData *) p_token->d_stack->data)->name), element_name) != 0) {
			g_set_error(error, G_MARKUP_ERROR, G_MARKUP_ERROR_INVALID_CONTENT,
				"Waiting for the end of Data Element %s, found a closing: %s",
				AdmRefStr_str(((AdmData *) p_token->d_stack)->name), element_name);
			break;
		}
		/* pop the data list */
		_adm_parse_pop_data(p_token);
		break;

	case ADM_PSTATE_CONTEXT_ENDWAIT:
		if (strcmp("adm_context", element_name) != 0) {
			g_set_error(error, G_MARKUP_ERROR, G_MARKUP_ERROR_INVALID_CONTENT,
				"Waiting for the end of an <adm_context>, found a closing: %s", element_name);
			break;
		}
		/* we've consumed the ADM_CONTEXT, done */
		break;
		
	case ADM_PSTATE_DELETE_ENDWAIT:
		if (strcmp("mask", element_name) != 0)
			g_set_error(error, G_MARKUP_ERROR, G_MARKUP_ERROR_INVALID_CONTENT,
				"Waiting for the end of a <mask> element, found a closing: %s", element_name);
		break;

	case ADM_PSTATE_FINAL:
		/* we've had a fatal error, raise the error */
		g_set_error(error, G_MARKUP_ERROR, G_MARKUP_ERROR_INVALID_CONTENT,
				"Data past document end.");
		break;

	default:
#ifdef __ADM_DEBUG__
		g_error("Undefined state reached: %d\n", state);
#endif
		g_set_error(error, G_MARKUP_ERROR, G_MARKUP_ERROR_INVALID_CONTENT,
				"Undefined state reached: %d", state);
		break;
	}

	return;
}

void _adm_parse_error          (GMarkupParseContext *context,
                          GError              *error,
                          gpointer             user_data)
{
	AdmParseRefStr *p_token;

#ifdef __ADM_DEBUG__
	g_assert(context != NULL);
	g_assert(user_data != NULL);
#endif

	/* disenter the p_token from user_data*/
	p_token = (AdmParseRefStr *) user_data;

	/* all errors are fatal, so we need to unref the AdmContext if it exists. */
	if (p_token->context != NULL) {
		AdmContext_unref(p_token->context, NULL);
		p_token->context = NULL;
	}

	/* put an error on the state stack */
	_adm_parse_push_state(p_token, ADM_PSTATE_ERROR);
	
	return;
}


/***************************************************************************
 * Public Functions
 ***************************************************************************/


/**
 * AdmContext_fromXML: decode an XML string into an @AdmContext element
 * @xml: a pointer to the xml string
 * @error: an optional GError
 *
 * This function parses an XML string into the AdmContext which it encodes.
 * All errors are fatal, and will cause it to return NULL.
 **/

const GMarkupParser _adm_fromxml_parser = {
start_element:	_adm_parse_start_element,
end_element:	_adm_parse_end_element,
text:		NULL,
passthrough:	NULL,
error:		_adm_parse_error,
};

AdmContext *
AdmContext_fromXML (const char *xml,
		    AdmError **error)
{
	guint k;
	GMarkupParseContext *parse_context;
	AdmParseRefStr p_token;
	GError *g_err;

	/* sanity check */
	_adm_abort_if_fail(xml != NULL);

	/* init the parse refstr */
	p_token.context = NULL;
	p_token.s_stack = NULL;
	p_token.d_stack = NULL;
	
	/* scan the string for its length */
	k = strlen(xml);

	/* Create a new parse context */
	parse_context = g_markup_parse_context_new(&_adm_fromxml_parser, 0, (gpointer) &p_token, NULL);

	/* run the parse */
	g_err = NULL;
	g_markup_parse_context_parse(parse_context, xml, k, &g_err);

	/* free the context */
	g_markup_parse_context_free(parse_context);

	/* examine the p_token */
	if (p_token.s_stack != NULL) {
		if (p_token.s_stack->data != (gpointer) ADM_PSTATE_FINAL) {
			/* something is broken, kill the context */
			if (p_token.context != NULL) {
				AdmContext_unref(p_token.context, NULL);
				p_token.context = NULL;
			}
	
		}
		
		/* consume the state stack */
		g_slist_free(p_token.s_stack);
		p_token.s_stack = NULL;
	}

	/* consume the data stack */
	if (p_token.d_stack != NULL) {
		g_slist_free(p_token.d_stack);
		p_token.d_stack = NULL;
	}

	if (g_err != NULL) {
		AdmError_create(error, ADM_EPARSE, g_err->message);
		g_error_free(g_err);
	}
	
	return p_token.context;
}

